/*
 * Copyright (c) 2022-2024 NVIDIA CORPORATION & AFFILIATES. All rights reserved.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at:
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

#include <config.h>

#include "batch.h"
#include "conntrack.h"
#include "conntrack-offload.h"
#include "conntrack-private.h"
#include "conntrack-tp.h"
#include "dp-packet.h"
#include "dpdk-ext.h"
#include "fatal-signal.h"
#include "netdev-offload.h"
#include "netdev-offload-dpdk-ext.h"
#include "openvswitch/vlog.h"
#include "ovs-doca.h"
#include "ovs-rcu.h"
#include "timeval.h"
#include "uuid.h"

VLOG_DEFINE_THIS_MODULE(conntrack_offload);

#define CT_OFFLOAD_DEFAULT_SIZE 250000
static unsigned int ct_offload_size = CT_OFFLOAD_DEFAULT_SIZE;

static bool ct_offload_ipv6_enabled = false;

static bool ct_e2e_cache_enabled = false;

static struct ovs_dpdk_mempool *handles;

struct ct_offload_handle *
conntrack_offload_get(struct conn *conn)
{
    return conn_master(conn)->offloads.coh;
}

static uintptr_t
conntrack_offload_get_ctid_key(struct conn *conn)
{
    return (uintptr_t) (conn->master_conn ? conn->master_conn : conn);
}

static bool
conntrack_offload_fill_item_common(struct ct_flow_offload_item *item,
                                   struct conntrack *ct,
                                   struct conn *conn,
                                   int dir)
{
    item->ufid = conn->offloads.dir_info[dir].ufid;
    item->ct_match.odp_port = conn->offloads.dir_info[dir].port;
    item->ct = ct;
    if (ct != NULL) {
        item->dp = ct->dp;
    }
    item->ctid_key = conntrack_offload_get_ctid_key(conn);
    item->ct_actions_set = false;
    item->actions = NULL;
    item->actions_size = 0;

    /* This field is set by the offload provider once offloading is completed. */
    item->offload_data = ct_dir_info_data_get(&conn->offloads.dir_info[dir]);

    return dir == CT_DIR_REP
           ? !!(conn->offloads.flags & CT_OFFLOAD_REP)
           : !!(conn->offloads.flags & CT_OFFLOAD_INIT);
}

void
conntrack_offload_build_del_items(struct conntrack *ct, struct conn *conn,
                                  struct ct_flow_offload_item items[CT_DIR_NUM])
{
    /* This function is also used when deleting e2e-offloaded connections.
     * In this case, the 'ct_offload_handle' might not be allocated. */
    for (int dir = 0; dir < CT_DIR_NUM; dir++) {
        struct conn *conn_dir;

        if (conn->nat_conn &&
            conn->nat_conn->offloads.dir_info[dir].dp) {
            conn_dir = conn->nat_conn;
        } else {
            conn_dir = conn;
        }
        if (!conntrack_offload_fill_item_common(&items[dir], ct, conn_dir, dir)) {
            continue;
        }
    }
}

void
conntrack_offload_free(struct conn *conn)
{
    struct ct_offload_handle *coh = conntrack_offload_get(conn);

    if (coh) {
        if (coh->lock.where) {
            ovs_spin_destroy(&coh->lock);
        }
        if (handles) {
            ovs_dpdk_mempool_free(handles, coh);
        } else {
            free(coh);
        }
        conn_master(conn)->offloads.coh = NULL;
    }
}

void
conntrack_offload_del_conn(struct conntrack *ct,
                           struct conn *conn,
                           bool flush)
    OVS_REQUIRES(conn->lock)
{
    struct conntrack_offload_class *offload_class;
    struct ct_flow_offload_item item[CT_DIR_NUM];
    struct conn *conn_dir;
    long long int now = time_usec();
    void *dp;
    int dir;

    if (!conntrack_offload_is_enabled()) {
        return;
    }

    if (conn_master(conn)->offloads.flags & CT_OFFLOAD_DELETED ||
        !(conn_master(conn)->offloads.flags & CT_OFFLOAD_REQUESTED)) {
        return;
    }

    offload_class = ovsrcu_get(struct conntrack_offload_class *,
                               &ct->offload_class);

    if (offload_class) {
        /* 'conn_del' must be defined if an offload handle is set. */
        ovs_assert(!conntrack_offload_get(conn) || offload_class->conn_del);
    } else {
        /* The datapath is being destroyed. If the offload handle is set,
         * do not enqueue a conn-del: directly free the offload handle now.
         * The offload resources will be freed when flushing each of the
         * port previously part of the datapath. */
        conntrack_conn_unref(conn);
        conntrack_conn_unref(conn);
        goto mark_deleted;
    }

    conn = conn_master(conn);

    if (ct_e2e_cache_enabled) {
        conntrack_offload_build_del_items(ct, conn, item);
    }
    for (dir = 0; dir < CT_DIR_NUM; dir ++) {
        if (conn->nat_conn &&
            conn->nat_conn->offloads.dir_info[dir].dp) {
            conn_dir = conn->nat_conn;
        } else {
            conn_dir = conn;
        }
        if (ct_e2e_cache_enabled) {
            dp = conn_dir->offloads.dir_info[dir].dp;
            offload_class->conn_e2e_del(&item[dir].ufid, dp, conn, now);
        }
        /* Set connection's status to terminated to indicate that the offload
         * of the connection is deleted, but should still bypass tcp seq
         * checking.
         */
        if (flush) {
            /* For flush, keep only terminated flag, so further TCP packets
             * will still bypass seq checking.
             */
            conn_dir->offloads.flags = CT_OFFLOAD_TERMINATED;
        } else {
            conn_dir->offloads.flags |= CT_OFFLOAD_TERMINATED | CT_OFFLOAD_SKIP;
        }
        if (dir == CT_DIR_REP) {
            conn_dir->offloads.flags &= ~CT_OFFLOAD_REP;
        } else {
            conn_dir->offloads.flags &= ~CT_OFFLOAD_INIT;
        }
        conn_dir->offloads.flags |= CT_OFFLOAD_DELETED;
    }

    if (!ct_e2e_cache_enabled) {
        offload_class->conn_del(ct, conn);
        conntrack_stats_inc(ct, CT_DPIF_CONN_STATE_OFFLOAD_DEL_REQUESTED);
    }
mark_deleted:
    conn_master(conn)->offloads.flags |= CT_OFFLOAD_DELETED;
}

static void
conntrack_swap_conn_key(const struct conn_key *key,
                        struct conn_key *swapped)
{
    memcpy(swapped, key, sizeof *swapped);
    swapped->src = key->dst;
    swapped->dst = key->src;
}

static void
conntrack_offload_fill_item_add(struct ct_flow_offload_item *item,
                                struct conntrack *ct,
                                struct conn *conn,
                                int dir,
                                long long int now)
{
    /* nat_conn has opposite directions. */
    bool reply = !!conn->master_conn ^ dir;

    if (reply) {
        item->ct_match.key = conn->rev_key;
        conntrack_swap_conn_key(&conn->key, &item->nat.key);
    } else {
        item->ct_match.key = conn->key;
        conntrack_swap_conn_key(&conn->rev_key, &item->nat.key);
    }

    item->nat.mod_flags = 0;
    if (memcmp(&item->nat.key.src.addr, &item->ct_match.key.src.addr,
               sizeof item->nat.key.src)) {
        item->nat.mod_flags |= NAT_ACTION_SRC;
    }
    if (item->nat.key.src.port != item->ct_match.key.src.port) {
        item->nat.mod_flags |= NAT_ACTION_SRC_PORT;
    }
    if (memcmp(&item->nat.key.dst.addr, &item->ct_match.key.dst.addr,
               sizeof item->nat.key.dst)) {
        item->nat.mod_flags |= NAT_ACTION_DST;
    }
    if (item->nat.key.dst.port != item->ct_match.key.dst.port) {
        item->nat.mod_flags |= NAT_ACTION_DST_PORT;
    }

    item->conn_dir_info = &conn->offloads.dir_info[dir];
    conntrack_offload_fill_item_common(item, ct, conn, dir);
    item->ct_state = conn->offloads.dir_info[dir].pkt_ct_state;
    item->mark_key = conn->offloads.dir_info[dir].pkt_ct_mark;
    item->label_key = conn->offloads.dir_info[dir].pkt_ct_label;
    item->timestamp = now;
}

static inline void
e2e_cache_trace_add_ct(struct conntrack *ct,
                       struct dp_packet *p,
                       struct conn *conn,
                       bool reply,
                       long long int now)
{
    struct conntrack_offload_class *offload_class;
    uint32_t e2e_trace_size = p->e2e_trace_size;
    struct ct_flow_offload_item item;
    struct ct_dir_info *dir_info;
    uint8_t e2e_seen_pkts;
    int dir;

    offload_class = ovsrcu_get(struct conntrack_offload_class *,
                               &ct->offload_class);
    if (!offload_class || !offload_class->conn_get_ufid ||
        !offload_class->conn_e2e_add) {
        return;
    }

    if (conn->key.dl_type == htons(ETH_TYPE_IPV6) &&
        !conntrack_offload_ipv6_is_enabled()) {
        /* If conn type is IPv6 and IPv6 offload is disabled,
         * do not generate requests. */
        p->e2e_trace_flags |= E2E_CACHE_TRACE_FLAG_ABORT;
        return;
    }


    if (OVS_UNLIKELY(e2e_trace_size >= E2E_CACHE_MAX_TRACE)) {
        p->e2e_trace_flags |= E2E_CACHE_TRACE_FLAG_OVERFLOW;
        return;
    }

    dir = ct_get_packet_dir(reply);
    if (conn->nat_conn &&
        conn->nat_conn->offloads.dir_info[dir].dp) {
        conn = conn->nat_conn;
    } else if (conn->master_conn &&
               conn->master_conn->offloads.dir_info[dir].dp) {
        conn = conn->master_conn;
    }
    conntrack_offload_fill_item_add(&item, ct, conn, dir, now);
    item.ct_match.odp_port = p->md.in_port.odp_port;
    item.ct_match.orig_in_port = p->md.orig_in_port;

    dir_info = &conn->offloads.dir_info[dir];
    dir = ct_get_packet_dir(!reply);
    if (conn->nat_conn &&
        conn->nat_conn->offloads.dir_info[dir].dp) {
        conn = conn->nat_conn;
    } else if (conn->master_conn &&
               conn->master_conn->offloads.dir_info[dir].dp) {
        conn = conn->master_conn;
    }
    if (!dir_info->e2e_flow) {
        dir_info->e2e_flow = true;
        offload_class->conn_get_ufid(&dir_info->ufid);
        item.ufid = dir_info->ufid;
        offload_class->conn_e2e_add(&item);
    }

    /* Prevent sending E2E trace messages for every packet. Send only
     * when number of seen packets is equal to 2^x.
     */
    e2e_seen_pkts = dir_info->e2e_seen_pkts++;
    if ((e2e_seen_pkts & (e2e_seen_pkts - 1u)) != 0) {
        p->e2e_trace_flags |= E2E_CACHE_TRACE_FLAG_THROTTLED;
        return;
    }

    p->e2e_trace_ct_ufids |= 1 << e2e_trace_size;
    p->e2e_trace[e2e_trace_size] = dir_info->ufid;
    p->e2e_trace_size = e2e_trace_size + 1;

    if (!conn->offloads.dir_info[dir].e2e_flow) {
        p->e2e_trace_flags |= E2E_CACHE_TRACE_FLAG_ABORT;
        return;
    }

    e2e_trace_size++;
    if (OVS_UNLIKELY(e2e_trace_size >= E2E_CACHE_MAX_TRACE)) {
        p->e2e_trace_flags |= E2E_CACHE_TRACE_FLAG_OVERFLOW;
        return;
    }

    p->e2e_trace_ct_ufids |= 1 << e2e_trace_size;
    p->e2e_trace[e2e_trace_size] = conn->offloads.dir_info[dir].ufid;
    p->e2e_trace_size = e2e_trace_size + 1;
}

void
conntrack_offload_build_add_items(struct conntrack *ct, struct conn *conn,
                                  struct ct_flow_offload_item items[CT_DIR_NUM],
                                  long long int now_us)
{
    struct ct_offload_handle *coh = conntrack_offload_get(conn);

    ovs_assert(coh);
    for (int dir = 0; dir < CT_DIR_NUM; dir++) {
        if (conn->nat_conn &&
            conn->nat_conn->offloads.dir_info[dir].dp) {
            conn = conn->nat_conn;
        } else if (conn->master_conn &&
                   conn->master_conn->offloads.dir_info[dir].dp) {
            conn = conn->master_conn;
        }
        conntrack_offload_fill_item_add(&items[dir], ct, conn, dir, now_us);
        items[dir].offload_data = &coh->dir[dir];
        items[dir].ufid = conn->offloads.dir_info[dir].ufid;
    }
}

void
conntrack_offload_prepare(struct conntrack *ct, struct conn *conn,
                          struct dp_packet *packet, bool reply)
{
    enum ct_offload_flag dir_flag = reply ? CT_OFFLOAD_REP : CT_OFFLOAD_INIT;
    int dir = reply ? CT_DIR_REP : CT_DIR_INIT;
    struct conn *actual;

    if (!conn || !conntrack_offload_is_enabled()) {
        return;
    }

    /* Always write offload info in the actual conn. */
    actual = conn_master(conn);

    if ((actual->key.nw_proto != IPPROTO_UDP && actual->key.nw_proto != IPPROTO_TCP) ||
        !(packet->md.ct_state & CS_ESTABLISHED) ||
        (actual->offloads.flags & CT_OFFLOAD_SKIP)) {
        return;
    }

    if (actual->offloads.flags & dir_flag) {
        /* Write metadata only once per direction. */
        return;
    }

    actual->offloads.flags |= dir_flag;
    conn->offloads.flags |= dir_flag;
    conn->offloads.dir_info[dir].dp = ct->dp;
    conn->offloads.dir_info[dir].pkt_ct_state = packet->md.ct_state;
    conn->offloads.dir_info[dir].pkt_ct_mark = packet->md.ct_mark;
    conn->offloads.dir_info[dir].pkt_ct_label = packet->md.ct_label;
    conn->offloads.dir_info[dir].port = packet->md.orig_in_port;
}

void
conntrack_offload_add(struct conntrack *ct, struct batch *conns)
{
    struct conntrack_offload_class *offload_class;
    struct conn *conn;

    if (!conntrack_offload_is_enabled() || ct_e2e_cache_enabled) {
        return;
    }

    /* CT doesn't handle alg */
    offload_class = ovsrcu_get(struct conntrack_offload_class *,
                               &ct->offload_class);
    if (!offload_class || !offload_class->conns_add) {
        return;
    }

    /* Filter out connections that are either not
     * yet ready to be offloaded or won't ever be offloaded.
     * They are removed from the batch if 'continue' is used,
     * only if they reach the end of the loop will they
     * remain inside. */
    BATCH_FOREACH_POP (conn, conns) {
        struct ct_offload_handle *coh;

        if (!conn) {
            continue;
        }

        conn = conn_master(conn);
        if ((conn->offloads.flags & CT_OFFLOAD_SKIP)) {
            continue;
        }

        if ((conn->offloads.flags & CT_OFFLOAD_BOTH) != CT_OFFLOAD_BOTH) {
            continue;
        }

        conn->offloads.flags |= CT_OFFLOAD_SKIP;

        if (conn->key.dl_type == htons(ETH_TYPE_IPV6) &&
            !conntrack_offload_ipv6_is_enabled()) {
            /* If conn type is IPv6 and IPv6 offload is disabled,
             * do not generate requests. */
            continue;
        }

        if (conn->alg || conn->alg_related) {
            continue;
        }

        if (!conntrack_conn_ref(conn)) {
            /* No point in offloading a connection already being deleted. */
            continue;
        }
        /* Take a reference per directions. */
        ignore(conntrack_conn_ref(conn));

        if (handles) {
            if (ovs_dpdk_mempool_alloc(handles, (void **) &coh)) {
                continue;
            }
        } else {
            coh = xzalloc(sizeof *coh);
        }

        conn->offloads.coh = coh;
        conn->offloads.flags |= CT_OFFLOAD_REQUESTED;
        batch_add(conns, conn);
    }

    conntrack_stats_add(ct, CT_DPIF_CONN_STATE_OFFLOAD_ADD_REQUESTED, batch_size(conns));
    offload_class->conns_add(ct, conns);
}

void
conntrack_offload_e2e_add(struct conntrack *ct, struct conn *conn,
                          struct dp_packet *packet, bool reply, long long now_us)
{
    enum ct_offload_flag dir_flag = reply ? CT_OFFLOAD_REP : CT_OFFLOAD_INIT;

    if (!conn || !ct_e2e_cache_enabled) {
        return;
    }

    if (!(conn_master(conn)->offloads.flags & dir_flag)) {
        /* This direction is not yet ready: the state might not
         * be established, the protocol not (UDP|TCP), etc. */
        return;
    }

    if (!(conn_master(conn)->offloads.flags & CT_OFFLOAD_REQUESTED)) {
        struct conn *actual = conn_master(conn);

        /* No offloads were yet requested for this conn.
         * We are preparing to do so for E2E: take the appropriate
         * references and mark it as such. */
        if (!conntrack_conn_ref(actual)) {
            return;
        }
        /* Take a reference per directions. */
        ignore(conntrack_conn_ref(actual));
        actual->offloads.flags |= CT_OFFLOAD_REQUESTED;
    }

    e2e_cache_trace_add_ct(ct, packet, conn, reply, now_us);
}

static bool
conntrack_offload_inserted(struct conn *conn, enum ct_direction dir)
{
    struct ct_offload_handle *coh = conntrack_offload_get(conn);
    enum ct_offload_flag dir_flag;

    if (dir == CT_DIR_INIT) {
        dir_flag = CT_OFFLOAD_INIT;
    } else if (dir == CT_DIR_REP) {
        dir_flag = CT_OFFLOAD_REP;
    } else {
        dir_flag = CT_OFFLOAD_BOTH;
    }

    return ct_offload_handle_active(coh, dir_flag);
}

bool
conntrack_offload_build_stats_items(struct conntrack *ct, struct conn *conn,
                                    struct ct_flow_offload_item items[CT_DIR_NUM])
{
    /* This function returns 'false' if either of the two direction
     * is not yet ready for processing. In such case, the 'items' are not
     * filled with the required information and querying statistics should not
     * proceed.
     */

    conn = conn_master(conn);
    for (int dir = 0; dir < CT_DIR_NUM; dir++) {
        struct conn *conn_dir = conn;

        if (conn->nat_conn &&
            conn->nat_conn->offloads.dir_info[dir].dp) {
            conn_dir = conn->nat_conn;
        }

        if (!conntrack_offload_inserted(conn_dir, dir) &&
            !conn_dir->offloads.dir_info[dir].e2e_flow) {
            return false;
        }

        if (!conntrack_offload_fill_item_common(&items[dir], ct, conn, dir)) {
            return false;
        }
    }

    return true;
}

static bool
conntrack_offload_is_e2e(struct conn *conn)
{
    conn = conn_master(conn);

    for (int dir = 0; dir < CT_DIR_NUM; dir++) {
        if (conn->offloads.dir_info[dir].e2e_flow ||
            (conn->nat_conn &&
             conn->nat_conn->offloads.dir_info[dir].e2e_flow)) {
            return true;
        }
    }
    return false;
}

static bool
conntrack_offload_active(struct conntrack *ct,
                         struct conntrack_offload_class *offload_class,
                         struct conn *conn, long long int now, int *ret)
{
    struct ct_flow_offload_item items[CT_DIR_NUM];

    conn = conn_master(conn);

    if (!conntrack_offload_is_e2e(conn)) {
        struct ct_offload_handle *coh = conntrack_offload_get(conn);

        if (!ct_offload_handle_active(coh, CT_OFFLOAD_BOTH)) {
            *ret = ENODATA;
        } else {
            *ret = offload_class->conn_active(ct, conn, now, conn->prev_query);
        }

        return (*ret == 0);
    }

    if (!conntrack_offload_build_stats_items(ct, conn, items)) {
        *ret = 0;
        return false;
    }

    for (int dir = 0; dir < CT_DIR_NUM; dir++) {
        *ret = offload_class->conn_e2e_active(&items[dir], now, conn->prev_query);
        if (*ret == 0 ||
            *ret == EAGAIN) {
            break;
        }
    }
    return (*ret == 0);
}

int
conn_hw_update(struct conntrack *ct,
               struct conntrack_offload_class *offload_class,
               struct conn *conn,
               enum ct_timeout *ptm,
               long long now)
{
    enum ct_timeout tm = *ptm;
    int ret = 0;

    if (conntrack_offload_active(ct, offload_class, conn, now, &ret)) {
        conn_lock(conn);
        conn_update_expiration(ct, conn, tm, now);
        conn_unlock(conn);
    }

    ignore(atomic_flag_test_and_set(&conn->exp.reschedule));
    conn->prev_query = now;
    return ret;
}

void
conntrack_set_offload_class(struct conntrack *ct,
                            struct conntrack_offload_class *cls)
{
    ovsrcu_set(&ct->offload_class, cls);
    ct_e2e_cache_enabled = netdev_is_e2e_cache_enabled();
}

unsigned int
conntrack_offload_size(void)
{
    return ct_offload_size;
}

bool
conntrack_offload_is_enabled(void)
{
    return netdev_is_flow_api_enabled()
        && (conntrack_offload_size() > 0);
}

bool
conntrack_offload_ipv6_is_enabled(void)
{
    return conntrack_offload_is_enabled() && ct_offload_ipv6_enabled;
}

static void
destroy_handle_pool(void *aux OVS_UNUSED)
{
    if (handles) {
        ovs_dpdk_mempool_destroy(handles);
        handles = NULL;
    }
}

void
conntrack_offload_config(const struct smap *other_config)
{
    static struct ovsthread_once once = OVSTHREAD_ONCE_INITIALIZER;
    unsigned int req_ct_size;

    if (!ovsthread_once_start(&once)) {
        return;
    }

    req_ct_size = smap_get_uint(other_config, "hw-offload-ct-size",
                                CT_OFFLOAD_DEFAULT_SIZE);
    if (req_ct_size != ct_offload_size) {
        ct_offload_size = req_ct_size;
        if (conntrack_offload_is_enabled()) {
            VLOG_INFO("Conntrack offload size set to %u",
                      ct_offload_size);
        } else {
            VLOG_INFO("Conntrack offloading is disabled");
        }
    }

    ct_offload_ipv6_enabled = smap_get_bool(other_config,
                                            "hw-offload-ct-ipv6-enabled",
                                            false);
    if (conntrack_offload_is_enabled()) {
        VLOG_INFO("Conntrack IPv6 offloading is %s",
                  conntrack_offload_ipv6_is_enabled() ?
                  "enabled" : "disabled");
    }

    if (conntrack_offload_is_enabled() && ovs_doca_initialized()) {
        handles = ovs_dpdk_mempool_create(ct_offload_size, sizeof(struct ct_offload_handle));
        ovs_assert(handles);
        fatal_signal_add_hook(destroy_handle_pool, NULL, NULL, true);
    }

    ovsthread_once_done(&once);
}

static bool
conntrack_offload_on_netdev(struct conn *conn, struct netdev *netdev)
{
    struct ct_offload_handle *coh;

    if (!netdev || !conn) {
        return false;
    }

    coh = conn->offloads.coh;
    return coh &&
        (coh->dir[CT_DIR_INIT].netdev == netdev ||
         coh->dir[CT_DIR_REP].netdev == netdev);
}

void
conntrack_offload_netdev_flush(struct conntrack *ct, struct netdev *netdev)
{
    struct conn *conn;

    CMAP_FOR_EACH (conn, cm_node, &ct->conns) {
        if (conntrack_offload_on_netdev(conn, netdev)) {
            conn_lock(conn);
            conntrack_offload_del_conn(ct, conn, true);
            conn_unlock(conn);
        }
    }
}

void
conntrack_offload_dump(struct ds *ds, struct conn *conn, int dir,
                       uint32_t ct_match_zone_id,
                       uint32_t ct_action_label_id,
                       uint32_t ct_miss_ctx_id)
{
    struct ct_flow_offload_item items[CT_DIR_NUM];
    struct ct_flow_offload_item *ct_offload;
    struct conn_key *nat_key;
    struct ct_match *c;

    conntrack_offload_build_add_items(NULL, conn, items, 0);
    ct_offload = &items[dir];
    c = &ct_offload->ct_match;
    nat_key = &ct_offload->nat.key;

    ds_put_format(ds, "zone %u proto %u ",
                  c->key.zone, c->key.nw_proto);
    if (c->key.dl_type == htons(ETH_TYPE_IP)) {
        ds_put_format(ds, IP_FMT":%u->"IP_FMT":%u",
                      IP_ARGS(c->key.src.addr.ipv4), ntohs(c->key.src.port),
                      IP_ARGS(c->key.dst.addr.ipv4), ntohs(c->key.dst.port));
    } else {
        char ip6_s[INET6_ADDRSTRLEN], ip6_d[INET6_ADDRSTRLEN];

        inet_ntop(AF_INET6, &c->key.src.addr.ipv6, ip6_s, sizeof ip6_s);
        inet_ntop(AF_INET6, &c->key.dst.addr.ipv6, ip6_d, sizeof ip6_d);
        ds_put_format(ds, "%s:%u->%s:%u",
                      ip6_s, ntohs(c->key.src.port),
                      ip6_d, ntohs(c->key.dst.port));
    }

    if (ct_offload->nat.mod_flags) {
        if (c->key.dl_type == htons(ETH_TYPE_IP)) {
            ds_put_format(ds, " nat "IP_FMT":%u->"IP_FMT":%u",
                          IP_ARGS(nat_key->src.addr.ipv4), ntohs(nat_key->src.port),
                          IP_ARGS(nat_key->dst.addr.ipv4), ntohs(nat_key->dst.port));
        } else {
            char ip6_s[INET6_ADDRSTRLEN], ip6_d[INET6_ADDRSTRLEN];

            inet_ntop(AF_INET6, &nat_key->src.addr.ipv6, ip6_s, sizeof ip6_s);
            inet_ntop(AF_INET6, &nat_key->dst.addr.ipv6, ip6_d, sizeof ip6_d);
            ds_put_format(ds, " nat %s:%u->%s:%u",
                          ip6_s, ntohs(nat_key->src.port),
                          ip6_d, ntohs(nat_key->dst.port));
        }
    }

    ds_put_format(ds, " zone_map=%d", ct_match_zone_id);

    /* CT MARK */
    ds_put_format(ds, " mark=0x%08x", ct_offload->mark_key);

    /* CT LABEL */
    ds_put_format(ds, " label=0x%08x", ct_action_label_id);

    /* CT STATE */
    ds_put_format(ds, " state=0x%02x", ct_offload->ct_state);

    /* CT CTX */
    ds_put_format(ds, " ctx=0x%02x", ct_miss_ctx_id);
}

void
conntrack_offload_get_ports(struct conn *conn, odp_port_t ports[CT_DIR_NUM])
{
    ovs_assert(!conn->master_conn);
    for (int dir = 0; dir < CT_DIR_NUM; dir++) {
        struct conn *conn_dir = conn;

        if (conn->nat_conn &&
            conn->nat_conn->offloads.dir_info[dir].dp) {
            conn_dir = conn->nat_conn;
        }

        ports[dir] = conn_dir->offloads.dir_info[dir].port;
    }
}
