/*
 * Copyright (c) 2023-2024 NVIDIA CORPORATION & AFFILIATES. All rights reserved.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at:
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

#include <config.h>
#include "dpdk.h"

#include <stdio.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <rte_devargs.h>
#include <rte_flow.h>
#include <rte_mempool.h>
#include <rte_pci.h>
#include <rte_version.h>

#include "openvswitch/vlog.h"
#include "smap.h"

VLOG_DEFINE_THIS_MODULE(dpdk_ext);

const char *
dpdk_get_version(void)
{
    return rte_version();
}

bool
dpdk_parse_devargs_pci(const char *devargs, struct rte_pci_addr *rte_pci)
{
    struct rte_devargs da;
    bool rv = true;

    if (rte_devargs_parse(&da, devargs)) {
        return false; /* Error has already been printed by the RTE function. */
    }

    if (rte_pci_addr_parse(da.name, rte_pci)) {
        rv = false;
        goto out;
    }

out:
    rte_devargs_reset(&da);
    return rv;
}

static void
dump_mempool_stats(struct rte_mempool *mp, void *arg)
{
    FILE *stream = arg;

    rte_mempool_dump(stream, mp);
}

void
malloc_dump_mempool_stats_wrapper(FILE *stream)
{
    rte_mempool_walk(dump_mempool_stats, stream);
}

void
malloc_dump_memzone_stats_wrapper(FILE *stream)
{
    rte_memzone_dump(stream);
}

void
dpdk_init_max_memzones(const struct smap *ovs_other_config)
{
    uint32_t max_memzones;

    max_memzones = smap_get_uint(ovs_other_config, "dpdk-max-memzones",
                                 UINT32_MAX);
    if (max_memzones != UINT32_MAX) {
        int rv;

        rv = rte_memzone_max_set(max_memzones);
        if (rv) {
            VLOG_WARN("Failed to set max memzones to %"PRIu32, max_memzones);
        } else {
            VLOG_INFO("Setting max memzones to %"PRIu32, max_memzones);
        }
    }
}

#define DMP_CACHE_SIZE 32

void
ovs_dpdk_mempool_destroy(struct ovs_dpdk_mempool *odmp)
{
    rte_mempool_free((struct rte_mempool *) odmp);
}

struct ovs_dpdk_mempool *
ovs_dpdk_mempool_create(unsigned n, unsigned elt_size)
{
    char mp_name[RTE_MEMPOOL_NAMESIZE];
    static int mp_count = 0;
    struct rte_mempool *mp;

    snprintf(mp_name, sizeof mp_name, "ovs_dpdk_mempool%d", mp_count++);
    /* The worst case is that all the caches (per l-core) are full except the
     * current one. Increase the number of elements by this number to handle
     * this case.
     */
    n += (RTE_MAX_LCORE - 1) * DMP_CACHE_SIZE;
    mp = rte_mempool_create(mp_name, n, elt_size, DMP_CACHE_SIZE, 0, NULL,
                            NULL, NULL, NULL, SOCKET_ID_ANY, 0);

    return (struct ovs_dpdk_mempool *) mp;
}

void
ovs_dpdk_mempool_free(struct ovs_dpdk_mempool *odmp, void *obj)
{
    rte_mempool_put((struct rte_mempool *) odmp, obj);
}

int
ovs_dpdk_mempool_alloc(struct ovs_dpdk_mempool *odmp, void **obj_p)
{
    return rte_mempool_get((struct rte_mempool *) odmp, obj_p);
}
