/*
 * Copyright (c) 2023-2024 NVIDIA CORPORATION & AFFILIATES. All rights reserved.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at:
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

#include <config.h>
#include "dpif-netdev.h"
#include "dpif-netdev-private.h"
#include "dpif-netdev-private-dfc.h"

#include <ctype.h>
#include <errno.h>
#include <fcntl.h>
#include <inttypes.h>
#include <net/if.h>
#include <sys/types.h>
#include <netinet/in.h>
#include <stdint.h>
#include <stdlib.h>
#include <string.h>
#include <sys/ioctl.h>
#include <sys/socket.h>
#include <sys/stat.h>
#include <unistd.h>

#include "batch.h"
#include "bitmap.h"
#include "ccmap.h"
#include "cmap.h"
#include "conntrack.h"
#include "conntrack-offload.h"
#include "conntrack-tp.h"
#include "coverage.h"
#include "ct-dpif.h"
#include "csum.h"
#include "dp-packet.h"
#include "dpif.h"
#include "dpif-netdev-lookup.h"
#include "dpif-netdev-perf.h"
#include "dpif-netdev-private-extract.h"
#include "dpif-provider.h"
#include "dummy.h"
#include "fat-rwlock.h"
#include "flow.h"
#include "histogram.h"
#include "hmapx.h"
#include "id-pool.h"
#include "id-fpool.h"
#include "ipf.h"
#include "metrics.h"
#include "mov-avg.h"
#include "mpsc-queue.h"
#include "netdev.h"
#include "netdev-offload.h"
#include "netdev-offload-dpdk-ext.h"
#include "netdev-offload-ext.h"
#include "netdev-provider.h"
#include "netdev-vport.h"
#include "netdev-dpdk.h"
#include "netlink.h"
#include "odp-execute.h"
#include "odp-util.h"
#include "openvswitch/dynamic-string.h"
#include "openvswitch/list.h"
#include "openvswitch/match.h"
#include "openvswitch/ofp-parse.h"
#include "openvswitch/ofp-print.h"
#include "openvswitch/ofpbuf.h"
#include "openvswitch/shash.h"
#include "openvswitch/vlog.h"
#include "ovs-doca.h"
#include "ovs-numa.h"
#include "ovs-rcu.h"
#include "packets.h"
#include "openvswitch/poll-loop.h"
#include "pvector.h"
#include "random.h"
#include "seq.h"
#include "smap.h"
#include "sset.h"
#include "timeval.h"
#include "tnl-neigh-cache.h"
#include "tnl-ports.h"
#include "unixctl.h"
#include "util.h"
#include "uuid.h"

VLOG_DEFINE_THIS_MODULE(dpif_netdev_ext);

extern struct odp_support dp_netdev_support;

static struct vlog_rate_limit upcall_rl = VLOG_RATE_LIMIT_INIT(600, 600);

static atomic_bool dump_packets_enabled = false;
static unsigned int n_pmd_threads;

bool dp_netdev_e2e_cache_enabled = false;
static uint32_t dp_netdev_e2e_cache_size = 0;
#define E2E_CACHE_MAX_TRACE_Q_SIZE   (10000u)
static uint32_t dp_netdev_e2e_cache_trace_q_size = E2E_CACHE_MAX_TRACE_Q_SIZE;
#define INVALID_OFFLOAD_THREAD_NB (MAX_OFFLOAD_THREAD_NB + 1)

static struct ovs_mutex flows_map_mutex = OVS_MUTEX_INITIALIZER;
static struct hmap flows_map OVS_GUARDED_BY(flows_map_mutex) =
    HMAP_INITIALIZER(&flows_map);
static atomic_count flows_map_count = ATOMIC_COUNT_INIT(0);
static struct ovs_mutex merged_flows_map_mutex = OVS_MUTEX_INITIALIZER;
static struct hmap merged_flows_map OVS_GUARDED_BY(merged_flows_map_mutex) =
    HMAP_INITIALIZER(&merged_flows_map);
static atomic_count merged_flows_map_count = ATOMIC_COUNT_INIT(0);

COVERAGE_DEFINE(flow_offload_200ms_latency);
COVERAGE_DEFINE(ct_offload_30us_latency);
COVERAGE_DEFINE(ct_offload_50us_latency);
COVERAGE_DEFINE(ct_offload_100us_latency);

#define DEFAULT_MAX_RECIRC_DEPTH 8
unsigned int max_recirc_depth = DEFAULT_MAX_RECIRC_DEPTH;

struct ovs_numa_dump *
dp_netdev_pmd_cmask2cores(const char *pmd_cmask)
{
    struct ovs_numa_dump *pmd_cores;

    if (pmd_cmask && pmd_cmask[0]) {
        pmd_cores = ovs_numa_dump_cores_with_cmask(pmd_cmask);
    } else {
        pmd_cores = ovs_numa_dump_n_cores_per_numa(NR_PMD_THREADS);
    }

    return pmd_cores;
}

unsigned int
dpif_netdev_get_n_pmd_threads(void)
{
    return n_pmd_threads;
}

void
dpif_netdev_set_n_pmd_threads(const char *pmd_cmask)
{
    struct ovs_numa_dump *pmd_cores;

    pmd_cores = dp_netdev_pmd_cmask2cores(pmd_cmask);
    n_pmd_threads = ovs_numa_dump_count(pmd_cores);
    ovs_numa_dump_destroy(pmd_cores);
}

void
dp_netdev_dump_packets_toggle(struct unixctl_conn *conn, int argc,
                              const char *argv[], void *aux OVS_UNUSED)
{
    bool flag = false;

    if (argc == 1) {
        flag = true;
    } else {
        if (!strcmp(argv[1], "on")) {
            flag = true;
        } else if (!strcmp(argv[1], "off")) {
            flag = false;
        } else {
            unixctl_command_reply_error(conn, "Invalid parameters");
            return;
        }
    }

    atomic_store_relaxed(&dump_packets_enabled, flag);
    unixctl_command_reply(conn, flag ? "ON" : "OFF");
}

void
dp_netdev_read_dump_packets_enabled(bool *flag)
{
    atomic_read_relaxed(&dump_packets_enabled, flag);
}

struct dp_offload_thread *
dp_netdev_offload_thread_next(struct dp_offload_thread *start,
                              unsigned int *tid, bool include_main)
{
    size_t array_size = ARRAY_SIZE(dp_offload_threads);
    struct dp_offload_thread *next;
    bool active;

    if (start == NULL) {
        unsigned int first = NETDEV_OFFLOAD_THREAD_MAIN;

        if (!include_main) {
            first += 1;
        }
        next = &dp_offload_threads[first];
    } else {
        next = start + 1;
    }

    while (next != &dp_offload_threads[array_size]) {
        atomic_read(&next->active, &active);
        if (active) {
            if (tid) {
                *tid = (unsigned int) (ptrdiff_t) (next - dp_offload_threads);
            }
            return next;
        }
        next++;
    }

    return NULL;
}

void
dp_netdev_offload_thread_enqueue(struct dp_offload_thread *thread,
                                 struct dp_offload_thread_item *offload)
{
    dp_netdev_offload_init();

    mpsc_queue_insert(&thread->offload_queue, &offload->node);
    atomic_count_inc64(&thread->enqueued_offload);
}

#define CT_ADD_DEFAULT_QUEUE_SIZE 200000
static unsigned int offload_ct_add_queue_size = CT_ADD_DEFAULT_QUEUE_SIZE;

void
packet_enqueue_to_flow_map(struct dp_packet *packet,
                           struct dp_netdev_flow *flow,
                           uint16_t tcp_flags,
                           struct dp_packet_flow_map *flow_map,
                           size_t index)
{
    struct dp_packet_flow_map *map = &flow_map[index];
    map->flow = flow;
    map->packet = packet;
    map->tcp_flags = tcp_flags;
}

static struct hmap counter_map = HMAP_INITIALIZER(&counter_map);

static inline int
e2e_cache_counter_cmp_key(const struct e2e_cache_counter_item *item,
                          const struct flows_counter_key *key)
{
    if (item->is_ct) {
        /* In case of CT compare only first 128 bits where 'ptr_key'
         * resides. It's not enough to compare only 'ptr_key' - second
         * argument can be not CT but first 64 bits of its key can be
         * equal to 'ptr_key' value of the first argument. In case of CT
         * next 64 bits after 'ptr_key' must always be 0, which cannot
         * happen in case of UFID.
         */
        const ovs_u128 *key0 = &item->key.ufid_key[0];
        const ovs_u128 *key1 = &key->ufid_key[0];

        return ovs_u128_equals(*key0, *key1) ? 0 : 1;
    }
    return memcmp(&item->key, key, sizeof *key);
}

static struct e2e_cache_counter_item *
e2e_cache_counter_find(size_t hash, const struct flows_counter_key *key)
{
    struct e2e_cache_counter_item *data;

    HMAP_FOR_EACH_WITH_HASH (data, node, hash, &counter_map) {
        if (data->hash == hash && !e2e_cache_counter_cmp_key(data, key)) {
            return data;
        }
    }
    return NULL;
}

static struct e2e_cache_counter_item *
e2e_cache_counter_alloc(const struct flows_counter_key *key, size_t hash,
                        bool is_ct)
{
    struct e2e_cache_counter_item *item;

    item = (struct e2e_cache_counter_item *) xmalloc(sizeof *item);
    item->hash = hash;
    item->is_ct = is_ct;
    ovs_list_init(&item->merged_flows);
    memcpy(&item->key, key, sizeof *key);
    return item;
}

#define merged_match_to_match_field(dst, src, field) \
    memcpy(&dst->flow.field, &src->spec.field, sizeof dst->flow.field); \
    memcpy(&dst->wc.masks.field, &src->mask.field, sizeof dst->wc.masks.field);

static void
merged_match_to_match(struct match *match,
                      struct merged_match *merged_match)
{
    memset(match, 0, sizeof *match);

    merged_match_to_match_field(match, merged_match, in_port);
    merged_match_to_match_field(match, merged_match, tunnel.ip_dst);
    merged_match_to_match_field(match, merged_match, tunnel.ipv6_dst);
    merged_match_to_match_field(match, merged_match, tunnel.ip_src);
    merged_match_to_match_field(match, merged_match, tunnel.ipv6_src);
    merged_match_to_match_field(match, merged_match, tunnel.tun_id);
    merged_match_to_match_field(match, merged_match, tunnel.tp_dst);

    merged_match_to_match_field(match, merged_match, dl_dst);
    merged_match_to_match_field(match, merged_match, dl_src);
    merged_match_to_match_field(match, merged_match, dl_type);

    merged_match_to_match_field(match, merged_match, vlans[0].tci);

    merged_match_to_match_field(match, merged_match, nw_src);
    merged_match_to_match_field(match, merged_match, nw_dst);
    merged_match_to_match_field(match, merged_match, ipv6_src);
    merged_match_to_match_field(match, merged_match, ipv6_dst);
    merged_match_to_match_field(match, merged_match, nw_frag);
    merged_match_to_match_field(match, merged_match, nw_proto);

    merged_match_to_match_field(match, merged_match, tp_src);
    merged_match_to_match_field(match, merged_match, tp_dst);

    merged_match_to_match_field(match, merged_match, ct_zone);
}

void
dpif_netdev_dump_e2e_flows(struct hmap *portno_names,
                           struct ofputil_port_map *port_map, struct ds *s)
{
    struct e2e_cache_merged_flow *merged_flow;
    struct dp_netdev_flow netdev_flow;
    struct match match;

    memset(&netdev_flow, 0, sizeof netdev_flow);

    ovs_mutex_lock(&merged_flows_map_mutex);

    HMAP_FOR_EACH (merged_flow, node.in_hmap, &merged_flows_map) {
        merged_match_to_match(&match, &merged_flow->merged_match);
        odp_format_ufid(&merged_flow->ufid, s);
        ds_put_cstr(s, ", ");
        match_format(&match, port_map, s, OFP_DEFAULT_PRIORITY);
        *CONST_CAST(ovs_u128 *, &netdev_flow.mega_ufid) = merged_flow->ufid;
        CONST_CAST(struct flow *, &netdev_flow.flow)->in_port =
            match.flow.in_port;
        ds_put_cstr(s, ", actions:");
        format_odp_actions(s, merged_flow->actions, merged_flow->actions_size,
                           portno_names);
        ds_put_cstr(s, "\n");
    }

    ovs_mutex_unlock(&merged_flows_map_mutex);
}

void
e2e_cache_trace_add_flow(struct dp_packet *p,
                         const ovs_u128 *ufid)
{
    uint32_t e2e_trace_size = p->e2e_trace_size;

    if (OVS_UNLIKELY(e2e_trace_size >= E2E_CACHE_MAX_TRACE)) {
        p->e2e_trace_flags |= E2E_CACHE_TRACE_FLAG_OVERFLOW;
        return;
    }
    p->e2e_trace[e2e_trace_size] = *ufid;
    p->e2e_trace_size = e2e_trace_size + 1;
}

static inline void
e2e_cache_trace_msg_enqueue(struct e2e_cache_trace_message *msg,
                            unsigned int tid)
{
    struct e2e_cache_stats *e2e_stats = &dp_offload_threads[tid].e2e_stats;

    mpsc_queue_insert(&dp_offload_threads[tid].trace_queue, &msg->node);
    atomic_count_inc(&e2e_stats->queue_trcs);
}

/* Associate the merged flow to each of its composing flows,
 * to allow accessing:
 * - From the merged flow to all its composing flows.
 * - From each flow to all the merged flow it is part of.
 */
static void
e2e_cache_associate_merged_flow(struct e2e_cache_merged_flow *merged_flow,
                                struct e2e_cache_ovs_flow *flows[],
                                uint16_t num_flows)
{
    uint16_t i, j;

    ovs_mutex_lock(&flows_map_mutex);

    for (j = 0, i = 0; j < num_flows; j++) {
        if (flows[j]->offload_state != E2E_OL_STATE_FLOW && j > 0 &&
            flows[j - 1]->offload_state != E2E_OL_STATE_FLOW) {
            continue;
        }
        merged_flow->associated_flows[i].index = i;
        ovs_list_push_back(&flows[j]->associated_merged_flows,
                           &merged_flow->associated_flows[i].list);
        merged_flow->associated_flows[i].mt_flow = flows[j];
        i++;
    }
    merged_flow->associated_flows_len = i;

    ovs_mutex_unlock(&flows_map_mutex);
}

static void
e2e_cache_disassociate_merged_flow(struct e2e_cache_merged_flow *merged_flow)
{
    uint16_t i, num_flows = merged_flow->associated_flows_len;

    for (i = 0; i < num_flows; i++) {
        ovs_list_remove(&merged_flow->associated_flows[i].list);
    }
}

/* Find e2e_cache_merged_flow with @ufid.
 * merged_flows_map_mutex mutex must be locked.
 */
static inline struct e2e_cache_merged_flow *
e2e_cache_merged_flow_find(const ovs_u128 *ufid, uint32_t hash)
{
    struct e2e_cache_merged_flow *merged_flow;

    HMAP_FOR_EACH_WITH_HASH (merged_flow, node.in_hmap, hash,
                             &merged_flows_map) {
        if (ovs_u128_equals(*ufid, merged_flow->ufid)) {
            return merged_flow;
        }
    }

    return NULL;
}

static inline struct e2e_cache_ovs_flow *
e2e_cache_flow_alloc(bool is_ct)
{
    struct e2e_cache_ovs_flow *flow;
    size_t alloc_bytes;

    alloc_bytes = sizeof *flow;
    alloc_bytes += is_ct ? sizeof flow->ct_match[0] : sizeof flow->match[0];

    flow = (struct e2e_cache_ovs_flow *) xzalloc(alloc_bytes);
    return flow;
}

static void
e2e_cache_flow_free(void *arg)
{
    struct e2e_cache_ovs_flow *flow = (struct e2e_cache_ovs_flow *) arg;

    if (flow->actions) {
        free(flow->actions);
    }
    free(flow);
}

static inline struct e2e_cache_ufid_msg *
e2e_cache_ufid_msg_alloc(int op, bool is_ct, size_t actions_len,
                         long long int now)
{
    struct e2e_cache_ufid_msg *msg;
    struct nlattr *actions = NULL;
    size_t alloc_size;

    alloc_size = sizeof *msg;
    if (op == E2E_UFID_MSG_PUT) {
        if (actions_len) {
            actions = (struct nlattr *) xmalloc(actions_len);
            if (OVS_UNLIKELY(!actions)) {
                return NULL;
            }
        }
        alloc_size += is_ct ? sizeof msg->ct_match[0] :
                              sizeof msg->match[0];
    }

    msg = (struct e2e_cache_ufid_msg *) xmalloc(alloc_size);
    if (OVS_UNLIKELY(!msg)) {
        goto err;
    }

    msg->op = op;
    msg->is_ct = is_ct;
    msg->actions = actions;
    msg->actions_len = actions_len;
    msg->timestamp = now;
    return msg;

err:
    if (actions) {
        free(actions);
    }
    return NULL;
}

static inline void
e2e_cache_ufid_msg_free(struct e2e_cache_ufid_msg *msg)
{
    if (msg->actions) {
        free(msg->actions);
    }
    free(msg);
}

static void
e2e_cache_disassociate_counters(struct e2e_cache_merged_flow *merged_flow);

static void
e2e_cache_merged_flow_free(struct e2e_cache_merged_flow *merged_flow)
{
    e2e_cache_disassociate_counters(merged_flow);
    if (merged_flow->actions) {
        free(merged_flow->actions);
    }
    free(merged_flow);
}

static void
e2e_cache_merged_flow_db_rem(struct e2e_cache_merged_flow *merged_flow)
{
    e2e_cache_disassociate_merged_flow(merged_flow);

    ovs_mutex_lock(&merged_flows_map_mutex);
    hmap_remove(&merged_flows_map, &merged_flow->node.in_hmap);
    ovs_mutex_unlock(&merged_flows_map_mutex);
    atomic_count_dec(&merged_flows_map_count);
}

static void
e2e_cache_merged_flow_db_del(struct e2e_cache_merged_flow *merged_flow)
{
    /* Lock/unlock to prevent race condition with
     * e2e_cache_get_merged_flows_stats()
     */
    ovs_mutex_lock(&flows_map_mutex);
    e2e_cache_merged_flow_db_rem(merged_flow);
    ovs_mutex_unlock(&flows_map_mutex);

    e2e_cache_merged_flow_free(merged_flow);
}

static int
e2e_cache_merged_flow_offload_del(struct e2e_cache_merged_flow *merged_flow);

static inline int
e2e_cache_merged_flow_db_put(struct e2e_cache_merged_flow *merged_flow)
{
    uint32_t hash =
        hash_bytes(&merged_flow->ufid, sizeof merged_flow->ufid, 0);
    struct e2e_cache_merged_flow *old_merged_flow;

    ovs_mutex_lock(&merged_flows_map_mutex);

    old_merged_flow = e2e_cache_merged_flow_find(&merged_flow->ufid, hash);
    /* In case the merged flow exists do nothing. */
    if (old_merged_flow) {
        uint16_t actions_size = merged_flow->actions_size;

        if (old_merged_flow->actions_size == actions_size &&
            !memcmp(old_merged_flow->actions, merged_flow->actions,
                    actions_size)) {
            ovs_mutex_unlock(&merged_flows_map_mutex);
            return -1;
        }

        /* Must unlock merged_flows_map_mutex before calling next functions */
        ovs_mutex_unlock(&merged_flows_map_mutex);

        /* In case it's a flow modification delete the current flow
         * before inserting the updated one.
         */
        e2e_cache_merged_flow_offload_del(old_merged_flow);
        e2e_cache_merged_flow_db_del(old_merged_flow);

        ovs_mutex_lock(&merged_flows_map_mutex);
    }

    hmap_insert(&merged_flows_map, &merged_flow->node.in_hmap, hash);

    ovs_mutex_unlock(&merged_flows_map_mutex);
    atomic_count_inc(&merged_flows_map_count);
    return 0;
}

/* Find e2e_cache_ovs_flow with @ufid and calculated @hash */
static inline struct e2e_cache_ovs_flow *
e2e_cache_flow_find(const ovs_u128 *ufid, uint32_t hash)
    OVS_REQUIRES(flows_map_mutex)
{
    struct e2e_cache_ovs_flow *flow;

    HMAP_FOR_EACH_WITH_HASH (flow, node, hash, &flows_map) {
        if (ovs_u128_equals(*ufid, flow->ufid)) {
            return flow;
        }
    }

    return NULL;
}

static void
e2e_cache_update_ct_stats(struct e2e_cache_ovs_flow *mt_flow, int op)
{
    unsigned int tid = netdev_offload_thread_id();
    struct dp_offload_thread *ofl_thread;
    struct e2e_cache_ovs_flow *ct_peer;

    ofl_thread = &dp_offload_threads[tid];

    ct_peer = mt_flow->ct_peer;
    if (op == DP_NETDEV_FLOW_OFFLOAD_OP_ADD) {
        if (ct_peer &&
            (ct_peer->offload_state == E2E_OL_STATE_CT_HW ||
             ct_peer->offload_state == E2E_OL_STATE_CT_MT ||
             ct_peer->offload_state == E2E_OL_STATE_CT2CT)) {
            atomic_count_inc64(&ofl_thread->ct_bi_dir_connections);
            atomic_count_dec64(&ofl_thread->ct_uni_dir_connections);
        } else {
            atomic_count_inc64(&ofl_thread->ct_uni_dir_connections);
        }
    } else if (op == DP_NETDEV_FLOW_OFFLOAD_OP_DEL) {
        if (ct_peer &&
            (ct_peer->offload_state == E2E_OL_STATE_CT_HW ||
             ct_peer->offload_state == E2E_OL_STATE_CT_MT ||
             ct_peer->offload_state == E2E_OL_STATE_CT2CT)) {
            atomic_count_dec64(&ofl_thread->ct_bi_dir_connections);
            atomic_count_inc64(&ofl_thread->ct_uni_dir_connections);
        } else {
            atomic_count_dec64(&ofl_thread->ct_uni_dir_connections);
        }
    } else {
        OVS_NOT_REACHED();
    }
}

static void
e2e_cache_del_associated_merged_flows(struct e2e_cache_ovs_flow *flow,
                                      struct ovs_list *merged_flows_to_delete)
{
    struct flow2flow_item *associated_flow_item, *next_item;
    unsigned int tid = netdev_offload_thread_id();
    struct e2e_cache_merged_flow *merged_flow;

    LIST_FOR_EACH_SAFE (associated_flow_item, next_item, list,
                        &flow->associated_merged_flows) {
        merged_flow =
            CONTAINER_OF(associated_flow_item,
                         struct e2e_cache_merged_flow,
                         associated_flows[associated_flow_item->index]);
        if (merged_flow->tid != tid) {
            continue;
        }

        e2e_cache_merged_flow_db_rem(merged_flow);
        ovs_list_push_back(merged_flows_to_delete, &merged_flow->node.in_list);
    }
}

static void
e2e_cache_del_merged_flows(struct ovs_list *merged_flows_to_delete)
{
    struct e2e_cache_merged_flow *merged_flow;
    struct ovs_list *l;

    while (!ovs_list_is_empty(merged_flows_to_delete)) {
        l = ovs_list_pop_front(merged_flows_to_delete);

        merged_flow =
            CONTAINER_OF(l, struct e2e_cache_merged_flow, node.in_list);

        e2e_cache_merged_flow_offload_del(merged_flow);
        e2e_cache_merged_flow_free(merged_flow);
    }
}

static struct e2e_cache_ovs_flow *
e2e_cache_flow_db_del_protected(const ovs_u128 *ufid, uint32_t hash,
                                struct ovs_list *merged_flows_to_delete,
                                struct ovs_refcount *del_refcnt)
    OVS_REQUIRES(flows_map_mutex)
{
    struct e2e_cache_ovs_flow *flow;

    flow = e2e_cache_flow_find(ufid, hash);
    if (OVS_UNLIKELY(!flow)) {
        return NULL;
    }
    e2e_cache_del_associated_merged_flows(flow, merged_flows_to_delete);
    if (flow->offload_state == E2E_OL_STATE_FLOW) {
        if (del_refcnt && ovs_refcount_unref(del_refcnt) > 1) {
            return NULL;
        }
        hmap_remove(&flows_map, &flow->node);
        atomic_count_dec(&flows_map_count);
        ovsrcu_postpone(e2e_cache_flow_free, flow);
        flow = NULL;
        ovsrcu_postpone(free, del_refcnt);
    }
    return flow;
}

static inline void
e2e_cache_populate_offload_item(struct dp_offload_thread_item *offload_item,
                                int op,
                                struct dp_netdev *dp,
                                struct dp_netdev_flow *flow,
                                long long now);

static int
e2e_cache_ct_flow_offload_del_mt(struct dp_netdev *dp,
                                 struct e2e_cache_ovs_flow *ct_flow)
{
    unsigned int tid = netdev_offload_thread_id();
    struct dp_offload_thread_item *offload_item;
    struct e2e_cache_stats *e2e_stats;
    struct dp_netdev_flow flow;
    long long now = time_usec();
    int ret;

    e2e_stats = &dp_offload_threads[tid].e2e_stats;
    memset(&flow, 0, sizeof flow);
    *CONST_CAST(ovs_u128 *, &flow.mega_ufid) = ct_flow->ufid;
    CONST_CAST(struct flow *, &flow.flow)->in_port.odp_port =
        ct_flow->ct_match[0].odp_port;

    offload_item = dp_netdev_alloc_flow_offload(dp, &flow,
                                                DP_NETDEV_FLOW_OFFLOAD_OP_DEL,
                                                now);
    e2e_cache_populate_offload_item(offload_item,
                                    DP_NETDEV_FLOW_OFFLOAD_OP_DEL, dp, &flow,
                                    now);

    ret = dp_netdev_flow_offload_del(offload_item);
    free(offload_item);
    if (!ret) {
        e2e_stats->del_ct_mt_flow_hw++;
    } else {
        e2e_stats->del_ct_mt_flow_err++;
    }
    return ret;
}

static void
e2e_cache_flow_state_set_at(struct e2e_cache_ovs_flow *flow,
                            enum e2e_offload_state next_state,
                            const char *where)
{
    static struct vlog_rate_limit rl = VLOG_RATE_LIMIT_INIT(100, 100);
    static const int op[E2E_OL_STATE_NUM][E2E_OL_STATE_NUM] = {
        [E2E_OL_STATE_CT_SW] = {
            [E2E_OL_STATE_CT_HW] = DP_NETDEV_FLOW_OFFLOAD_OP_ADD,
            [E2E_OL_STATE_CT2CT] = DP_NETDEV_FLOW_OFFLOAD_OP_ADD,
            [E2E_OL_STATE_CT_MT] = DP_NETDEV_FLOW_OFFLOAD_OP_ADD,
        },
        [E2E_OL_STATE_CT_HW] = {
            [E2E_OL_STATE_CT_SW] = DP_NETDEV_FLOW_OFFLOAD_OP_DEL,
            [E2E_OL_STATE_CT_ERR] = DP_NETDEV_FLOW_OFFLOAD_OP_DEL,
        },
        [E2E_OL_STATE_CT_MT] = {
            [E2E_OL_STATE_CT_SW] = DP_NETDEV_FLOW_OFFLOAD_OP_DEL,
            [E2E_OL_STATE_CT2CT] = DP_NETDEV_FLOW_OFFLOAD_OP_DEL,
            [E2E_OL_STATE_CT_ERR] = DP_NETDEV_FLOW_OFFLOAD_OP_DEL,
        },
        [E2E_OL_STATE_CT2CT] = {
            [E2E_OL_STATE_CT_SW] = DP_NETDEV_FLOW_OFFLOAD_OP_DEL,
            [E2E_OL_STATE_CT_MT] = DP_NETDEV_FLOW_OFFLOAD_OP_DEL,
            [E2E_OL_STATE_CT_ERR] = DP_NETDEV_FLOW_OFFLOAD_OP_DEL,
        },
        [E2E_OL_STATE_CT_ERR] = {
            [E2E_OL_STATE_CT_HW] = DP_NETDEV_FLOW_OFFLOAD_OP_ADD,
            [E2E_OL_STATE_CT_MT] = DP_NETDEV_FLOW_OFFLOAD_OP_ADD,
            [E2E_OL_STATE_CT2CT] = DP_NETDEV_FLOW_OFFLOAD_OP_ADD,
        },
    };
    enum e2e_offload_state prev_state = flow->offload_state;
    int flow_op;

    if (prev_state < E2E_OL_STATE_FLOW || prev_state >= E2E_OL_STATE_NUM) {
        /* If flow state was not yet initialized, assume a start state that
         * is assured to be a no-op regarding CT stats.
         */
        prev_state = E2E_OL_STATE_FLOW;
    }

    ovs_assert(next_state >= E2E_OL_STATE_FLOW &&
               next_state < E2E_OL_STATE_NUM);

    VLOG_DBG_RL(&rl, "%s: e2e-flow " UUID_FMT " state is %s", where,
                UUID_ARGS((struct uuid *) &flow->ufid),
                e2e_offload_state_names[next_state]);
    flow->offload_state = next_state;

    flow_op = op[prev_state][next_state];
    if (flow_op == DP_NETDEV_FLOW_OFFLOAD_OP_NONE) {
        return;
    }

    e2e_cache_update_ct_stats(flow, flow_op);
}
#define e2e_cache_flow_state_set(f, s) \
    e2e_cache_flow_state_set_at(f, s, __func__)

static int
dp_netdev_ct_offload_add_cb(struct ct_flow_offload_item *ct_offload,
                            struct ct_match *ct_match, struct nlattr *actions,
                            int actions_len);

static struct nlattr *
e2e_cache_ct_flow_offload_add_mt(struct dp_netdev *dp,
                                 struct e2e_cache_ovs_flow *ct_flow,
                                 struct nlattr *actions,
                                 uint16_t *actions_size)
{
    unsigned int tid = netdev_offload_thread_id();
    uint16_t max_actions_len = *actions_size;
    struct ct_flow_offload_item offload;
    struct e2e_cache_stats *e2e_stats;
    int ret;

    e2e_stats = &dp_offload_threads[tid].e2e_stats;

    memset(&offload, 0, sizeof offload);
    offload.dp = dp;

    /* Only non-offloaded CTs. Either to MT or cache. */
    if (ct_flow->offload_state != E2E_OL_STATE_CT_SW ||
        !ovs_list_is_empty(&ct_flow->associated_merged_flows)) {
        return actions;
    }

    offload.ufid = ct_flow->ufid;
    if (!actions || max_actions_len < ct_flow->actions_size) {
        if (actions) {
            free(actions);
        }
        max_actions_len = ct_flow->actions_size;
        actions = xmalloc(max_actions_len);
    }
    memcpy(actions, ct_flow->actions, ct_flow->actions_size);
    ret = dp_netdev_ct_offload_add_cb(&offload, ct_flow->ct_match, actions,
                                      ct_flow->actions_size);

    if (OVS_LIKELY(ret == 0)) {
        e2e_cache_flow_state_set(ct_flow, E2E_OL_STATE_CT_MT);
        e2e_stats->add_ct_mt_flow_hw++;
    } else {
        e2e_cache_flow_state_set(ct_flow, E2E_OL_STATE_CT_ERR);
        e2e_stats->add_ct_mt_flow_err++;
    }

    *actions_size = max_actions_len;
    return actions;
}

static void
e2e_cache_flow_db_del(struct e2e_cache_ufid_msg *ufid_msg)
{
    struct ovs_list merged_flows_to_delete =
        OVS_LIST_INITIALIZER(&merged_flows_to_delete);
    size_t hash = hash_bytes(&ufid_msg->ufid, sizeof ufid_msg->ufid, 0);
    struct e2e_cache_ovs_flow *ct_flow, *iter_flow;
    struct e2e_cache_merged_flow *merged_flow;
    uint16_t i;

    ovs_mutex_lock(&flows_map_mutex);
    ct_flow = e2e_cache_flow_db_del_protected(&ufid_msg->ufid, hash,
                                              &merged_flows_to_delete,
                                              ufid_msg->del_refcnt);
    ovs_mutex_unlock(&flows_map_mutex);

    /* Update CT stats affected by deletion of the merged flows. */
    LIST_FOR_EACH (merged_flow, node.in_list, &merged_flows_to_delete) {
        for (i = 0; i < merged_flow->associated_flows_len; i++) {
            iter_flow = merged_flow->associated_flows[i].mt_flow;
            if (iter_flow->offload_state == E2E_OL_STATE_FLOW) {
                continue;
            }
            if (ovs_list_is_empty(&iter_flow->associated_merged_flows)) {
                e2e_cache_flow_state_set(iter_flow, E2E_OL_STATE_CT_SW);
            }
        }
    }
    if (ct_flow) {
        if (ovs_refcount_unref(ufid_msg->del_refcnt) == 1) {
            ovs_mutex_lock(&flows_map_mutex);
            hmap_remove(&flows_map, &ct_flow->node);
            ovs_mutex_unlock(&flows_map_mutex);
            atomic_count_dec(&flows_map_count);
            ovsrcu_postpone(e2e_cache_flow_free, ct_flow);
            if (ufid_msg->del_refcnt) {
                ovsrcu_postpone(free, ufid_msg->del_refcnt);
            }
            conntrack_conn_unref(ufid_msg->conn);
        }
        /* Only the thread that merged ct_flow should should change statistics
         * etc.
         */
        if (ct_flow->merge_tid == netdev_offload_thread_id()) {
            /* This is a CT MT flow that is deleted. If it is offloaded using
             * MT remove it and update CT stats.
             */
            if (ct_flow->offload_state == E2E_OL_STATE_CT_MT) {
                e2e_cache_ct_flow_offload_del_mt(ufid_msg->dp, ct_flow);
            }
            e2e_cache_flow_state_set(ct_flow, E2E_OL_STATE_CT_SW);
            if (ct_flow->ct_peer) {
                ct_flow->ct_peer->ct_peer = NULL;
            }
        }
    }
    e2e_cache_del_merged_flows(&merged_flows_to_delete);
}

static int
e2e_cache_flow_db_put(struct e2e_cache_ufid_msg *ufid_msg)
{
    struct e2e_cache_ovs_flow *flow_prev, *flow;
    struct ovs_list merged_flows_to_delete =
        OVS_LIST_INITIALIZER(&merged_flows_to_delete);
    const ovs_u128 *ufid;
    size_t hash;

    flow = e2e_cache_flow_alloc(ufid_msg->is_ct);
    if (OVS_UNLIKELY(!flow)) {
        return -1;
    }

    flow->ufid = ufid_msg->ufid;
    if (ufid_msg->is_ct) {
        flow->ct_match[0] = ufid_msg->ct_match[0];
        e2e_cache_flow_state_set(flow, E2E_OL_STATE_CT_SW);
    } else {
        flow->match[0] = ufid_msg->match[0];
        e2e_cache_flow_state_set(flow, E2E_OL_STATE_FLOW);
    }
    flow->actions = ufid_msg->actions;
    ufid_msg->actions = NULL;
    flow->actions_size = ufid_msg->actions_len;
    ovs_list_init(&flow->associated_merged_flows);
    flow->merge_tid = INVALID_OFFLOAD_THREAD_NB;
    hmap_init(&flow->merged_counters);

    ufid = &flow->ufid;
    hash = hash_bytes(ufid, sizeof *ufid, 0);

    ovs_mutex_lock(&flows_map_mutex);

    flow_prev = e2e_cache_flow_find(ufid, hash);
    if (flow_prev) {
        e2e_cache_flow_db_del_protected(ufid, hash, &merged_flows_to_delete,
                                        NULL);
    }

    hmap_insert(&flows_map, &flow->node, hash);

    ovs_mutex_unlock(&flows_map_mutex);
    atomic_count_inc(&flows_map_count);

    e2e_cache_del_merged_flows(&merged_flows_to_delete);
    return 0;
}

int
e2e_cache_flow_del(const ovs_u128 *ufid, struct dp_netdev *dp,
                   struct conn *conn, long long int now)
{
    static struct vlog_rate_limit rl = VLOG_RATE_LIMIT_INIT(10, 10);
    struct e2e_cache_ufid_msg *del_msg;
    struct e2e_cache_stats *e2e_stats;
    struct dp_offload_thread *thread;
    struct e2e_cache_ovs_flow *flow;
    struct ovs_refcount *del_refcnt;
    uint32_t hash;

    VLOG_DBG_RL(&rl, "%s: ufid="UUID_FMT, __FUNCTION__,
                UUID_ARGS((struct uuid *)ufid));

    ovs_mutex_lock(&flows_map_mutex);
    hash = hash_bytes(ufid, sizeof *ufid, 0);
    flow = e2e_cache_flow_find(ufid, hash);
    ovs_mutex_unlock(&flows_map_mutex);
    if (!flow) {
      return -1;
    }
    del_refcnt = xmalloc(sizeof *del_refcnt);
    ovs_refcount_init(del_refcnt);
    DP_NETDEV_OFFLOAD_FOREACH_THREAD_NO_MAIN (thread) {
        ovs_refcount_ref(del_refcnt);
    }
    /* Compensate for the reference taken at initialization. */
    ovs_refcount_unref(del_refcnt);
    DP_NETDEV_OFFLOAD_FOREACH_THREAD_NO_MAIN (thread) {
        del_msg = e2e_cache_ufid_msg_alloc(E2E_UFID_MSG_DEL, false, 0, now);
        if (OVS_UNLIKELY(!del_msg)) {
            free(del_refcnt);
            return -1;
        }
        del_msg->ufid = *ufid;
        del_msg->dp = dp;
        del_msg->del_refcnt = del_refcnt;
        del_msg->conn = conn;

        /* Insert message into queue, e2e_cache_ufid_msg_dequeue()
         * is used to dequeue it from there.
         */
        mpsc_queue_insert(&thread->ufid_queue, &del_msg->node);
        e2e_stats = &thread->e2e_stats;
        atomic_count_inc(&e2e_stats->flow_del_msgs);
    }
    return 0;
}

int
e2e_cache_flow_put(bool is_ct, const ovs_u128 *ufid, const void *match,
                   const struct nlattr *actions, size_t actions_len,
                   long long int now)
{
    static struct vlog_rate_limit rl = VLOG_RATE_LIMIT_INIT(10, 10);
    struct e2e_cache_ufid_msg *put_msg;
    struct e2e_cache_stats *e2e_stats;
    unsigned int tid;

    VLOG_DBG_RL(&rl, "%s: ufid="UUID_FMT, __FUNCTION__,
                UUID_ARGS((struct uuid *)ufid));
    put_msg = e2e_cache_ufid_msg_alloc(E2E_UFID_MSG_PUT, is_ct, actions_len,
                                       now);
    if (OVS_UNLIKELY(!put_msg)) {
        return -1;
    }

    put_msg->ufid = *ufid;
    if (actions_len) {
        memcpy(put_msg->actions, actions, actions_len);
    }
    if (is_ct) {
        put_msg->ct_match[0] = *((const struct ct_match *) match);
    } else {
        put_msg->match[0] = *((const struct match *) match);
    }

    /* Insert message into queue, e2e_cache_ufid_msg_dequeue()
     * is used to dequeue it from there.
     */
    tid = netdev_offload_ufid_to_thread_id(*ufid);
    mpsc_queue_insert(&dp_offload_threads[tid].ufid_queue, &put_msg->node);
    e2e_stats = &dp_offload_threads[tid].e2e_stats;
    atomic_count_inc(&e2e_stats->flow_add_msgs);
    return 0;
}

static int
e2e_cache_ufids_to_flows(const ovs_u128 *ufids,
                         uint16_t num_elements,
                         struct e2e_cache_ovs_flow *flows[]);
static unsigned int
netdev_offload_trace_to_thread_id(ovs_u128 *ufids,
                                  uint16_t num_elements)
{
    struct e2e_cache_ovs_flow *mt_flows[E2E_CACHE_MAX_TRACE];
    uint32_t ufid_hash;
    unsigned int tid;
    uint16_t i;

    ovs_mutex_lock(&flows_map_mutex);
    e2e_cache_ufids_to_flows(ufids, num_elements, mt_flows);
    /* If a previous trace already determined the tid to handle, send it to
     * the same one.
     */
    tid = INVALID_OFFLOAD_THREAD_NB;
    for (i = 0; i < num_elements; i++) {
        if (!mt_flows[i]) {
            ovs_mutex_unlock(&flows_map_mutex);
            return INVALID_OFFLOAD_THREAD_NB;
        }
    }
    for (i = 0; i < num_elements; i++) {
        /* Skip megaflows. */
        if (i % 3 == 0) {
            continue;
        }
        if (mt_flows[i]->merge_tid != INVALID_OFFLOAD_THREAD_NB) {
            tid = mt_flows[i]->merge_tid;
            break;
        }
    }

    if (tid == INVALID_OFFLOAD_THREAD_NB) {
        ufid_hash = 1;
        for (i = 0; i < num_elements; i++) {
            /* Skip ct peers. */
            if (i % 3 == 2) {
                continue;
            }
            ufid_hash = hash_words64(
                (const uint64_t [2]){ ufids[i].u64.lo,
                                      ufids[i].u64.hi }, 2, ufid_hash);
        }
        /* Skip main thread at ID 0. */
        tid = 1 + ufid_hash % netdev_offload_thread_nb();
    }

    /* Set the selected tid to CTs. */
    for (i = 0; i < num_elements; i++) {
        /* Skip megaflows. */
        if (i % 3 == 0) {
            continue;
        }
        mt_flows[i]->merge_tid = tid;
    }
    ovs_mutex_unlock(&flows_map_mutex);
    return tid;
}

void
e2e_cache_dispatch_trace_message(struct dp_netdev *dp,
                                 struct dp_packet_batch *batch,
                                 long long int now)
{
    struct e2e_cache_trace_info *cur_trace_info[MAX_OFFLOAD_THREAD_NB];
    struct e2e_cache_trace_message *buffer[MAX_OFFLOAD_THREAD_NB];
    struct e2e_cache_stats *e2e_stats[MAX_OFFLOAD_THREAD_NB];
    unsigned int first_tid = NETDEV_OFFLOAD_THREAD_MAIN;
    uint32_t num_elements[MAX_OFFLOAD_THREAD_NB];
    uint32_t cur_q_size[MAX_OFFLOAD_THREAD_NB];
    struct dp_offload_thread *thread;
    struct dp_packet *packet;
    size_t buffer_size;
    unsigned int tid;

    buffer_size = sizeof(struct e2e_cache_trace_message) +
                         2 * batch->count * sizeof(struct e2e_cache_trace_info);

    DP_NETDEV_OFFLOAD_FOREACH_THREAD_NO_MAIN (thread, tid) {
        if (first_tid == NETDEV_OFFLOAD_THREAD_MAIN) {
            first_tid = tid;
        }
        buffer[tid] =
            (struct e2e_cache_trace_message *) xmalloc_cacheline(buffer_size);
        num_elements[tid] = 0;
        cur_trace_info[tid] = &buffer[tid]->data[0];
        cur_q_size[tid] =
            atomic_count_get(&dp_offload_threads[tid].e2e_stats.queue_trcs);
        e2e_stats[tid] = &dp_offload_threads[tid].e2e_stats;
    }

    DP_PACKET_BATCH_FOR_EACH (i, packet, batch) {
        uint32_t e2e_trace_size = packet->e2e_trace_size;
        ovs_u128 *e2e_trace = &packet->e2e_trace[0];

        /* Don't send untraced packets. */
        if (!e2e_trace_size) {
            continue;
        }

        /* Don't send aborted traces */
        if (OVS_UNLIKELY(packet->e2e_trace_flags &
                         E2E_CACHE_TRACE_FLAG_ABORT)) {
            atomic_count_inc(&e2e_stats[first_tid]->aborted_trcs);
            continue;
        }
        /* In case the packet had tnl_pop, we split the trace to the tnl_pop
         * ufid (the 1st one in the trace), and the rest of the trace,
         * representing the path of the packet with the virtual port. Once the
         * tnl_pop flow is offloaded, we will get only the virtual port path.
         */
        if (packet->e2e_trace_flags & E2E_CACHE_TRACE_FLAG_TNL_POP) {
            tid = netdev_offload_ufid_to_thread_id(e2e_trace[0]);
            if (tid == INVALID_OFFLOAD_THREAD_NB) {
                continue;
            }
            if (dp_netdev_e2e_cache_trace_q_size &&
                cur_q_size[tid] >= dp_netdev_e2e_cache_trace_q_size) {
                atomic_count_inc(&e2e_stats[tid]->overflow_trcs);
                continue;
            }

            cur_trace_info[tid]->num_elements = 1;
            cur_trace_info[tid]->e2e_trace_ct_ufids = 0;
            packet->e2e_trace_ct_ufids >>= 1;

            memcpy(&cur_trace_info[tid]->ufids[0], e2e_trace,
                   sizeof *e2e_trace);

            e2e_trace_size--;
            e2e_trace++;
            num_elements[tid]++;
            cur_trace_info[tid]++;
            packet->e2e_trace_flags &= ~E2E_CACHE_TRACE_FLAG_TNL_POP;
            if (!packet->e2e_trace_ct_ufids) {
                continue;
            }
        }
        /* If the trace is marked as "throttled" this means that it must be
         * omitted from sending due to high messages rate.
         */
        if (packet->e2e_trace_flags & E2E_CACHE_TRACE_FLAG_THROTTLED) {
            atomic_count_inc(&e2e_stats[first_tid]->throttled_trcs);
            continue;
        }
        /* Don't send "partial" traces due to overflow of the trace storage */
        if (OVS_UNLIKELY(packet->e2e_trace_flags &
                         E2E_CACHE_TRACE_FLAG_OVERFLOW)) {
            atomic_count_inc(&e2e_stats[first_tid]->discarded_trcs);
            continue;
        }
        /* Send only traces for packet that passed conntrack */
        if (!packet->e2e_trace_ct_ufids) {
            atomic_count_inc(&e2e_stats[first_tid]->discarded_trcs);
            continue;
        }

        tid = netdev_offload_trace_to_thread_id(e2e_trace, e2e_trace_size);
        if (tid == INVALID_OFFLOAD_THREAD_NB) {
            continue;
        }

        if (dp_netdev_e2e_cache_trace_q_size &&
            cur_q_size[tid] >= dp_netdev_e2e_cache_trace_q_size) {
            atomic_count_inc(&e2e_stats[tid]->overflow_trcs);
            continue;
        }

        cur_trace_info[tid]->e2e_trace_ct_ufids = packet->e2e_trace_ct_ufids;
        cur_trace_info[tid]->num_elements = e2e_trace_size;
        cur_trace_info[tid]->orig_in_port = packet->md.orig_in_port;

        memcpy(&cur_trace_info[tid]->ufids[0], e2e_trace,
               e2e_trace_size * sizeof *e2e_trace);

        num_elements[tid]++;
        cur_trace_info[tid]++;
    }

    DP_NETDEV_OFFLOAD_FOREACH_THREAD_NO_MAIN (thread, tid) {
        if (num_elements[tid] == 0) {
            free_cacheline(buffer[tid]);
            continue;
        }

        buffer[tid]->dp = dp;
        buffer[tid]->num_elements = num_elements[tid];
        buffer[tid]->timestamp = now;

        e2e_cache_trace_msg_enqueue(buffer[tid], tid);
        atomic_count_inc(&e2e_stats[tid]->generated_trcs);
    }
}

void
e2e_cache_trace_tnl_pop(struct dp_packet *packet)
{
    packet->e2e_trace_flags |= E2E_CACHE_TRACE_FLAG_TNL_POP;
}

static int
e2e_cache_ufids_to_flows(const ovs_u128 *ufids,
                         uint16_t num_elements,
                         struct e2e_cache_ovs_flow *flows[])
    OVS_REQUIRES(flows_map_mutex)
{
    static struct vlog_rate_limit rl = VLOG_RATE_LIMIT_INIT(10, 10);
    const ovs_u128 *ufid;
    uint32_t hash;
    uint16_t i;

    for (i = 0; i < num_elements; i++) {
        ufid = &ufids[i];
        hash = hash_bytes(ufid, sizeof *ufid, 0);
        flows[i] = e2e_cache_flow_find(ufid, hash);
        VLOG_DBG_RL(&rl, "%s: ufids[%d]="UUID_FMT" flows[%d]=%p", __func__,
                    i, UUID_ARGS((struct uuid *)ufid), i, flows[i]);
        if (OVS_UNLIKELY(!flows[i])) {
            return -1;
        }
        if (i > 0 && flows[i - 1]->offload_state != E2E_OL_STATE_FLOW &&
            flows[i]->offload_state != E2E_OL_STATE_FLOW) {
            if (!flows[i - 1]->ct_peer) {
                flows[i - 1]->ct_peer = flows[i];
            }
            if (!flows[i]->ct_peer) {
                flows[i]->ct_peer = flows[i - 1];
            }
        }
    }
    return 0;
}

static inline void
e2e_cache_populate_offload_item(struct dp_offload_thread_item *offload_item,
                                int op,
                                struct dp_netdev *dp,
                                struct dp_netdev_flow *flow,
                                long long now)
{
    struct dp_offload_flow_item *flow_offload;
    flow_offload = &offload_item->data->flow;

    memset(offload_item, 0, sizeof *offload_item);
    offload_item->type = DP_OFFLOAD_FLOW;
    offload_item->dp = dp;
    offload_item->timestamp = now;
    flow_offload->flow = flow;
    flow_offload->op = op;
    flow_offload->is_e2e_cache_flow = true;
    flow_offload->orig_in_port = flow->orig_in_port;
}

static void
e2e_cache_disassociate_counters(struct e2e_cache_merged_flow *merged_flow)
{
    size_t counter_hash, ct_counter_hash, flows_counter_hash;
    struct flows_counter_key counter_key_on_stack;
    const struct flows_counter_key *counter_key;
    struct e2e_cache_counter_item *counter_item;
    struct e2e_cache_ovs_flow *mt_flow;
    struct ovs_list *next_counter;
    uint16_t i;

    /* If flow_counter_list is empty this means e2e_cache_associate_counters
     * was not executed for this e2e_cache_merged_flow.
     * In such case ct_counter_list must also empty.
     */
    if (OVS_UNLIKELY(ovs_list_is_empty(&merged_flow->flow_counter_list))) {
        return;
    }

    memset(&counter_key_on_stack, 0, sizeof counter_key_on_stack);
    ct_counter_hash = merged_flow->ct_counter_key;
    counter_key_on_stack.ptr_key = ct_counter_hash;
    flows_counter_hash = hash_bytes(&merged_flow->flows_counter_key,
                                    sizeof merged_flow->flows_counter_key,
                                    0);

    ovs_mutex_lock(&flows_map_mutex);
    for (i = 0; i < merged_flow->associated_flows_len; i++) {
        mt_flow = merged_flow->associated_flows[i].mt_flow;
        if (mt_flow->offload_state == E2E_OL_STATE_FLOW) {
            counter_hash = flows_counter_hash;
            counter_key = &merged_flow->flows_counter_key;
        } else {
            counter_hash = ct_counter_hash;
            counter_key = &counter_key_on_stack;
        }
        HMAP_FOR_EACH_WITH_HASH (counter_item, node, counter_hash,
                                 &mt_flow->merged_counters) {
            if (counter_item->hash == counter_hash &&
                !e2e_cache_counter_cmp_key(counter_item, counter_key)) {
                break;
            }
        }
        if (OVS_LIKELY(counter_item)) {
            hmap_remove(&mt_flow->merged_counters, &counter_item->node);
            free(counter_item);
        }
    }
    next_counter = ovs_list_front(&merged_flow->flow_counter_list);
    ovs_list_remove(&merged_flow->flow_counter_list);
    if (ovs_list_is_empty(next_counter)) {
        counter_item = CONTAINER_OF(next_counter,
                                    struct e2e_cache_counter_item,
                                    merged_flows);
        hmap_remove(&counter_map, &counter_item->node);
        free(counter_item);
    }
    next_counter = ovs_list_front(&merged_flow->ct_counter_list);
    ovs_list_remove(&merged_flow->ct_counter_list);
    if (ovs_list_is_empty(next_counter)) {
        counter_item = CONTAINER_OF(next_counter,
                                    struct e2e_cache_counter_item,
                                    merged_flows);
        hmap_remove(&counter_map, &counter_item->node);
        free(counter_item);
    }
    ovs_mutex_unlock(&flows_map_mutex);
}

static void
e2e_cache_associate_counters(struct e2e_cache_merged_flow *merged_flow,
                             struct e2e_cache_ovs_flow *mt_flows[],
                             const struct e2e_cache_trace_info *trc_info,
                             uint16_t num_elements)
{
    size_t counter_hash, ct_counter_key_hash, flows_counter_key_hash;
    struct flows_counter_key counter_key_on_stack;
    const struct flows_counter_key *counter_key;
    struct e2e_cache_counter_item *counter_item;
    uint16_t mt_index, flows_index = 0;

    BUILD_ASSERT_DECL(sizeof(size_t) >= sizeof(uintptr_t));

    memset(&counter_key_on_stack, 0, sizeof counter_key_on_stack);
    ct_counter_key_hash = merged_flow->ct_counter_key;
    counter_key_on_stack.ptr_key = ct_counter_key_hash;
    flows_counter_key_hash = hash_bytes(&merged_flow->flows_counter_key,
                                        sizeof merged_flow->flows_counter_key,
                                        0);

    ovs_mutex_lock(&flows_map_mutex);

    for (mt_index = 0; mt_index < num_elements; mt_index++) {
        bool is_ct = trc_info->e2e_trace_ct_ufids & (1 << mt_index);
        bool counter_found = false;

        if (is_ct) {
            counter_key = &counter_key_on_stack;
            counter_hash = ct_counter_key_hash;
        } else {
            counter_key = &merged_flow->flows_counter_key;
            counter_hash = flows_counter_key_hash;
        }
        /* Search if this counter is already used by this flow. */
        HMAP_FOR_EACH_WITH_HASH (counter_item, node, counter_hash,
                                 &mt_flows[flows_index]->merged_counters) {
            if (counter_item->hash == counter_hash &&
                !e2e_cache_counter_cmp_key(counter_item, counter_key)) {
                if (OVS_UNLIKELY(counter_item->is_ct != is_ct)) {
                    OVS_NOT_REACHED();
                }
                counter_found = true;
                break;
            }
        }
        /* If this counter is not in use by this flow, add it. */
        if (!counter_found) {
            counter_item = e2e_cache_counter_alloc(counter_key, counter_hash,
                                                   is_ct);
            hmap_insert(&mt_flows[flows_index]->merged_counters,
                        &counter_item->node, counter_hash);
        }
        flows_index++;
    }

    /* Search for an already existing CT counter item, or create if not. */
    counter_hash = ct_counter_key_hash;
    counter_key = &counter_key_on_stack;
    counter_item = e2e_cache_counter_find(counter_hash, counter_key);
    if (!counter_item) {
        counter_item = e2e_cache_counter_alloc(counter_key, counter_hash,
                                               true);
        hmap_insert(&counter_map, &counter_item->node, counter_hash);
    }
    /* Add the merged flow to the counter item. */
    ovs_list_push_back(&counter_item->merged_flows,
                       &merged_flow->ct_counter_list);

    /* Search for an already existing flows counter item, or create if not. */
    counter_hash = flows_counter_key_hash;
    counter_key = &merged_flow->flows_counter_key;
    counter_item = e2e_cache_counter_find(counter_hash, counter_key);
    if (!counter_item) {
        counter_item = e2e_cache_counter_alloc(counter_key, counter_hash,
                                               false);
        hmap_insert(&counter_map, &counter_item->node, counter_hash);
    }
    /* Add the merged flow to the counter item. */
    ovs_list_push_back(&counter_item->merged_flows,
                       &merged_flow->flow_counter_list);

    ovs_mutex_unlock(&flows_map_mutex);
}

static int
e2e_cache_merged_flow_offload_del(struct e2e_cache_merged_flow *merged_flow)
{
    unsigned int tid = netdev_offload_thread_id();
    struct dp_offload_thread_item *offload_item;
    struct dp_netdev *dp = merged_flow->dp;
    struct e2e_cache_stats *e2e_stats;
    struct dp_netdev_flow flow;
    int rv;

    e2e_stats = &dp_offload_threads[tid].e2e_stats;

    ovs_assert(dp);

    memset(&flow, 0, sizeof flow);
    *CONST_CAST(ovs_u128 *, &flow.mega_ufid) = merged_flow->ufid;
    CONST_CAST(struct flow *, &flow.flow)->in_port =
        merged_flow->merged_match.spec.in_port;

    offload_item = xmalloc(sizeof *offload_item +
                           sizeof offload_item->data->flow);
    e2e_cache_populate_offload_item(offload_item,
                                    DP_NETDEV_FLOW_OFFLOAD_OP_DEL, dp, &flow,
                                    time_usec());

    merged_flow->dp = NULL;
    e2e_stats->del_merged_flow_hw++;
    rv = dp_netdev_flow_offload_del(offload_item);
    free(offload_item);
    return rv;
}

static void
e2e_cache_calc_counters(struct e2e_cache_merged_flow *merged_flow,
                        struct e2e_cache_ovs_flow *mt_flows[],
                        const struct e2e_cache_trace_info *trc_info,
                        uint16_t num_elements)
{
    uintptr_t ptr, ct_counter_key = UINTPTR_MAX;
    uint16_t mt_index, flows_index = 0;

    merged_flow->ct_counter_key = 0;
    memset(&merged_flow->flows_counter_key, 0,
           sizeof merged_flow->flows_counter_key);

    for (mt_index = 0; mt_index < num_elements; mt_index++) {
        if (trc_info->e2e_trace_ct_ufids & (1 << mt_index)) {
            if (trc_info->e2e_trace_ct_ufids & (1 << (mt_index + 1))) {
                continue;
            }
            /* CT are traced only for both directions, adjacent. Calc
             * ct_counter as the lowest value among all pointers to DB items
             * for all CT in the trace.
             */
            ptr = (uintptr_t) mt_flows[mt_index];
            if (ptr < ct_counter_key) {
                ct_counter_key = ptr;
            }
            ptr = (uintptr_t) mt_flows[mt_index - 1];
            if (ptr < ct_counter_key) {
                ct_counter_key = ptr;
            }
        } else {
            merged_flow->flows_counter_key.ufid_key[flows_index++] =
                trc_info->ufids[mt_index];
        }
    }

    if (trc_info->e2e_trace_ct_ufids) {
        merged_flow->ct_counter_key = ct_counter_key;
    }
}

static int
e2e_cache_merged_flow_offload_put(struct dp_netdev *dp,
                                  struct e2e_cache_merged_flow *merged_flow,
                                  struct e2e_cache_ovs_flow *mt_flows[],
                                  const struct e2e_cache_trace_info *trc_info)
{
    unsigned int tid = netdev_offload_thread_id();
    struct dp_offload_thread_item *offload_item;
    struct dp_offload_flow_item *flow_offload;
    struct e2e_cache_stats *e2e_stats;
    struct dp_netdev_flow flow;
    union flow_in_port in_port;
    uint16_t num_elements;
    int err;

    e2e_stats = &dp_offload_threads[tid].e2e_stats;

    in_port = merged_flow->merged_match.spec.in_port;

    memset(&flow, 0, sizeof flow);
    flow.mark = merged_flow->flow_mark;
    flow.dead = false;
    *CONST_CAST(ovs_u128 *, &flow.mega_ufid) = merged_flow->ufid;
    CONST_CAST(struct flow *, &flow.flow)->in_port = in_port;
    flow.orig_in_port = trc_info->orig_in_port;

    offload_item = xmalloc(sizeof *offload_item +
                           sizeof offload_item->data->flow);
    e2e_cache_populate_offload_item(offload_item,
                                    DP_NETDEV_FLOW_OFFLOAD_OP_ADD, dp, &flow,
                                    time_usec());

    num_elements = trc_info->num_elements;
    /* For CT2CT, don't associate the last megaflow. */
    if (flow.mark != INVALID_FLOW_MARK) {
        num_elements--;
    }
    e2e_cache_calc_counters(merged_flow, mt_flows, trc_info, num_elements);
    e2e_cache_associate_counters(merged_flow, mt_flows, trc_info,
                                 num_elements);

    flow_offload = &offload_item->data->flow;
    merged_match_to_match(&flow_offload->match, &merged_flow->merged_match);
    flow_offload->actions = xmalloc(merged_flow->actions_size);
    if (OVS_UNLIKELY(!flow_offload->actions)) {
        err = -1;
        goto error;
    }

    memcpy(flow_offload->actions, merged_flow->actions,
           merged_flow->actions_size);
    flow_offload->actions_len = merged_flow->actions_size;
    flow_offload->ct_counter_key = merged_flow->ct_counter_key;
    memcpy(&flow_offload->flows_counter_key, &merged_flow->flows_counter_key,
           sizeof flow_offload->flows_counter_key);
    err = dp_netdev_flow_offload_put(offload_item);
    free(flow_offload->actions);
    free(offload_item);
    if (OVS_UNLIKELY(err != 0)) {
        goto error;
    }

    merged_flow->dp = dp;
    e2e_stats->add_merged_flow_hw++;
    return 0;

error:
    return err;
}

static void
e2e_cache_offload_ct_mt_flows(struct dp_netdev *dp,
                              struct e2e_cache_ovs_flow *mt_flows[],
                              uint16_t num_flows)
{
    struct nlattr *actions = NULL;
    uint16_t max_actions_len = 0;
    uint16_t i;

    for (i = 0; i < num_flows; i++) {
        actions = e2e_cache_ct_flow_offload_add_mt(dp, mt_flows[i],
                                                   actions, &max_actions_len);
    }
    if (actions) {
        free(actions);
    }
}

static int
e2e_cache_merge_flows(struct e2e_cache_ovs_flow **flows,
                      uint16_t num_flows,
                      struct e2e_cache_merged_flow *merged_flow,
                      struct ofpbuf *merged_actions);
static int
ct2ct_merge_flows(struct e2e_cache_ovs_flow **flows,
                  uint16_t num_flows,
                  struct e2e_cache_merged_flow *merged_flow,
                  struct ofpbuf *merged_actions);

/* CT rules should be offloaded either to MT or cache, but not both. For a
 * merged flow created, remove from HW the MT rules for its composing CT
 * flows.
 */
static void
e2e_cache_purge_ct_flows_from_mt(struct dp_netdev *dp,
                                 struct e2e_cache_ovs_flow *mt_flows[],
                                 uint16_t num_flows)
{
    uint16_t i;

    for (i = 0; i < num_flows; i++) {
        /* When an e2e flow is created, it should remove its MT HW rule if
         * exists, but not its peer MT HW rule. Skip if not in HW or peer
         * CT flows.
         */
        if (mt_flows[i]->offload_state != E2E_OL_STATE_CT_MT ||
            (i > 0 && mt_flows[i - 1]->offload_state != E2E_OL_STATE_FLOW)) {
            continue;
        }
        e2e_cache_ct_flow_offload_del_mt(dp, mt_flows[i]);
        e2e_cache_flow_state_set(mt_flows[i], E2E_OL_STATE_CT_SW);
    }
}

int
e2e_cache_process_trace_info(struct dp_netdev *dp,
                             const struct e2e_cache_trace_info *trc_info,
                             unsigned int tid)
{
    struct e2e_cache_merged_flow *merged_flow;
    struct nlattr *actions;
    size_t actions_len;
    struct ofpbuf merged_actions;
    struct e2e_cache_ovs_flow *mt_flows[E2E_CACHE_MAX_TRACE];
    uint64_t merged_actions_buf[1024 / sizeof(uint64_t)];
    uint32_t merged_flows_count;
    uint16_t i, num_flows;
    bool ct2ct;
    int err;

    ovs_mutex_lock(&flows_map_mutex);
    err = e2e_cache_ufids_to_flows(trc_info->ufids, trc_info->num_elements,
                                   mt_flows);
    ovs_mutex_unlock(&flows_map_mutex);
    if (OVS_UNLIKELY(err)) {
        return -1;
    }
    num_flows = trc_info->num_elements;
    merged_flows_count = atomic_count_get(&merged_flows_map_count);
    ct2ct = num_flows > 4 && merged_flows_count >= dp_netdev_e2e_cache_size;
    if (!ct2ct && merged_flows_count >= dp_netdev_e2e_cache_size) {
        e2e_cache_offload_ct_mt_flows(dp, mt_flows, num_flows);
        return -1;
    }

    merged_flow = xzalloc(sizeof *merged_flow +
                          num_flows * sizeof merged_flow->associated_flows[0]);
    if (OVS_UNLIKELY(!merged_flow)) {
        return -1;
    }
    merged_flow->tid = tid;
    ovs_list_init(&merged_flow->flow_counter_list);
    ovs_list_init(&merged_flow->ct_counter_list);
    for (i = 0; i < num_flows; i++) {
        ovs_list_init(&merged_flow->associated_flows[i].list);
    }

    ofpbuf_use_stack(&merged_actions, &merged_actions_buf,
                     sizeof merged_actions_buf);

    err = ct2ct
          ? ct2ct_merge_flows(mt_flows, num_flows, merged_flow,
                              &merged_actions)
          : e2e_cache_merge_flows(mt_flows, num_flows, merged_flow,
                                  &merged_actions);
    if (OVS_UNLIKELY(err)) {
        goto free_merged_flow;
    }

    actions = (struct nlattr *) ofpbuf_at_assert(&merged_actions, 0,
                                                 sizeof(struct nlattr));
    actions_len = merged_actions.size;
    merged_flow->actions = (struct nlattr *) xmalloc(actions_len);
    if (OVS_UNLIKELY(!merged_flow->actions)) {
        goto free_merged_flow;
    }
    memcpy(merged_flow->actions, actions, actions_len);
    merged_flow->actions_size = actions_len;

    e2e_cache_associate_merged_flow(merged_flow, mt_flows,
                                    ct2ct ? num_flows - 1 : num_flows);

    err = e2e_cache_merged_flow_db_put(merged_flow);
    if (OVS_UNLIKELY(err)) {
        goto disassociate_merged_flow;
    }

    err = e2e_cache_merged_flow_offload_put(dp, merged_flow, mt_flows,
                                            trc_info);
    if (OVS_UNLIKELY(err)) {
        goto remove_flow_from_db;
    }
    e2e_cache_purge_ct_flows_from_mt(dp, mt_flows, num_flows);
    for (i = 0; i < num_flows; i++) {
        if (mt_flows[i]->offload_state == E2E_OL_STATE_FLOW ||
            (i > 0 && mt_flows[i - 1]->offload_state != E2E_OL_STATE_FLOW)) {
            continue;
        }
        /* Update CT stats affected by offloading the merged flow. */
        e2e_cache_flow_state_set(mt_flows[i], ct2ct ? E2E_OL_STATE_CT2CT
                                                    : E2E_OL_STATE_CT_HW);
    }
    return 0;

remove_flow_from_db:
    e2e_cache_merged_flow_db_del(merged_flow);
    return err;
disassociate_merged_flow:
    /* Lock/unlock to prevent race condition with
     * e2e_cache_get_merged_flows_stats()
     */
    ovs_mutex_lock(&flows_map_mutex);
    e2e_cache_disassociate_merged_flow(merged_flow);
    ovs_mutex_unlock(&flows_map_mutex);
free_merged_flow:
    e2e_cache_merged_flow_free(merged_flow);
    return err;
}

bool
e2e_cache_get_merged_flows_stats(struct netdev *netdev,
                                 struct match *match,
                                 struct nlattr **actions,
                                 const ovs_u128 *mt_ufid,
                                 struct dpif_flow_stats *stats,
                                 struct ofpbuf *buf,
                                 long long now,
                                 long long prev_now)
{
    struct e2e_cache_counter_item *mt_counter_item, *mapped_counter_item;
    static struct vlog_rate_limit rl = VLOG_RATE_LIMIT_INIT(10, 10);
    struct e2e_cache_merged_flow *merged_flow;
    struct dpif_flow_stats merged_stats;
    struct dpif_flow_attrs merged_attr;
    struct ovs_list *merged_flow_node;
    struct e2e_cache_ovs_flow *flow;
    bool rv = false;
    uint32_t hash;
    int ret;

    hash = hash_bytes(mt_ufid, sizeof *mt_ufid, 0);

    ovs_mutex_lock(&flows_map_mutex);
    flow = e2e_cache_flow_find(mt_ufid, hash);
    if (OVS_UNLIKELY(!flow) ||
        ovs_list_is_empty(&flow->associated_merged_flows)) {
        ovs_mutex_unlock(&flows_map_mutex);
        return false;
    }
    if (flow->offload_state == E2E_OL_STATE_CT2CT) {
        struct flow2flow_item *associated_flow_item;

        associated_flow_item =
            CONTAINER_OF(ovs_list_front(&flow->associated_merged_flows),
                         struct flow2flow_item, list);
        merged_flow =
            CONTAINER_OF(associated_flow_item,
                         struct e2e_cache_merged_flow,
                         associated_flows[associated_flow_item->index]);
        /* Query the counter. */
        ret = netdev_flow_get(netdev, match, actions, &merged_flow->ufid,
                              &merged_stats, &merged_attr, buf, now);
        if (ret) {
            VLOG_ERR_RL(&rl, "Failed to get merged flow ufid "UUID_FMT,
                        UUID_ARGS((struct uuid *) &merged_flow->ufid));
            goto out;
        }
        stats->n_bytes += merged_stats.n_bytes;
        stats->n_packets += merged_stats.n_packets;
        stats->used = MAX(stats->used, merged_stats.used);
        rv = true;
        goto out;
    }

    HMAP_FOR_EACH (mt_counter_item, node, &flow->merged_counters) {
        if (flow->offload_state == E2E_OL_STATE_FLOW) {
            /* Get the counter item from the global map. */
            mapped_counter_item =
                e2e_cache_counter_find(mt_counter_item->hash,
                                       &mt_counter_item->key);
            if (OVS_UNLIKELY(!mapped_counter_item)) {
                VLOG_ERR_RL(&rl, "Failed to get counter item for ufid "
                            UUID_FMT, UUID_ARGS((struct uuid *) &flow->ufid));
                continue;
            }

            /* Get one of the merged flows using this counter. */
            merged_flow_node =
                ovs_list_front(&mapped_counter_item->merged_flows);
            merged_flow = CONTAINER_OF(merged_flow_node,
                                       struct e2e_cache_merged_flow,
                                       flow_counter_list);
            /* Query the counter. */
            ret = netdev_flow_get(netdev, match, actions, &merged_flow->ufid,
                                  &merged_stats, &merged_attr, buf, now);
            if (ret) {
                VLOG_ERR_RL(&rl, "Failed to get merged flow ufid "UUID_FMT,
                            UUID_ARGS((struct uuid *) &merged_flow->ufid));
                continue;
            }
            stats->n_bytes += merged_stats.n_bytes;
            stats->n_packets += merged_stats.n_packets;
            stats->used = MAX(stats->used, merged_stats.used);
            rv = true;
        } else {
            ret = netdev_ct_counter_query(netdev, mt_counter_item->key.ptr_key,
                                          now, prev_now, stats);
            if (ret) {
                VLOG_ERR_RL(&rl, "Failed to query ct counter netdev=%s, "
                            "ptr_key=%"PRIxPTR, netdev_get_name(netdev),
                            mt_counter_item->key.ptr_key);
                continue;
            }
            rv |= stats->used == now;
        }
    }
out:
    ovs_mutex_unlock(&flows_map_mutex);
    return rv;
}

int
e2e_cache_flow_del(const ovs_u128 *ufid, struct dp_netdev *dp,
                   struct conn *conn, long long int now);

static void
dp_netdev_ct_offload_e2e_del(ovs_u128 *ufid, void *dp,
                             struct conn *conn, long long int now)
{
    e2e_cache_flow_del(ufid, dp, conn, now);
}

static int
dp_netdev_ct_offload_e2e_active(struct ct_flow_offload_item *offload,
                                long long now, long long prev_now);

static void
pmd_thread_offload_netdevs(struct dp_netdev_pmd_thread *pmd, odp_port_t port_nos[CT_DIR_NUM],
                           struct netdev *netdevs[CT_DIR_NUM]);

static void
dp_netdev_offload_conns(struct dp_netdev_pmd_thread *pmd,
                        struct conntrack *ct, struct batch *conns, int op)
    OVS_NO_THREAD_SAFETY_ANALYSIS
{
    static struct vlog_rate_limit rl = VLOG_RATE_LIMIT_INIT(1, 5);
    struct dp_netdev *dp = conntrack_datapath(ct);
    struct conn *first = batch_first(conns);
    struct netdev *netdevs[CT_DIR_NUM];
    odp_port_t ports[CT_DIR_NUM];
    int n = batch_size(conns);
    const char *desc = NULL;
    int ret = 0;

    conntrack_offload_get_ports(first, ports);
    if (pmd) {
        memset(netdevs, 0, sizeof netdevs);
        pmd_thread_offload_netdevs(pmd, ports, netdevs);
        /* No need to lock the datapath port_rwlock here: we are in the context
         * of a polling thread, which will safely pause itself whenever a thread
         * has to modify the port map.
         *
         * Similarly, the netdev references are not kept, as they are
         * already held within the thread poll_list. */
        for (int i = CT_DIR_INIT; i < CT_DIR_NUM; i++) {
            if (!netdevs[i]) {
                netdevs[i] = netdev_ports_get_short(ports[i], dpif_normalize_type(dp->class->type));
            }
        }
    } else {
        for (int i = CT_DIR_INIT; i < CT_DIR_NUM; i++) {
            netdevs[i] = netdev_ports_get(ports[i], dpif_normalize_type(dp->class->type));
        }
        dp_netdev_port_rdlock_limit(dp, 10);
    }

    if (op == DP_NETDEV_FLOW_OFFLOAD_OP_ADD) {
        desc = "add";
        ret = netdev_conns_add(netdevs, conns);
        conntrack_stats_add(ct, CT_DPIF_CONN_STATE_OFFLOAD_ADD_PROCESSED, n);
    } else if (op == DP_NETDEV_FLOW_OFFLOAD_OP_DEL) {
        struct conn *conn;

        desc = "delete";
        ret = netdev_conns_del(netdevs, conns);
        conntrack_stats_add(ct, CT_DPIF_CONN_STATE_OFFLOAD_DEL_PROCESSED, n);

        /* Make sure that the offload provider left the batch intact, we
         * must release all connection references. */
        ovs_assert(n == batch_size(conns));

        BATCH_FOREACH (conn, conns) {
            conntrack_conn_unref(conn);
            conntrack_conn_unref(conn);
        }
    }

    if (!pmd) {
        ovs_rwlock_unlock(&dp->port_rwlock);
        netdev_close(netdevs[CT_DIR_INIT]);
        netdev_close(netdevs[CT_DIR_REP]);
    }

    if (ret) {
        VLOG_ERR_RL(&rl, "failed to %s %d connection offloads", desc, n);
    } else {
        int nb_ok = batch_size(conns);
        int nb_fail = n - nb_ok;

        if (nb_fail > 0) {
            VLOG_ERR_RL(&rl, "failed to %s %d connection offloads", desc, nb_fail);
        }
        if (nb_ok > 0 && VLOG_IS_DBG_ENABLED()) {
            VLOG_DBG_RL(&rl, "succeeded to %s %d connection offloads", desc, nb_ok);
        }
    }
}

static void
dp_netdev_offload_conns_add(struct conntrack *ct, struct batch *conns)
{
    struct dp_netdev *dp = conntrack_datapath(ct);
    struct dp_netdev_pmd_thread *pmd;

    if (dp_netdev_e2e_cache_enabled) {
        return;
    }

    pmd = ovsthread_getspecific(dp->per_pmd_key);

    while (!batch_is_empty(conns)) {
        struct conn *first = batch_first(conns);
        odp_port_t ports[CT_DIR_NUM];
        struct batch same_ports;
        struct conn *conn;

        batch_init(&same_ports);
        conntrack_offload_get_ports(first, ports);
        BATCH_FOREACH_POP (conn, conns) {
            odp_port_t it_ports[CT_DIR_NUM];

            /* All connection handles in 'conns' must share the exact same ports,
             * in the same order. */
            conntrack_offload_get_ports(conn, it_ports);
            if (it_ports[CT_DIR_INIT] == ports[CT_DIR_INIT] &&
                it_ports[CT_DIR_REP] == ports[CT_DIR_REP]) {
                batch_add(&same_ports, conn);
            } else {
                batch_add(conns, conn);
            }
        }
        dp_netdev_offload_conns(pmd, ct, &same_ports, DP_NETDEV_FLOW_OFFLOAD_OP_ADD);
    }
}

static void
dp_netdev_offload_conn_del(struct conntrack *ct, struct conn *conn)
{
    struct ct_offload_handle *coh;
    unsigned int insertion_tid;

    if (dp_netdev_e2e_cache_enabled) {
        return;
    }

    coh = conntrack_offload_get(conn);
    insertion_tid = coh->insertion_tid;

    if (insertion_tid != netdev_offload_thread_id()) {
        struct dp_netdev *dp = conntrack_datapath(ct);
        struct dp_offload_thread_item *msg;

        msg = xzalloc(sizeof *msg + sizeof(struct dp_offload_conn_item));
        msg->timestamp = time_usec();
        msg->type = DP_OFFLOAD_CONN;
        msg->dp = dp;
        msg->data->conn.op = DP_NETDEV_FLOW_OFFLOAD_OP_DEL;
        msg->data->conn.ct = ct;
        msg->data->conn.conn = conn;
        dp_netdev_offload_thread_enqueue(&dp_offload_threads[insertion_tid], msg);
    } else {
        struct dp_netdev *dp = conntrack_datapath(ct);
        struct dp_netdev_pmd_thread *pmd = ovsthread_getspecific(dp->per_pmd_key);
        struct batch conns = batch_init_one(conn);

        dp_netdev_offload_conns(pmd, ct, &conns, DP_NETDEV_FLOW_OFFLOAD_OP_DEL);
    }
}

static int
dp_netdev_offload_conn_active(struct conntrack *ct, struct conn *conn,
                              long long now, long long prev_now)
{
    struct netdev *netdevs[CT_DIR_NUM];
    odp_port_t ports[CT_DIR_NUM];
    struct dpif_flow_stats stats;
    const struct dp_netdev *dp;
    int ret = 0;

    if (!conntrack_offload_is_enabled()) {
        return EINVAL;
    }

    dp = conntrack_datapath(ct);
    conntrack_offload_get_ports(conn, ports);
    for (int i = CT_DIR_INIT; i < CT_DIR_NUM; i++) {
        netdevs[i] = netdev_ports_get(ports[i], dpif_normalize_type(dp->class->type));
    }
    ret = netdev_conn_stats(netdevs, conn, &stats, NULL, now);
    netdev_close(netdevs[CT_DIR_INIT]);
    netdev_close(netdevs[CT_DIR_REP]);

    if (ret) {
        return ret;
    }

    return stats.used > prev_now ? 0 : EINVAL;
}

static void
dp_netdev_ct_offload_get_ufid(ovs_u128 *ufid)
{
    ufid->u64.hi = ufid->u64.lo = hash_pointer(ufid, 0);
    uuid_set_bits_v4((struct uuid *) ufid, UUID_ATTR_1);
}

static int
dp_netdev_ct_e2e_add_cb(struct ct_flow_offload_item *offload,
                        struct ct_match *match, struct nlattr *actions,
                        int actions_len)
{
    return e2e_cache_flow_put(true, &offload->ufid, match, actions,
                              actions_len, offload->timestamp);
}

static void
dp_netdev_ct_offload_e2e_add(struct ct_flow_offload_item *offload);

static struct conntrack_offload_class dpif_ct_offload_class = {
    .conn_get_ufid = dp_netdev_ct_offload_get_ufid,
    .conns_add = dp_netdev_offload_conns_add,
    .conn_del = dp_netdev_offload_conn_del,
    .conn_active = dp_netdev_offload_conn_active,
    .conn_e2e_add = dp_netdev_ct_offload_e2e_add,
    .conn_e2e_del = dp_netdev_ct_offload_e2e_del,
    .conn_e2e_active = dp_netdev_ct_offload_e2e_active,
};

void
dp_netdev_ct_offload_init(struct dp_netdev *dp)
{
    dp->conntrack = conntrack_init(dp);
    conntrack_set_offload_class(dp->conntrack, &dpif_ct_offload_class);
}

void
dp_netdev_ct_offload_uninit(struct dp_netdev *dp)
{
    /* Set offload_class to NULL to mark dp is being destroyed. */
    conntrack_set_offload_class(dp->conntrack, NULL);
    conntrack_destroy(dp->conntrack);
}

unsigned int
dp_netdev_offload_thread_nb(void)
{
    struct dp_offload_thread *t;
    unsigned int count = 0;

    DP_NETDEV_OFFLOAD_FOREACH_THREAD (t) {
        count++;
    }
    return count;
}

static void
dp_netdev_e2e_offload_init(struct e2e_cache_stats *e2e_stats)
{
    atomic_count_init(&e2e_stats->generated_trcs, 0);
    e2e_stats->processed_trcs = 0;
    atomic_count_init(&e2e_stats->discarded_trcs, 0);
    atomic_count_init(&e2e_stats->aborted_trcs, 0);
    atomic_count_init(&e2e_stats->throttled_trcs, 0);
    atomic_count_init(&e2e_stats->queue_trcs, 0);
    atomic_count_init(&e2e_stats->overflow_trcs, 0);
    atomic_count_init(&e2e_stats->flow_add_msgs, 0);
    atomic_count_init(&e2e_stats->flow_del_msgs, 0);
    e2e_stats->flush_flow_msgs = 0;
    e2e_stats->succ_merged_flows = 0;
    e2e_stats->merge_rej_flows = 0;
    e2e_stats->add_merged_flow_hw = 0;
    e2e_stats->del_merged_flow_hw = 0;
    e2e_stats->add_ct_mt_flow_hw = 0;
    e2e_stats->del_ct_mt_flow_hw = 0;
    e2e_stats->add_ct_mt_flow_err = 0;
    e2e_stats->del_ct_mt_flow_err = 0;
    e2e_stats->succ_ct2ct_merges = 0;
    e2e_stats->rej_ct2ct_merges = 0;
    e2e_stats->add_ct2ct_flows = 0;
    e2e_stats->del_ct2ct_flows = 0;
}

static void
dp_netdev_offload_thread_reset(struct dp_offload_thread *thread)
{
    dp_netdev_e2e_offload_init(&thread->e2e_stats);
    atomic_init(&thread->ct_uni_dir_connections, 0);
    atomic_init(&thread->ct_bi_dir_connections, 0);
    atomic_init(&thread->enqueued_offload, 0);
    atomic_init(&thread->enqueued_ct_add, 0);
    mov_avg_cma_init(&thread->cma);
    mov_avg_ema_init(&thread->ema, 100);
}

void
dp_netdev_offload_thread_uninit(struct dp_offload_thread *thread)
{
    bool active;

    atomic_read(&thread->active, &active);
    if (!active) {
        return;
    }

    atomic_store_relaxed(&thread->active, false);
    dp_netdev_offload_thread_reset(thread);
}

void
dp_netdev_offload_thread_init(struct dp_offload_thread *thread)
{
    mpsc_queue_init(&thread->offload_queue);
    mpsc_queue_init(&thread->ufid_queue);
    mpsc_queue_init(&thread->trace_queue);
    cmap_init(&thread->megaflow_to_mark);
    cmap_init(&thread->mark_to_flow);
    histogram_walls_set_log(&thread->latency, 1, 2000);

    for (int i = 0; i < DP_OFFLOAD_TYPE_NUM; i++) {
        struct dp_offload_queue_metrics *m;

        m = &thread->queue_metrics[i];
        histogram_walls_set_log(&m->wait_time, 1, 2000);
        histogram_walls_set_log(&m->service_time, 1, 10000);
        histogram_walls_set_log(&m->sojourn_time, 1, 2000);
    }

    dp_netdev_offload_thread_reset(thread);
    atomic_store_relaxed(&thread->active, true);
}

int
dpif_netdev_offload_stats_get(struct dpif *dpif,
                              struct netdev_custom_stats *stats,
                              bool verbose)
{
    enum {
        DP_NETDEV_HW_OFFLOADS_STATS_ENQUEUED_OFFLOADS,
        DP_NETDEV_HW_OFFLOADS_STATS_INSERTED,
        DP_NETDEV_HW_OFFLOADS_STATS_CT_UNI_DIR_CONNS,
        DP_NETDEV_HW_OFFLOADS_STATS_CT_BI_DIR_CONNS,
        DP_NETDEV_HW_OFFLOADS_STATS_LAT_CMA_MEAN,
        DP_NETDEV_HW_OFFLOADS_STATS_LAT_CMA_STDDEV,
        DP_NETDEV_HW_OFFLOADS_STATS_LAT_EMA_MEAN,
        DP_NETDEV_HW_OFFLOADS_STATS_LAT_EMA_STDDEV,
        DP_NETDEV_HW_OFFLOADS_STATS_LAST,
    };
    enum {
        DP_NETDEV_E2E_STATS_GENERATED_TRCS,
        DP_NETDEV_E2E_STATS_PROCESSED_TRCS,
        DP_NETDEV_E2E_STATS_DISCARRDED_TRCS,
        DP_NETDEV_E2E_STATS_ABORTED_TRCS,
        DP_NETDEV_E2E_STATS_THROTTLED_TRCS,
        DP_NETDEV_E2E_STATS_QUEUE_TRCS,
        DP_NETDEV_E2E_STATS_OVERFLOW_TRCS,
        DP_NETDEV_E2E_STATS_FLOW_ADDS,
        DP_NETDEV_E2E_STATS_FLOW_DELS,
        DP_NETDEV_E2E_STATS_FLOW_FLUSHS,
        DP_NETDEV_E2E_STATS_SUC_MERGES,
        DP_NETDEV_E2E_STATS_REJ_MERGES,
        DP_NETDEV_E2E_STATS_HW_ADD_E2E_FLOWS,
        DP_NETDEV_E2E_STATS_HW_DEL_E2E_FLOWS,
        DP_NETDEV_E2E_STATS_MERGED_FLOWS,
        DP_NETDEV_E2E_STATS_DB_FLOWS,
        DP_NETDEV_E2E_STATS_CT_MT_ADDS,
        DP_NETDEV_E2E_STATS_CT_MT_DELS,
        DP_NETDEV_E2E_STATS_FAILED_CT_MT_ADDS,
        DP_NETDEV_E2E_STATS_FAILED_CT_MT_DELS,
        DP_NETDEV_E2E_STATS_SUC_CT2CT_MERGES,
        DP_NETDEV_E2E_STATS_REJ_CT2CT_MERGES,
        DP_NETDEV_E2E_STATS_CT2CT_ADDS,
        DP_NETDEV_E2E_STATS_CT2CT_DELS,
        DP_NETDEV_E2E_STATS_LAST,
    };
    struct {
        const char *name;
        uint64_t total;
    } hwol_stats[] = {
        [DP_NETDEV_HW_OFFLOADS_STATS_ENQUEUED_OFFLOADS] =
            { "                Enqueued offloads", 0 },
        [DP_NETDEV_HW_OFFLOADS_STATS_INSERTED] =
            { "                Inserted offloads", 0 },
        [DP_NETDEV_HW_OFFLOADS_STATS_CT_UNI_DIR_CONNS] =
            { "           CT uni-dir Connections", 0 },
        [DP_NETDEV_HW_OFFLOADS_STATS_CT_BI_DIR_CONNS] =
            { "            CT bi-dir Connections", 0 },
        [DP_NETDEV_HW_OFFLOADS_STATS_LAT_CMA_MEAN] =
            { "  Cumulative Average latency (us)", 0 },
        [DP_NETDEV_HW_OFFLOADS_STATS_LAT_CMA_STDDEV] =
            { "   Cumulative Latency stddev (us)", 0 },
        [DP_NETDEV_HW_OFFLOADS_STATS_LAT_EMA_MEAN] =
            { " Exponential Average latency (us)", 0 },
        [DP_NETDEV_HW_OFFLOADS_STATS_LAT_EMA_STDDEV] =
            { "  Exponential Latency stddev (us)", 0 },
    }, e2e_stats[] = {
        [DP_NETDEV_E2E_STATS_GENERATED_TRCS] =
            { "                 Generated traces", 0 },
        [DP_NETDEV_E2E_STATS_PROCESSED_TRCS] =
            { "                 Processed traces", 0 },
        [DP_NETDEV_E2E_STATS_DISCARRDED_TRCS] =
            { "                 Discarded traces", 0 },
        [DP_NETDEV_E2E_STATS_ABORTED_TRCS] =
            { "                   Aborted traces", 0 },
        [DP_NETDEV_E2E_STATS_THROTTLED_TRCS] =
            { "                 Throttled traces", 0 },
        [DP_NETDEV_E2E_STATS_QUEUE_TRCS] =
            { "                     Queue traces", 0 },
        [DP_NETDEV_E2E_STATS_OVERFLOW_TRCS] =
            { "                  Overflow traces", 0 },
        [DP_NETDEV_E2E_STATS_FLOW_ADDS] =
            { "                Flow add messages", 0 },
        [DP_NETDEV_E2E_STATS_FLOW_DELS] =
            { "                Flow del messages", 0 },
        [DP_NETDEV_E2E_STATS_FLOW_FLUSHS] =
            { "              Flow flush messages", 0 },
        [DP_NETDEV_E2E_STATS_SUC_MERGES] =
            { "                Successful merges", 0 },
        [DP_NETDEV_E2E_STATS_REJ_MERGES] =
            { "                  Rejected merges", 0 },
        [DP_NETDEV_E2E_STATS_HW_ADD_E2E_FLOWS] =
            { "                 HW add e2e flows", 0 },
        [DP_NETDEV_E2E_STATS_HW_DEL_E2E_FLOWS] =
            { "                 HW del e2e flows", 0 },
        [DP_NETDEV_E2E_STATS_MERGED_FLOWS] =
            { "                 Merged e2e flows", 0 },
        [DP_NETDEV_E2E_STATS_DB_FLOWS] =
            { "                     e2e DB flows", 0 },
        [DP_NETDEV_E2E_STATS_CT_MT_ADDS] =
            { "                       CT MT Adds", 0 },
        [DP_NETDEV_E2E_STATS_CT_MT_DELS] =
            { "                       CT MT Dels", 0 },
        [DP_NETDEV_E2E_STATS_FAILED_CT_MT_ADDS] =
            { "                Failed CT MT Adds", 0 },
        [DP_NETDEV_E2E_STATS_FAILED_CT_MT_DELS] =
            { "                Failed CT MT Dels", 0 },
        [DP_NETDEV_E2E_STATS_SUC_CT2CT_MERGES] =
            { "            Successful CT2CT mrgs", 0 },
        [DP_NETDEV_E2E_STATS_REJ_CT2CT_MERGES] =
            { "            Rejected CT2CT merges", 0 },
        [DP_NETDEV_E2E_STATS_CT2CT_ADDS] =
            { "                       CT2CT Adds", 0 },
        [DP_NETDEV_E2E_STATS_CT2CT_DELS] =
            { "                       CT2CT Dels", 0 },
    }, *cur_stats;

    struct netdev_offload_stats per_port_nos[MAX_OFFLOAD_THREAD_NB];
    struct netdev_offload_stats total_nos[MAX_OFFLOAD_THREAD_NB];
    struct dp_netdev *dp = get_dp_netdev(dpif);
    struct dp_offload_thread *thread;
    struct dp_netdev_port *port;
    unsigned int nb_thread;
    unsigned int nb_counts;
    unsigned int tid;
    size_t i;
#define DP_NETDEV_STATS_TOTAL_COUNTS \
    (ARRAY_SIZE(hwol_stats) + ARRAY_SIZE(e2e_stats))

    if (!netdev_is_flow_api_enabled()) {
        return EINVAL;
    }

    /* /!\ NOTE:
     * It is safe-ish to read nb_thread here and allocate
     * counters, before iterating on the active threads.
     * This function executes in the main thread. A change in
     * active offload threads could only result from a configuration
     * change executing in the main thread. As such, it is impossible
     * for this thread number to change between the two reads.
     *
     * This is only valid if this function executes in the main thread.
     */
    nb_thread = dp_netdev_offload_thread_nb();

    ovs_assert(nb_thread > 0);
    /* nb_thread counters for the overall total as well. */
    nb_counts = ARRAY_SIZE(hwol_stats);
    if (netdev_is_e2e_cache_enabled() && verbose) {
        nb_counts += ARRAY_SIZE(e2e_stats);
    }
    stats->size = (nb_thread + 1) * nb_counts;
    stats->counters = xcalloc(stats->size, sizeof *stats->counters);

    memset(total_nos, 0, sizeof total_nos);

    dp_netdev_port_rdlock(dp);
    HMAP_FOR_EACH (port, node, &dp->ports) {
        memset(per_port_nos, 0, sizeof per_port_nos);
        /* Do not abort on read error from a port, just report 0. */
        if (!netdev_offload_get_stats(port->netdev, per_port_nos)) {
            for (i = 0; i < nb_thread; i++) {
                netdev_offload_stats_add(&total_nos[i], per_port_nos[i]);
            }
        }
    }
    ovs_rwlock_unlock(&dp->port_rwlock);

    DP_NETDEV_OFFLOAD_FOREACH_THREAD (thread, tid) {
        uint64_t counts[DP_NETDEV_STATS_TOTAL_COUNTS];
        uint64_t *e2e_counts = &counts[DP_NETDEV_HW_OFFLOADS_STATS_LAST];
        size_t idx = (tid + 1) * nb_counts;
        struct e2e_cache_stats *cur_e2e_stats;

        memset(counts, 0, sizeof counts);
        counts[DP_NETDEV_HW_OFFLOADS_STATS_INSERTED] =
            total_nos[tid].n_inserted;

        atomic_read_relaxed(&thread->enqueued_offload,
                            &counts[DP_NETDEV_HW_OFFLOADS_STATS_ENQUEUED_OFFLOADS]);
        atomic_read_relaxed(&thread->ct_uni_dir_connections,
                            &counts[DP_NETDEV_HW_OFFLOADS_STATS_CT_UNI_DIR_CONNS]);
        atomic_read_relaxed(&thread->ct_bi_dir_connections,
                            &counts[DP_NETDEV_HW_OFFLOADS_STATS_CT_BI_DIR_CONNS]);
        counts[DP_NETDEV_HW_OFFLOADS_STATS_CT_BI_DIR_CONNS] +=
            total_nos[tid].n_conns / 2;
        counts[DP_NETDEV_HW_OFFLOADS_STATS_LAT_CMA_MEAN] =
            mov_avg_cma(&thread->cma);
        counts[DP_NETDEV_HW_OFFLOADS_STATS_LAT_CMA_STDDEV] =
            mov_avg_cma_std_dev(&thread->cma);
        counts[DP_NETDEV_HW_OFFLOADS_STATS_LAT_EMA_MEAN] =
            mov_avg_ema(&thread->ema);
        counts[DP_NETDEV_HW_OFFLOADS_STATS_LAT_EMA_STDDEV] =
            mov_avg_ema_std_dev(&thread->ema);
        cur_e2e_stats = &thread->e2e_stats;
        e2e_counts[DP_NETDEV_E2E_STATS_GENERATED_TRCS] =
            atomic_count_get(&cur_e2e_stats->generated_trcs);
        e2e_counts[DP_NETDEV_E2E_STATS_PROCESSED_TRCS] =
            cur_e2e_stats->processed_trcs;
        e2e_counts[DP_NETDEV_E2E_STATS_DISCARRDED_TRCS] =
            atomic_count_get(&cur_e2e_stats->discarded_trcs);
        e2e_counts[DP_NETDEV_E2E_STATS_ABORTED_TRCS] =
            atomic_count_get(&cur_e2e_stats->aborted_trcs);
        e2e_counts[DP_NETDEV_E2E_STATS_THROTTLED_TRCS] =
            atomic_count_get(&cur_e2e_stats->throttled_trcs);
        e2e_counts[DP_NETDEV_E2E_STATS_QUEUE_TRCS] =
            atomic_count_get(&cur_e2e_stats->queue_trcs);
        e2e_counts[DP_NETDEV_E2E_STATS_OVERFLOW_TRCS] =
            atomic_count_get(&cur_e2e_stats->overflow_trcs);
        e2e_counts[DP_NETDEV_E2E_STATS_FLOW_ADDS] =
            atomic_count_get(&cur_e2e_stats->flow_add_msgs);
        e2e_counts[DP_NETDEV_E2E_STATS_FLOW_DELS] =
            atomic_count_get(&cur_e2e_stats->flow_del_msgs);
        e2e_counts[DP_NETDEV_E2E_STATS_FLOW_FLUSHS] =
            cur_e2e_stats->flush_flow_msgs;
        e2e_counts[DP_NETDEV_E2E_STATS_SUC_MERGES] =
            cur_e2e_stats->succ_merged_flows;
        e2e_counts[DP_NETDEV_E2E_STATS_REJ_MERGES] =
            cur_e2e_stats->merge_rej_flows;
        e2e_counts[DP_NETDEV_E2E_STATS_HW_ADD_E2E_FLOWS] =
            cur_e2e_stats->add_merged_flow_hw;
        e2e_counts[DP_NETDEV_E2E_STATS_HW_DEL_E2E_FLOWS] =
            cur_e2e_stats->del_merged_flow_hw;
        e2e_counts[DP_NETDEV_E2E_STATS_MERGED_FLOWS] =
            atomic_count_get(&merged_flows_map_count);
        e2e_counts[DP_NETDEV_E2E_STATS_DB_FLOWS] =
            atomic_count_get(&flows_map_count);
        e2e_counts[DP_NETDEV_E2E_STATS_CT_MT_ADDS] =
            cur_e2e_stats->add_ct_mt_flow_hw;
        e2e_counts[DP_NETDEV_E2E_STATS_CT_MT_DELS] =
            cur_e2e_stats->del_ct_mt_flow_hw;
        e2e_counts[DP_NETDEV_E2E_STATS_FAILED_CT_MT_ADDS] =
            cur_e2e_stats->add_ct_mt_flow_err;
        e2e_counts[DP_NETDEV_E2E_STATS_FAILED_CT_MT_DELS] =
            cur_e2e_stats->del_ct_mt_flow_err;
        e2e_counts[DP_NETDEV_E2E_STATS_SUC_CT2CT_MERGES] =
            cur_e2e_stats->succ_ct2ct_merges;
        e2e_counts[DP_NETDEV_E2E_STATS_REJ_CT2CT_MERGES] =
            cur_e2e_stats->rej_ct2ct_merges;
        e2e_counts[DP_NETDEV_E2E_STATS_CT2CT_ADDS] =
            cur_e2e_stats->add_ct2ct_flows;
        e2e_counts[DP_NETDEV_E2E_STATS_CT2CT_DELS] =
            cur_e2e_stats->del_ct2ct_flows;

        for (i = 0; i < nb_counts; i++) {
            cur_stats = i < DP_NETDEV_HW_OFFLOADS_STATS_LAST
                        ? &hwol_stats[i]
                        : &e2e_stats[i - DP_NETDEV_HW_OFFLOADS_STATS_LAST];
            snprintf(stats->counters[idx + i].name,
                     sizeof(stats->counters[idx + i].name),
                     "  [%3u] %s", tid, cur_stats->name);
            stats->counters[idx + i].value = counts[i];
            cur_stats->total += counts[i];
        }
        e2e_stats[DP_NETDEV_E2E_STATS_MERGED_FLOWS].total =
            atomic_count_get(&merged_flows_map_count);
        e2e_stats[DP_NETDEV_E2E_STATS_DB_FLOWS].total =
            atomic_count_get(&flows_map_count);
    }

    /* Do an average of the average for the aggregate. */
    hwol_stats[DP_NETDEV_HW_OFFLOADS_STATS_LAT_CMA_MEAN].total /= nb_thread;
    hwol_stats[DP_NETDEV_HW_OFFLOADS_STATS_LAT_CMA_STDDEV].total /= nb_thread;
    hwol_stats[DP_NETDEV_HW_OFFLOADS_STATS_LAT_EMA_MEAN].total /= nb_thread;
    hwol_stats[DP_NETDEV_HW_OFFLOADS_STATS_LAT_EMA_STDDEV].total /= nb_thread;

    for (i = 0; i < nb_counts; i++) {
        cur_stats = i < DP_NETDEV_HW_OFFLOADS_STATS_LAST
                    ? &hwol_stats[i]
                    : &e2e_stats[i - DP_NETDEV_HW_OFFLOADS_STATS_LAST];
        snprintf(stats->counters[i].name, sizeof(stats->counters[i].name),
                 "  Total %s", cur_stats->name);
        stats->counters[i].value = cur_stats->total;
    }

    return 0;
}

/* Equivalent to 'dpif_is_netdev' but usable in the
 * metrics context. */
static bool
metrics_dpif_is_netdev(void *it)
{
    return dpif_is_netdev(it);
}

METRICS_COND(foreach_dpif, foreach_dpif_netdev,
             metrics_dpif_is_netdev);

METRICS_COND(foreach_dpif_netdev, foreach_dpif_netdev_ext,
             metrics_ext_enabled);

static bool
dpif_netdev_offload_enabled(void *it OVS_UNUSED)
{
    return netdev_is_flow_api_enabled();
}

METRICS_COND(foreach_dpif_netdev, dpif_netdev_offload,
             dpif_netdev_offload_enabled);

struct hw_offload_it {
    struct dp_netdev *dp;
    struct dp_offload_thread *thread;
    unsigned int tid;
};

static void
do_foreach_hw_offload_threads(metrics_visitor_fn visitor,
                              struct metrics_visitor_context *ctx,
                              struct metrics_node *node,
                              struct metrics_label *labels,
                              size_t n OVS_UNUSED)
{
    struct dp_offload_thread *thread;
    struct hw_offload_it it;
    unsigned int tid;
    char id[50];

    it.dp = get_dp_netdev(ctx->it);
    labels[0].value = id;
    DP_NETDEV_OFFLOAD_FOREACH_THREAD (thread, tid) {
        snprintf(id, sizeof id, "%u", tid);
        it.thread = thread;
        it.tid = tid;
        ctx->it = &it;
        visitor(ctx, node);
    }
}

METRICS_COLLECTION(dpif_netdev_offload, foreach_hw_offload_threads,
                   do_foreach_hw_offload_threads, "thread_num");
METRICS_COND(foreach_hw_offload_threads, foreach_hw_offload_threads_dbg,
             metrics_dbg_enabled);

enum {
    HWOL_METRICS_ENQUEUED,
    HWOL_METRICS_INSERTED,
    HWOL_METRICS_CT_UNIDIR,
    HWOL_METRICS_CT_BIDIR,
    HWOL_METRICS_N_ENTRIES,
};

#define HWOL_METRICS_ENTRIES \
    [HWOL_METRICS_ENQUEUED] = METRICS_GAUGE(n_enqueued,                  \
        "Number of hardware offload requests waiting to be processed."), \
    [HWOL_METRICS_INSERTED] = METRICS_GAUGE(n_inserted,                  \
        "Number of hardware offload rules currently inserted."),         \
    [HWOL_METRICS_CT_UNIDIR] = METRICS_GAUGE(n_ct_unidir,                \
        "Number of uni-directional connections offloaded in hardware."), \
    [HWOL_METRICS_CT_BIDIR] = METRICS_GAUGE(n_ct_bidir,                  \
        "Number of bi-directional connections offloaded in hardware."),

static void
hw_offload_read_value(double *values, void *_it)
{
    struct netdev_offload_stats per_port_nos[MAX_OFFLOAD_THREAD_NB];
    struct netdev_offload_stats total_nos[MAX_OFFLOAD_THREAD_NB];
    struct hw_offload_it *it = _it;
    unsigned int tid = it->tid;
    struct dp_netdev *dp = it->dp;
    struct dp_offload_thread *t = it->thread;
    struct dp_netdev_port *port;
    uint64_t count;

    atomic_read_relaxed(&t->enqueued_offload, &count);
    values[HWOL_METRICS_ENQUEUED] = count;

    memset(total_nos, 0, sizeof total_nos);
    dp_netdev_port_rdlock(dp);
    HMAP_FOR_EACH (port, node, &dp->ports) {
        memset(per_port_nos, 0, sizeof per_port_nos);
        if (!netdev_offload_get_stats(port->netdev, per_port_nos)) {
            netdev_offload_stats_add(&total_nos[tid], per_port_nos[tid]);
        }
    }
    ovs_rwlock_unlock(&dp->port_rwlock);

    values[HWOL_METRICS_INSERTED] = total_nos[tid].n_inserted;

    atomic_read_relaxed(&t->ct_uni_dir_connections, &count);
    values[HWOL_METRICS_CT_UNIDIR] = count;

    atomic_read_relaxed(&t->ct_bi_dir_connections, &count);
    values[HWOL_METRICS_CT_BIDIR] = total_nos[tid].n_conns / 2 + count;
}

METRICS_ENTRIES(foreach_hw_offload_threads_dbg, hw_offload_threads_dbg_entries,
                "hw_offload", hw_offload_read_value, HWOL_METRICS_ENTRIES);

static struct histogram *
hw_offload_latency_get(void *_it)
{
    struct hw_offload_it *it = _it;

    return &it->thread->latency;
}

METRICS_HISTOGRAM(foreach_hw_offload_threads_dbg, hw_offload_latency,
                  "Latency in milliseconds between an offload request and its "
                  "completion.", hw_offload_latency_get);

static void
do_foreach_hw_offload_types(metrics_visitor_fn visitor,
                            struct metrics_visitor_context *ctx,
                            struct metrics_node *node,
                            struct metrics_label *labels,
                            size_t n OVS_UNUSED)
{
    const char *hw_offload_type_names[] = {
        [DP_OFFLOAD_FLOW] = "flow",
        [DP_OFFLOAD_FLUSH] = "flush",
        [DP_OFFLOAD_CONN] = "conn",
        [DP_OFFLOAD_STATS_CLEAR] = "stats_clear",
    };
    struct hw_offload_it *ctx_it = ctx->it;

    for (int i = 0; i < DP_OFFLOAD_TYPE_NUM; i++) {
        labels[0].value = hw_offload_type_names[i];
        ctx->it = &ctx_it->thread->queue_metrics[i];
        visitor(ctx, node);
    }
}

/* Iterates on offload (thread x type). */
METRICS_COLLECTION(foreach_hw_offload_threads_dbg, foreach_hw_offload_types,
                   do_foreach_hw_offload_types, "type");

static struct histogram *
hw_offload_queue_sojourn_time_get(void *_it)
{
    struct dp_offload_queue_metrics *m = _it;

    return &m->sojourn_time;
}

METRICS_HISTOGRAM(foreach_hw_offload_types, hw_offload_queue_sojourn_time,
                  "Distribution of sojourn time for an offload request "
                  "in milliseconds", hw_offload_queue_sojourn_time_get);

static struct histogram *
hw_offload_queue_wait_time_get(void *_it)
{
    struct dp_offload_queue_metrics *m = _it;

    return &m->wait_time;
}

METRICS_HISTOGRAM(foreach_hw_offload_types, hw_offload_queue_wait_time,
                  "Distribution of wait time for an offload request "
                  "in milliseconds", hw_offload_queue_wait_time_get);

static struct histogram *
hw_offload_queue_service_time_get(void *_it)
{
    struct dp_offload_queue_metrics *m = _it;

    return &m->service_time;
}

METRICS_HISTOGRAM(foreach_hw_offload_types, hw_offload_queue_service_time,
                  "Distribution of service time for an offload request "
                  "in microseconds", hw_offload_queue_service_time_get);

static void
datapath_hw_offload_read_value(double *values, void *_dp)
{
    double t_values[HWOL_METRICS_N_ENTRIES];
    struct dp_offload_thread *thread;
    struct hw_offload_it it;
    size_t i;

    for (i = 0; i < HWOL_METRICS_N_ENTRIES; i++) {
        values[i] = 0.0;
    }

    it.dp = get_dp_netdev(_dp);
    DP_NETDEV_OFFLOAD_FOREACH_THREAD (thread, it.tid) {
        it.thread = thread;
        hw_offload_read_value(t_values, &it);
        for (i = 0; i < HWOL_METRICS_N_ENTRIES; i++) {
            values[i] += t_values[i];
        }
    }
}

METRICS_ENTRIES(dpif_netdev_offload, datapath_hw_offload_entries,
                "datapath_hw_offload", datapath_hw_offload_read_value,
                HWOL_METRICS_ENTRIES);

int
dpif_netdev_offload_stats_clear(struct dpif *dpif OVS_UNUSED)
{
    struct dp_netdev *dp = get_dp_netdev(dpif);
    struct dp_offload_thread *thread;

    if (!netdev_is_flow_api_enabled()) {
        return EINVAL;
    }

    DP_NETDEV_OFFLOAD_FOREACH_THREAD_NO_MAIN (thread) {
        struct dp_offload_thread_item *item;

        item = xmalloc(sizeof *item);
        item->type = DP_OFFLOAD_STATS_CLEAR;
        item->dp = dp;
        item->timestamp = time_usec();

        dp_netdev_offload_thread_enqueue(thread, item);
    }

    return 0;
}

void
dpif_netdev_set_static_config_ct_add_queue_size(const struct smap *other_config)
{
    if (conntrack_offload_is_enabled()) {
        offload_ct_add_queue_size =
            smap_get_uint(other_config, "hw-offload-ct-add-queue-size",
                          CT_ADD_DEFAULT_QUEUE_SIZE);
        if (offload_ct_add_queue_size == 0) {
            offload_ct_add_queue_size = CT_ADD_DEFAULT_QUEUE_SIZE;
            VLOG_WARN("The size of hw-offload-ct-add-queue-size must be "
                      "greater than 0");
        } else if (conntrack_offload_size() < offload_ct_add_queue_size) {
            offload_ct_add_queue_size = conntrack_offload_size();
            VLOG_INFO("Limiting hw-offload-ct-add-queue-size to the "
                      "conntrack offload size %u",
                      offload_ct_add_queue_size);
        }
        VLOG_INFO("hw-offload-ct-add-queue-size = %"PRIi32,
                  offload_ct_add_queue_size);
    } else {
        offload_ct_add_queue_size = 0;
    }
}

void
dpif_netdev_set_config_e2e_cache(const struct smap *other_config)
{
    dp_netdev_e2e_cache_enabled = netdev_is_e2e_cache_enabled();
    dp_netdev_e2e_cache_size = netdev_get_e2e_cache_size();
    if (dp_netdev_e2e_cache_enabled) {
        static bool done = false;
        int i_value = smap_get_int(other_config, "e2e-cache-trace-q-size",
                                   E2E_CACHE_MAX_TRACE_Q_SIZE);
        if (i_value < 0) {
            i_value = 0;
        }
        if (!done || dp_netdev_e2e_cache_trace_q_size != (uint32_t) i_value) {
            dp_netdev_e2e_cache_trace_q_size = (uint32_t) i_value;
            if (dp_netdev_e2e_cache_trace_q_size) {
                VLOG_INFO("E2E cache trace queue size %u",
                        dp_netdev_e2e_cache_trace_q_size);
            } else {
                VLOG_INFO("E2E cache trace queue unlimited");
            }
            done = true;
        }
    }
}

static inline bool
e2e_cache_set_action_is_valid(struct nlattr *a)
{
    const struct nlattr *set_action = nl_attr_get(a);
    const size_t set_len = nl_attr_get_size(a);
    const struct nlattr *sa;
    unsigned int sleft;

    NL_ATTR_FOR_EACH (sa, sleft, set_action, set_len) {
        enum ovs_key_attr type = nl_attr_type(sa);

        if (!(type == OVS_KEY_ATTR_ETHERNET ||
              type == OVS_KEY_ATTR_IPV4 ||
              type == OVS_KEY_ATTR_IPV6 ||
              type == OVS_KEY_ATTR_TCP ||
              type == OVS_KEY_ATTR_UDP)) {
            VLOG_DBG("Unsupported set action type %d", type);
            /* TODO: add statistic counter */
            return false;
        }
    }
    return true;
}

static inline bool
e2e_cache_flows_are_valid(struct e2e_cache_ovs_flow **netdev_flows,
                          uint16_t num)
{
    struct e2e_cache_ovs_flow *flow;
    const struct match *match;
    unsigned int left;
    struct nlattr *a;
    uint16_t i;

    for (i = 0; i < num; i++) {
        flow = netdev_flows[i];
        if (flow->offload_state != E2E_OL_STATE_FLOW) {
            continue;
        }

        match = &flow->match[0];
        /* validate match */
        if ((match->flow.ipv6_label & match->wc.masks.ipv6_label) ||
            (match->flow.nw_tos & match->wc.masks.nw_tos) ||
            (match->flow.tcp_flags & match->wc.masks.tcp_flags) ||
            (match->flow.igmp_group_ip4 & match->wc.masks.igmp_group_ip4)) {
            /* TODO: add statistic counter */
            return false;
        }

        /* validate actions */
        NL_ATTR_FOR_EACH (a, left, flow->actions, flow->actions_size) {
            enum ovs_action_attr type = nl_attr_type(a);
            if (type == OVS_ACTION_ATTR_USERSPACE ||
                type == OVS_ACTION_ATTR_HASH ||
                type == OVS_ACTION_ATTR_TRUNC ||
                type == OVS_ACTION_ATTR_PUSH_NSH ||
                type == OVS_ACTION_ATTR_POP_NSH ||
                type == OVS_ACTION_ATTR_CT_CLEAR ||
                type == OVS_ACTION_ATTR_CHECK_PKT_LEN ||
                type == OVS_ACTION_ATTR_SAMPLE ||
                ((type == OVS_ACTION_ATTR_OUTPUT ||
                  type == OVS_ACTION_ATTR_CLONE) &&
                 left > NLA_ALIGN(a->nla_len)) ||
                ((type == OVS_ACTION_ATTR_SET ||
                  type == OVS_ACTION_ATTR_SET_MASKED) &&
                 !e2e_cache_set_action_is_valid(a))) {
                 /* TODO: add statistic counter */
                return false;
            }
        }
    }
    return true;
}

#define e2e_save_set_attr(mfield, field, flag)                         \
    ovs_assert(key);                                                   \
    if (mask) {                                                        \
        if (!is_all_zeros(&mask->field, sizeof mask->field)) {         \
            if (!is_all_ones(&mask->field, sizeof mask->field)) {      \
                VLOG_DBG_RL(&rl, "HW partial mask is not supported");  \
            }                                                          \
            merged->flags |= flag;                                     \
            merged->mfield.field = key->field;                         \
        }                                                              \
    } else if (!is_all_zeros(&key->field, sizeof key->field)) {        \
        merged->flags |= flag;                                         \
        merged->mfield.field = key->field;                             \
    }

static inline void
e2e_cache_save_set_actions(struct e2e_cache_merged_set *merged, bool masked,
                           const struct nlattr *set_action,
                           const size_t set_len)
{
    const struct nlattr *sa;
    unsigned int sleft;
    static struct vlog_rate_limit rl = VLOG_RATE_LIMIT_INIT(10, 10);

    NL_ATTR_FOR_EACH (sa, sleft, set_action, set_len) {
        if (nl_attr_type(sa) == OVS_KEY_ATTR_ETHERNET) {
            const struct ovs_key_ethernet *key = nl_attr_get(sa);
            const struct ovs_key_ethernet *mask = masked ? key + 1 : NULL;

            e2e_save_set_attr(eth, eth_src, E2E_SET_ETH_SRC);
            e2e_save_set_attr(eth, eth_dst, E2E_SET_ETH_DST);
        } else if (nl_attr_type(sa) == OVS_KEY_ATTR_IPV4) {
            const struct ovs_key_ipv4 *key = nl_attr_get(sa);
            const struct ovs_key_ipv4 *mask = masked ? key + 1 : NULL;

            e2e_save_set_attr(ipv4, ipv4_src, E2E_SET_IPV4_SRC);
            e2e_save_set_attr(ipv4, ipv4_dst, E2E_SET_IPV4_DST);
            e2e_save_set_attr(ipv4, ipv4_ttl, E2E_SET_IPV4_TTL);
        } else if (nl_attr_type(sa) == OVS_KEY_ATTR_IPV6) {
            const struct ovs_key_ipv6 *key = nl_attr_get(sa);
            const struct ovs_key_ipv6 *mask = masked ? key + 1 : NULL;

            e2e_save_set_attr(ipv6, ipv6_src, E2E_SET_IPV6_SRC);
            e2e_save_set_attr(ipv6, ipv6_dst, E2E_SET_IPV6_DST);
            e2e_save_set_attr(ipv6, ipv6_hlimit, E2E_SET_IPV6_HLMT);
        } else if (nl_attr_type(sa) == OVS_KEY_ATTR_TCP) {
            const struct ovs_key_tcp *key = nl_attr_get(sa);
            const struct ovs_key_tcp *mask = masked ? key + 1 : NULL;

            e2e_save_set_attr(tcp, tcp_src, E2E_SET_TCP_SRC);
            e2e_save_set_attr(tcp, tcp_dst, E2E_SET_TCP_DST);
        } else if (nl_attr_type(sa) == OVS_KEY_ATTR_UDP) {
            const struct ovs_key_udp *key = nl_attr_get(sa);
            const struct ovs_key_udp *mask = masked ? key + 1 : NULL;

            e2e_save_set_attr(udp, udp_src, E2E_SET_UDP_SRC);
            e2e_save_set_attr(udp, udp_dst, E2E_SET_UDP_DST);
        }
    }
}

#define e2e_construct_set_attr(mfield, field, flag)       \
    if (merged->flags & flag) {                           \
        key->field = merged->mfield.field;                \
        memset(&mask->field, 0xFF, sizeof mask->field);   \
    }

static inline void
e2e_cache_attach_merged_set_action(struct ofpbuf *buf, size_t tnl_offset,
                                   struct e2e_cache_merged_set *merged)
{
    size_t offset;
    struct ofpbuf tmpbuf;

    ofpbuf_init(&tmpbuf, 0);
    offset = nl_msg_start_nested(&tmpbuf, OVS_ACTION_ATTR_SET_MASKED);
    if (merged->flags & E2E_SET_ETH) {
        struct ovs_key_ethernet *key = NULL;
        struct ovs_key_ethernet *mask = NULL;

        key = nl_msg_put_unspec_zero(&tmpbuf, OVS_KEY_ATTR_ETHERNET,
                                     2 * sizeof *key);
        mask = key + 1;
        e2e_construct_set_attr(eth, eth_src, E2E_SET_ETH_SRC);
        e2e_construct_set_attr(eth, eth_dst, E2E_SET_ETH_DST);
    }
    if (merged->flags & E2E_SET_IPV4) {
        struct ovs_key_ipv4 *key = NULL;
        struct ovs_key_ipv4 *mask = NULL;

        key = nl_msg_put_unspec_zero(&tmpbuf, OVS_KEY_ATTR_IPV4,
                                     2 * sizeof *key);
        mask = key + 1;
        e2e_construct_set_attr(ipv4, ipv4_src, E2E_SET_IPV4_SRC);
        e2e_construct_set_attr(ipv4, ipv4_dst, E2E_SET_IPV4_DST);
        e2e_construct_set_attr(ipv4, ipv4_ttl, E2E_SET_IPV4_TTL);
    }
    if (merged->flags & E2E_SET_IPV6) {
        struct ovs_key_ipv6 *key = NULL;
        struct ovs_key_ipv6 *mask = NULL;

        key = nl_msg_put_unspec_zero(&tmpbuf, OVS_KEY_ATTR_IPV6,
                                     2 * sizeof *key);
        mask = key + 1;
        e2e_construct_set_attr(ipv6, ipv6_src, E2E_SET_IPV6_SRC);
        e2e_construct_set_attr(ipv6, ipv6_dst, E2E_SET_IPV6_DST);
        e2e_construct_set_attr(ipv6, ipv6_hlimit, E2E_SET_IPV6_HLMT);
    }
    if (merged->flags & E2E_SET_TCP) {
        struct ovs_key_tcp *key = NULL;
        struct ovs_key_tcp *mask = NULL;

        key = nl_msg_put_unspec_zero(&tmpbuf, OVS_KEY_ATTR_TCP,
                                     2 * sizeof *key);
        mask = key + 1;
        e2e_construct_set_attr(tcp, tcp_src, E2E_SET_TCP_SRC);
        e2e_construct_set_attr(tcp, tcp_dst, E2E_SET_TCP_DST);
    }
    if (merged->flags & E2E_SET_UDP) {
        struct ovs_key_udp *key = NULL;
        struct ovs_key_udp *mask = NULL;

        key = nl_msg_put_unspec_zero(&tmpbuf, OVS_KEY_ATTR_UDP,
                                     2 * sizeof *key);
        mask = key + 1;
        e2e_construct_set_attr(udp, udp_src, E2E_SET_UDP_SRC);
        e2e_construct_set_attr(udp, udp_dst, E2E_SET_UDP_DST);
    }
    nl_msg_end_nested(&tmpbuf, offset);
    /* insert the set action after tnl_pop in the buf */
    ofpbuf_insert(buf, tnl_offset, tmpbuf.data, tmpbuf.size);
}

static void
e2e_cache_merge_actions(struct e2e_cache_ovs_flow **netdev_flows,
                        uint16_t num, struct ofpbuf *buf,
                        const struct nlattr **last_ct)
{
    uint16_t i = 0;
    unsigned int left;
    const struct nlattr *a;
    uint16_t num_set = 0;
    struct e2e_cache_merged_set merged_set;
    size_t tnl_offset = 0;
    size_t pad;

    memset(&merged_set, 0, sizeof merged_set);
    for (i = 0; i < num; i++) {
        if (i > 0 && netdev_flows[i]->offload_state != E2E_OL_STATE_FLOW &&
            netdev_flows[i - 1]->offload_state != E2E_OL_STATE_FLOW) {
            continue;
        }
        NL_ATTR_FOR_EACH (a, left, netdev_flows[i]->actions,
                          netdev_flows[i]->actions_size) {
            enum ovs_action_attr type = nl_attr_type(a);

            if (type == OVS_ACTION_ATTR_CT && last_ct) {
                *last_ct = a;
            }
            if (type == OVS_ACTION_ATTR_CT ||
                type == OVS_ACTION_ATTR_RECIRC) {
                continue;
            }
            if (type == OVS_ACTION_ATTR_SET ||
                type == OVS_ACTION_ATTR_SET_MASKED) {
                const struct nlattr *set_action = nl_attr_get(a);
                const size_t set_len = nl_attr_get_size(a);
                bool masked = (type == OVS_ACTION_ATTR_SET_MASKED);

                e2e_cache_save_set_actions(&merged_set, masked,
                                           set_action, set_len);
                num_set++;
                continue;
            }
            if (type == OVS_ACTION_ATTR_TUNNEL_POP) {
                tnl_offset = buf->size + a->nla_len;
            }
            pad = PAD_SIZE(a->nla_len, NLMSG_ALIGNTO);
            ofpbuf_put(buf, a, a->nla_len + pad);
        }
    }
    if (num_set) {
        e2e_cache_attach_merged_set_action(buf, tnl_offset, &merged_set);
    }
}

#define merge_flow_match(field, src, dst)                   \
    if (!is_all_zeros(&src->wc.masks.field,                 \
                      sizeof src->wc.masks.field) &&        \
        is_all_zeros(&dst->mask.field,                      \
                     sizeof dst->mask.field)) {             \
        memcpy(&dst->spec.field, &src->flow.field,          \
               sizeof src->flow.field);                     \
        memcpy(&dst->mask.field, &src->wc.masks.field,      \
               sizeof src->wc.masks.field);                 \
    }

static void
dp_netdev_fill_ct_match(struct match *match, const struct ct_match *ct_match);

static void
e2e_cache_merge_match(struct e2e_cache_ovs_flow **netdev_flows,
                      uint16_t num, struct merged_match *merged_match)
{
    struct e2e_cache_ovs_flow *flow;
    struct match match_on_stack;
    const struct match *match;
    uint16_t i = 0;

    memset(merged_match, 0, sizeof *merged_match);

    for (i = 0; i < num; i++) {
        flow = netdev_flows[i];
        if (i > 0 && flow->offload_state != E2E_OL_STATE_FLOW &&
            netdev_flows[i - 1]->offload_state != E2E_OL_STATE_FLOW) {
            continue;
        }
        /* parse match */
        if (flow->offload_state == E2E_OL_STATE_FLOW) {
            match = &flow->match[0];
        } else {
            dp_netdev_fill_ct_match(&match_on_stack, &flow->ct_match[0]);
            match = &match_on_stack;
        }
        /* merge in_port */
        merge_flow_match(in_port, match, merged_match);

        /* merge tunnel outer */
        merge_flow_match(tunnel.ip_src, match, merged_match);
        merge_flow_match(tunnel.ip_dst, match, merged_match);
        merge_flow_match(tunnel.ipv6_src, match, merged_match);
        merge_flow_match(tunnel.ipv6_dst, match, merged_match);
        merge_flow_match(tunnel.tun_id, match, merged_match);
        merge_flow_match(tunnel.tp_dst, match, merged_match);

        /* merge inner/non-tnl */
        merge_flow_match(dl_src, match, merged_match);
        merge_flow_match(dl_dst, match, merged_match);
        merge_flow_match(dl_type, match, merged_match);
        merge_flow_match(nw_src, match, merged_match);
        merge_flow_match(nw_dst, match, merged_match);
        merge_flow_match(ipv6_src, match, merged_match);
        merge_flow_match(ipv6_dst, match, merged_match);
        merge_flow_match(nw_frag, match, merged_match);
        merge_flow_match(nw_proto, match, merged_match);
        merge_flow_match(tp_src, match, merged_match);
        merge_flow_match(tp_dst, match, merged_match);
        if (match->flow.vlans[0].tci) {
            merge_flow_match(vlans[0].tci, match, merged_match);
        }
        merge_flow_match(ct_zone, match, merged_match);
    }
}

static int
e2e_cache_merge_flows(struct e2e_cache_ovs_flow **flows,
                      uint16_t num_flows,
                      struct e2e_cache_merged_flow *merged_flow,
                      struct ofpbuf *merged_actions)
{
    unsigned int tid = netdev_offload_thread_id();
    struct e2e_cache_stats *e2e_stats;
    struct match match;

    e2e_stats = &dp_offload_threads[tid].e2e_stats;
    if (!e2e_cache_flows_are_valid(flows, num_flows)) {
        e2e_stats->merge_rej_flows++;
        return -1;
    }
    e2e_cache_merge_match(flows, num_flows, &merged_flow->merged_match);
    merged_flow->merged_match.mask.ct_zone = 0;
    merged_match_to_match(&match, &merged_flow->merged_match);
    dp_netdev_get_mega_ufid(&match, &merged_flow->ufid);
    uuid_set_bits_v4((struct uuid *) &merged_flow->ufid, UUID_ATTR_3);
    e2e_cache_merge_actions(flows, num_flows, merged_actions, NULL);
    if (OVS_UNLIKELY(merged_actions->size < sizeof(struct nlattr))) {
        e2e_stats->merge_rej_flows++;
        return -1;
    }
    merged_flow->flow_mark = INVALID_FLOW_MARK;
    e2e_stats->succ_merged_flows++;
    return 0;
}

static int
ct2ct_merge_flows(struct e2e_cache_ovs_flow **flows,
                  uint16_t num_flows,
                  struct e2e_cache_merged_flow *merged_flow,
                  struct ofpbuf *merged_actions)
{
    unsigned int tid = netdev_offload_thread_id();
    const struct nlattr *last_ct = NULL;
    struct e2e_cache_stats *e2e_stats;
    struct match match;

    ovs_assert(num_flows > 4);

    e2e_stats = &dp_offload_threads[tid].e2e_stats;

    /* Trace is:
     * 0              Flow1
     * 1              CT1
     * ...
     * num_flows - 4  Flow(N-1)
     * num_flows - 3  CTN
     * num_flows - 2  CTN-peer
     * num_flows - 1  FlowN
     *
     * Matches are merged from CT1 (in [1]) until CTN included (in
     * [num_flows - 3]).
     * Actions are merged from Flow1 (in [0]) until CTN included (in
     * [num_flows - 3]).
     * The mark should be of Flow(N-1) (in [num_flows - 4]).
     */
    e2e_cache_merge_match(&flows[1], num_flows - 3,
                          &merged_flow->merged_match);
    merged_match_to_match(&match, &merged_flow->merged_match);
    dp_netdev_get_mega_ufid(&match, &merged_flow->ufid);
    uuid_set_bits_v4((struct uuid *) &merged_flow->ufid, UUID_ATTR_4);
    e2e_cache_merge_actions(flows, num_flows - 2, merged_actions, &last_ct);
    if (!last_ct) {
        return -1;
    }
    ofpbuf_put(merged_actions, last_ct, last_ct->nla_len);
    if (OVS_UNLIKELY(merged_actions->size < sizeof(struct nlattr))) {
        ofpbuf_uninit(merged_actions);
        e2e_stats->rej_ct2ct_merges++;
        return -1;
    }
    /* Set the mark to the last flow in the CT2CT section. */
    merged_flow->flow_mark =
        megaflow_to_mark_find(&flows[num_flows - 4]->ufid);
    if (merged_flow->flow_mark == INVALID_FLOW_MARK) {
        ofpbuf_uninit(merged_actions);
        e2e_stats->rej_ct2ct_merges++;
        return -1;
    }
    e2e_stats->succ_ct2ct_merges++;
    return 0;
}

static void
dp_netdev_fill_ct_match(struct match *match, const struct ct_match *ct_match)
{
    memset(match, 0, sizeof *match);
    if (ct_match->key.dl_type == htons(ETH_TYPE_IP)) {
        /* Fill in ipv4 5-tuples */
        match->flow.nw_src = ct_match->key.src.addr.ipv4;
        match->flow.nw_dst = ct_match->key.dst.addr.ipv4;
        match->wc.masks.nw_src = OVS_BE32_MAX;
        match->wc.masks.nw_dst = OVS_BE32_MAX;
    } else {
        /* Fill in ipv6 5-tuples */
        memcpy(&match->flow.ipv6_src,
               &ct_match->key.src.addr.ipv6,
               sizeof match->flow.ipv6_src);
        memcpy(&match->flow.ipv6_dst,
               &ct_match->key.dst.addr.ipv6,
               sizeof match->flow.ipv6_dst);
        memset(&match->wc.masks.ipv6_src, 0xFF,
                sizeof match->wc.masks.ipv6_src);
        memset(&match->wc.masks.ipv6_dst, 0xFF,
                sizeof match->wc.masks.ipv6_dst);
    }
    match->flow.dl_type = ct_match->key.dl_type;
    match->flow.nw_proto = ct_match->key.nw_proto;
    match->wc.masks.dl_type = OVS_BE16_MAX;
    match->wc.masks.nw_proto = UINT8_MAX;
    if (match->flow.nw_proto == IPPROTO_TCP) {
        match->wc.masks.tcp_flags = htons(TCP_SYN | TCP_RST | TCP_FIN);
    }
    if (match->flow.nw_proto == IPPROTO_TCP ||
        match->flow.nw_proto == IPPROTO_UDP) {
        match->flow.tp_src = ct_match->key.src.port;
        match->flow.tp_dst = ct_match->key.dst.port;
        match->wc.masks.tp_src = OVS_BE16_MAX;
        match->wc.masks.tp_dst = OVS_BE16_MAX;
    }
    match->flow.ct_zone = ct_match->key.zone;
    match->wc.masks.ct_zone = UINT16_MAX;
    match->flow.in_port.odp_port = ct_match->odp_port;
    match->wc.masks.in_port.odp_port = u32_to_odp(UINT32_MAX);
}

static void
dp_netdev_set_ct_mark_labels_attr(struct ofpbuf *buf,
                                  uint16_t attr,
                                  void *offload_key,
                                  size_t size)
{
    uint8_t *key, *mask;

    key = nl_msg_put_unspec_zero(buf, attr, 2 * size);
    mask = key + size;
    memcpy(key, offload_key, size);
    memset(mask, 0xFF, size);
}

static void
dp_netdev_create_ct_actions(struct ofpbuf *buf,
                            struct ct_flow_offload_item *offload)
{
    size_t offset;
    char helper[] = "offl,st(0x  ),id_key(0x                )";
    char s[17];
    char *end;

    if (offload->nat.mod_flags) {
        offset = nl_msg_start_nested(buf, OVS_ACTION_ATTR_SET_MASKED);
        if (offload->ct_match.key.dl_type == htons(ETH_TYPE_IP)) {
            struct ovs_key_ipv4 *ipv4_key = NULL, *ipv4_mask = NULL;

            if (offload->nat.mod_flags & NAT_ACTION_SRC ||
                offload->nat.mod_flags & NAT_ACTION_DST) {
                ipv4_key = nl_msg_put_unspec_zero(buf, OVS_KEY_ATTR_IPV4,
                                                  2 * sizeof *ipv4_key);
                ipv4_mask = ipv4_key + 1;
            }
            if (offload->nat.mod_flags & NAT_ACTION_SRC) {
                ipv4_key->ipv4_src = offload->nat.key.src.addr.ipv4;
                ipv4_mask->ipv4_src = OVS_BE32_MAX;
            }
            if (offload->nat.mod_flags & NAT_ACTION_DST) {
                ipv4_key->ipv4_dst = offload->nat.key.dst.addr.ipv4;
                ipv4_mask->ipv4_dst = OVS_BE32_MAX;
            }
        } else {
            struct ovs_key_ipv6 *ipv6_key = NULL, *ipv6_mask = NULL;

            if (offload->nat.mod_flags & NAT_ACTION_SRC ||
                offload->nat.mod_flags & NAT_ACTION_DST) {
                ipv6_key = nl_msg_put_unspec_zero(buf, OVS_KEY_ATTR_IPV6,
                                                  2 * sizeof *ipv6_key);
                ipv6_mask = ipv6_key + 1;
            }
            if (offload->nat.mod_flags & NAT_ACTION_SRC) {
                ipv6_key->ipv6_src = offload->nat.key.src.addr.ipv6;
                memset(&ipv6_mask->ipv6_src, 0xFF, sizeof ipv6_mask->ipv6_src);
            }
            if (offload->nat.mod_flags & NAT_ACTION_DST) {
                ipv6_key->ipv6_dst = offload->nat.key.dst.addr.ipv6;
                memset(&ipv6_mask->ipv6_dst, 0xFF, sizeof ipv6_mask->ipv6_dst);
            }
        }
        if (offload->nat.mod_flags & NAT_ACTION_SRC_PORT ||
            offload->nat.mod_flags & NAT_ACTION_DST_PORT) {
            if (offload->ct_match.key.nw_proto == IPPROTO_TCP) {
                struct ovs_key_tcp *tcp_key, *tcp_mask;

                tcp_key = nl_msg_put_unspec_zero(buf, OVS_KEY_ATTR_TCP,
                                                 2 * sizeof *tcp_key);
                tcp_mask = tcp_key + 1;
                if (offload->nat.mod_flags & NAT_ACTION_SRC_PORT) {
                    tcp_key->tcp_src = offload->nat.key.src.port;
                    tcp_mask->tcp_src = OVS_BE16_MAX;
                }
                if (offload->nat.mod_flags & NAT_ACTION_DST_PORT) {
                    tcp_key->tcp_dst = offload->nat.key.dst.port;
                    tcp_mask->tcp_dst = OVS_BE16_MAX;
                }
            }
            if (offload->ct_match.key.nw_proto == IPPROTO_UDP) {
                struct ovs_key_udp *udp_key, *udp_mask;

                udp_key = nl_msg_put_unspec_zero(buf, OVS_KEY_ATTR_UDP,
                                                 2 * sizeof *udp_key);
                udp_mask = udp_key + 1;
                if (offload->nat.mod_flags & NAT_ACTION_SRC_PORT) {
                    udp_key->udp_src = offload->nat.key.src.port;
                    udp_mask->udp_src = OVS_BE16_MAX;
                }
                if (offload->nat.mod_flags & NAT_ACTION_DST_PORT) {
                    udp_key->udp_dst = offload->nat.key.dst.port;
                    udp_mask->udp_dst = OVS_BE16_MAX;
                }
            }
        }
        nl_msg_end_nested(buf, offset);
    }
    offset = nl_msg_start_nested(buf, OVS_ACTION_ATTR_CT);
    dp_netdev_set_ct_mark_labels_attr(buf, OVS_CT_ATTR_MARK,
                                      &offload->mark_key, sizeof(uint32_t));
    dp_netdev_set_ct_mark_labels_attr(buf, OVS_CT_ATTR_LABELS,
                                      &offload->label_key, sizeof(ovs_u128));
    nl_msg_put_u16(buf, OVS_CT_ATTR_ZONE, offload->ct_match.key.zone);

    end = helper;
    ovs_strcat(helper, sizeof helper, &end, "offl,st(0x");
    ovs_strcat(helper, sizeof helper, &end, u32_to_hex(s, offload->ct_state));
    ovs_strcat(helper, sizeof helper, &end, "),id_key(0x");
    ovs_strcat(helper, sizeof helper, &end, uintptr_to_hex(s,
                                                           offload->ctid_key));
    ovs_strcat(helper, sizeof helper, &end, ")");

    nl_msg_put_string(buf, OVS_CT_ATTR_HELPER, helper);
    nl_msg_end_nested(buf, offset);
}

static int
dp_netdev_ct_offload_add_cb(struct ct_flow_offload_item *ct_offload,
                            struct ct_match *ct_match, struct nlattr *actions,
                            int actions_len)
{
    struct dp_netdev *dp = ct_offload->dp;
    const char *dpif_type_str = dpif_normalize_type(dp->class->type);
    struct offload_info info = { .flow_mark = INVALID_FLOW_MARK, };
    struct netdev *port;
    struct match match;
    int ret;

    port = netdev_ports_get(ct_match->odp_port, dpif_type_str);
    if (OVS_UNLIKELY(!port)) {
        return ENODEV;
    }

    dp_netdev_fill_ct_match(&match, ct_match);

    dp_netdev_port_rdlock_limit(dp, 10);
    if (OVS_UNLIKELY(!VLOG_DROP_DBG((&upcall_rl)))) {
        struct ds ds = DS_EMPTY_INITIALIZER;
        struct ofpbuf key_buf, mask_buf;
        struct odp_flow_key_parms odp_parms = {
            .flow = &match.flow,
            .mask = &match.wc.masks,
            .support = dp_netdev_support,
        };

        ofpbuf_init(&key_buf, 0);
        ofpbuf_init(&mask_buf, 0);

        odp_flow_key_from_flow(&odp_parms, &key_buf);
        odp_parms.key_buf = &key_buf;
        odp_flow_key_from_mask(&odp_parms, &mask_buf);

        ds_put_cstr(&ds, "ct_add: ");
        odp_format_ufid(&ct_offload->ufid, &ds);
        ds_put_cstr(&ds, " ");
        odp_flow_format(key_buf.data, key_buf.size,
                        mask_buf.data, mask_buf.size,
                        NULL, &ds, false);
        ds_put_cstr(&ds, ", actions:");
        format_odp_actions(&ds, actions, actions_len, NULL);

        VLOG_DBG("%s", ds_cstr(&ds));

        ofpbuf_uninit(&key_buf);
        ofpbuf_uninit(&mask_buf);

        ds_destroy(&ds);
    }
    info.is_ct_conn = true;
    info.orig_in_port = ct_match->orig_in_port;
    ret = netdev_flow_put(port, &match, actions, actions_len, &ct_offload->ufid,
                          &info, NULL);
    ovs_rwlock_unlock(&dp->port_rwlock);
    netdev_close(port);

    return ret;
}

typedef int
(*dp_netdev_ct_add_cb)(struct ct_flow_offload_item *ct_offload,
                       struct ct_match *match, struct nlattr *actions,
                       int actions_len);

static int
dp_netdev_ct_add(struct ct_flow_offload_item *ct_offload,
                 dp_netdev_ct_add_cb cb)
{
    struct nlattr *actions;
    size_t actions_size;
    struct ofpbuf buf;
    int ret;

    /* Bypass actions building if the work is already done.
     *
     * When e2e is enabled, the datapath will create the ct_actions and
     * send them ready to the e2e thread. There, if the e2e-cache is not
     * yet full, they will be consumed directly. Otherwise, an offload
     * request will be emitted to the regular offload threads.
     *
     * In this case, those OFL-threads will call again this function,
     * but the actions will already have been created.
     */

    if (!ct_offload->ct_actions_set) {
        ofpbuf_init(&buf, 0);
        dp_netdev_create_ct_actions(&buf, ct_offload);
        actions = ofpbuf_at_assert(&buf, 0, sizeof(struct nlattr));
        actions_size = buf.size;
    } else {
        actions = ct_offload->actions;
        actions_size = ct_offload->actions_size;
    }

    ret = cb(ct_offload, &ct_offload->ct_match, actions, actions_size);

    if (!ct_offload->ct_actions_set) {
        ofpbuf_uninit(&buf);
    }

    return ret;
}

static void
dp_offload_conn(struct dp_offload_thread_item *msg)
{
    struct dp_offload_conn_item *doci = &msg->data->conn;
    struct dp_netdev *dp = conntrack_datapath(doci->ct);
    struct dp_netdev_pmd_thread *pmd = ovsthread_getspecific(dp->per_pmd_key);
    struct batch conns = batch_init_one(doci->conn);

    dp_netdev_offload_conns(pmd, doci->ct, &conns, doci->op);
}

void
dp_netdev_port_rdlock_at(struct dp_netdev *dp, unsigned long long int limit_ms,
                         const char *where)
    OVS_ACQ_RDLOCK(dp->port_rwlock)
{
    static struct vlog_rate_limit rl = VLOG_RATE_LIMIT_INIT(1, 5);
    unsigned long long int start = time_msec();

    if (ovs_rwlock_tryrdlock(&dp->port_rwlock)) {
        const char *holder = dp->port_rwlock.where;
        unsigned long long int elapsed;

        ovs_rwlock_rdlock(&dp->port_rwlock);
        elapsed = time_msec() - start;
        if (elapsed > limit_ms) {
            VLOG_WARN_RL(&rl, "%s: Unreasonably long %llums port_rwlock wait, "
                         "held from %s", where, elapsed, holder);
        }
    }
}

static void
pmd_thread_offload_netdevs(struct dp_netdev_pmd_thread *pmd, odp_port_t port_nos[CT_DIR_NUM],
                           struct netdev *netdevs[CT_DIR_NUM])
{
    for (int i = 0; i < pmd->rx_port_count; i++) {
        struct dp_netdev_port *port = pmd->rx_port_cache[i];

        for (int dir = 0; dir < CT_DIR_NUM; dir++) {
            if (!netdevs[dir]) {
                if (port->port_no == port_nos[dir] &&
                    !port->offload_disabled) {
                    netdevs[dir] = port->netdev;
                }
            }
        }
    }
}

void
pmd_thread_offload_disable(struct dp_netdev_pmd_thread *pmd, struct netdev *netdev)
{
    for (int i = 0; i < pmd->rx_port_count; i++) {
        struct dp_netdev_port *port = pmd->rx_port_cache[i];

        if (port->netdev == netdev) {
            port->offload_disabled = true;
            return;
        }
    }
}

#define DP_PMD_OFFLOAD_UPKEEP_PERIOD_MS (64)
#define DP_OFFLOAD_UPKEEP_PERIOD_MS (256)
/* Number of max-backoff to roughly reach the upkeep period. */
#define DP_OFFLOAD_UPKEEP_N_BACKOFF \
    (DP_OFFLOAD_UPKEEP_PERIOD_MS / DP_NETDEV_OFFLOAD_BACKOFF_MAX)
BUILD_ASSERT_DECL(IS_POW2(DP_OFFLOAD_UPKEEP_N_BACKOFF));

void
pmd_thread_offload_upkeep(struct dp_netdev_pmd_thread *pmd, bool force)
{
    if (!force && pmd->offload_next_upkeep > pmd->ctx.now) {
        return;
    }

    dp_netdev_pmd_idle_end(pmd);
    for (int i = 0; i < pmd->rx_port_count; i++) {
        netdev_offload_upkeep(pmd->rx_port_cache[i]->netdev, true);
    }
    pmd->offload_next_upkeep = pmd->ctx.now + (DP_PMD_OFFLOAD_UPKEEP_PERIOD_MS * 1000);
}

/* Eschew safety analysis as the PMD thread implicitly took ownership
 * of the offload thread by allocating its netdev-offload thread ID.
 * This allocation cannot be expressed in a way safety analysis understands.
 */
unsigned int
pmd_thread_offload_process(struct dp_netdev_pmd_thread *pmd, unsigned int limit)
    OVS_REQUIRES(dp_offload_threads[pmd->offload_thread_id].offload_queue.read_lock)
{
    struct dp_offload_thread *offload_thread;
    struct dp_offload_thread_item *offload;
    struct mpsc_queue_node *node;
    unsigned int n_msgs = 0;

    if (pmd->offload_thread_id == OVSTHREAD_ID_UNSET) {
        return 0;
    }

    offload_thread = &dp_offload_threads[pmd->offload_thread_id];
    MPSC_QUEUE_FOR_EACH_POP (node, &dp_offload_threads[pmd->offload_thread_id].offload_queue) {
        dp_netdev_pmd_idle_end(pmd);
        n_msgs++;
        atomic_count_dec64(&offload_thread->enqueued_offload);
        offload = CONTAINER_OF(node, struct dp_offload_thread_item, node);
        dp_offload_process(offload_thread, offload);
        /* Using limit == 0 means that every message still in the queue
         * must be processed. It is used when terminating the polling thread
         * to clean up requests. */
        if (limit != 0 && n_msgs >= limit) {
            break;
        }
    }

    pmd_thread_offload_upkeep(pmd, false);

    return n_msgs;
}

void
pmd_thread_offload_init(struct dp_netdev_pmd_thread *pmd)
    OVS_ACQUIRES(dp_offload_threads[pmd->offload_thread_id].offload_queue.read_lock)
{
    pmd->offload_thread_id = netdev_offload_thread_id();
    if (pmd->core_id == NON_PMD_CORE_ID) {
        ovs_assert(pmd->offload_thread_id == NETDEV_OFFLOAD_THREAD_MAIN);
    }
    dp_netdev_offload_thread_init(&dp_offload_threads[pmd->offload_thread_id]);
    mpsc_queue_acquire(&dp_offload_threads[pmd->offload_thread_id].offload_queue);
}

void
pmd_thread_offload_uninit(struct dp_netdev_pmd_thread *pmd)
    OVS_RELEASES(dp_offload_threads[pmd->offload_thread_id].offload_queue.read_lock)
{
    pmd_thread_offload_process(pmd, 0);
    mpsc_queue_release(&dp_offload_threads[pmd->offload_thread_id].offload_queue);

    dp_netdev_offload_thread_uninit(&dp_offload_threads[pmd->offload_thread_id]);
    pmd->offload_thread_id = OVSTHREAD_ID_UNSET;
    netdev_offload_thread_uninit();
}

void
dp_netdev_pmd_idle_begin(struct dp_netdev_pmd_thread *pmd)
{
    if (pmd->core_id != NON_PMD_CORE_ID &&
        !pmd->idle) {
        ovsrcu_quiesce_start();
        pmd->idle = true;
    }
}

void
dp_netdev_pmd_idle_end(struct dp_netdev_pmd_thread *pmd)
{
    if (pmd->idle) {
        ovsrcu_quiesce_end();
        pmd->idle = false;
        pmd->next_rcu_quiesce =
            pmd->ctx.now + PMD_RCU_QUIESCE_INTERVAL;
    }
}

static int
dp_netdev_ct_offload_e2e_active(struct ct_flow_offload_item *offload,
                                long long now, long long prev_now)
{
    struct dpif_flow_stats stats;
    const struct dp_netdev *dp;
    struct netdev *netdev;
    int ret;

    if (!conntrack_offload_is_enabled()) {
        return EINVAL;
    }

    dp = offload->dp;
    netdev = netdev_ports_get(offload->ct_match.odp_port,
                              dpif_normalize_type(dp->class->type));
    if (!netdev) {
        return ENODEV;
    }

    memset(&stats, 0, sizeof stats);
    ret = !e2e_cache_get_merged_flows_stats(netdev, NULL, NULL, &offload->ufid, &stats, NULL,
                                            now, prev_now);

    netdev_close(netdev);
    if (ret) {
        return ret;
    }

    return stats.used > prev_now ? 0 : EINVAL;
}

static void
dp_netdev_ct_offload_e2e_add(struct ct_flow_offload_item *offload)
{
    dp_netdev_ct_add(offload, dp_netdev_ct_e2e_add_cb);
}

void
dp_netdev_flow_format(const char *prefix,
                      struct ds *s,
                      const struct dp_netdev_flow *dp_flow)
{
    struct dp_netdev_actions *dp_actions;

    ds_init(s);
    ds_put_format(s, "%s: ", prefix);
    odp_format_ufid(&dp_flow->ufid, s);
    ds_put_cstr(s, " mega_");
    odp_format_ufid(&dp_flow->mega_ufid, s);
    ds_put_cstr(s, " ");

    flow_format(s, &dp_flow->flow, NULL);

    dp_actions = dp_netdev_flow_get_actions(dp_flow);
    ds_put_cstr(s, ", actions:");
    if (dp_actions) {
        format_odp_actions(s, dp_actions->actions, dp_actions->size, NULL);
    } else {
        ds_put_cstr(s, "(nil)");
    }
}

long long int
e2e_cache_flow_db_handle_ufid_msg(struct e2e_cache_ufid_msg *ufid_msg)
{
    long long int enqueue_time_us = 0;

    if (ufid_msg->op == E2E_UFID_MSG_PUT) {
        e2e_cache_flow_db_put(ufid_msg);
    } else if (ufid_msg->op == E2E_UFID_MSG_DEL) {
        e2e_cache_flow_db_del(ufid_msg);
    } else {
        OVS_NOT_REACHED();
    }

    enqueue_time_us = ufid_msg->timestamp;
    e2e_cache_ufid_msg_free(ufid_msg);
    return enqueue_time_us;
}

void
dp_netdev_offload_poll_queues(struct dp_offload_thread *ofl_thread,
                              struct e2e_cache_ufid_msg **ufid_msg,
                              struct dp_offload_thread_item **offload_item,
                              struct e2e_cache_trace_message **trace_msg)
    OVS_REQUIRES(ofl_thread->ufid_queue.read_lock,
                 ofl_thread->offload_queue.read_lock,
                 ofl_thread->trace_queue.read_lock)
{
    struct mpsc_queue_node *queue_node;
    unsigned int n_backoff;
    uint64_t backoff;

    *ufid_msg = NULL;
    *offload_item = NULL;
    *trace_msg = NULL;

    backoff = DP_NETDEV_OFFLOAD_BACKOFF_MIN;
    n_backoff = 0;

    while (1) {
        queue_node = mpsc_queue_pop(&ofl_thread->ufid_queue);
        if (queue_node != NULL) {
            /* ufid message is high priority. if we have it we are done. */
            *ufid_msg = CONTAINER_OF(queue_node, struct e2e_cache_ufid_msg,
                                     node);
            return;
        }

        queue_node = mpsc_queue_pop(&ofl_thread->offload_queue);
        if (queue_node != NULL) {
            *offload_item = CONTAINER_OF(queue_node,
                                         struct dp_offload_thread_item, node);
            atomic_count_dec64(&ofl_thread->enqueued_offload);
            return;
        }

        queue_node = mpsc_queue_pop(&ofl_thread->trace_queue);
        if (queue_node != NULL) {
            *trace_msg = CONTAINER_OF(queue_node,
                                      struct e2e_cache_trace_message, node);
            atomic_count_dec(&ofl_thread->e2e_stats.queue_trcs);
            return;
        }

        /* Execute upkeep if
         *
         *   + we are waiting for work for the first time
         *     -> We have just stopped a streak of offloading,
         *        some remaining things might need cleanup.
         *
         *   + we have waited roughly the amount of time
         *     between upkeep period.
         */
        if ((n_backoff & (DP_OFFLOAD_UPKEEP_N_BACKOFF - 1)) == 0) {
            /* Signal 'quiescing' only on the first backoff. */
            netdev_ports_upkeep(n_backoff == 0);
        }
        n_backoff += 1;

        /* The thread is flagged as quiescent during xnanosleep(). */
        xnanosleep(backoff * 1E6);
        if (backoff < DP_NETDEV_OFFLOAD_BACKOFF_MAX) {
            backoff <<= 1;
        }
    }
}

long long int
dp_offload_measure_latency(struct dp_offload_thread *thread,
                           long long int enqueue_time_us,
                           long long int finish_time_us)
{
    long long int latency_us;

    latency_us = finish_time_us - enqueue_time_us;
    mov_avg_cma_update(&thread->cma, latency_us);
    mov_avg_ema_update(&thread->ema, latency_us);
    histogram_add_sample(&thread->latency, latency_us / 1000);

    return latency_us;
}

void
dp_offload_process(struct dp_offload_thread *thread,
                   struct dp_offload_thread_item *msg)
{
    struct dp_offload_queue_metrics *m;
    long long int enqueue_time_us;
    long long int dequeue_time_us;
    long long int service_time_us;
    long long int finish_time_us;
    long long int wait_time_ms;
    long long int latency_us;

    enqueue_time_us = msg->timestamp;
    dequeue_time_us = time_usec();

    switch (msg->type) {
    case DP_OFFLOAD_FLOW:
        dp_offload_flow(msg);
        break;
    case DP_OFFLOAD_CONN:
        dp_offload_conn(msg);
        break;
    case DP_OFFLOAD_STATS_CLEAR:
        mov_avg_cma_init(&thread->cma);
        mov_avg_ema_init(&thread->ema, 100);
        break;
    case DP_OFFLOAD_FLUSH:
        dp_offload_flush(msg);
        break;
    default:
        OVS_NOT_REACHED();
    }

    finish_time_us = time_usec();
    latency_us = dp_offload_measure_latency(thread,
                                            enqueue_time_us,
                                            finish_time_us);

    wait_time_ms = (dequeue_time_us - enqueue_time_us) / 1000;
    service_time_us = finish_time_us - dequeue_time_us;
    m = &thread->queue_metrics[msg->type];
    histogram_add_sample(&m->wait_time, wait_time_ms);
    histogram_add_sample(&m->service_time, service_time_us);
    histogram_add_sample(&m->sojourn_time, latency_us / 1000);

    switch (msg->type) {
    case DP_OFFLOAD_FLOW:
        if (!thread->high_latency_event &&
            latency_us >= 200000) {
            thread->high_latency_event = true;
            COVERAGE_INC(flow_offload_200ms_latency);
        }
        break;
    case DP_OFFLOAD_CONN:
        if (!thread->high_latency_event) {
            thread->high_latency_event = true;
            if (latency_us >= 100) {
                COVERAGE_INC(ct_offload_100us_latency);
            } else if (latency_us >= 50) {
                COVERAGE_INC(ct_offload_50us_latency);
            } else if (latency_us > 30) {
                COVERAGE_INC(ct_offload_30us_latency);
            } else {
                thread->high_latency_event = false;
            }
        }
        break;
    case DP_OFFLOAD_STATS_CLEAR:
        /* Fallthrough */
    case DP_OFFLOAD_FLUSH:
        /* Fallthrough */
    default:
        break;
    }

    dp_netdev_free_offload(msg);
}

static void
poll_threads_n_read_value(double *values, void *it)
{
    struct dp_netdev *dp = get_dp_netdev(it);

    values[0] = cmap_count(&dp->poll_threads);
}

METRICS_ENTRIES(foreach_dpif_netdev, poll_threads_n,
    "poll_threads", poll_threads_n_read_value,
    METRICS_GAUGE(n, "Number of polling threads."),
);

static void
do_foreach_poll_threads(metrics_visitor_fn visitor,
                        struct metrics_visitor_context *ctx,
                        struct metrics_node *node,
                        struct metrics_label *labels,
                        size_t n OVS_UNUSED)
{
    struct dp_netdev *dp = get_dp_netdev(ctx->it);
    struct dp_netdev_pmd_thread *pmd;
    char core[50];
    char numa[50];

    labels[0].value = core;
    labels[1].value = numa;
    CMAP_FOR_EACH (pmd, node, &dp->poll_threads) {
        if (pmd->core_id == NON_PMD_CORE_ID &&
            !metrics_dbg_enabled(NULL)) {
            /* By definition, if the core ID is not one of a PMD,
             * then it is not a poll thread (i.e. 'main').
             * Do not iterate on it as if it was one. */
            continue;
        }
        snprintf(core, sizeof core, "%u", pmd->core_id);
        snprintf(numa, sizeof numa, "%d", pmd->numa_id);
        if (pmd->core_id == NON_PMD_CORE_ID) {
            snprintf(core, sizeof core, "main");
            snprintf(numa, sizeof numa, "0");
        }
        ctx->it = pmd;
        visitor(ctx, node);
    }
}

METRICS_COLLECTION(foreach_dpif_netdev, foreach_poll_threads,
    do_foreach_poll_threads, "core", "numa");

METRICS_COND(foreach_poll_threads, foreach_poll_threads_ext,
             metrics_ext_enabled);

METRICS_COND(foreach_poll_threads, foreach_poll_threads_dbg,
             metrics_dbg_enabled);

enum {
    PMD_METRICS_PACKETS,
    PMD_METRICS_RECIRC,
    PMD_METRICS_HIT,
    PMD_METRICS_MISSED,
    PMD_METRICS_LOST,
    PMD_METRICS_AVG_LOOKUPS_PER_HIT,
    PMD_METRICS_AVG_PACKETS_PER_BATCH,
    PMD_METRICS_AVG_RECIRC_PER_PACKET,
    PMD_METRICS_AVG_PASSES_PER_PACKET,
    PMD_METRICS_AVG_CYCLES_PER_PACKET,
    PMD_METRICS_AVG_BUSY_CYCLES_PER_PACKET,
    PMD_METRICS_PERCENT_BUSY_CYCLES,
    PMD_METRICS_PERCENT_IDLE_CYCLES,
};

static void
poll_threads_read_value(double *values, void *it)
{
    struct dp_netdev_pmd_thread *pmd = it;
    uint64_t total_cycles, total_packets;
    uint64_t stats[PMD_N_STATS];
    double busy_cycles_per_pkt;
    double packets_per_batch;
    double avg_busy_cycles;
    double avg_idle_cycles;
    double lookups_per_hit;
    double recirc_per_pkt;
    double passes_per_pkt;
    double cycles_per_pkt;
    uint64_t n_hit;

    /* Do not use 'pmd_perf_read_counters'. Counters are supposed to
     * always be increasing, while the pmd perf module is made
     * for debugging purpose and offers a 'clear' operation.
     * Read the counters exactly as they are.
     */
    for (int i = 0; i < PMD_N_STATS; i++) {
        atomic_read_relaxed(&pmd->perf_stats.counters.n[i], &stats[i]);
    }

    n_hit = 0;
    n_hit += stats[PMD_STAT_PHWOL_HIT];
    n_hit += stats[PMD_STAT_SIMPLE_HIT];
    n_hit += stats[PMD_STAT_EXACT_HIT];
    n_hit += stats[PMD_STAT_SMC_HIT];
    n_hit += stats[PMD_STAT_MASKED_HIT];

    total_cycles = stats[PMD_CYCLES_ITER_IDLE] +
                   stats[PMD_CYCLES_ITER_BUSY];
    total_packets = stats[PMD_STAT_RECV];

    lookups_per_hit = 0;
    if (stats[PMD_STAT_MASKED_HIT] > 0) {
        lookups_per_hit = (double) stats[PMD_STAT_MASKED_LOOKUP] /
                          (double) stats[PMD_STAT_MASKED_HIT];
    }

    packets_per_batch = 0;
    if (stats[PMD_STAT_SENT_BATCHES] > 0) {
        packets_per_batch = (double) stats[PMD_STAT_SENT_PKTS] /
                            (double) stats[PMD_STAT_SENT_BATCHES];
    }

    avg_idle_cycles = 0;
    avg_busy_cycles = 0;
    if (total_cycles > 0) {
        avg_idle_cycles = (double) stats[PMD_CYCLES_ITER_IDLE] /
                          (double) total_cycles * 100.0;
        avg_busy_cycles = (double) stats[PMD_CYCLES_ITER_BUSY] /
                          (double) total_cycles * 100.0;
    }

    recirc_per_pkt = 0;
    passes_per_pkt = 0;
    cycles_per_pkt = 0;
    busy_cycles_per_pkt = 0;
    if (total_packets > 0) {
        recirc_per_pkt = (double) stats[PMD_STAT_RECIRC] /
                         (double) total_packets;
        passes_per_pkt = (double) (total_packets + stats[PMD_STAT_RECIRC]) /
                         (double) total_packets;
        cycles_per_pkt = (double) total_cycles / (double) total_packets;
        busy_cycles_per_pkt = (double) stats[PMD_CYCLES_ITER_BUSY] /
                              (double) total_packets;
    }

    values[PMD_METRICS_PACKETS] = stats[PMD_STAT_RECV];
    values[PMD_METRICS_RECIRC] = stats[PMD_STAT_RECIRC];
    values[PMD_METRICS_HIT] = n_hit;
    values[PMD_METRICS_MISSED] = stats[PMD_STAT_MISS];
    values[PMD_METRICS_LOST] = stats[PMD_STAT_LOST];

    values[PMD_METRICS_AVG_LOOKUPS_PER_HIT] = lookups_per_hit;
    values[PMD_METRICS_AVG_PACKETS_PER_BATCH] = packets_per_batch;
    values[PMD_METRICS_AVG_RECIRC_PER_PACKET] = recirc_per_pkt;
    values[PMD_METRICS_AVG_PASSES_PER_PACKET] = passes_per_pkt;
    values[PMD_METRICS_AVG_CYCLES_PER_PACKET] = cycles_per_pkt;
    values[PMD_METRICS_AVG_BUSY_CYCLES_PER_PACKET] = busy_cycles_per_pkt;
    values[PMD_METRICS_PERCENT_BUSY_CYCLES] = avg_busy_cycles;
    values[PMD_METRICS_PERCENT_IDLE_CYCLES] = avg_idle_cycles;
}

METRICS_ENTRIES(foreach_poll_threads, poll_threads_entries,
    "poll_threads", poll_threads_read_value,
    [PMD_METRICS_PACKETS] = METRICS_COUNTER(packets,
        "Number of received packets."),
    [PMD_METRICS_RECIRC] = METRICS_COUNTER(recirculations,
        "Number of executed packet recirculations."),
    [PMD_METRICS_HIT] = METRICS_COUNTER(hit,
        "Number of flow table matches."),
    [PMD_METRICS_MISSED] = METRICS_COUNTER(missed,
        "Number of flow table misses and upcall succeeded."),
    [PMD_METRICS_LOST] = METRICS_COUNTER(lost,
        "Number of flow table misses and upcall failed."),
    [PMD_METRICS_AVG_LOOKUPS_PER_HIT] = METRICS_GAUGE(lookups_per_hit,
        "Average number of lookups per flow table hit."),
    [PMD_METRICS_AVG_PACKETS_PER_BATCH] = METRICS_GAUGE(packets_per_batch,
        "Average number of packets per batch."),
    [PMD_METRICS_AVG_RECIRC_PER_PACKET] = METRICS_GAUGE(recirc_per_packet,
        "Average number of recirculations per packet."),
    [PMD_METRICS_AVG_PASSES_PER_PACKET] = METRICS_GAUGE(passes_per_packet,
        "Average number of datapath passes per packet."),
    [PMD_METRICS_AVG_CYCLES_PER_PACKET] = METRICS_GAUGE(cycles_per_packet,
        "Average number of CPU cycles per packet."),
    [PMD_METRICS_AVG_BUSY_CYCLES_PER_PACKET] = METRICS_GAUGE(
            busy_cycles_per_packet,
        "Average number of active CPU cycles per packet."),
    [PMD_METRICS_PERCENT_BUSY_CYCLES] = METRICS_GAUGE(busy_cycles,
        "Percent of useful CPU cycles."),
    [PMD_METRICS_PERCENT_IDLE_CYCLES] = METRICS_GAUGE(idle_cycles,
        "Percent of idle CPU cycles."),
);

enum {
    PMD_METRICS_SIMPLE_N_ENTRIES,
    PMD_METRICS_SIMPLE_HIT,
    PMD_METRICS_SIMPLE_MISS,
    PMD_METRICS_SIMPLE_UPDATE,
    PMD_METRICS_EMC_N_ENTRIES,
    PMD_METRICS_EMC_HIT,
    PMD_METRICS_EMC_MISS,
    PMD_METRICS_EMC_UPDATE,
    PMD_METRICS_SMC_N_ENTRIES,
    PMD_METRICS_SMC_HIT,
    PMD_METRICS_SMC_MISS,
    PMD_METRICS_SMC_UPDATE,
    PMD_METRICS_CLS_N_ENTRIES,
    PMD_METRICS_CLS_HIT,
    PMD_METRICS_CLS_MISS,
    PMD_METRICS_CLS_UPDATE,
    PMD_METRICS_N_CACHE_ENTRIES,
};

static unsigned int
dpcls_count(struct dpcls *cls)
{
    struct dpcls_subtable *subtable;
    unsigned int count = 0;

    CMAP_FOR_EACH (subtable, cmap_node, &cls->subtables_map) {
        count += cmap_count(&subtable->rules);
    }

    return count;
}

static void
poll_threads_cache_read_value(double *values, void *it)
{
    struct dp_netdev_pmd_thread *pmd = it;
    uint64_t stats[PMD_N_STATS];
    unsigned int pmd_n_cls_rules;
    struct dpcls *cls;

    for (int i = 0; i < PMD_N_STATS; i++) {
        atomic_read_relaxed(&pmd->perf_stats.counters.n[i], &stats[i]);
    }

    values[PMD_METRICS_SIMPLE_N_ENTRIES] =
        cmap_count(&pmd->simple_match_table);
    values[PMD_METRICS_SIMPLE_HIT] = stats[PMD_STAT_SIMPLE_HIT];
    values[PMD_METRICS_SIMPLE_MISS] = stats[PMD_STAT_SIMPLE_MISS];
    values[PMD_METRICS_SIMPLE_UPDATE] = stats[PMD_STAT_SIMPLE_UPDATE];

    values[PMD_METRICS_EMC_N_ENTRIES] =
        emc_cache_count(&(pmd->flow_cache).emc_cache);
    values[PMD_METRICS_EMC_HIT] = stats[PMD_STAT_EXACT_HIT];
    values[PMD_METRICS_EMC_MISS] = stats[PMD_STAT_EXACT_MISS];
    values[PMD_METRICS_EMC_UPDATE] = stats[PMD_STAT_EXACT_UPDATE];

    values[PMD_METRICS_SMC_N_ENTRIES] =
        smc_cache_count(&(pmd->flow_cache).smc_cache);
    values[PMD_METRICS_SMC_HIT] = stats[PMD_STAT_SMC_HIT];
    values[PMD_METRICS_SMC_MISS] = stats[PMD_STAT_SMC_MISS];
    values[PMD_METRICS_SMC_UPDATE] = stats[PMD_STAT_SMC_UPDATE];

    pmd_n_cls_rules = 0;
    CMAP_FOR_EACH (cls, node, &pmd->classifiers) {
        pmd_n_cls_rules += dpcls_count(cls);
    }

    values[PMD_METRICS_CLS_N_ENTRIES] = pmd_n_cls_rules;
    values[PMD_METRICS_CLS_HIT] = stats[PMD_STAT_MASKED_HIT];
    values[PMD_METRICS_CLS_MISS] = stats[PMD_STAT_MASKED_LOOKUP] -
                                     stats[PMD_STAT_MASKED_HIT];
    values[PMD_METRICS_CLS_UPDATE] = stats[PMD_STAT_MASKED_UPDATE];
}

/* Use a single point of definition for the cache entries to enforce
 * strict alignment between 'datapath_cache' and 'poll_threads_cache'
 * metrics. */
#define PMD_METRICS_CACHE_ENTRIES                                      \
    /* Simple match cache. */                                          \
    [PMD_METRICS_SIMPLE_N_ENTRIES] = METRICS_GAUGE(simple_n_entries,   \
        "Number of entries in the simple match cache."),               \
    [PMD_METRICS_SIMPLE_HIT] = METRICS_COUNTER(simple_hit,             \
        "Number of lookup hit in the simple match cache."),            \
    [PMD_METRICS_SIMPLE_MISS] = METRICS_COUNTER(simple_miss,           \
        "Number of lookup miss in the simple match cache."),           \
    [PMD_METRICS_SIMPLE_UPDATE] = METRICS_COUNTER(simple_update,       \
        "Number of updates of the simple match cache."),               \
    /* Exact match cache. */                                           \
    [PMD_METRICS_EMC_N_ENTRIES] = METRICS_GAUGE(emc_n_entries,         \
        "Number of entries in the exact match cache."),                \
    [PMD_METRICS_EMC_HIT] = METRICS_COUNTER(emc_hit,                   \
        "Number of lookup hit in the exact match cache."),             \
    [PMD_METRICS_EMC_MISS] = METRICS_COUNTER(emc_miss,                 \
        "Number of lookup miss in the exact match cache."),            \
    [PMD_METRICS_EMC_UPDATE] = METRICS_COUNTER(emc_update,             \
        "Number of updates of the exact match cache."),                \
    /* Signature match cache. */                                       \
    [PMD_METRICS_SMC_N_ENTRIES] = METRICS_GAUGE(smc_n_entries,         \
        "Number of entries in the signature match cache."),            \
    [PMD_METRICS_SMC_HIT] = METRICS_COUNTER(smc_hit,                   \
        "Number of lookup hit in the signature match cache."),         \
    [PMD_METRICS_SMC_MISS] = METRICS_COUNTER(smc_miss,                 \
        "Number of lookup miss in the signature match cache."),        \
    [PMD_METRICS_SMC_UPDATE] = METRICS_COUNTER(smc_update,             \
        "Number of updates of the signature match cache."),            \
    /* Datapath classifiers. */                                        \
    [PMD_METRICS_CLS_N_ENTRIES] = METRICS_GAUGE(cls_n_entries,         \
        "Number of entries in the datapath classifiers."),             \
    [PMD_METRICS_CLS_HIT] = METRICS_COUNTER(cls_hit,                   \
        "Number of lookup hit in the datapath classifiers."),          \
    [PMD_METRICS_CLS_MISS] = METRICS_COUNTER(cls_miss,                 \
        "Number of lookup miss in the datapath classifiers."),         \
    [PMD_METRICS_CLS_UPDATE] = METRICS_COUNTER(cls_update,             \
        "Number of updates of the datapath classifiers."),

METRICS_ENTRIES(foreach_poll_threads_dbg, poll_threads_cache_dbg_entries,
    "poll_threads_cache", poll_threads_cache_read_value, PMD_METRICS_CACHE_ENTRIES);

static void
datapath_cache_read_value(double *values, void *it)
{
    double pmd_values[PMD_METRICS_N_CACHE_ENTRIES];
    struct dp_netdev *dp = get_dp_netdev(it);
    struct dp_netdev_pmd_thread *pmd;
    int i;

    for (i = 0; i < ARRAY_SIZE(pmd_values); i++) {
        values[i] = 0.0;
    }

    CMAP_FOR_EACH (pmd, node, &dp->poll_threads) {
        poll_threads_cache_read_value(pmd_values, pmd);
        for (i = 0; i < ARRAY_SIZE(pmd_values); i++) {
            values[i] += pmd_values[i];
        }
    }
}

METRICS_ENTRIES(foreach_dpif_netdev_ext, datapath_cache_ext_entries,
    "datapath_cache", datapath_cache_read_value, PMD_METRICS_CACHE_ENTRIES);

METRICS_DECLARE(hw_offload_threads_dbg_entries);
METRICS_DECLARE(hw_offload_latency);
METRICS_DECLARE(hw_offload_queue_sojourn_time);
METRICS_DECLARE(hw_offload_queue_wait_time);
METRICS_DECLARE(hw_offload_queue_service_time);
METRICS_DECLARE(datapath_hw_offload_entries);

void
dpif_netdev_metrics_register(void)
{
    METRICS_REGISTER(datapath_cache_ext_entries);
    METRICS_REGISTER(poll_threads_entries);
    METRICS_REGISTER(poll_threads_cache_dbg_entries);
    METRICS_REGISTER(hw_offload_threads_dbg_entries);
    METRICS_REGISTER(hw_offload_latency);
    METRICS_REGISTER(hw_offload_queue_sojourn_time);
    METRICS_REGISTER(hw_offload_queue_wait_time);
    METRICS_REGISTER(hw_offload_queue_service_time);
    METRICS_REGISTER(datapath_hw_offload_entries);
}

static void
dpif_netdev_set_static_config(const struct smap *other_config)
{
    static struct ovsthread_once once = OVSTHREAD_ONCE_INITIALIZER;

    if (!ovsthread_once_start(&once)) {
        return;
    }

    conntrack_offload_config(other_config);
    dpif_netdev_set_static_config_ct_add_queue_size(other_config);

    ovsthread_once_done(&once);
}

void
dpif_netdev_set_ext_config(struct dp_netdev *dp, const struct smap *other_config,
                           bool first_set_config)
{
    bool pmd_quiet_idle, cur_pmd_quiet_idle;

    if (smap_get_node(other_config, "max-recirc-depth")) {
        unsigned int read_depth;

        read_depth = smap_get_uint(other_config, "max-recirc-depth",
                                   DEFAULT_MAX_RECIRC_DEPTH);
        if (read_depth < DEFAULT_MAX_RECIRC_DEPTH) {
            read_depth = DEFAULT_MAX_RECIRC_DEPTH;
        }
        if (netdev_is_e2e_cache_enabled()
            && read_depth > E2E_CACHE_MAX_TRACE) {
            VLOG_INFO("max recirc depth is %d if e2e-cache is enabled",
                      E2E_CACHE_MAX_TRACE);
            read_depth = E2E_CACHE_MAX_TRACE;
        }
        if (max_recirc_depth != read_depth) {
            max_recirc_depth = read_depth;
            VLOG_INFO("max recirc depth set to %u", read_depth);
        }
    }

    bool sleep_changed = set_all_pmd_max_sleeps(dp, other_config);
    if (first_set_config || sleep_changed) {
        log_all_pmd_sleeps(dp);
    }

    pmd_quiet_idle = smap_get_bool(other_config, "pmd-quiet-idle", false);
    atomic_read_relaxed(&dp->pmd_quiet_idle, &cur_pmd_quiet_idle);
    if (first_set_config || pmd_quiet_idle != cur_pmd_quiet_idle) {
        atomic_store_relaxed(&dp->pmd_quiet_idle, pmd_quiet_idle);
        VLOG_INFO("PMD quiescent idling mode %s.",
                  pmd_quiet_idle ? "enabled" : "disabled");
    }

    dpif_netdev_set_static_config(other_config);
}

void
dp_netdev_doca_affinity_list(struct dp_netdev *dp,
                             struct netdev *netdev,
                             char **affinity_list)
{
    struct ds doca_affinity_s = DS_EMPTY_INITIALIZER;
    int numa_id = netdev_get_numa_id(netdev);
    struct ovs_numa_dump *pmd_cores;
    struct ovs_numa_info_core *core;
    bool same_numa;
    uint8_t q;
    int iter;

    pmd_cores = dp_netdev_pmd_cmask2cores(dp->pmd_cmask);
    same_numa = true;
    q = 0;
    /* In the first iteration, assign the same numa cores, and the rest of
     * the queues to other numa cores in the 2nd one.
    */
    for (iter = 0; iter < 2; iter++) {
        FOR_EACH_CORE_ON_DUMP(core, pmd_cores) {
            if ((same_numa && numa_id != core->numa_id) ||
                (!same_numa && numa_id == core->numa_id)) {
                continue;
            }
            if (q > 0) {
                ds_put_cstr(&doca_affinity_s, ",");
            }
            ds_put_format(&doca_affinity_s, "%d:%d", q, core->core_id);
            q++;
        }
        same_numa = false;
    }
    ovs_numa_dump_destroy(pmd_cores);
    free(*affinity_list);
    *affinity_list = ds_steal_cstr(&doca_affinity_s);
    ds_destroy(&doca_affinity_s);
}

int
dpif_netdev_ct_get_stats(struct dpif *dpif,
                         struct ct_dpif_stats *stats)
{
    struct dp_netdev *dp = get_dp_netdev(dpif);

    return conntrack_get_stats(dp->conntrack, stats);
}

void
dp_netdev_esw_ports_set_disabled(struct dp_netdev *dp, struct netdev *esw_mgr, bool value)
    OVS_REQ_WRLOCK(dp->port_rwlock)
{
    struct dp_netdev_port *port;
    int esw_mgr_pid;

    esw_mgr_pid = netdev_dpdk_get_esw_mgr_port_id(esw_mgr);

    if (esw_mgr_pid == -1) {
        return;
    }

    HMAP_FOR_EACH (port, node, &dp->ports) {
        if (esw_mgr_pid == netdev_dpdk_get_esw_mgr_port_id(port->netdev)) {
            port->disabled = value;
        }
    }
}

int
dp_netdev_offload_netdev_meter_set(uint32_t meter_id_,
                                   struct ofputil_meter_config *config)
{
    /* Compensate for ovs-ofctl (meter_ID - 1) adjustment */
    ofproto_meter_id id = { .uint32 = meter_id_ + 1 };
    return netdev_dpdk_meter_set(id, config);
}

int
dp_netdev_offload_netdev_meter_get(uint32_t meter_id_,
                                   struct ofputil_meter_stats *stats,
                                   uint16_t n_bands)
{
    /* Compensate for ovs-ofctl (meter_ID - 1) adjustment */
    ofproto_meter_id id = { .uint32 = meter_id_ + 1 };
    return netdev_dpdk_meter_get(id, stats, n_bands);
}

int
dp_netdev_offload_netdev_meter_del(uint32_t meter_id_,
                                   struct ofputil_meter_stats *stats,
                                   uint16_t n_bands)
{
    /* Compensate for ovs-ofctl (meter_ID - 1) adjustment */
    ofproto_meter_id id = { .uint32 = meter_id_ + 1 };
    return netdev_dpdk_meter_del(id, stats, n_bands);
}

bool
dp_netdev_port_exists(struct netdev *netdev)
{
    struct ifreq ifr;
    int sock;
    int err;

    sock = socket(AF_INET, SOCK_STREAM, IPPROTO_IP);
    if (sock < 0) {
        return false;
    }
    ovs_strlcpy(ifr.ifr_name, netdev_get_name(netdev), sizeof ifr.ifr_name);

    /* Try to get some property of the port, to check it exists. */
    err = ioctl(sock, SIOCGIFMTU, &ifr);

    close(sock);

    return err == 0;
}
