/*
 * Copyright (c) 2023-2024 NVIDIA CORPORATION & AFFILIATES. All rights reserved.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at:
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

#ifndef DPIF_NETDEV_EXT_H
#define DPIF_NETDEV_EXT_H 1

#include <stdbool.h>
#include <stddef.h>
#include <stdint.h>
#include "dpif.h"
#include "dp-packet.h"
#include "netdev-offload-ext.h"
#include "openvswitch/match.h"
#include "openvswitch/types.h"
#include "packets.h"

#ifdef  __cplusplus
extern "C" {
#endif

struct unixctl_conn;
struct dp_offload_thread;
struct dp_offload_thread_item;
struct dp_packet_flow_map;

extern bool dp_netdev_e2e_cache_enabled;
extern unsigned int max_recirc_depth;

struct ovs_numa_dump *dp_netdev_pmd_cmask2cores(const char *pmd_cmask);
unsigned int dpif_netdev_get_n_pmd_threads(void);
void dpif_netdev_set_n_pmd_threads(const char *pmd_cmask);

void
dp_netdev_dump_packets_toggle(struct unixctl_conn *conn, int argc,
                              const char *argv[], void *aux OVS_UNUSED);

void
dp_netdev_read_dump_packets_enabled(bool *flag);

/* This struct holds the e2e-cache statistic counters
 * generated_trcs = Amount of trace messages generated/dispatched to E2E cache.
 * processed_trcs = Amount of trace messages processed by E2E cache.
 * discarded_trcs = Amount of trace messages discarded by E2E cache.
 * aborted_trcs = Amount of trace messages aborted by E2E cache.
 * throttled_trcs = Amount of trace messages throttled due to high message
 *                  rate.
 * queue_trcs = Amount of trace messages in E2E cache queue.
 * overflow_trcs = Amount of trace messages dropped due to
 *                             queue overflow.
 * flow_add_msgs = Amount of new flow messages received by E2E cache.
 * flow_del_msgs = Amount of delete flow messages received by E2E cache.
 * flush_flow_msgs = Amount of flush flow messages received by E2E cache.
 * succ_merged_flows = Amount of successfully merged flows.
 * merge_rej_flows = Amount of flows rejected by the merge engine.
 * add_merged_flow_hw = Amount of add merged flow messages dispatched to
 *                      HW offload.
 * del_merged_flow_hw = Amount of delete merged flow messages dispatched to
 *                      HW offload.
 * add_ct_flow_hw = Amount of successful CT offload operations to MT.
 * add_ct_flow_err = Amount of failed CT offload operations MT.
 * succ_ct2ct_merges = Amount of successfully ct2ct merges.
 * rej_ct2ct_merges = Amount of merges rejected by the ct2ct merge engine.
 * add_ct2ct_flows = Amount of CT2CT offload add operations.
 * del_ct2ct_flows = Amount of CT2CT offload del operations.
 */
struct e2e_cache_stats {
    atomic_count generated_trcs;
    uint32_t processed_trcs;
    atomic_count discarded_trcs;
    atomic_count aborted_trcs;
    atomic_count throttled_trcs;
    atomic_count queue_trcs;
    atomic_count overflow_trcs;
    atomic_count flow_add_msgs;
    atomic_count flow_del_msgs;
    uint32_t flush_flow_msgs;
    uint32_t succ_merged_flows;
    uint32_t merge_rej_flows;
    uint32_t add_merged_flow_hw;
    uint32_t del_merged_flow_hw;
    uint32_t add_ct_mt_flow_hw;
    uint32_t del_ct_mt_flow_hw;
    uint32_t add_ct_mt_flow_err;
    uint32_t del_ct_mt_flow_err;
    uint32_t succ_ct2ct_merges;
    uint32_t rej_ct2ct_merges;
    uint32_t add_ct2ct_flows;
    uint32_t del_ct2ct_flows;
};

enum {
    E2E_UFID_MSG_PUT = 1,
    E2E_UFID_MSG_DEL = 2,
};

struct e2e_cache_ufid_msg {
    struct mpsc_queue_node node;
    ovs_u128 ufid;
    int op;
    bool is_ct;
    struct nlattr *actions;
    struct dp_netdev *dp;
    struct conn *conn;
    struct netdev *netdev;
    struct ovs_barrier *barrier;
    struct ovs_refcount *del_refcnt;
    size_t actions_len;
    long long int timestamp;
    union {
        struct match match[0];
        struct ct_match ct_match[0];
    };
};

enum e2e_offload_state {
    E2E_OL_STATE_FLOW,
    E2E_OL_STATE_CT_SW,
    E2E_OL_STATE_CT_HW,
    E2E_OL_STATE_CT_MT,
    E2E_OL_STATE_CT2CT,
    E2E_OL_STATE_CT_ERR,
    E2E_OL_STATE_NUM,
};

static const char * const e2e_offload_state_names[] = {
    [E2E_OL_STATE_FLOW] = "E2E_OL_STATE_FLOW",
    [E2E_OL_STATE_CT_SW] = "E2E_OL_STATE_CT_SW",
    [E2E_OL_STATE_CT_HW] = "E2E_OL_STATE_CT_HW",
    [E2E_OL_STATE_CT_MT] = "E2E_OL_STATE_CT_MT",
    [E2E_OL_STATE_CT2CT] = "E2E_OL_STATE_CT2CT",
    [E2E_OL_STATE_CT_ERR] = "E2E_OL_STATE_CT_ERR",
    [E2E_OL_STATE_NUM] = "Unknown",
};

struct merged_match_fields {
    union flow_in_port in_port; /* Input port.*/
    struct {
        ovs_be32 ip_dst;
        struct in6_addr ipv6_dst;
        ovs_be32 ip_src;
        struct in6_addr ipv6_src;
        ovs_be64 tun_id;
        ovs_be16 tp_dst;
    } tunnel;

    /* L2. */
    struct eth_addr dl_dst;     /* Ethernet destination address. */
    struct eth_addr dl_src;     /* Ethernet source address. */
    ovs_be16 dl_type;           /* Ethernet frame type. */

    /* VLANs. */
    union flow_vlan_hdr vlans[1]; /* VLANs */

    /* L3. */
    ovs_be32 nw_src;            /* IPv4 source address or ARP SPA. */
    ovs_be32 nw_dst;            /* IPv4 destination address or ARP TPA. */
    struct in6_addr ipv6_src;   /* IPv6 source address. */
    struct in6_addr ipv6_dst;   /* IPv6 destination address. */
    uint8_t nw_frag;            /* FLOW_FRAG_* flags. */
    uint8_t nw_proto;           /* IP protocol or low 8 bits of ARP opcode. */

    /* L4. */
    ovs_be16 tp_src;            /* TCP/UDP/SCTP source port/ICMP type. */
    ovs_be16 tp_dst;            /* TCP/UDP/SCTP destination port/ICMP code. */

    /* MD. */
    uint16_t ct_zone;           /* Connection tracking zone. */
};

struct merged_match {
    struct merged_match_fields spec;
    struct merged_match_fields mask;
};

/*
 * A mapping from ufid to flow for e2e cache.
 */
struct e2e_cache_ovs_flow {
    struct hmap_node node;
    ovs_u128 ufid;
    unsigned int merge_tid;
    struct nlattr *actions;
    struct e2e_cache_ovs_flow *ct_peer;
    enum e2e_offload_state offload_state;
    uint16_t actions_size;
    struct hmap merged_counters; /* Map of merged flows counters
                                    it is part of. */
    struct ovs_list associated_merged_flows;
    union {
        struct match match[0];
        struct ct_match ct_match[0];
    };
};

/* Helper struct for accessing a struct containing ovs_list array.
 * Containing struct
 *   |- Helper array
 *      [0] Helper item 0
 *          |- ovs_list item 0
 *          |- index (0)
 *      [1] Helper item 1
 *          |- ovs_list item 1
 *          |- index (1)
 * To access the containing struct from one of the ovs_list items:
 * 1. Get the helper item from the ovs_list item using
 *    helper item =
      CONTAINER_OF(ovs_list item, helper struct type, ovs_list field)
 * 2. Get the contining struct from the helper item and its index in the array:
 *    containing struct =
 *    CONTAINER_OF(helper item, containing struct type, helper field[index])
 */
struct flow2flow_item {
    struct ovs_list list;
    struct e2e_cache_ovs_flow *mt_flow;
    uint16_t index;
};

enum {
    E2E_SET_ETH_SRC = 1 << 0,
    E2E_SET_ETH_DST = 1 << 1,
    E2E_SET_ETH = E2E_SET_ETH_SRC | E2E_SET_ETH_DST,

    E2E_SET_IPV4_SRC = 1 << 2,
    E2E_SET_IPV4_DST = 1 << 3,
    E2E_SET_IPV4_TTL = 1 << 4,
    E2E_SET_IPV4 = E2E_SET_IPV4_SRC | E2E_SET_IPV4_DST | \
                   E2E_SET_IPV4_TTL,

    E2E_SET_IPV6_SRC = 1 << 5,
    E2E_SET_IPV6_DST = 1 << 6,
    E2E_SET_IPV6_HLMT = 1 << 7,
    E2E_SET_IPV6 = E2E_SET_IPV6_SRC | E2E_SET_IPV6_DST | \
                   E2E_SET_IPV6_HLMT,

    E2E_SET_UDP_SRC = 1 << 8,
    E2E_SET_UDP_DST = 1 << 9,
    E2E_SET_UDP = E2E_SET_UDP_SRC | E2E_SET_UDP_DST,

    E2E_SET_TCP_SRC = 1 << 10,
    E2E_SET_TCP_DST = 1 << 11,
    E2E_SET_TCP = E2E_SET_TCP_SRC | E2E_SET_TCP_DST,
};

struct e2e_cache_merged_set {
    struct ovs_key_ethernet eth;
    struct ovs_key_ipv4 ipv4;
    struct ovs_key_ipv6 ipv6;
    struct ovs_key_tcp tcp;
    struct ovs_key_udp udp;
    uint32_t flags;
};

/*
 * Merged flow structure.
 */
struct e2e_cache_merged_flow {
    union {
        struct hmap_node in_hmap;
        struct ovs_list  in_list;
    } node;
    ovs_u128 ufid;
    unsigned int tid;
    struct dp_netdev *dp;
    struct nlattr *actions;
    uint16_t actions_size;
    uint16_t associated_flows_len;
    struct ovs_list flow_counter_list; /* Anchor for list of merged flows
                                          using the same flow counter. */
    struct ovs_list ct_counter_list; /* Anchor for list of merged flows
                                        using the same CT counter. */
    uintptr_t ct_counter_key;
    struct flows_counter_key flows_counter_key;
    struct merged_match merged_match;
    uint32_t flow_mark;
    struct flow2flow_item associated_flows[0];
};

/* Counter object. */
struct e2e_cache_counter_item {
    struct hmap_node node;
    struct ovs_list merged_flows; /* List of merged flows using this counter. */
    size_t hash;
    bool is_ct;
    struct flows_counter_key key;
};

int
dpif_netdev_offload_stats_get(struct dpif *dpif,
                              struct netdev_custom_stats *stats,
                              bool verbose);

int
dpif_netdev_offload_stats_clear(struct dpif *dpif OVS_UNUSED);

void
dpif_netdev_set_static_config_ct_add_queue_size(const struct smap *other_config);

void
dpif_netdev_set_config_e2e_cache(const struct smap *other_config);

void
dp_netdev_offload_thread_uninit(struct dp_offload_thread *thread);

void
dp_netdev_offload_thread_init(struct dp_offload_thread *thread);

struct dp_offload_thread *
dp_netdev_offload_thread_next(struct dp_offload_thread *start,
                              unsigned int *tid, bool include_main);

void
e2e_cache_trace_add_flow(struct dp_packet *p,
                         const ovs_u128 *ufid);

int
e2e_cache_flow_del(const ovs_u128 *ufid, struct dp_netdev *dp,
                   struct conn *conn, long long int now);

void
e2e_cache_trace_tnl_pop(struct dp_packet *packet);

void
e2e_cache_dispatch_trace_message(struct dp_netdev *dp,
                                 struct dp_packet_batch *batch,
                                 long long int now);

int
e2e_cache_flow_put(bool is_ct, const ovs_u128 *ufid, const void *match,
                   const struct nlattr *actions, size_t actions_len,
                   long long int now);

int
e2e_cache_process_trace_info(struct dp_netdev *dp,
                             const struct e2e_cache_trace_info *trc_info,
                             unsigned int tid);

bool
e2e_cache_get_merged_flows_stats(struct netdev *netdev,
                                 struct match *match,
                                 struct nlattr **actions,
                                 const ovs_u128 *mt_ufid,
                                 struct dpif_flow_stats *stats,
                                 struct ofpbuf *buf,
                                 long long now,
                                 long long prev_now);

void
dpif_netdev_dump_e2e_flows(struct hmap *portno_names,
                           struct ofputil_port_map *port_map, struct ds *s);

void
packet_enqueue_to_flow_map(struct dp_packet *packet,
                           struct dp_netdev_flow *flow,
                           uint16_t tcp_flags,
                           struct dp_packet_flow_map *flow_map,
                           size_t index);

unsigned int
dp_netdev_offload_thread_nb(void);

void
pmd_thread_offload_upkeep(struct dp_netdev_pmd_thread *pmd, bool force);

unsigned int
pmd_thread_offload_process(struct dp_netdev_pmd_thread *pmd, unsigned int limit);

void
pmd_thread_offload_init(struct dp_netdev_pmd_thread *pmd);

void
pmd_thread_offload_uninit(struct dp_netdev_pmd_thread *pmd);

void
pmd_thread_offload_disable(struct dp_netdev_pmd_thread *pmd, struct netdev *netdev);

void
dp_netdev_pmd_idle_begin(struct dp_netdev_pmd_thread *pmd);

void
dp_netdev_pmd_idle_end(struct dp_netdev_pmd_thread *pmd);

void
dp_netdev_flow_format(const char *prefix,
                      struct ds *s,
                      const struct dp_netdev_flow *dp_flow);

long long int
e2e_cache_flow_db_handle_ufid_msg(struct e2e_cache_ufid_msg *ufid_msg);

long long int
dp_offload_measure_latency(struct dp_offload_thread *thread,
                           long long int enqueue_time_us,
                           long long int finish_time_us);

void
dp_offload_process(struct dp_offload_thread *thread,
                   struct dp_offload_thread_item *msg);

void
dp_netdev_offload_poll_queues(struct dp_offload_thread *ofl_thread,
                              struct e2e_cache_ufid_msg **ufid_msg,
                              struct dp_offload_thread_item **offload_item,
                              struct e2e_cache_trace_message **trace_msg);

void
dp_netdev_ct_offload_init(struct dp_netdev *dp);

void
dp_netdev_ct_offload_uninit(struct dp_netdev *dp);

/* Iterator over all active offload threads.
 *
 * THREAD (struct dp_offload_thread *):
 *   Pointer to a thread object. Will be set to each valid
 *   thread in succession and then NULL after completion.
 */
#define DP_NETDEV_OFFLOAD_FOREACH_THREAD_SHORT(INCLUDE_MAIN, THREAD) \
    for ((THREAD) = dp_netdev_offload_thread_next(NULL, NULL, INCLUDE_MAIN); (THREAD); \
         (THREAD) = dp_netdev_offload_thread_next((THREAD), NULL, INCLUDE_MAIN))

/* Iterator over all active offload threads.
 *
 * THREAD (struct dp_offload_thread *):
 *   Pointer to a thread object. Will be set to each valid
 *   thread in succession and then NULL after completion.
 *
 * ID (unsigned int):
 *   Variable that will contain the thread-id during each
 *   step of the iteration.
 */
#define DP_NETDEV_OFFLOAD_FOREACH_THREAD_LONG(INCLUDE_MAIN, THREAD, ID) \
    for ((THREAD) = dp_netdev_offload_thread_next(NULL, &(ID), INCLUDE_MAIN); (THREAD); \
         (THREAD) = dp_netdev_offload_thread_next((THREAD), &(ID), INCLUDE_MAIN))

/* Iterator over all active offload threads.
 *
 * This is a variadic macro that can take either SHORT or LONG
 * form: passing only (THREAD), or (THREAD, ID)) according to each
 * variant signature above.
 *
 */
#define DP_NETDEV_OFFLOAD_FOREACH_THREAD(...) \
    OVERLOAD_SAFE_MACRO(DP_NETDEV_OFFLOAD_FOREACH_THREAD_LONG, \
                        DP_NETDEV_OFFLOAD_FOREACH_THREAD_SHORT, \
                        3, true, __VA_ARGS__)

/* Iterator over all active offload threads except 'main'.
 *
 * This is a variadic macro that can take either SHORT or LONG
 * form: passing only (THREAD), or (THREAD, ID)) according to each
 * variant signature above.
 *
 * The reason to exclude the main sometimes is that this special thread
 * does not handle offload messages. If the iteration is done to emit
 * messages, the main thread must be excluded.
 *
 */
#define DP_NETDEV_OFFLOAD_FOREACH_THREAD_NO_MAIN(...) \
    OVERLOAD_SAFE_MACRO(DP_NETDEV_OFFLOAD_FOREACH_THREAD_LONG, \
                        DP_NETDEV_OFFLOAD_FOREACH_THREAD_SHORT, \
                        3, false, __VA_ARGS__)

void
dpif_netdev_metrics_register(void);

void
dp_netdev_offload_thread_enqueue(struct dp_offload_thread *thread,
                                 struct dp_offload_thread_item *offload);

void
dpif_netdev_set_ext_config(struct dp_netdev *dp, const struct smap *other_config,
                           bool first_set_config);

void
dp_netdev_doca_affinity_list(struct dp_netdev *dp,
                             struct netdev *netdev,
                             char **affinity_list);

int
dpif_netdev_ct_get_stats(struct dpif *dpif,
                         struct ct_dpif_stats *stats);

int
dp_netdev_offload_netdev_meter_set(uint32_t meter_id_,
                                   struct ofputil_meter_config *config);

int
dp_netdev_offload_netdev_meter_get(uint32_t meter_id_,
                                   struct ofputil_meter_stats *stats,
                                   uint16_t n_bands);
int
dp_netdev_offload_netdev_meter_del(uint32_t meter_id_,
                                   struct ofputil_meter_stats *stats,
                                   uint16_t n_bands);

bool
dp_netdev_port_exists(struct netdev *netdev);

#ifdef  __cplusplus
}
#endif

#endif /* dpif-netdev-ext.h */
