/*
 * Copyright (c) 2023-2024 NVIDIA CORPORATION & AFFILIATES. All rights reserved.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at:
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

#ifndef NETDEV_DPDK_EXT_H
#define NETDEV_DPDK_EXT_H

#include <config.h>
#include <dirent.h>
#include <errno.h>
#include <net/if.h>

#include "openvswitch/netdev.h"
#include "openvswitch/ofp-meter.h"
#include "smap.h"

struct doca_flow_port;
struct dp_packet;
struct netdev;

#ifdef DPDK_NETDEV

#include <rte_flow.h>
#include <rte_ethdev.h>

#define NR_HAIRPINQ (ovs_doca_initialized() ? 0 : 1)

typedef uint16_t dpdk_port_t;

bool
netdev_dpdk_is_ethdev(struct netdev *netdev);

struct doca_flow_port *
netdev_dpdk_doca_port_get(struct netdev *netdev);

bool
netdev_dpdk_is_esw_mgr(struct netdev *netdev);

struct rte_flow_action_handle *
netdev_dpdk_indirect_action_create(struct netdev *,
                                   int port_id,
                                   const struct rte_flow_action *,
                                   struct rte_flow_error *);
int
netdev_dpdk_indirect_action_destroy(int port_id,
                                    struct rte_flow_action_handle *,
                                    struct rte_flow_error *);
int
netdev_dpdk_indirect_action_query(int port_id,
                                  struct rte_flow_action_handle *,
                                  void *,
                                  struct rte_flow_error *);

int
netdev_dpdk_get_esw_mgr_port_id(struct netdev *netdev);

enum netdev_dpdk_port_dir {
    NETDEV_DPDK_PORT_DIR_RX,
    NETDEV_DPDK_PORT_DIR_TX,
    NUM_NETDEV_DPDK_PORT_DIR,
};

enum netdev_dpdk_port_dir
netdev_dpdk_get_port_dir(struct netdev *netdev);

int
netdev_dpdk_meter_set(ofproto_meter_id meter_id,
                      struct ofputil_meter_config *config);
int
netdev_dpdk_meter_get(ofproto_meter_id meter_id,
                      struct ofputil_meter_stats *stats, uint16_t n_bands);
int
netdev_dpdk_meter_del(ofproto_meter_id meter_id,
                      struct ofputil_meter_stats *stats, uint16_t n_bands);
void
netdev_dpdk_meters_init(void);
bool
netdev_dpdk_meter_ref(uint32_t meter_id);
void
netdev_dpdk_meter_unref(uint32_t meter_id);

bool
netdev_dpdk_sw_meter_get(struct netdev *netdev,
                         struct ofputil_meter_config *config,
                         struct ofputil_meter_band *band);

bool
netdev_dpdk_sw_meter_get_dry_run(struct netdev *netdev);

struct netdev_dpdk_sw_meter {
    struct ofputil_meter_config config;
    struct ofputil_meter_band band;
    bool dry_run;
};

#define NETDEV_DPDK_METER_PORT_ID 0
#define NETDEV_DPDK_METER_POLICY_ID 25
#define NETDEV_DPDK_MAX_METERS (MAX_METERS + RTE_MAX_ETHPORTS)

int
netdev_dpdk_sw_meter_parse(const char *s,
                           struct netdev_dpdk_sw_meter *m,
                           const char **errp);

void
netdev_dpdk_sw_meter_copy(struct netdev_dpdk_sw_meter *dst,
                          const struct netdev_dpdk_sw_meter *src);

bool
netdev_dpdk_sw_meter_equal(const struct netdev_dpdk_sw_meter *m1,
                           const struct netdev_dpdk_sw_meter *m2);

void
add_meter_policy(dpdk_port_t port_id, uint32_t policy_id);

void
netdev_dpdk_get_config_sw_meter(struct smap *args,
                                struct netdev_dpdk_sw_meter *requested_sw_meter,
                                struct netdev_dpdk_sw_meter *sw_meter);

int
netdev_dpdk_set_config_sw_meter(struct netdev *netdev, const struct smap *args,
                                struct netdev_dpdk_sw_meter *requested_sw_meter,
                                bool dev_started,
                                char **errp);

void
netdev_dpdk_set_config_port_dir(const char *devargs,
                                enum netdev_dpdk_port_dir *port_dir);

char *
netdev_dpdk_generate_devargs(const char *name, char *devargs, size_t maxlen,
                             char iface[IFNAMSIZ]);

int
get_phys_port_name(const char *devname, char *outp, size_t maxlen);

int
netdev_dpdk_setup_hairpin_queues(struct netdev *netdev, dpdk_port_t port_id,
                                 int rxq_size, int txq_size,
                                 int n_rxq, int n_txq,
                                 struct rte_eth_hairpin_conf *hairpin_conf);

const char *
adapt_doca_mlx5_devargs(const char *devargs, char *new_devargs, size_t maxlen);

void
netdev_dpdk_get_custom_stats_sw_meter(const struct netdev *netdev,
                                      struct netdev_custom_stats *custom_stats);

#else

static inline int
netdev_dpdk_meter_set(ofproto_meter_id meter_id OVS_UNUSED,
                      struct ofputil_meter_config *config OVS_UNUSED)
{
    return EOPNOTSUPP;
}

static inline int
netdev_dpdk_meter_get(ofproto_meter_id meter_id OVS_UNUSED,
                      struct ofputil_meter_stats *stats OVS_UNUSED,
                      uint16_t n_bands OVS_UNUSED)
{
    return EOPNOTSUPP;
}

static inline int
netdev_dpdk_meter_del(ofproto_meter_id meter_id OVS_UNUSED,
                      struct ofputil_meter_stats *stats OVS_UNUSED,
                      uint16_t n_bands OVS_UNUSED)
{
    return EOPNOTSUPP;
}

static inline bool
netdev_dpdk_is_esw_mgr(struct netdev *netdev OVS_UNUSED)
{
    return false;
}

static inline int
netdev_dpdk_get_esw_mgr_port_id(struct netdev *netdev OVS_UNUSED)
{
    return -1;
}

static inline bool
netdev_dpdk_is_ethdev(struct netdev *netdev OVS_UNUSED)
{
    return false;
}

#endif /* DPDK_NETDEV */

#endif /* netdev-dpdk-ext.h */
