/*
 * Copyright (c) 2019-2024 NVIDIA CORPORATION & AFFILIATES. All rights reserved.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at:
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
#include <config.h>

#include <sys/types.h>
#include <netinet/ip6.h>
#include <rte_flow.h>
#include <rte_gre.h>

#include "batch.h"
#include "cmap.h"
#include "conntrack-offload.h"
#include "dpdk-offload-provider.h"
#include "dpif-netdev.h"
#include "id-fpool.h"
#include "id-refmap.h"
#include "netdev-offload.h"
#include "netdev-offload-dpdk-ext.h"
#include "netdev-offload-provider.h"
#include "netdev-provider.h"
#include "netdev-vport.h"
#include "ovs-atomic.h"
#include "ovs-doca.h"
#include "odp-util.h"
#include "offload-metadata.h"
#include "openvswitch/match.h"
#include "openvswitch/vlog.h"
#include "ovs-rcu.h"
#include "packets.h"
#include "refmap.h"
#include "salloc.h"
#include "uuid.h"
#include "unixctl.h"

VLOG_DEFINE_THIS_MODULE(netdev_offload_dpdk_ext);
static struct vlog_rate_limit rl = VLOG_RATE_LIMIT_INIT(600, 600);

static bool netdev_offload_dpdk_ct_labels_mapping = false;
static bool netdev_offload_dpdk_disable_zone_tables = false;

struct per_thread {
PADDED_MEMBERS(CACHE_LINE_SIZE,
    char scratch[10000];
    struct salloc *s;
    struct ovs_list e2e_conn_list;
    struct ovs_list conn_list;
);
};

struct per_thread per_threads[MAX_OFFLOAD_THREAD_NB];

void
per_thread_init(void)
{
    struct per_thread *pt = &per_threads[netdev_offload_thread_id()];

    if (pt->s == NULL) {
        pt->s = salloc_init(pt->scratch, sizeof pt->scratch);
        ovs_list_init(&pt->conn_list);
        ovs_list_init(&pt->e2e_conn_list);
    }
    salloc_reset(pt->s);
}

static void *
per_thread_xmalloc(size_t n)
{
    struct per_thread *pt = &per_threads[netdev_offload_thread_id()];
    void *p = salloc(pt->s, n);

    if (p == NULL) {
        p = xmalloc(n);
    }

    return p;
}

void *
per_thread_xzalloc(size_t n)
{
    struct per_thread *pt = &per_threads[netdev_offload_thread_id()];
    void *p = szalloc(pt->s, n);

    if (p == NULL) {
        p = xzalloc(n);
    }

    return p;
}

void *
per_thread_xcalloc(size_t n, size_t sz)
{
    struct per_thread *pt = &per_threads[netdev_offload_thread_id()];
    void *p = scalloc(pt->s, n, sz);

    if (p == NULL) {
        p = xcalloc(n, sz);
    }

    return p;
}

void *
per_thread_xrealloc(void *old_p, size_t old_size, size_t new_size)
{
    struct per_thread *pt = &per_threads[netdev_offload_thread_id()];
    void *new_p = NULL;

    if (salloc_contains(pt->s, old_p)) {
        new_p = srealloc(pt->s, old_p, new_size);
        if (new_p == NULL) {
            new_p = xmalloc(new_size);
            if (new_p) {
                memcpy(new_p, old_p, old_size);
            }
        }
    } else {
        new_p = xrealloc(old_p, new_size);
    }

    return new_p;
}

void
per_thread_free(void *p)
{
    struct per_thread *pt = &per_threads[netdev_offload_thread_id()];

    if (salloc_contains(pt->s, p)) {
        /* The only freeing done in the scratch allocator is when resetting it.
         * However, realloc has a chance to shrink, so still attempt it. */
        srealloc(pt->s, p, 0);
    } else {
        free(p);
    }
}

struct ovs_list *
per_thread_get_e2e_conn_list(int tid)
{
    return &per_threads[tid].e2e_conn_list;
}

struct ovs_list *
per_thread_get_conn_list(int tid)
{
    return &per_threads[tid].conn_list;
}

struct esw_members_aux {
    struct netdev *esw_netdev;
    int ret;
    /* Operation to perform on ESW members */
    bool (*op)(struct netdev *, odp_port_t, void *);
};

static struct reg_field reg_fields[] = {
    [REG_FIELD_CT_STATE] = {
        .type = REG_TYPE_TAG,
        .index = 0,
        .offset = 0,
        .mask = 0x000000FF,
    },
    [REG_FIELD_CT_ZONE] = {
        .type = REG_TYPE_TAG,
        .index = 0,
        .offset = 8,
        .mask = 0x000000FF,
    },
    [REG_FIELD_TUN_INFO] = {
        .type = REG_TYPE_TAG,
        .index = 0,
        .offset = 16,
        .mask = 0x0000FFFF,
    },
    [REG_FIELD_CT_MARK] = {
        .type = REG_TYPE_TAG,
        .index = 1,
        .offset = 0,
        .mask = 0xFFFFFFFF,
    },
    [REG_FIELD_CT_LABEL_ID] = {
        .type = REG_TYPE_TAG,
        .index = 2,
        .offset = 0,
        .mask = 0xFFFFFFFF,
    },
    [REG_FIELD_CT_CTX] = {
        .type = REG_TYPE_META,
        .index = 0,
        .offset = 0,
        .mask = 0x0000FFFF,
    },
    /* Since sFlow and CT will not work concurrently is it safe
     * to have the reg_fields use the same bits for SFLOW_CTX and CT_CTX.
     */
    [REG_FIELD_SFLOW_CTX] = {
        .type = REG_TYPE_META,
        .index = 0,
        .offset = 0,
        .mask = 0x0000FFFF,
    },
    /* Since recirc-reg and CT will not work concurrently is it safe
     * to have the reg_fields use the same bits for RECIRC and CT_MARK.
     */
    [REG_FIELD_RECIRC] = {
        .type = REG_TYPE_TAG,
        .index = 1,
        .offset = 0,
        .mask = 0xFFFFFFFF,
    },
    [REG_FIELD_FLOW_INFO] = {
        .type = REG_TYPE_MARK,
        .index = 0,
        .offset = 0,
        .mask = 0x00FFFFFF,
    },
};

static struct reg_field *
rte_get_reg_fields(void)
{
    return reg_fields;
}

static bool
esw_members_cb(struct netdev *netdev,
               odp_port_t odp_port OVS_UNUSED,
               void *aux_);
static bool
flush_esw_members_op(struct netdev *netdev,
                     odp_port_t odp_port OVS_UNUSED,
                     void *aux_);
static bool
init_esw_members_op(struct netdev *netdev,
                    odp_port_t odp_port OVS_UNUSED,
                    void *aux_ OVS_UNUSED);
static bool
uninit_esw_members_op(struct netdev *netdev,
                      odp_port_t odp_port OVS_UNUSED,
                      void *aux_ OVS_UNUSED);

#define MIN_LABEL_ID 1
#define MAX_LABEL_ID (dpdk_offload_api->reg_fields()[REG_FIELD_CT_LABEL_ID].mask - 1)
#define NUM_LABEL_ID (MAX_LABEL_ID - MIN_LABEL_ID + 1)

static struct id_refmap *label_id_irfm;

static struct ds *
format_label_id_key(struct ds *s, void *key)
{
    ovs_u128 not_mapped_ct_label = *(ovs_u128 *) key;

    ds_put_format(s, "label = %x%x%x%x", not_mapped_ct_label.u32[3],
                                         not_mapped_ct_label.u32[2],
                                         not_mapped_ct_label.u32[1],
                                         not_mapped_ct_label.u32[0]);
    return s;
}

static void
label_id_init(void)
{
    static struct ovsthread_once init_once = OVSTHREAD_ONCE_INITIALIZER;

    if (ovsthread_once_start(&init_once)) {
        label_id_irfm = id_refmap_create(MAX_OFFLOAD_THREAD_NB, "label-id", sizeof(ovs_u128),
                                         format_label_id_key, MIN_LABEL_ID, NUM_LABEL_ID,
                                         0, false);
        ovsthread_once_done(&init_once);
    }
}

static int
get_label_id(ovs_u128 *ct_label, uint32_t *ct_label_id)
{
    if (is_all_zeros(ct_label, sizeof *ct_label)) {
        *ct_label_id = 0;
        return 0;
    }

    return !id_refmap_ref(label_id_irfm, netdev_offload_thread_id(), ct_label, ct_label_id);
}

static void
put_label_id(uint32_t label_id)
{
    id_refmap_unref(label_id_irfm, netdev_offload_thread_id(), label_id);
}

static int
find_ct_label(int label_id, ovs_u128 *ct_label)
{
    ovs_u128 *key = id_refmap_find_key(label_id_irfm, label_id);

    if (key) {
        memcpy(ct_label, key, sizeof *ct_label);
        return 0;
    }
    return -1;
}

static struct id_refmap *zone_id_irfm;

static struct ds *
format_zone_id_key(struct ds *s, void *key)
{
    uint16_t not_mapped_ct_zone = *(uint16_t *) key;

    ds_put_format(s, "zone = %d", not_mapped_ct_zone);
    return s;
}

static void
zone_id_init(void)
{
    static struct ovsthread_once init_once = OVSTHREAD_ONCE_INITIALIZER;

    if (ovsthread_once_start(&init_once)) {
        zone_id_irfm = id_refmap_create(MAX_OFFLOAD_THREAD_NB, "zone-id",
                                        sizeof(uint16_t), format_zone_id_key,
                                        MIN_ZONE_ID, NUM_ZONE_ID, 0, false);
        ovsthread_once_done(&init_once);
    }
}

int
get_zone_id(uint16_t ct_zone, uint32_t *ct_zone_id)
{
    if (ovs_doca_initialized()) {
        *ct_zone_id = ct_zone;
        return 0;
    }

    return !id_refmap_ref(zone_id_irfm, netdev_offload_thread_id(),
                          &ct_zone, ct_zone_id);
}

static void
put_zone_id(uint32_t zone_id)
{
    if (ovs_doca_initialized()) {
        return;
    }

    id_refmap_unref(zone_id_irfm, netdev_offload_thread_id(), zone_id);
}

static struct id_fpool *table_id_pool;
static struct refmap *table_id_rfm;

OVS_ASSERT_PACKED(struct table_ctx_key,
    struct netdev *netdev;
    odp_port_t vport;
    uint32_t recirc_id;
);

static int
add_miss_flow(struct netdev *netdev,
              uint32_t src_table_id,
              uint32_t dst_table_id,
              uint32_t recirc_id,
              uint32_t mark_id,
              struct dpdk_offload_handle *doh);

static int
table_ctx_init(void *ctx_, void *arg_)
{
    struct table_ctx_key *arg = arg_;
    struct table_ctx *ctx = ctx_;
    uint32_t e2e_table_id;

    /* Allocate a dedicated ID always if a table ID context has been created. */
    if (!id_fpool_new_id(table_id_pool, netdev_offload_thread_id(), &ctx->id)) {
        return -1;
    }

    if (!netdev_is_e2e_cache_enabled() || arg->recirc_id != 0) {
        return 0;
    }

    if (arg->vport == ODPP_NONE && arg->recirc_id == 0 && netdev_is_e2e_cache_enabled()) {
        e2e_table_id = 0;
    } else {
        e2e_table_id = E2E_BASE_TABLE_ID | (OVS_FORCE uint32_t) arg->vport;
    }

    ctx->netdev = netdev_ref(arg->netdev);
    if (add_miss_flow(ctx->netdev, e2e_table_id, ctx->id, 0, 0,
                      &ctx->miss_flow)) {
        ctx->netdev = NULL;
        netdev_close(ctx->netdev);
        return -1;
    }
    return 0;
}

static void
table_ctx_uninit(void *ctx_)
{
    struct table_ctx *ctx = ctx_;

    if (!netdev_is_e2e_cache_enabled() || !ctx->netdev) {
       return;
    }

    id_fpool_free_id(table_id_pool, netdev_offload_thread_id(), ctx->id);
    ctx->id = 0;
    netdev_offload_dpdk_destroy_flow(ctx->netdev, &ctx->miss_flow, NULL, true);
    netdev_close(ctx->netdev);
    ctx->netdev = NULL;
}

static struct ds *
table_ctx_format(struct ds *s, void *key_,
                 void *ctx_ OVS_UNUSED, void *arg_ OVS_UNUSED)
{
    struct table_ctx_key *key = key_;
    const char *netdev_name;

    netdev_name = key->netdev
                  ? netdev_get_name(key->netdev)
                  : NULL;
    ds_put_format(s, "%s, vport=%"PRIu32", recirc_id=%"PRIu32,
                  netdev_name, key->vport, key->recirc_id);
    return s;
}

static void
table_id_init(void)
{
    static struct ovsthread_once init_once = OVSTHREAD_ONCE_INITIALIZER;

    if (ovsthread_once_start(&init_once)) {
        table_id_pool = id_fpool_create(MAX_OFFLOAD_THREAD_NB, MIN_TABLE_ID, NUM_TABLE_ID);
        table_id_rfm = refmap_create("table-id",
                                     sizeof(struct table_ctx_key), sizeof(struct table_ctx),
                                     table_ctx_init, table_ctx_uninit, table_ctx_format, false);
        ovsthread_once_done(&init_once);
    }
}

static struct table_ctx table_zero_ctx;

int
netdev_offload_dpdk_get_table_ctx(odp_port_t vport,
                                  uint32_t recirc_id,
                                  struct netdev *physdev,
                                  bool is_e2e_cache,
                                  struct table_ctx **ctx)
{
    struct table_ctx_key key = {
        .netdev = physdev,
        .vport = vport,
        .recirc_id = recirc_id,
    };
    struct table_ctx_key arg = key;

    if (vport == ODPP_NONE && recirc_id == 0 &&
        !(netdev_is_e2e_cache_enabled() && !is_e2e_cache)) {
        *ctx = &table_zero_ctx;
        return 0;
    }

    if (is_e2e_cache) {
        *ctx = xzalloc(sizeof(struct table_ctx));
        (*ctx)->id = E2E_BASE_TABLE_ID | (OVS_FORCE uint32_t) vport;
        return 0;
    }

    if (vport != ODPP_NONE && ovs_doca_enabled()) {
        key.netdev = NULL;
    }

    if (ovs_doca_enabled() && (recirc_id & MATCH_RECIRC_DP_HASH_BIT)) {
        key.recirc_id = MATCH_RECIRC_DP_HASH_BIT;
    }

    *ctx = refmap_ref(table_id_rfm, &key, &arg);
    return (*ctx) ? 0 : -1;
}

static void
put_table_ctx(struct table_ctx *ctx)
{
    if (ctx == &table_zero_ctx || ctx == NULL) {
        return;
    }

    if (ctx->id > MAX_TABLE_ID) {
        /* This table context was allocated for an E2E flow offload.
         * It has not been managed by the refmap, only allocated on the heap. */
        free(ctx);
    } else {
        refmap_unref(table_id_rfm, ctx);
    }
}

#define MIN_SFLOW_ID 1
#define MAX_SFLOW_ID (dpdk_offload_api->reg_fields()[REG_FIELD_SFLOW_CTX].mask - 1)
#define NUM_SFLOW_ID (MAX_SFLOW_ID - MIN_SFLOW_ID + 1)

static struct id_refmap *sflow_id_irfm;

OVS_ASSERT_PACKED(struct sflow_ctx,
    struct dpif_sflow_attr sflow_attr;
    struct user_action_cookie cookie;
    struct flow_tnl sflow_tnl;
);

static struct ds *
format_sflow_key(struct ds *s, void *key)
{
    struct sflow_ctx *sflow_ctx = key;
    struct user_action_cookie *cookie;

    cookie = &sflow_ctx->cookie;
    ds_put_format(s, "sFlow cookie %p, ofproto_uuid "UUID_FMT,
                  cookie, UUID_ARGS(&cookie->ofproto_uuid));
    return s;
}

static void
sflow_id_init(void)
{
    static struct ovsthread_once init_once = OVSTHREAD_ONCE_INITIALIZER;

    if (ovsthread_once_start(&init_once)) {
        sflow_id_irfm = id_refmap_create(MAX_OFFLOAD_THREAD_NB, "sflow-id",
                                         sizeof(struct sflow_ctx), format_sflow_key,
                                         MIN_SFLOW_ID, NUM_SFLOW_ID, 0, false);
        ovsthread_once_done(&init_once);
    }
}

static int
get_sflow_id(struct sflow_ctx *sflow_ctx, uint32_t *sflow_id)
{
    return !id_refmap_ref(sflow_id_irfm, netdev_offload_thread_id(),
                          sflow_ctx, sflow_id);
}

static void
put_sflow_id(uint32_t sflow_id)
{
    id_refmap_unref(sflow_id_irfm, netdev_offload_thread_id(), sflow_id);
}

static int
find_sflow_ctx(int sflow_id, struct sflow_ctx *ctx)
{
    struct sflow_ctx *key = id_refmap_find_key(sflow_id_irfm, sflow_id);

    if (key) {
        memcpy(ctx, key, sizeof *ctx);
        return 0;
    }
    return -1;
}

#define MIN_CT_CTX_ID 1
#define MAX_CT_CTX_ID (dpdk_offload_api->reg_fields()[REG_FIELD_CT_CTX].mask - 1)
#define NUM_CT_CTX_ID (MAX_CT_CTX_ID - MIN_CT_CTX_ID + 1)

static struct id_refmap *ct_ctx_irfm;

OVS_ASSERT_PACKED(struct ct_miss_ctx,
    ovs_u128 label;
    uint32_t mark;
    uint16_t zone;
    uint8_t state;
    /* Manual padding must be used instead of PADDED_MEMBERS. */
    uint8_t pad[1];
);

static struct ds *
format_ct_miss_ctx(struct ds *s, void *ctx_)
{
    struct ct_miss_ctx *ctx = ctx_;
    ovs_be128 label;

    label = hton128(ctx->label);
    ds_put_format(s, "ct_state=0x%" PRIx8 ", ct_zone=%d, ct_mark=0x%" PRIx32 ", ct_label=",
                  ctx->state, ctx->zone, ctx->mark);
    ds_put_hex(s, &label, sizeof label);
    return s;
}

/* In ofproto/ofproto-dpif-rid.c, function recirc_run. Timeout for expired
 * flows is 250 msec. Set this timeout the same.
 */
#define DELAYED_RELEASE_TIMEOUT_MS 250

static void
ct_ctx_init(void)
{
    static struct ovsthread_once init_once = OVSTHREAD_ONCE_INITIALIZER;

    if (ovsthread_once_start(&init_once)) {
        ct_ctx_irfm = id_refmap_create(MAX_OFFLOAD_THREAD_NB, "ct-miss-ctx",
                                       sizeof(struct ct_miss_ctx), format_ct_miss_ctx,
                                       MIN_CT_CTX_ID, NUM_CT_CTX_ID,
                                       DELAYED_RELEASE_TIMEOUT_MS, false);
        ovsthread_once_done(&init_once);
    }
}

static int
get_ct_ctx_id(struct ct_miss_ctx *ctx, uint32_t *ct_ctx_id)
{
    if (ovs_doca_initialized()) {
        *ct_ctx_id = 0;
        return 0;
    }

    return !id_refmap_ref(ct_ctx_irfm, netdev_offload_thread_id(),
                          ctx, ct_ctx_id);
}

static void
put_ct_ctx_id(uint32_t ct_ctx_id)
{
    if (ovs_doca_initialized()) {
        return;
    }

    id_refmap_unref(ct_ctx_irfm, netdev_offload_thread_id(), ct_ctx_id);
}

static int
find_ct_miss_ctx(int ct_ctx_id, struct ct_miss_ctx *ctx)
{
    struct ct_miss_ctx *key = id_refmap_find_key(ct_ctx_irfm, ct_ctx_id);

    if (key) {
        memcpy(ctx, key, sizeof *ctx);
        return 0;
    }
    return -1;
}

#define MIN_TUNNEL_ID 1
#define MAX_TUNNEL_ID (dpdk_offload_api->reg_fields()[REG_FIELD_TUN_INFO].mask - 1)
#define NUM_TUNNEL_ID (MAX_TUNNEL_ID - MIN_TUNNEL_ID + 1)

static struct id_refmap *tnl_irfm;

static struct ds *
format_tnl_key(struct ds *s, void *key)
{
    struct flow_tnl *tnl = key;

    ds_put_format(s, IP_FMT" -> "IP_FMT", tun_id=%"PRIu64,
                  IP_ARGS(tnl->ip_src), IP_ARGS(tnl->ip_dst),
                  ntohll(tnl->tun_id));
    return s;
}

static void
tnl_id_init(void)
{
    static struct ovsthread_once init_once = OVSTHREAD_ONCE_INITIALIZER;

    if (ovsthread_once_start(&init_once)) {
        tnl_irfm = id_refmap_create(MAX_OFFLOAD_THREAD_NB, "tunnel",
                                    2 * sizeof(struct flow_tnl), format_tnl_key,
                                    MIN_TUNNEL_ID, NUM_TUNNEL_ID, 0, false);
        ovsthread_once_done(&init_once);
    }
}

static void
get_tnl_masked(struct flow_tnl *dst_key, struct flow_tnl *dst_mask,
               struct flow_tnl *src_key, struct flow_tnl *src_mask)
{
    char *psrc_key;
    char *pdst_key, *pdst_mask;
    int i;

    if (dst_mask) {
        memcpy(dst_mask, src_mask, sizeof *dst_mask);
        memset(&dst_mask->metadata, 0, sizeof dst_mask->metadata);

        pdst_key = (char *)dst_key;
        psrc_key = (char *)src_key;
        pdst_mask = (char *)dst_mask;
        for (i = 0; i < sizeof *dst_key; i++) {
            *pdst_key++ = *psrc_key++ & *pdst_mask++;
        }
    } else {
        memcpy(dst_key, src_key, sizeof *dst_key);
    }
}

static int
get_tnl_id(struct flow_tnl *tnl_key, struct flow_tnl *tnl_mask,
           uint32_t *tnl_id)
{
    struct flow_tnl tnl_tmp[2];

    get_tnl_masked(&tnl_tmp[0], &tnl_tmp[1], tnl_key, tnl_mask);
    if (is_all_zeros(&tnl_tmp, sizeof tnl_tmp)) {
        *tnl_id = 0;
        return 0;
    }

    return !id_refmap_ref(tnl_irfm, netdev_offload_thread_id(), tnl_tmp, tnl_id);
}

static void
put_tnl_id(uint32_t tnl_id)
{
    id_refmap_unref(tnl_irfm, netdev_offload_thread_id(), tnl_id);
}

static struct ds *
dump_flow_ctx_id(struct ds *s, void *data,
                 void *priv OVS_UNUSED, void *priv_arg OVS_UNUSED)
{
    struct flow_miss_ctx *flow_ctx_data = data;

    ds_put_format(s, "vport=%"PRIu32", recirc_id=%"PRIu32", ",
                  flow_ctx_data->vport, flow_ctx_data->recirc_id);
    format_tnl_key(s, &flow_ctx_data->tnl);

    return s;
}

struct flow_miss_ctx_priv_arg {
    struct netdev *netdev;
    uint32_t table_id;
    uint32_t recirc_id;
};

struct flow_miss_ctx_priv {
    struct netdev *netdev;
    struct dpdk_offload_handle miss_flow;
};

static int
flow_miss_ctx_priv_init(void *priv_, void *priv_arg_, uint32_t mark_id)
{
    struct flow_miss_ctx_priv_arg *priv_arg = priv_arg_;
    struct flow_miss_ctx_priv *priv = priv_;

    if (priv->netdev != NULL) {
        return 0;
    }

    priv->netdev = netdev_ref(priv_arg->netdev);
    if (add_miss_flow(priv->netdev, priv_arg->table_id, MISS_TABLE_ID,
                      priv_arg->recirc_id, mark_id, &priv->miss_flow)) {
        netdev_close(priv->netdev);
        priv->netdev = NULL;
        return -1;
    }

    return 0;
}

static void
flow_miss_ctx_priv_uninit(void *priv_)
{
    struct flow_miss_ctx_priv *priv = priv_;

    if (!priv->netdev) {
       return;
    }

    netdev_offload_dpdk_destroy_flow(priv->netdev, &priv->miss_flow, NULL, true);
    netdev_close(priv->netdev);
    priv->netdev = NULL;
}

static struct offload_metadata *flow_miss_ctx_md;

static void
flow_miss_ctx_init(void)
{
    static struct ovsthread_once init_once = OVSTHREAD_ONCE_INITIALIZER;

    if (ovsthread_once_start(&init_once)) {
        struct offload_metadata_parameters params = {
            .id_alloc = netdev_offload_flow_mark_alloc,
            .id_free = netdev_offload_flow_mark_free,
            .priv_size = sizeof(struct flow_miss_ctx_priv),
            .priv_init = flow_miss_ctx_priv_init,
            .priv_uninit = flow_miss_ctx_priv_uninit,
            .release_delay_ms = DELAYED_RELEASE_TIMEOUT_MS,
        };

        flow_miss_ctx_md = offload_metadata_create(MAX_OFFLOAD_THREAD_NB, "flow_miss_ctx",
                                                   sizeof(struct flow_miss_ctx),
                                                   dump_flow_ctx_id, params);

        ovsthread_once_done(&init_once);
    }
}

int
get_flow_miss_ctx_id(struct flow_miss_ctx *flow_ctx_data,
                     struct netdev *netdev,
                     uint32_t table_id,
                     uint32_t recirc_id,
                     uint32_t *miss_ctx_id)
{
    struct flow_miss_ctx_priv_arg priv_arg = {
        .netdev = netdev,
        .table_id = table_id,
        .recirc_id = recirc_id,
    };

    return offload_metadata_id_ref(flow_miss_ctx_md, flow_ctx_data, &priv_arg,
                                   miss_ctx_id);
}

static void
put_flow_miss_ctx_id(uint32_t flow_ctx_id)
{
    offload_metadata_id_unref(flow_miss_ctx_md,
                              netdev_offload_thread_id(),
                              flow_ctx_id);
}

static int
associate_flow_id(uint32_t flow_id, struct flow_miss_ctx *flow_ctx_data)
{
    offload_metadata_id_set(flow_miss_ctx_md, flow_ctx_data, flow_id);
    return 0;
}

static int
disassociate_flow_id(uint32_t flow_id)
{
    offload_metadata_id_unset(flow_miss_ctx_md,
                              netdev_offload_thread_id(),
                              flow_id);
    return 0;
}

static int
shared_ctx_init(void *ctx_, void *arg_)
{
    struct ovs_shared_arg *arg = arg_;
    struct ovs_shared_ctx *ctx = ctx_;
    struct rte_flow_error error;

    ctx->res_type = arg->type;
    ctx->port_id = netdev_dpdk_get_esw_mgr_port_id(arg->netdev);

    if (dpdk_offload_api->shared_create(ctx, arg, &error)) {
        ctx->port_id = -1;
        return -1;
    }

    ctx->netdev = arg->netdev;

    return 0;
}

static void
shared_ctx_uninit(void *ctx_)
{
    struct ovs_shared_ctx *ctx = ctx_;
    struct rte_flow_error error;

    if (!ctx) {
        return;
    }

    dpdk_offload_api->shared_destroy(ctx, &error);
    ctx->port_id = -1;
}

static struct ds *
dump_shared_age(struct ds *s, void *data,
                void *priv, void *priv_arg)
{
    struct ovs_shared_arg *arg = priv_arg;
    struct ovs_shared_ctx *ctx = priv;
    uintptr_t *key = data;

    if (key) {
        ds_put_format(s, "netdev=%s, key=0x%"PRIxPTR", ",
                      arg ? netdev_get_name(arg->netdev) : "nil",
                      *((uintptr_t *) key));
    }
    if (ctx) {
        ds_put_format(s, "ctx->port_id=%d, ctx->act_hdl=%p",
                      ctx->port_id, ctx->act_hdl);
    } else {
        ds_put_cstr(s, "ctx=NULL");
    }
    return s;
}

static struct refmap *shared_age_rfm;

static void
shared_age_init(void)
{
    static struct ovsthread_once init_once = OVSTHREAD_ONCE_INITIALIZER;

    if (ovsthread_once_start(&init_once)) {
        shared_age_rfm = refmap_create("shared-age",
                                       sizeof(uintptr_t), sizeof(struct ovs_shared_ctx),
                                       shared_ctx_init, shared_ctx_uninit, dump_shared_age,
                                       false);
        ovsthread_once_done(&init_once);
    }
}

struct ovs_shared_ctx *
get_shared_age_ctx(struct netdev *netdev,
                     uintptr_t app_counter_key,
                     bool create)
{
    struct rte_flow_action_age age_conf = {
        .timeout = 0xFFFFFF,
    };
    struct rte_flow_action action = {
        .type = RTE_FLOW_ACTION_TYPE_AGE,
        .conf = &age_conf,
    };

    if (create) {
        struct ovs_shared_arg arg = {
            .netdev = netdev,
            .action = &action,
        };

        return refmap_ref(shared_age_rfm, &app_counter_key, &arg);
    } else {
        return refmap_find(shared_age_rfm, &app_counter_key);
    }
}

static void
free_shared_age_ctx(void *ctx)
{
    refmap_unref(shared_age_rfm, ctx);
}

static struct ds *
dump_shared_count(struct ds *s, void *data, void *priv, void *priv_arg)
{
    struct ovs_shared_arg *arg = priv_arg;
    struct flows_counter_key *fck = data;
    struct ovs_shared_ctx *ctx = priv;
    const char *type;

    ovs_assert(arg);

    switch (arg->type) {
    case OVS_SHARED_COUNT:
        type = "FLOW";
        break;
    case OVS_SHARED_CT_COUNT:
        type = "CT";
        break;
    case OVS_SHARED_CT_ACTIONS:
        type = "CT_ACTIONS";
        break;
    case OVS_SHARED_UNDEFINED:
        type = "UNDEFINED";
        break;
    default:
        OVS_NOT_REACHED();
    }

    if (fck) {
        ovs_u128 *last_ufid_key;

        last_ufid_key = &fck->ufid_key[OFFLOAD_FLOWS_COUNTER_KEY_SIZE - 1];
        if (ovs_u128_is_zero(*last_ufid_key)) {
            ds_put_format(s, "netdev=%s, type=%s key=0x%"PRIxPTR", ",
                          arg ? netdev_get_name(arg->netdev) : "nil",
                          type, fck->ptr_key);
        } else {
            char key_str[OFFLOAD_FLOWS_COUNTER_KEY_STRING_SIZE];

            netdev_flow_counter_key_to_string(fck, key_str, sizeof key_str);
            ds_put_format(s, "netdev=%s, type=%s, key=%s, ",
                          arg ? netdev_get_name(arg->netdev) : "nil",
                          type, key_str);
        }
    }
    if (ctx) {
        ds_put_format(s, "ctx->port_id=%d, type=%s, ctx->act_hdl=%p",
                      ctx->port_id, type, ctx->act_hdl);
    } else {
        ds_put_cstr(s, "ctx=NULL");
    }
    return s;
}

static struct refmap *shared_count_rfm;

static void
shared_count_init(void)
{
    static struct ovsthread_once init_once = OVSTHREAD_ONCE_INITIALIZER;

    if (ovsthread_once_start(&init_once)) {
        shared_count_rfm = refmap_create("shared-count", sizeof(struct flows_counter_key),
                                         sizeof(struct ovs_shared_ctx), shared_ctx_init,
                                         shared_ctx_uninit, dump_shared_count, true);
        ovsthread_once_done(&init_once);
    }
}


struct ovs_shared_ctx *
get_shared_count_ctx(struct netdev *netdev,
                     struct flows_counter_key *key,
                     enum ovs_shared_type type,
                     bool create)
{
    struct rte_flow_action action = {
        .type = RTE_FLOW_ACTION_TYPE_COUNT,
    };

    if (create) {
        struct ovs_shared_arg arg = {
            .netdev = netdev,
            .action = &action,
            .type = type,
        };

        return refmap_ref(shared_count_rfm, key, &arg);
    } else {
        return refmap_find(shared_count_rfm, key);
    }
}

static void
free_shared_count_ctx(void *ctx)
{
    refmap_unref(shared_count_rfm, ctx);
}

static struct refmap *shared_ct_actions_rfm;

static struct ds *
dump_shared_ct_actions(struct ds *s, void *key_, void *ctx_, void *arg_)
{
    struct shared_ct_actions_key *key = key_;
    struct ovs_shared_ctx *ctx = ctx_;
    struct ovs_shared_arg *arg = arg_;

    ovs_assert(arg->type == OVS_SHARED_CT_ACTIONS);

    ds_put_format(s, "mark=0x%08" PRIx32 ", label=0x%08" PRIx32 ", state=0x%02" PRIx8 ", ",
                  key->ct_mark, key->ct_label, key->ct_state);

    if (ctx) {
        ds_put_format(s, "port-id=%d, id=%" PRIu32, ctx->port_id, ctx->res_id);
    } else {
        ds_put_cstr(s, "ctx=NULL");
    }

    return s;
}

static void
shared_ct_actions_init(void)
{
    static struct ovsthread_once init_once = OVSTHREAD_ONCE_INITIALIZER;

    if (ovsthread_once_start(&init_once)) {
        shared_ct_actions_rfm = refmap_create("shared-ct-actions",
                                              sizeof(struct shared_ct_actions_key),
                                              sizeof(struct ovs_shared_ctx), shared_ctx_init,
                                              shared_ctx_uninit, dump_shared_ct_actions, false);
        ovsthread_once_done(&init_once);
    }
}


static struct ovs_shared_ctx *
get_shared_ct_actions_ctx(struct netdev *netdev,
                          uint32_t ct_label, uint32_t ct_mark, uint8_t ct_state)
{
    struct shared_ct_actions_key key = {
        .ct_label = ct_label,
        .ct_mark = ct_mark,
        .ct_state = ct_state,
    };
    struct ovs_shared_arg arg = {
        .type = OVS_SHARED_CT_ACTIONS,
        .netdev = netdev,
        .ct_actions_key = &key,
    };

    return refmap_ref(shared_ct_actions_rfm, &key, &arg);
}

static void
free_shared_ct_actions_ctx(void *ctx)
{
    refmap_unref(shared_ct_actions_rfm, ctx);
}

void
put_action_resources(struct act_resources *act_resources)
{
    int i;

    put_table_ctx(act_resources->self_table_ctx);
    put_table_ctx(act_resources->next_table_ctx);
    put_flow_miss_ctx_id(act_resources->flow_miss_ctx_id);
    put_tnl_id(act_resources->tnl_id);
    if (act_resources->associated_flow_id) {
        disassociate_flow_id(act_resources->flow_id);
    }
    put_ct_ctx_id(act_resources->ct_miss_ctx_id);
    put_zone_id(act_resources->ct_match_zone_id);
    put_zone_id(act_resources->ct_action_zone_id);
    put_label_id(act_resources->ct_match_label_id);
    put_label_id(act_resources->ct_action_label_ctx_id);
    free_shared_age_ctx(act_resources->shared_age_ctx);
    free_shared_count_ctx(act_resources->shared_count_ctx);
    free_shared_ct_actions_ctx(act_resources->shared_ct_actions_ctx[CT_DIR_INIT]);
    free_shared_ct_actions_ctx(act_resources->shared_ct_actions_ctx[CT_DIR_REP]);
    put_sflow_id(act_resources->sflow_id);
    for (i = 0; i < DPDK_OFFLOAD_MAX_METERS_PER_FLOW &&
         act_resources->meter_ids[i] != 0; i++) {
        netdev_dpdk_meter_unref(act_resources->meter_ids[i]);
    }
}

static int
find_flow_miss_ctx(int flow_ctx_id, struct flow_miss_ctx *ctx)
{
    return offload_metadata_data_from_id(flow_miss_ctx_md, flow_ctx_id, ctx);
}

void
netdev_offload_dpdk_upkeep(struct netdev *netdev, bool quiescing)
{
    unsigned int tid = netdev_offload_thread_id();
    long long int now = time_last_usec() / 1000;

    dpdk_offload_api->upkeep(netdev, quiescing);

    offload_metadata_upkeep(flow_miss_ctx_md, tid, now);
}

static void
parse_geneve_opt_match(struct flow *consumed_masks,
                       struct flow_patterns *patterns,
                       struct match *match,
                       struct act_vars *act_vars)
{
    int len, opt_idx;
    uint8_t idx;
    struct geneve_opt curr_opt_spec, curr_opt_mask;
    struct gnv_opts {
        struct rte_flow_item_geneve_opt opts[TUN_METADATA_NUM_OPTS];
        uint32_t options_data[TUN_METADATA_NUM_OPTS];
    } *gnv_opts;
    BUILD_ASSERT_DECL(offsetof(struct gnv_opts, opts) == 0);
    struct gnv_opts_mask {
        struct rte_flow_item_geneve_opt opts_mask[TUN_METADATA_NUM_OPTS];
        uint32_t options_data_mask[TUN_METADATA_NUM_OPTS];
    } *gnv_opts_mask;
    BUILD_ASSERT_DECL(offsetof(struct gnv_opts_mask, opts_mask) == 0);

    len = match->flow.tunnel.metadata.present.len;
    idx = 0;
    opt_idx = 0;
    curr_opt_spec = match->flow.tunnel.metadata.opts.gnv[opt_idx];
    curr_opt_mask = match->wc.masks.tunnel.metadata.opts.gnv[opt_idx];

    if (!is_all_zeros(match->wc.masks.tunnel.metadata.opts.gnv,
                      sizeof *match->wc.masks.tunnel.metadata.opts.gnv) &&
        match->flow.tunnel.metadata.present.len) {
        while (len) {
            gnv_opts = per_thread_xzalloc(sizeof *gnv_opts);
            gnv_opts_mask = per_thread_xzalloc(sizeof *gnv_opts_mask);
            memcpy(&gnv_opts->opts[idx].option_class,
                   &curr_opt_spec.opt_class, sizeof curr_opt_spec.opt_class);
            memcpy(&gnv_opts_mask->opts_mask[idx].option_class,
                   &curr_opt_mask.opt_class,
                   sizeof curr_opt_mask.opt_class);

            gnv_opts->opts[idx].option_type = curr_opt_spec.type;
            gnv_opts_mask->opts_mask[idx].option_type = curr_opt_mask.type;

            gnv_opts->opts[idx].option_len = curr_opt_spec.length;
            gnv_opts_mask->opts_mask[idx].option_len = curr_opt_mask.length;

            /* According to the Geneve protocol
            * https://tools.ietf.org/html/draft-gross-geneve-00#section-3.1
            * Length (5 bits):  Length of the option, expressed in four byte
            * multiples excluding the option header
            * (tunnel.metadata.opts.gnv.length).
            * Opt Len (6 bits):  The length of the options fields, expressed
            * in four byte multiples, not including the eight byte
            * fixed tunnel header (tunnel.metadata.present.len).
            */
            opt_idx++;
            memcpy(&gnv_opts->options_data[opt_idx - 1],
                   &match->flow.tunnel.metadata.opts.gnv[opt_idx],
                   sizeof gnv_opts->options_data[opt_idx - 1] *
                   curr_opt_spec.length * 4);
            memcpy(&gnv_opts_mask->options_data_mask[opt_idx - 1],
                   &match->wc.masks.tunnel.metadata.opts.gnv[opt_idx],
                   sizeof gnv_opts_mask->options_data_mask[opt_idx - 1] *
                   curr_opt_spec.length * 4);

            gnv_opts->opts[opt_idx - 1].data = gnv_opts->options_data;
            gnv_opts_mask->opts_mask[opt_idx - 1].data =
            gnv_opts_mask->options_data_mask;

            add_flow_pattern(patterns, RTE_FLOW_ITEM_TYPE_GENEVE_OPT,
                             &gnv_opts->opts[idx],
                             &gnv_opts_mask->opts_mask[idx], NULL);

            len -= sizeof(struct geneve_opt) + curr_opt_spec.length * 4;
            opt_idx += sizeof(struct geneve_opt) / 4 +
                curr_opt_spec.length - 1;
            idx++;
        }
        memset(&consumed_masks->tunnel.metadata.opts.gnv, 0,
               sizeof consumed_masks->tunnel.metadata.opts.gnv);
    }
    act_vars->gnv_opts_cnt = idx;
}

int
parse_geneve_match(struct flow_patterns *patterns,
                   struct match *match,
                   struct act_vars *act_vars)
{
    struct rte_flow_item_geneve *gnv_spec, *gnv_mask;
    struct flow *consumed_masks;
    int ret;

    ret = parse_tnl_ip_match(patterns, match, IPPROTO_UDP);
    if (ret) {
        return -1;
    }

    parse_tnl_udp_match(patterns, match);

    consumed_masks = &match->wc.masks;
    /* GENEVE */
    gnv_spec = per_thread_xzalloc(sizeof *gnv_spec);
    gnv_mask = per_thread_xzalloc(sizeof *gnv_mask);

    put_unaligned_be32(ALIGNED_CAST(ovs_be32 *, gnv_spec->vni),
                       htonl(ntohll(match->flow.tunnel.tun_id) << 8));
    put_unaligned_be32(ALIGNED_CAST(ovs_be32 *, gnv_mask->vni),
                       htonl(ntohll(match->wc.masks.tunnel.tun_id) << 8));

    consumed_masks->tunnel.tun_id = 0;
    consumed_masks->tunnel.flags = 0;

    add_flow_pattern(patterns, RTE_FLOW_ITEM_TYPE_GENEVE, gnv_spec, gnv_mask,
                     NULL);
    parse_geneve_opt_match(consumed_masks, patterns, match, act_vars);

    /* tunnel.metadata.present.len value indicates the number of
     * options, it's mask does not indicate any match on the packet,
     * thus masked.
     */
    memset(&consumed_masks->tunnel.metadata.present, 0,
           sizeof consumed_masks->tunnel.metadata.present);

    return 0;
}

uint32_t
reg_field_spec(enum dpdk_reg_id reg_id, uint32_t val)
{
    struct reg_field *reg = &dpdk_offload_api->reg_fields()[reg_id];
    uint32_t offset = reg->offset;
    uint32_t mask = reg->mask;

    return (val & mask) << offset;
}

uint32_t
reg_field_mask(enum dpdk_reg_id reg_id, uint32_t mask)
{
    return reg_field_spec(reg_id, mask);
}

bool
dpdk_offload_get_reg_field(struct dp_packet *packet,
                           enum dpdk_reg_id reg_id,
                           uint32_t *val)
{
    struct reg_field *reg_field = &dpdk_offload_api->reg_fields()[reg_id];
    uint32_t mark = 0;
    uint32_t meta;

    if (reg_field->type == REG_TYPE_META) {
        if (!dp_packet_get_meta(packet, &meta)) {
            return false;
        }

        /* An error should be returned above if meta is 0.
         * We perform another validation and alert on unexpected
         * DPDK behavior.
         */
        if (meta == 0) {
            dp_packet_has_flow_mark(packet, &mark);
            VLOG_ERR_RL(&rl, "port %d, recirc=%d, mark=%d, has meta 0",
                        packet->md.in_port.odp_port, packet->md.recirc_id,
                        mark);
            return false;
        }

        meta >>= reg_field->offset;
        meta &= reg_field->mask;

        *val = meta;
        return true;
    }

    if (reg_field->type == REG_TYPE_MARK) {
        if (dp_packet_has_flow_mark(packet, &mark)) {
            *val = mark;
            return true;
        }
        return false;
    }

    OVS_NOT_REACHED();
}

int
add_pattern_match_reg_field(struct flow_patterns *patterns,
                            uint8_t reg_field_id, uint32_t val, uint32_t mask)
{
    struct rte_flow_item_meta *meta_spec, *meta_mask;
    struct rte_flow_item_tag *tag_spec, *tag_mask;
    struct reg_field *reg_field;
    uint32_t reg_spec, reg_mask;
    uint8_t reg_index;

    if (reg_field_id >= REG_FIELD_NUM) {
        VLOG_ERR("unkonwn reg id %d", reg_field_id);
        return -1;
    }
    reg_field = &dpdk_offload_api->reg_fields()[reg_field_id];
    if (val != (val & reg_field->mask)) {
        VLOG_ERR("value 0x%"PRIx32" is out of range for reg id %d", val,
                 reg_field_id);
        return -1;
    }

    reg_spec = reg_field_spec(reg_field_id, val);
    reg_mask = reg_field_mask(reg_field_id, mask);
    reg_index = reg_field->index;
    ovs_assert(reg_index < REG_TAG_INDEX_NUM);
    switch (reg_field->type) {
    case REG_TYPE_TAG:
        if (patterns->tag_spec[reg_index] == NULL) {
            tag_spec = per_thread_xzalloc(sizeof *tag_spec);
            tag_spec->index = reg_index;
            patterns->tag_spec[reg_index] = tag_spec;

            tag_mask = per_thread_xzalloc(sizeof *tag_mask);
            tag_mask->index = 0xFF;
            patterns->tag_mask[reg_index] = tag_mask;

            add_flow_pattern(patterns, RTE_FLOW_ITEM_TYPE_TAG,
                             tag_spec, tag_mask, NULL);
        } else {
            tag_spec = patterns->tag_spec[reg_index];
            tag_mask = patterns->tag_mask[reg_index];
        }
        tag_spec->data |= reg_spec;
        tag_mask->data |= reg_mask;
        break;
    case REG_TYPE_META:
        meta_spec = per_thread_xzalloc(sizeof *meta_spec);
        meta_spec->data = reg_spec;

        meta_mask = per_thread_xzalloc(sizeof *meta_mask);
        meta_mask->data = reg_mask;

        add_flow_pattern(patterns, RTE_FLOW_ITEM_TYPE_META, meta_spec,
                         meta_mask, NULL);
        break;
    case REG_TYPE_MARK:
    default:
        VLOG_ERR("unkonwn reg type (%d) for reg field %d", reg_field->type,
                 reg_field_id);
        return -1;
    }

    return 0;
}

static int
add_action_set_reg_field(struct flow_actions *actions,
                         uint8_t reg_field_id, uint32_t val, uint32_t mask)
{
    struct rte_flow_action_set_meta *set_meta;
    struct rte_flow_action_set_tag *set_tag;
    struct reg_field *reg_field;
    uint32_t reg_spec, reg_mask;

    if (reg_field_id >= REG_FIELD_NUM) {
        VLOG_ERR("unkonwn reg id %d", reg_field_id);
        return -1;
    }
    reg_field = &dpdk_offload_api->reg_fields()[reg_field_id];
    if (val != (val & reg_field->mask)) {
        VLOG_ERR_RL(&rl, "value 0x%"PRIx32" is out of range for reg id %d",
                          val, reg_field_id);
        return -1;
    }

    reg_spec = reg_field_spec(reg_field_id, val);
    reg_mask = reg_field_mask(reg_field_id, mask);
    switch (reg_field->type) {
    case REG_TYPE_TAG:
        set_tag = per_thread_xzalloc(sizeof *set_tag);
        set_tag->index = reg_field->index;
        set_tag->data = reg_spec;
        set_tag->mask = reg_mask;
        add_flow_action(actions, RTE_FLOW_ACTION_TYPE_SET_TAG, set_tag);
        break;
    case REG_TYPE_META:
        set_meta = per_thread_xzalloc(sizeof *set_meta);
        set_meta->data = reg_spec;
        set_meta->mask = reg_mask;
        add_flow_action(actions, RTE_FLOW_ACTION_TYPE_SET_META, set_meta);
        break;
    case REG_TYPE_MARK:
    default:
        VLOG_ERR("unkonwn reg type (%d) for reg field %d", reg_field->type,
                 reg_field_id);
        return -1;
    }

    return 0;
}

int
parse_tnl_match_recirc(struct flow_patterns *patterns,
                       struct match *match,
                       struct act_resources *act_resources)
{
    if (get_tnl_id(&match->flow.tunnel, &match->wc.masks.tunnel,
                   &act_resources->tnl_id)) {
        return -1;
    }
    if (add_pattern_match_reg_field(patterns, REG_FIELD_TUN_INFO,
                                    act_resources->tnl_id, 0xFFFFFFFF)) {
        return -1;
    }
    memset(&match->wc.masks.tunnel, 0, sizeof match->wc.masks.tunnel);
    return 0;
}

void
add_empty_sample_action(int ratio,
                        struct flow_actions *actions)
{
    struct sample_data {
        struct rte_flow_action_sample sample;
        struct rte_flow_action end_action;
    } *sample_data;
    BUILD_ASSERT_DECL(offsetof(struct sample_data, sample) == 0);

    sample_data = xzalloc(sizeof *sample_data);
    sample_data->end_action.type = RTE_FLOW_ACTION_TYPE_END;
    sample_data->sample.actions = &sample_data->end_action;
    sample_data->sample.ratio = ratio;

    add_flow_action(actions, RTE_FLOW_ACTION_TYPE_SAMPLE,
                    sample_data);
}

static int
add_action_sflow(struct flow_actions *actions, uint32_t sflow_id)
{
    struct rte_flow_action_set_meta *set_meta;
    struct reg_field *reg_field;

    reg_field = &dpdk_offload_api->reg_fields()[REG_FIELD_SFLOW_CTX];
    if (sflow_id != (sflow_id & reg_field->mask)) {
        VLOG_ERR_RL(&rl, "value 0x%"PRIx32" is out of range for reg id %d",
                          sflow_id, REG_FIELD_SFLOW_CTX);
        return -1;
    }

    set_meta = per_thread_xzalloc(sizeof *set_meta);
    set_meta->data = (sflow_id & reg_field->mask) << reg_field->offset;
    set_meta->mask = reg_field->mask << reg_field->offset;
    add_flow_action(actions, OVS_RTE_FLOW_ACTION_TYPE(SFLOW), set_meta);

    return 0;
}

static void
add_action_userspace(struct flow_actions *actions)
{
    struct sample_data {
        struct rte_flow_action_sample sample;
        struct rte_flow_action end_action;
    } *sample_data;
    BUILD_ASSERT_DECL(offsetof(struct sample_data, sample) == 0);

    sample_data = xzalloc(sizeof *sample_data);
    /* Empty sample action list */
    sample_data->end_action.type = RTE_FLOW_ACTION_TYPE_END;
    sample_data->sample.actions = &sample_data->end_action;
    sample_data->sample.ratio = 1;

    add_flow_action(actions, OVS_RTE_FLOW_ACTION_TYPE(USERSPACE), sample_data);
}

static int
map_sflow_attr(struct flow_actions *actions,
               const struct nlattr *nl_actions,
               struct dpif_sflow_attr *sflow_attr,
               struct act_resources *act_resources,
               struct act_vars *act_vars)
{
    struct sflow_ctx sflow_ctx;
    const struct nlattr *nla;
    unsigned int left;

    NL_NESTED_FOR_EACH_UNSAFE (nla, left, nl_actions) {
        if (nl_attr_type(nla) == OVS_USERSPACE_ATTR_USERDATA) {
            const struct user_action_cookie *cookie;

            cookie = nl_attr_get(nla);
            if (cookie->type == USER_ACTION_COOKIE_SFLOW) {
                sflow_attr->userdata_len = nl_attr_get_size(nla);
                memset(&sflow_ctx, 0, sizeof sflow_ctx);
                sflow_ctx.sflow_attr = *sflow_attr;
                sflow_ctx.cookie = *cookie;
                if (act_vars->tun_type != TUN_TYPE_NONE) {
                    memcpy(&sflow_ctx.sflow_tnl, act_vars->tnl_key,
                           sizeof sflow_ctx.sflow_tnl);
                }
                if (!get_sflow_id(&sflow_ctx, &act_resources->sflow_id) &&
                    !add_action_sflow(actions, act_resources->sflow_id)) {
                    return 0;
                }
            }
        }
    }

    VLOG_DBG_RL(&rl, "no sFlow cookie");
    return -1;
}

static int
parse_userspace_action(struct flow_actions *actions,
                       const struct nlattr *nl_actions,
                       struct dpif_sflow_attr *sflow_attr,
                       struct act_resources *act_resources,
                       struct act_vars *act_vars)
{
    const struct nlattr *nla;
    unsigned int left;

    NL_NESTED_FOR_EACH_UNSAFE (nla, left, nl_actions) {
        if (nl_attr_type(nla) == OVS_ACTION_ATTR_USERSPACE) {
            return map_sflow_attr(actions, nla, sflow_attr,
                                  act_resources, act_vars);
        }
    }

    VLOG_DBG_RL(&rl, "no OVS_ACTION_ATTR_USERSPACE attribute");
    return -1;
}

int
parse_sample_action(struct flow_actions *actions,
                    const struct nlattr *nl_actions,
                    struct dpif_sflow_attr *sflow_attr,
                    struct act_resources *act_resources,
                    struct act_vars *act_vars)
{
    const struct nlattr *nla;
    unsigned int left;
    int ratio = 0;

    sflow_attr->sflow = nl_actions;
    sflow_attr->sflow_len = nl_actions->nla_len;

    NL_NESTED_FOR_EACH_UNSAFE (nla, left, nl_actions) {
        if (nl_attr_type(nla) == OVS_SAMPLE_ATTR_ACTIONS) {
            if (parse_userspace_action(actions, nla,
                                       sflow_attr, act_resources, act_vars)) {
                return -1;
            }
        } else if (nl_attr_type(nla) == OVS_SAMPLE_ATTR_PROBABILITY) {
            ratio = UINT32_MAX / nl_attr_get_u32(nla);
        } else {
            return -1;
        }
    }

    add_empty_sample_action(ratio, actions);
    return 0;
}

void
add_hairpin_action(struct flow_actions *actions)
{
    struct rte_flow_action_mark *mark = per_thread_xzalloc(sizeof *mark);
    struct rte_flow_action_jump *jump = per_thread_xzalloc (sizeof *jump);

    mark->id = HAIRPIN_FLOW_MARK;
    add_flow_action(actions, RTE_FLOW_ACTION_TYPE_MARK, mark);

    jump->group = MISS_TABLE_ID;
    add_flow_action(actions, RTE_FLOW_ACTION_TYPE_JUMP, jump);
}

int
get_netdev_by_port(struct netdev *netdev,
                   const struct nlattr *nla,
                   int *outdev_id,
                   struct netdev **outdev)
{
    odp_port_t port;

    port = nl_attr_get_odp_port(nla);
    *outdev = netdev_ports_get(port, netdev->dpif_type);
    if (!*outdev) {
        VLOG_DBG_RL(&rl, "Cannot find netdev for odp port %"PRIu32, port);
        return -1;
    }
    if (!netdev_flow_api_equals(netdev, *outdev)) {
        goto err;
    }
    if (!outdev_id) {
        return 0;
    }
    *outdev_id = netdev_dpdk_get_port_id(*outdev);
    if (*outdev_id < 0) {
        goto err;
    }
    return 0;
err:
    VLOG_DBG_RL(&rl, "%s: Output to port \'%s\' cannot be offloaded.",
                netdev_get_name(netdev), netdev_get_name(*outdev));
    netdev_close(*outdev);
    return -1;
}

int
add_meter_action(struct flow_actions *actions,
                 const struct nlattr *nla,
                 struct act_resources *act_resources)
{
    /* Compensate for ovs-ofctl (meter_ID - 1) adjustment */
    uint32_t mtr_id = nl_attr_get_u32(nla) + 1;
    struct meter_data *mtr_data;
    int mtr_cnt = 0;

    while (mtr_cnt < DPDK_OFFLOAD_MAX_METERS_PER_FLOW &&
           act_resources->meter_ids[mtr_cnt] != 0) {
        mtr_cnt++;
    }

    /* DPDK supports only single meter per flow. */
    if (mtr_cnt > 0 && !ovs_doca_enabled()) {
        return -1;
    }
    if (mtr_cnt >= DPDK_OFFLOAD_MAX_METERS_PER_FLOW) {
        VLOG_ERR("Failed to add meter action, max supported is %u",
                 DPDK_OFFLOAD_MAX_METERS_PER_FLOW);
        return -1;
    }

    if (!netdev_dpdk_meter_ref(mtr_id)) {
        return -1;
    }
    act_resources->meter_ids[mtr_cnt] = mtr_id;

    mtr_data = per_thread_xzalloc(sizeof *mtr_data);
    mtr_data->conf.mtr_id = mtr_id;
    mtr_data->flow_id = act_resources->flow_id;
    add_flow_action(actions, RTE_FLOW_ACTION_TYPE_METER, mtr_data);

    return 0;
}

int
add_sflow_action(struct flow_actions *actions,
                 const struct nlattr *nla,
                 struct act_resources *act_resources,
                 struct act_vars *act_vars)
{
    struct dpif_sflow_attr sflow_attr;

    memset(&sflow_attr, 0, sizeof sflow_attr);
    /* Cases where the sFlow sampling rate is 1 the ovs action
     * is translated into OVS_ACTION_ATTR_USERSPACE and not
     * OVS_ACTION_ATTR_SAMPLE, this requires only mapping the
     * sFlow cookie.
     */
    sflow_attr.sflow = nla;
    sflow_attr.sflow_len = nla->nla_len;
    if (map_sflow_attr(actions, nla, &sflow_attr,
                       act_resources, act_vars)) {
        return -1;
    }
    add_action_userspace(actions);
    return 0;
}

int
add_ct_clear_action(struct flow_actions *actions,
                    struct act_resources *act_resources)
{
    if (act_resources->ct_action_label_id) {
        put_label_id(act_resources->ct_action_label_id);
        act_resources->ct_action_label_id = 0;
    }
    if (act_resources->ct_action_zone_id) {
        put_zone_id(act_resources->ct_action_zone_id);
        act_resources->ct_action_zone_id = 0;
    }
    if (get_zone_id(0, &act_resources->ct_action_zone_id)) {
        return -1;
    }
    add_action_set_reg_field(actions, REG_FIELD_CT_STATE, 0,
                             dpdk_offload_api->reg_fields()[REG_FIELD_CT_STATE].mask);
    add_action_set_reg_field(actions, REG_FIELD_CT_ZONE,
                             act_resources->ct_action_zone_id,
                             dpdk_offload_api->reg_fields()[REG_FIELD_CT_ZONE].mask);
    add_action_set_reg_field(actions, REG_FIELD_CT_MARK, 0,
                             dpdk_offload_api->reg_fields()[REG_FIELD_CT_MARK].mask);
    add_action_set_reg_field(actions, REG_FIELD_CT_LABEL_ID, 0,
                             dpdk_offload_api->reg_fields()[REG_FIELD_CT_LABEL_ID].mask);
    return 0;
}

static int
add_miss_flow(struct netdev *netdev,
              uint32_t src_table_id,
              uint32_t dst_table_id,
              uint32_t recirc_id,
              uint32_t mark_id,
              struct dpdk_offload_handle *doh)
{
    struct rte_flow_attr miss_attr = {
        .transfer = 1,
        .priority = DPDK_OFFLOAD_PRIORITY_MISS,
    };
    struct rte_flow_item_port_id port_id;
    struct rte_flow_item_tag tag_spec;
    struct rte_flow_item_tag tag_mask;
    struct flow_patterns miss_patterns = {
        .items = (struct rte_flow_item []) {
            { .type = RTE_FLOW_ITEM_TYPE_TAG, .spec = &tag_spec,
              .mask = &tag_mask },
            { .type = RTE_FLOW_ITEM_TYPE_PORT_ID, .spec = &port_id, },
            { .type = RTE_FLOW_ITEM_TYPE_ETH, },
            { .type = RTE_FLOW_ITEM_TYPE_END, },
        },
        .cnt = 4,
    };
    struct rte_flow_action_jump miss_jump = { .group = dst_table_id, };
    struct rte_flow_action_mark miss_mark;
    struct flow_actions miss_actions = {
        .actions = (struct rte_flow_action []) {
            { .type = OVS_RTE_FLOW_ACTION_TYPE(FLOW_INFO), .conf = &miss_mark },
            { .type = RTE_FLOW_ACTION_TYPE_JUMP, .conf = &miss_jump },
            { .type = RTE_FLOW_ACTION_TYPE_END, },
        },
        .cnt = 3,
    };
    struct rte_flow_error error;
    struct reg_field *reg_field;

    miss_attr.group = src_table_id;
    miss_mark.id = mark_id;
    if (mark_id == INVALID_FLOW_MARK) {
        miss_actions.actions++;
        miss_actions.cnt--;
    }

    if (!ovs_doca_enabled() || (recirc_id & MATCH_RECIRC_DP_HASH_BIT) == 0) {
        miss_patterns.items++;
        miss_patterns.cnt--;
    } else {
        reg_field = &dpdk_offload_api->reg_fields()[REG_FIELD_RECIRC];
        memset(&tag_spec, 0, sizeof(tag_spec));
        memset(&tag_mask, 0, sizeof(tag_mask));
        tag_spec.index = reg_field->index;
        tag_spec.data = reg_field_spec(REG_FIELD_RECIRC, recirc_id);
        tag_mask.index = 0xFF;
        tag_mask.data = reg_field_mask(REG_FIELD_RECIRC, UINT32_MAX);
    }

    port_id.id = netdev_dpdk_get_port_id(netdev);
    if (!create_rte_flow(netdev, &miss_attr, &miss_patterns, &miss_actions,
                         doh, &error)) {
        return 0;
    }
    return -1;
}

int
add_recirc_action(struct netdev *netdev,
                  struct flow_actions *actions,
                  const struct nlattr *nla,
                  struct act_resources *act_resources,
                  struct act_vars *act_vars)
{
    struct flow_miss_ctx miss_ctx;

    if (act_resources->sflow_id && !ovs_doca_enabled()) {
        /* Reject flows with sample and jump actions */
        VLOG_DBG_RL(&rl, "cannot offload sFlow with jump");
        return -1;
    }
    memset(&miss_ctx, 0, sizeof miss_ctx);
    if (act_vars->tnl_push_out_port != ODPP_NONE) {
        miss_ctx.vport = act_vars->tnl_push_out_port;
    } else {
        miss_ctx.vport = act_vars->vport;
    }
    miss_ctx.recirc_id = nl_attr_get_u32(nla);
    miss_ctx.skip_actions = 0;
    miss_ctx.has_dp_hash = act_vars->has_dp_hash;
    miss_ctx.esw_mgr_port_id = netdev_dpdk_get_esw_mgr_port_id(netdev);
    if (act_vars->tun_type != TUN_TYPE_NONE) {
        get_tnl_masked(&miss_ctx.tnl, NULL, act_vars->tnl_key,
                       &act_vars->tnl_mask);
    }
    if (netdev_offload_dpdk_get_table_ctx(miss_ctx.vport, miss_ctx.recirc_id,
                                          netdev,
                                          act_vars->is_e2e_cache,
                                          &act_resources->next_table_ctx)) {
        return -1;
    }
    if (!act_vars->is_e2e_cache &&
        get_flow_miss_ctx_id(&miss_ctx, netdev, act_resources->next_table_ctx->id,
                             miss_ctx.recirc_id,
                             &act_resources->flow_miss_ctx_id)) {
        return -1;
    }
    if (act_vars->tun_type != TUN_TYPE_NONE && act_vars->recirc_id == 0) {
        if (get_tnl_id(act_vars->tnl_key, &act_vars->tnl_mask,
                       &act_resources->tnl_id)) {
            return -1;
        }
        if (add_action_set_reg_field(actions, REG_FIELD_TUN_INFO,
                                     act_resources->tnl_id, 0xFFFFFFFF)) {
            return -1;
        }
    }
    if (ovs_doca_enabled() && (miss_ctx.recirc_id & MATCH_RECIRC_DP_HASH_BIT)) {
        uint32_t val = miss_ctx.recirc_id & ~MATCH_RECIRC_DP_HASH_BIT;
        uint32_t mask = dpdk_offload_api->reg_fields()[REG_FIELD_RECIRC].mask;

        if (add_action_set_reg_field(actions, REG_FIELD_RECIRC, val, mask)) {
            return -1;
        }
    }
    add_jump_action(actions, act_resources->next_table_ctx->id);
    return 0;
}

static void
dump_raw_decap(struct ds *s_extra,
               struct act_vars *act_vars)
{
    int i;

    ds_init(s_extra);
    ds_put_format(s_extra, "set raw_decap eth / udp / ");
    if (act_vars->is_outer_ipv4) {
        ds_put_format(s_extra, "ipv4 / ");
    } else {
        ds_put_format(s_extra, "ipv6 / ");
    }
    ds_put_format(s_extra, "geneve / ");
    for (i = 0; i < act_vars->gnv_opts_cnt; i++) {
        ds_put_format(s_extra, "geneve-opt / ");
    }
    ds_put_format(s_extra, "end_set");
}

static int
add_vxlan_decap_action(struct flow_actions *actions)
{
    add_flow_action(actions, RTE_FLOW_ACTION_TYPE_VXLAN_DECAP, NULL);
    return 0;
}

static int
add_geneve_decap_action(struct flow_actions *actions,
                        struct act_vars *act_vars)
{
    struct rte_flow_action_raw_decap *conf;

    conf = per_thread_xmalloc(sizeof (struct rte_flow_action_raw_decap));
    /* MLX5 PMD supports only one option of size 32 bits
     * which is the minimum size of options (if exists)
     * in case a flow exists with an option decapsulate 32 bits
     * from the header for the geneve options.
     */
    conf->size = sizeof (struct eth_header) +
                 sizeof (struct udp_header) +
                 sizeof (struct geneve_opt) +
                 (act_vars->is_outer_ipv4 ?
                  sizeof (struct ip_header) :
                  sizeof (struct ovs_16aligned_ip6_hdr)) +
                 (act_vars->gnv_opts_cnt ?
                   sizeof (uint32_t) : 0);

    conf->data = NULL;

    add_flow_action(actions, RTE_FLOW_ACTION_TYPE_RAW_DECAP, conf);
    if (VLOG_IS_DBG_ENABLED()) {
        dump_raw_decap(&actions->s_tnl, act_vars);
    }
    return 0;
}

static int
add_gre_decap_action(struct flow_actions *actions)
{
    add_flow_action(actions, RTE_FLOW_ACTION_TYPE_NVGRE_DECAP, NULL);
    return 0;
}

int
add_tnl_decap_action(struct flow_actions *actions,
                     struct act_vars *act_vars)
{
    struct rte_flow_action *actions_iter;
    int cnt;

    if (actions->cnt > 0) {
        struct rte_flow_action *last_action;

        last_action = &actions->actions[actions->cnt - 1];
        if (last_action->type == RTE_FLOW_ACTION_TYPE_RAW_ENCAP) {
            actions->has_encap = false;
            actions->cnt--;
            return 0;
        }
    }

    if (actions->has_encap) {
        VLOG_DBG_RL(&rl, "Decap after Encap action is not supported");
        return -1;
    }

    /* Avoid rewrite actions before decap action, as they have no meaning as
     * the outer header is going to be decapsulated.
     */
    for (cnt = 0, actions_iter = actions->actions; cnt < actions->cnt;
         cnt++, actions_iter++) {
        if (!IS_REWRITE_ACTION(actions_iter->type)) {
            continue;
        }

        actions_iter->type = RTE_FLOW_ACTION_TYPE_VOID;
    }

    if (act_vars->tun_type == TUN_TYPE_VXLAN) {
        return add_vxlan_decap_action(actions);
    }
    if (act_vars->tun_type == TUN_TYPE_GENEVE) {
        return add_geneve_decap_action(actions, act_vars);
    }
    if (act_vars->tun_type == TUN_TYPE_GRE) {
        return add_gre_decap_action(actions);
    }
    return -1;
}

int
parse_ct_actions(struct netdev *netdev,
                 struct flow_actions *actions,
                 const struct nlattr *ct_actions,
                 const size_t ct_actions_len,
                 struct act_resources *act_resources,
                 struct act_vars *act_vars)
{
    struct ct_miss_ctx ct_miss_ctx;
    const struct nlattr *cta;
    unsigned int ctleft;

    /* Not Supported */
    if (act_vars->proto && act_vars->proto != IPPROTO_UDP &&
        act_vars->proto != IPPROTO_TCP) {
        VLOG_DBG_RL(&rl, "Unsupported CT offload for L4 protocol: 0x02%" PRIx8,
                    act_vars->proto);
        return -1;
    }

    memset(&ct_miss_ctx, 0, sizeof ct_miss_ctx);
    act_vars->ct_mode = CT_MODE_CT;
    NL_ATTR_FOR_EACH_UNSAFE (cta, ctleft, ct_actions, ct_actions_len) {
        if (nl_attr_type(cta) == OVS_CT_ATTR_ZONE) {
            if (!netdev_offload_dpdk_disable_zone_tables) {
                if (act_resources->ct_action_zone_id) {
                    put_zone_id(act_resources->ct_action_zone_id);
                    act_resources->ct_action_zone_id = 0;
                }
                if (act_resources->flow_id != INVALID_FLOW_MARK &&
                    get_zone_id(nl_attr_get_u16(cta),
                                &act_resources->ct_action_zone_id)) {
                    VLOG_DBG_RL(&rl, "Could not create zone id");
                    return -1;
                }
            } else {
                const uint32_t ct_zone_mask = dpdk_offload_api->reg_fields()[REG_FIELD_CT_ZONE].mask;

                if (act_resources->flow_id != INVALID_FLOW_MARK &&
                    (get_zone_id(nl_attr_get_u16(cta),
                                 &act_resources->ct_action_zone_id) ||
                     add_action_set_reg_field(actions, REG_FIELD_CT_ZONE,
                                              act_resources->ct_action_zone_id,
                                              ct_zone_mask))) {
                    VLOG_DBG_RL(&rl, "Could not create zone id");
                    return -1;
                }
            }

            ct_miss_ctx.zone = nl_attr_get_u16(cta);
        } else if (nl_attr_type(cta) == OVS_CT_ATTR_MARK) {
            const uint32_t *key = nl_attr_get(cta);
            const uint32_t *mask = key + 1;

            add_action_set_reg_field(actions, REG_FIELD_CT_MARK, *key, *mask);
            ct_miss_ctx.mark = *key;
        } else if (nl_attr_type(cta) == OVS_CT_ATTR_LABELS) {
            const ovs_32aligned_u128 *key = nl_attr_get(cta);
            const ovs_32aligned_u128 *mask = key + 1;
            uint32_t set_value, set_mask;

            if (netdev_offload_dpdk_ct_labels_mapping) {
                ovs_u128 tmp_key, tmp_mask;

                tmp_key.u32[0] = key->u32[0];
                tmp_key.u32[1] = key->u32[1];
                tmp_key.u32[2] = key->u32[2];
                tmp_key.u32[3] = key->u32[3];

                tmp_mask.u32[0] = mask->u32[0];
                tmp_mask.u32[1] = mask->u32[1];
                tmp_mask.u32[2] = mask->u32[2];
                tmp_mask.u32[3] = mask->u32[3];

                tmp_key.u64.lo &= tmp_mask.u64.lo;
                tmp_key.u64.hi &= tmp_mask.u64.hi;
                if (get_label_id(&tmp_key, &act_resources->ct_action_label_id)) {
                    return -1;
                }
                set_value = act_resources->ct_action_label_id;
                set_mask = dpdk_offload_api->reg_fields()[REG_FIELD_CT_LABEL_ID].mask;
            } else {
                if (!act_vars->is_ct_conn) {
                    if (key->u32[1] & mask->u32[1] ||
                        key->u32[2] & mask->u32[2] ||
                        key->u32[3] & mask->u32[3]) {
                        return -1;
                    }
                }
                set_value = key->u32[0] & mask->u32[0];
                set_mask = mask->u32[0];
            }

            if (add_action_set_reg_field(actions, REG_FIELD_CT_LABEL_ID,
                                         set_value, set_mask)) {
                VLOG_DBG_RL(&rl, "Could not create label id");
                return -1;
            }
            ct_miss_ctx.label.u32[0] = key->u32[0];
            ct_miss_ctx.label.u32[1] = key->u32[1];
            ct_miss_ctx.label.u32[2] = key->u32[2];
            ct_miss_ctx.label.u32[3] = key->u32[3];
        } else if (nl_attr_type(cta) == OVS_CT_ATTR_NAT) {
            act_vars->ct_mode = CT_MODE_CT_NAT;
        } else if (nl_attr_type(cta) == OVS_CT_ATTR_HELPER) {
            struct rte_flow_action_set_meta *set_meta;
            const char *helper = nl_attr_get(cta);
            uintptr_t ctid_key;

            if (strncmp(helper, "offl", strlen("offl"))) {
                continue;
            }

            if (!ovs_scan(helper, "offl,st(0x%"SCNx8"),id_key(0x%"SCNxPTR")",
                          &ct_miss_ctx.state, &ctid_key)) {
                VLOG_ERR("Invalid offload helper: '%s'", helper);
                return -1;
            }
            if (act_resources->flow_id == INVALID_FLOW_MARK) {
                struct flows_counter_key counter_id_key = {
                    .ptr_key = ctid_key,
                };
                struct ovs_shared_ctx *ctx;
                struct rte_flow_action *ia;

                ctx = get_shared_count_ctx(netdev, &counter_id_key,
                                           OVS_SHARED_CT_COUNT, true);
                if (!ctx) {
                    return -1;
                }
                act_resources->shared_count_ctx = ctx;
                ia = &actions->actions[actions->shared_count_action_pos];
                ovs_assert(ia->type == RTE_FLOW_ACTION_TYPE_INDIRECT &&
                           ia->conf == NULL);
                ia->conf = ctx->act_hdl;
            }

            act_vars->ct_mode = CT_MODE_CT_CONN;
            if (get_ct_ctx_id(&ct_miss_ctx, &act_resources->ct_miss_ctx_id)) {
                return -1;
            }
            add_action_set_reg_field(actions, REG_FIELD_CT_STATE,
                                     ct_miss_ctx.state, 0xFF);
            set_meta = per_thread_xzalloc(sizeof *set_meta);
            set_meta->data = act_resources->ct_miss_ctx_id;
            add_flow_action(actions, OVS_RTE_FLOW_ACTION_TYPE(CT_INFO), set_meta);
            if (act_resources->flow_id != INVALID_FLOW_MARK) {
                struct rte_flow_action_mark *mark =
                    per_thread_xzalloc(sizeof *mark);

                mark->id = act_resources->flow_id;
                add_flow_action(actions, OVS_RTE_FLOW_ACTION_TYPE(FLOW_INFO),
                                mark);
            }
            add_jump_action(actions, POSTCT_TABLE_ID);
            act_vars->has_known_fate = true;
        } else {
            VLOG_DBG_RL(&rl,
                        "Ignored nested action inside ct(), action type: %d",
                        nl_attr_type(cta));
            continue;
        }
    }
    return 0;
}

int
add_nested_action(struct netdev *flowdev,
                  struct netdev *netdev,
                  struct flow_actions *actions,
                  struct nlattr *nl_actions,
                  size_t nl_actions_len,
                  struct act_resources *act_resources,
                  struct act_vars *act_vars,
                  uint8_t nest_level)
{
    struct rte_flow_action_sample *sample;
    struct flow_actions *nested_actions;

    nested_actions = per_thread_xzalloc(sizeof *nested_actions);

    if (parse_flow_actions(flowdev, netdev, nested_actions, nl_actions,
                           nl_actions_len, act_resources, act_vars,
                           nest_level + 1)) {
        goto err;
    }
    add_flow_action(nested_actions, RTE_FLOW_ACTION_TYPE_END, NULL);
    sample = per_thread_xzalloc(sizeof *sample);
    sample->ratio = 1;
    sample->actions = nested_actions->actions;
    add_flow_action(actions, OVS_RTE_FLOW_ACTION_TYPE(NESTED), sample);

    return 0;
err:
    per_thread_free(nested_actions);
    return -1;
}

static void
set_ct_ctnat_conf(const struct rte_flow_action *actions,
                  struct rte_flow_action_set_tag *ct_state,
                  struct rte_flow_action_set_tag *ctnat_state,
                  const void **ct_conf, const void **ctnat_conf)
{
    const struct rte_flow_action_set_tag *set_tag = actions->conf;
    struct reg_field *rf = &dpdk_offload_api->reg_fields()[REG_FIELD_CT_STATE];

    *ct_conf = actions->conf;
    *ctnat_conf = actions->conf;

    /* In case this is not a ct-state set, no need for further changes. */
    if (actions->type != RTE_FLOW_ACTION_TYPE_SET_TAG ||
        set_tag->index != rf->index ||
        set_tag->mask != (rf->mask << rf->offset)) {
        return;
    }

    /* For ct-state set, clear NAT bits in ct_conf, and set in ctnat_conf.
     * Hops following this one will then be able to know that a CT-NAT action
     * has been executed in the past.
     */
    *ct_state = *(const struct rte_flow_action_set_tag *) *ct_conf;
    *ctnat_state =
        *(const struct rte_flow_action_set_tag *) *ctnat_conf;
    ct_state->data &= ((~(uint32_t) OVS_CS_F_NAT_MASK) << rf->offset);

    ctnat_state->data |= OVS_CS_F_NAT_MASK << rf->offset;
    *ct_conf = ct_state;
    *ctnat_conf = ctnat_state;
}

static void
split_ct_conn_actions(const struct rte_flow_action *actions,
                      struct flow_actions *ct_actions,
                      struct flow_actions *nat_actions,
                      struct rte_flow_action_set_tag *ct_state,
                      struct rte_flow_action_set_tag *ctnat_state)
{
    const void *ct_conf, *ctnat_conf;

    for (; actions && actions->type != RTE_FLOW_ACTION_TYPE_END; actions++) {
        /* This is only a dummy action used to split pre and post CT. It
         * should never be actually used.
         */
        if (actions->type == OVS_RTE_FLOW_ACTION_TYPE(PRE_CT_END)) {
            continue;
        }
        set_ct_ctnat_conf(actions, ct_state, ctnat_state, &ct_conf, &ctnat_conf);
        if (actions->type != RTE_FLOW_ACTION_TYPE_SET_IPV4_SRC &&
            actions->type != RTE_FLOW_ACTION_TYPE_SET_IPV4_DST &&
            actions->type != RTE_FLOW_ACTION_TYPE_SET_IPV6_SRC &&
            actions->type != RTE_FLOW_ACTION_TYPE_SET_IPV6_DST &&
            actions->type != RTE_FLOW_ACTION_TYPE_SET_TP_SRC &&
            actions->type != RTE_FLOW_ACTION_TYPE_SET_TP_DST) {
            add_flow_action(ct_actions, actions->type, ct_conf);
        }
        add_flow_action(nat_actions, actions->type, ctnat_conf);
    }
    add_flow_action(ct_actions, RTE_FLOW_ACTION_TYPE_END, NULL);
    add_flow_action(nat_actions, RTE_FLOW_ACTION_TYPE_END, NULL);
}

int
create_e2e_mt_ct_conn(struct netdev *netdev,
                      struct flow_patterns *flow_patterns,
                      struct flow_actions *flow_actions,
                      struct rte_flow_error *error,
                      struct act_resources *act_resources,
                      struct flow_item *fi)
{
    struct flow_actions nat_actions = { .actions = NULL, .cnt = 0 };
    struct flow_actions ct_actions = { .actions = NULL, .cnt = 0 };
    struct rte_flow_action_set_tag ct_state, ctnat_state;
    struct rte_flow_attr attr = { .transfer = 1 };
    int ret = -1;
    int pos = 0;
    bool is_ct;

    fi->doh[0].rte_flow = fi->doh[1].rte_flow = NULL;

    split_ct_conn_actions(flow_actions->actions, &ct_actions, &nat_actions,
                          &ct_state, &ctnat_state);
    is_ct = ct_actions.cnt == nat_actions.cnt;

    if (act_resources) {
        put_table_ctx(act_resources->self_table_ctx);
        act_resources->self_table_ctx = NULL;
    }
    pos = netdev_offload_ct_on_ct_nat;

    if (netdev_offload_ct_on_ct_nat || !is_ct) {
        attr.group = CTNAT_TABLE_ID;
        ret = create_rte_flow(netdev, &attr, flow_patterns, &nat_actions,
                              &fi->doh[pos], error);
        if (ret) {
            goto out;
        }
    }

    if (netdev_offload_ct_on_ct_nat || is_ct) {
        attr.group = CT_TABLE_ID;
        ret = create_rte_flow(netdev, &attr, flow_patterns, &ct_actions,
                              &fi->doh[0], error);
        if (ret) {
            goto ct_err;
        }
    }
    goto out;

ct_err:
    if (netdev_offload_ct_on_ct_nat) {
        netdev_offload_dpdk_destroy_flow(netdev, &fi->doh[1], NULL, true);
    }
out:
    free_flow_actions(&ct_actions, false);
    free_flow_actions(&nat_actions, false);
    return ret;
}

static int
create_ct_conn(struct netdev *netdev,
               struct flow_patterns *flow_patterns,
               struct flow_actions *flow_actions,
               struct rte_flow_error *error,
               struct act_resources *act_resources,
               struct conn_item *ci)
{
    struct flow_actions nat_actions = { .actions = NULL, .cnt = 0 };
    struct flow_actions ct_actions = { .actions = NULL, .cnt = 0 };
    struct rte_flow_action_set_tag ct_state, ctnat_state;
    struct rte_flow_attr attr = { .transfer = 1 };
    struct dpdk_offload_handle doh;
    int ret = -1;
    bool is_ct;

    ci->rte_flow = ci->ct_ct_nat = NULL;

    split_ct_conn_actions(flow_actions->actions, &ct_actions, &nat_actions,
                          &ct_state, &ctnat_state);
    is_ct = ct_actions.cnt == nat_actions.cnt;

    if (act_resources) {
        put_table_ctx(act_resources->self_table_ctx);
        act_resources->self_table_ctx = NULL;
    }

    if (netdev_offload_ct_on_ct_nat || !is_ct) {
        doh.rte_flow = NULL;
        attr.group = CTNAT_TABLE_ID;
        ret = create_rte_flow(netdev, &attr, flow_patterns, &nat_actions,
                              &doh, error);
        if (ret) {
            goto out;
        }
        if (netdev_offload_ct_on_ct_nat) {
            ci->ct_ct_nat = doh.rte_flow;
        } else {
            ci->rte_flow = doh.rte_flow;
        }
        ci->valid = doh.valid;
    }

    if (netdev_offload_ct_on_ct_nat || is_ct) {
        attr.group = CT_TABLE_ID;
        doh.rte_flow = NULL;
        ret = create_rte_flow(netdev, &attr, flow_patterns, &ct_actions,
                              &doh, error);
        if (ret) {
            goto ct_err;
        }
        ci->rte_flow = doh.rte_flow;
        ci->valid = doh.valid;
    }
    goto out;

ct_err:
    if (netdev_offload_ct_on_ct_nat) {
        doh.rte_flow = ci->ct_ct_nat;
        doh.valid = ci->valid;
        netdev_offload_dpdk_destroy_flow(netdev, &doh, NULL, true);
    }
out:
    free_flow_actions(&ct_actions, false);
    free_flow_actions(&nat_actions, false);
    return ret;
}

static void
split_pre_post_ct_actions(const struct rte_flow_action *actions,
                          struct flow_actions *pre_ct_actions,
                          struct flow_actions *post_ct_actions)
{
    struct flow_actions *split = pre_ct_actions;

    while (actions && actions->type != RTE_FLOW_ACTION_TYPE_END) {
        if (actions->type == RTE_FLOW_ACTION_TYPE_COUNT) {
            add_flow_action(post_ct_actions, actions->type, actions->conf);
        } else if (actions->type == OVS_RTE_FLOW_ACTION_TYPE(PRE_CT_END)) {
            split = post_ct_actions;
        } else {
            add_flow_action(split, actions->type, actions->conf);
        }
        actions++;
    }
}

int
create_pre_post_ct(struct netdev *netdev,
                   const struct rte_flow_attr *attr,
                   struct flow_patterns *flow_patterns,
                   struct flow_actions *flow_actions,
                   struct rte_flow_error *error,
                   struct act_resources *act_resources,
                   struct act_vars *act_vars,
                   struct flow_item *fi)
{
    struct flow_actions post_ct_actions = { .actions = NULL, .cnt = 0 };
    struct flow_actions pre_ct_actions = { .actions = NULL, .cnt = 0 };
    struct rte_flow_item_port_id port_id;
    struct rte_flow_item_mark post_ct_mark;
    struct flow_patterns post_ct_patterns = {
        .items = (struct rte_flow_item []) {
            { .type = RTE_FLOW_ITEM_TYPE_PORT_ID, .spec = &port_id, },
            { .type = OVS_RTE_FLOW_ITEM_TYPE(FLOW_INFO), .spec = &post_ct_mark, },
            { .type = RTE_FLOW_ITEM_TYPE_END, },
        },
        .cnt = 3,
    };
    struct rte_flow_action_mark pre_ct_mark;
    struct rte_flow_action_jump pre_ct_jump;
    struct flow_miss_ctx pre_ct_miss_ctx;
    struct rte_flow_attr post_ct_attr;
    uint32_t ct_table_id;
    int ret;

    port_id.id = netdev_dpdk_get_port_id(netdev);

    /* post-ct */
    post_ct_mark.id = act_resources->flow_id;
    memcpy(&post_ct_attr, attr, sizeof post_ct_attr);
    post_ct_attr.group = POSTCT_TABLE_ID;
    split_pre_post_ct_actions(flow_actions->actions, &pre_ct_actions,
                              &post_ct_actions);
    add_flow_action(&post_ct_actions, RTE_FLOW_ACTION_TYPE_END, NULL);
    ret = create_rte_flow(netdev, &post_ct_attr, &post_ct_patterns,
                          &post_ct_actions, &fi->doh[1], error);
    if (ret) {
        goto out;
    }

    /* pre-ct */
    if (act_vars->ct_mode == CT_MODE_CT) {
        ct_table_id = CT_TABLE_ID;
    } else {
        ct_table_id = CTNAT_TABLE_ID;
    }
    if (!netdev_offload_dpdk_disable_zone_tables) {
        ct_table_id += act_resources->ct_action_zone_id;
    }
    pre_ct_miss_ctx.vport = act_vars->vport;
    pre_ct_miss_ctx.recirc_id = act_vars->recirc_id;
    if (act_vars->vport != ODPP_NONE) {
        get_tnl_masked(&pre_ct_miss_ctx.tnl, NULL, act_vars->tnl_key,
                       &act_vars->tnl_mask);
    } else {
        memset(&pre_ct_miss_ctx.tnl, 0, sizeof pre_ct_miss_ctx.tnl);
    }
    pre_ct_miss_ctx.skip_actions = act_vars->n_hw_actions;
    if (!act_resources->associated_flow_id) {
        if (associate_flow_id(act_resources->flow_id, &pre_ct_miss_ctx)) {
            goto pre_ct_err;
        }
        act_resources->associated_flow_id = true;
    }
    pre_ct_mark.id = act_resources->flow_id;
    add_flow_action(&pre_ct_actions, OVS_RTE_FLOW_ACTION_TYPE(FLOW_INFO),
                    &pre_ct_mark);
    pre_ct_jump.group = ct_table_id;
    add_flow_action(&pre_ct_actions, RTE_FLOW_ACTION_TYPE_JUMP, &pre_ct_jump);
    add_flow_action(&pre_ct_actions, RTE_FLOW_ACTION_TYPE_END, NULL);
    ret = create_rte_flow(netdev, attr, flow_patterns, &pre_ct_actions,
                          &fi->doh[0], error);
    if (ret) {
        goto pre_ct_err;
    }
    goto out;

pre_ct_err:
    netdev_offload_dpdk_destroy_flow(netdev, &fi->doh[1], NULL, true);
out:
    free_flow_actions(&pre_ct_actions, false);
    free_flow_actions(&post_ct_actions, false);
    return ret;
}

void *
find_raw_encap_spec(const struct raw_encap_data *raw_encap_data,
                    enum rte_flow_item_type type)
{
    struct ovs_16aligned_ip6_hdr *ipv6;
    struct udp_header *udp = NULL;
    struct vlan_header *vlan;
    struct eth_header *eth;
    struct ip_header *ipv4;
    uint8_t *next_hdr;
    ovs_be16 proto;

    eth = ALIGNED_CAST(struct eth_header *, raw_encap_data->conf.data);
    if (type == RTE_FLOW_ITEM_TYPE_ETH) {
        return eth;
    }

    next_hdr = (uint8_t *) (eth + 1);
    proto = eth->eth_type;
    /* VLAN skipping */
    while (eth_type_vlan(proto)) {
        vlan = ALIGNED_CAST(struct vlan_header *, next_hdr);
        proto = vlan->vlan_next_type;
        next_hdr += sizeof *vlan;
    }

    if (proto == htons(RTE_ETHER_TYPE_IPV4)) {
        ipv4 = ALIGNED_CAST(struct ip_header *, next_hdr);
        if (type == RTE_FLOW_ITEM_TYPE_IPV4) {
            return ipv4;
        }
        if (ipv4->ip_proto != IPPROTO_UDP) {
            return NULL;
        }
        udp = (struct udp_header *) (ipv4 + 1);
    }

    if (proto == htons(RTE_ETHER_TYPE_IPV6)) {
        ipv6 = ALIGNED_CAST(struct ovs_16aligned_ip6_hdr *, next_hdr);
        if (type == RTE_FLOW_ITEM_TYPE_IPV6) {
            return ipv6;
        }
        if (ipv6->ip6_nxt != IPPROTO_UDP) {
            return NULL;
        }
        udp = (struct udp_header *) (ipv6 + 1);
    }

    if (udp && type == RTE_FLOW_ITEM_TYPE_UDP) {
        return udp;
    }

    return NULL;
}

static void *
find_encap_spec(struct raw_encap_data *raw_encap_data,
                enum rte_flow_item_type type)
{
    void *rv;

    if (raw_encap_data) {
        rv = find_raw_encap_spec(raw_encap_data, type);
        if (rv == NULL) {
            VLOG_DBG_RL(&rl, "Could not find raw_encap spec type=%d", type);
        }
        return rv;
    }

    OVS_NOT_REACHED();
}

static void
set_encap_field__(const void *value, void *hdr, void *mask, const size_t size)
{
    const uint8_t *v;
    uint8_t *h, *m;
    uint32_t i;

    if (!mask) {
        memcpy(hdr, value, size);
        return;
    }

    for (i = 0, h = hdr, v = value, m = mask; i < size; i++) {
        *h = (*h & ~*m) | (*v & *m);
        h++;
        v++;
        *m++ = 0;
    }
}

int
outer_encap_set_actions(struct raw_encap_data *raw_encap_data,
                        const struct nlattr *set_actions,
                        const size_t set_actions_len,
                        bool masked)
{
    const struct nlattr *sa;
    unsigned int sleft;

#define set_encap_field(field, hdr)                                           \
    set_encap_field__(&key->field, hdr,                                       \
                    mask ? CONST_CAST(void *, &mask->field) : NULL,           \
                    sizeof key->field)

    NL_ATTR_FOR_EACH_UNSAFE (sa, sleft, set_actions, set_actions_len) {
        if (nl_attr_type(sa) == OVS_KEY_ATTR_ETHERNET) {
            const struct ovs_key_ethernet *key = nl_attr_get(sa);
            const struct ovs_key_ethernet *mask = masked ? key + 1 : NULL;
            struct rte_flow_item_eth *spec;

            spec = find_encap_spec(raw_encap_data, RTE_FLOW_ITEM_TYPE_ETH);
            if (!spec) {
                return -1;
            }

            set_encap_field(eth_src, spec->src.addr_bytes);
            set_encap_field(eth_dst, spec->dst.addr_bytes);

            if (mask && !is_all_zeros(mask, sizeof *mask)) {
                VLOG_DBG_RL(&rl, "Unsupported ETHERNET set action");
                return -1;
            }
        } else if (nl_attr_type(sa) == OVS_KEY_ATTR_IPV4) {
            const struct ovs_key_ipv4 *key = nl_attr_get(sa);
            const struct ovs_key_ipv4 *mask = masked ? key + 1 : NULL;
            struct rte_flow_item_ipv4 *spec;

            spec = find_encap_spec(raw_encap_data, RTE_FLOW_ITEM_TYPE_IPV4);
            if (!spec) {
                return -1;
            }

            set_encap_field(ipv4_src, &spec->hdr.src_addr);
            set_encap_field(ipv4_dst, &spec->hdr.dst_addr);
            set_encap_field(ipv4_ttl, &spec->hdr.time_to_live);

            if (mask && !is_all_zeros(mask, sizeof *mask)) {
                VLOG_DBG_RL(&rl, "Unsupported IPv4 set action");
                return -1;
            }
        } else if (nl_attr_type(sa) == OVS_KEY_ATTR_IPV6) {
            const struct ovs_key_ipv6 *key = nl_attr_get(sa);
            const struct ovs_key_ipv6 *mask = masked ? key + 1 : NULL;
            struct rte_flow_item_ipv6 *spec;

            spec = find_encap_spec(raw_encap_data, RTE_FLOW_ITEM_TYPE_IPV6);
            if (!spec) {
                return -1;
            }

            set_encap_field(ipv6_src, &spec->hdr.src_addr);
            set_encap_field(ipv6_dst, &spec->hdr.dst_addr);
            set_encap_field(ipv6_hlimit, &spec->hdr.hop_limits);

            if (mask && !is_all_zeros(mask, sizeof *mask)) {
                VLOG_DBG_RL(&rl, "Unsupported IPv6 set action");
                return -1;
            }
        } else if (nl_attr_type(sa) == OVS_KEY_ATTR_UDP) {
            const struct ovs_key_udp *key = nl_attr_get(sa);
            const struct ovs_key_udp *mask = masked ? key + 1 : NULL;
            struct rte_flow_item_udp *spec;

            spec = find_encap_spec(raw_encap_data, RTE_FLOW_ITEM_TYPE_UDP);
            if (!spec) {
                return -1;
            }

            set_encap_field(udp_src, &spec->hdr.src_port);
            set_encap_field(udp_dst, &spec->hdr.dst_port);

            if (mask && !is_all_zeros(mask, sizeof *mask)) {
                VLOG_DBG_RL(&rl, "Unsupported UDP set action");
                return -1;
            }
        } else {
            VLOG_DBG_RL(&rl,
                        "Unsupported set action type %d", nl_attr_type(sa));
            return -1;
        }
    }

    return 0;
}

int
add_cont_in_sw_action(struct flow_actions *actions,
                      struct act_resources *act_resources,
                      struct act_vars *act_vars)
{
    struct rte_flow_action_jump *jump = per_thread_xzalloc (sizeof *jump);
    struct rte_flow_action_mark *mark = per_thread_xzalloc(sizeof *mark);
    struct flow_miss_ctx miss_ctx;
    struct rte_flow_action *ca;

    miss_ctx.vport = act_vars->vport;
    miss_ctx.recirc_id = act_vars->recirc_id;
    if (act_vars->vport != ODPP_NONE && act_vars->tnl_key) {
        get_tnl_masked(&miss_ctx.tnl, NULL, act_vars->tnl_key,
                       &act_vars->tnl_mask);
    } else {
        memset(&miss_ctx.tnl, 0, sizeof miss_ctx.tnl);
    }
    miss_ctx.skip_actions = act_vars->n_hw_actions;
    if (!act_resources->associated_flow_id) {
        if (associate_flow_id(act_resources->flow_id, &miss_ctx)) {
            return -1;
        }
        act_resources->associated_flow_id = true;
    }
    act_resources->cont_in_sw = true;

    mark->id = act_resources->flow_id;
    add_flow_action(actions, OVS_RTE_FLOW_ACTION_TYPE(FLOW_INFO), mark);

    jump->group = MISS_TABLE_ID;
    add_flow_action(actions, RTE_FLOW_ACTION_TYPE_JUMP, jump);

    /* Disable count action for partial offload. */
    ca = &actions->actions[actions->count_action_pos];
    if (ca->type == RTE_FLOW_ACTION_TYPE_COUNT) {
        ca->type = RTE_FLOW_ACTION_TYPE_VOID;
    }

    return 0;
}

static void
netdev_offload_dpdk_update_stats(struct dpif_flow_stats *stats,
                                 struct dpif_flow_attrs *attrs,
                                 struct rte_flow_query_count *query)
{
    if (attrs) {
        attrs->dp_layer = "dpdk";
    }

    if (!stats || !query) {
        return;
    }

    stats->n_packets += (query->hits_set) ? query->hits : 0;
    stats->n_bytes += (query->bytes_set) ? query->bytes : 0;
}

/* Upon flushing the ESW manager, its members netdevs should be flushed too,
 * as their offloads are done on it.
 */
static bool
flush_esw_members_op(struct netdev *netdev,
                     odp_port_t odp_port OVS_UNUSED,
                     void *aux_)
{
    struct esw_members_aux *aux = aux_;

    if (flush_netdev_flows_in_related(netdev, netdev, true)) {
        aux->ret = -1;
        return true;
    }

    return false;
}

static bool
esw_members_cb(struct netdev *netdev,
               odp_port_t odp_port OVS_UNUSED,
               void *aux_)
{
    struct esw_members_aux *aux = aux_;
    struct netdev *esw_netdev;
    int netdev_esw_mgr_pid;
    int esw_mgr_pid;

    esw_netdev = aux->esw_netdev;
    esw_mgr_pid = netdev_dpdk_get_esw_mgr_port_id(aux->esw_netdev);

    /* Skip the ESW netdev itself. */
    if (netdev == esw_netdev) {
        return false;
    }

    netdev_esw_mgr_pid = netdev_dpdk_get_esw_mgr_port_id(netdev);

    /* Skip a non-member. */
    if (netdev_esw_mgr_pid == -1 || netdev_esw_mgr_pid != esw_mgr_pid) {
        return false;
    }

    return aux->op(netdev, odp_port, aux);
}

static bool
uninit_esw_members_op(struct netdev *netdev,
                      odp_port_t odp_port OVS_UNUSED,
                      void *aux_ OVS_UNUSED)
{
    netdev_uninit_flow_api(netdev);

    return false;
}

static bool
init_esw_members_op(struct netdev *netdev,
                    odp_port_t odp_port OVS_UNUSED,
                    void *aux_ OVS_UNUSED)
{
    netdev_init_flow_api(netdev);

    return false;
}

#define PKT_DUMP_MAX_LEN    80

static void
log_packet_err(struct netdev *netdev, struct dp_packet *pkt, char *str)
{
    struct ds s;

    ds_init(&s);

    VLOG_ERR("%s: %s. %s", netdev_get_name(netdev), str,
             ds_cstr(dp_packet_ds_put_hex(&s, pkt, PKT_DUMP_MAX_LEN)));

    ds_destroy(&s);
}

static bool
rte_get_packet_recovery_info(struct dp_packet *packet,
                             struct dpdk_offload_recovery_info *info)
{
    memset(info, 0, sizeof *info);
    if (dpdk_offload_get_reg_field(packet, REG_FIELD_FLOW_INFO,
                                   &info->flow_miss_id)) {
        dpdk_offload_get_reg_field(packet, REG_FIELD_CT_CTX,
                                   &info->ct_miss_id);
    } else {
        dpdk_offload_get_reg_field(packet, REG_FIELD_SFLOW_CTX,
                                   &info->sflow_id);
    }

    return true;
}

int
netdev_offload_dpdk_packet_hw_hash(struct netdev *netdev,
                                   struct dp_packet *packet,
                                   uint32_t seed,
                                   uint32_t *hash)
{
    return dpdk_offload_api->packet_hw_hash(netdev, packet, seed, hash);
}

int
netdev_offload_dpdk_packet_hw_entropy(struct netdev *netdev,
                                      struct dp_packet *packet,
                                      uint16_t *entropy)
{
    return dpdk_offload_api->packet_hw_entropy(netdev, packet, entropy);
}

int
netdev_offload_dpdk_get_stats(struct netdev *netdev,
                              struct netdev_offload_stats *stats)
{
    struct netdev_offload_dpdk_data *data;
    unsigned int tid;

    data = (struct netdev_offload_dpdk_data *)
        ovsrcu_get(void *, &netdev->hw_info.offload_data);
    if (!data) {
        return -1;
    }

    for (tid = 0; tid < MAX_OFFLOAD_THREAD_NB; tid++) {
        stats[tid].n_inserted = data->offload_counters[tid];
        stats[tid].n_flows = data->flow_counters[tid];
        stats[tid].n_conns = data->conn_counters[tid];
    }

    return 0;
}

int
netdev_offload_dpdk_ct_counter_query(struct netdev *netdev,
                                     uintptr_t counter_key,
                                     long long now,
                                     long long prev_now,
                                     struct dpif_flow_stats *stats)
{
    struct rte_flow_query_age query_age;
    struct ovs_shared_ctx *ctx;
    struct rte_flow_error error;
    int ret;

    memset(stats, 0, sizeof *stats);

    ctx = get_shared_age_ctx(netdev, counter_key, false);
    if (ctx == NULL) {
        VLOG_ERR_RL(&rl, "Could not get shared age ctx for "
                    "counter_key=0x%"PRIxPTR, counter_key);
        return -1;
    }

    ret = netdev_dpdk_indirect_action_query(ctx->port_id, ctx->act_hdl,
                                            &query_age, &error);
    if (!ret && query_age.sec_since_last_hit_valid &&
        (query_age.sec_since_last_hit * 1000) <= (now - prev_now)) {
        stats->used = now;
    }
    return ret;
}

static void
fixed_rule_uninit(struct netdev *netdev, struct fixed_rule *fr, bool is_esw)
{
    if (!fr->doh.rte_flow) {
        return;
    }

    netdev_offload_dpdk_destroy_flow(netdev, &fr->doh, NULL, is_esw);
    fr->doh.rte_flow = NULL;
}

static void
ct_nat_miss_uninit(struct netdev *netdev, struct fixed_rule *fr)
{
    fixed_rule_uninit(netdev, fr, true);
}

static int
ct_nat_miss_init(struct netdev *netdev, struct fixed_rule *fr)
{
    if (add_miss_flow(netdev, CTNAT_TABLE_ID, CT_TABLE_ID, 0, 0, &fr->doh)) {
        return -1;
    }
    return 0;
}

static void
ct_zones_uninit(struct netdev *netdev, struct netdev_offload_dpdk_data *data)
{
    struct fixed_rule *fr;
    uint32_t zone_id;
    int nat, i;

    if (netdev_offload_dpdk_disable_zone_tables) {
        return;
    }

    for (nat = 0; nat < 2; nat++) {
        for (i = 0; i < 2; i++) {
            for (zone_id = MIN_ZONE_ID; zone_id <= MAX_ZONE_ID; zone_id++) {
                fr = &data->zone_flows[nat][i][zone_id];

                fixed_rule_uninit(netdev, fr, true);
            }
        }
    }
}

static int
ct_zones_init(struct netdev *netdev, struct netdev_offload_dpdk_data *data)
{
    struct rte_flow_action_set_tag set_tag;
    struct rte_flow_item_port_id port_id;
    struct rte_flow_item_tag tag_spec;
    struct rte_flow_item_tag tag_mask;
    struct rte_flow_action_jump jump;
    struct rte_flow_attr attr = {
        .transfer = 1,
    };
    struct flow_patterns patterns = {
        .items = (struct rte_flow_item []) {
            { .type = RTE_FLOW_ITEM_TYPE_PORT_ID, .spec = &port_id, },
            { .type = RTE_FLOW_ITEM_TYPE_ETH, },
            { .type = RTE_FLOW_ITEM_TYPE_TAG, .spec = &tag_spec,
              .mask = &tag_mask },
            { .type = RTE_FLOW_ITEM_TYPE_END, },
        },
        .cnt = 4,
    };
    struct dpdk_offload_handle doh;
    struct flow_actions actions = {
        .actions = (struct rte_flow_action []) {
            { .type = RTE_FLOW_ACTION_TYPE_SET_TAG, .conf = &set_tag, },
            { .type = RTE_FLOW_ACTION_TYPE_JUMP, .conf = &jump, },
            { .type = RTE_FLOW_ACTION_TYPE_END, .conf = NULL, },
        },
        .cnt = 3,
    };
    struct reg_field *reg_field;
    struct rte_flow_error error;
    struct fixed_rule *fr;
    uint32_t base_group;
    uint32_t zone_id;
    int nat;

    if (netdev_offload_dpdk_disable_zone_tables) {
        return 0;
    }

    memset(&set_tag, 0, sizeof(set_tag));
    memset(&jump, 0, sizeof(jump));
    memset(&port_id, 0, sizeof(port_id));
    memset(&tag_spec, 0, sizeof(tag_spec));
    memset(&tag_mask, 0, sizeof(tag_mask));

    port_id.id = netdev_dpdk_get_port_id(netdev);

    /* Merge the tag match for zone and state only if they are
     * at the same index. */
    ovs_assert(dpdk_offload_api->reg_fields()[REG_FIELD_CT_ZONE].index == dpdk_offload_api->reg_fields()[REG_FIELD_CT_STATE].index);

    for (nat = 0; nat < 2; nat++) {
        base_group = nat ? CTNAT_TABLE_ID : CT_TABLE_ID;

        for (zone_id = MIN_ZONE_ID; zone_id <= MAX_ZONE_ID; zone_id++) {
            uint32_t ct_zone_spec, ct_zone_mask;
            uint32_t ct_state_spec, ct_state_mask;

            attr.group = base_group + zone_id;
            jump.group = base_group;

            fr = &data->zone_flows[nat][0][zone_id];
            attr.priority = 0;
            /* If the zone is the same, and already visited ct/ct-nat, skip
             * ct/ct-nat and jump directly to post-ct.
             */
            reg_field = &dpdk_offload_api->reg_fields()[REG_FIELD_CT_ZONE];
            ct_zone_spec = reg_field_spec(REG_FIELD_CT_ZONE, zone_id);
            ct_zone_mask = reg_field_mask(REG_FIELD_CT_ZONE, UINT32_MAX);
            reg_field = &dpdk_offload_api->reg_fields()[REG_FIELD_CT_STATE];
            ct_state_spec = OVS_CS_F_TRACKED;
            if (nat) {
                ct_state_spec |= OVS_CS_F_NAT_MASK;
            }
            ct_state_spec = reg_field_spec(REG_FIELD_CT_STATE, ct_state_spec);
            ct_state_mask = ct_state_spec;

            /* Merge ct_zone and ct_state matches in a single item. */
            tag_spec.index = reg_field->index;
            tag_spec.data = ct_zone_spec | ct_state_spec;
            tag_mask.index = 0xFF;
            tag_mask.data = ct_zone_mask | ct_state_mask;
            patterns.items[2].type = RTE_FLOW_ITEM_TYPE_TAG;
            actions.actions[0].type = RTE_FLOW_ACTION_TYPE_VOID;
            jump.group = POSTCT_TABLE_ID;
            if (create_rte_flow(netdev, &attr, &patterns, &actions, &doh,
                                &error)) {
                goto err;
            }
            fr->doh.rte_flow = doh.rte_flow;

            fr = &data->zone_flows[nat][1][zone_id];
            attr.priority = 1;
            /* Otherwise, set the zone and go to CT/CT-NAT. */
            reg_field = &dpdk_offload_api->reg_fields()[REG_FIELD_CT_STATE];
            tag_spec.index = reg_field->index;
            tag_spec.data = 0;
            tag_mask.index = 0xFF;
            tag_mask.data = reg_field->mask << reg_field->offset;
            patterns.items[2].type = RTE_FLOW_ITEM_TYPE_VOID;
            reg_field = &dpdk_offload_api->reg_fields()[REG_FIELD_CT_ZONE];
            set_tag.index = reg_field->index;
            set_tag.data = reg_field_spec(REG_FIELD_CT_ZONE, zone_id);
            set_tag.mask = reg_field_mask(REG_FIELD_CT_ZONE, UINT32_MAX);
            actions.actions[0].type = RTE_FLOW_ACTION_TYPE_SET_TAG;
            jump.group = base_group;
            if (create_rte_flow(netdev, &attr, &patterns, &actions, &doh,
                                &error)) {
                goto err;
            }
            fr->doh.rte_flow = doh.rte_flow;
        }
    }

    return 0;

err:
    ct_zones_uninit(netdev, data);
    return -1;
}

static void
hairpin_uninit(struct netdev *netdev, struct fixed_rule *fr)
{
    fixed_rule_uninit(netdev, fr, false);
}

static int
hairpin_init(struct netdev *netdev, struct fixed_rule *fr)
{
    struct rte_flow_attr attr = { .ingress = 1, };
    struct rte_flow_item_mark hp_mark;
    struct flow_patterns patterns = {
        .items = (struct rte_flow_item []) {
            { .type = OVS_RTE_FLOW_ITEM_TYPE(FLOW_INFO), .spec = &hp_mark, },
            { .type = RTE_FLOW_ITEM_TYPE_END, },
        },
        .cnt = 2,
    };
    struct rte_flow_action_queue hp_queue;
    struct dpdk_offload_handle doh;
    struct flow_actions actions = {
        .actions = (struct rte_flow_action []) {
            { .type = RTE_FLOW_ACTION_TYPE_QUEUE, .conf = &hp_queue, },
            { .type = RTE_FLOW_ACTION_TYPE_END, },
        },
        .cnt = 2,
    };
    struct rte_flow_error error;

    hp_mark.id = HAIRPIN_FLOW_MARK;
    hp_queue.index = netdev->n_rxq;

    if (create_rte_flow(netdev, &attr, &patterns, &actions, &doh, &error)) {
        return -1;
    }
    fr->doh.rte_flow = doh.rte_flow;
    return 0;
}

static void
rte_aux_tables_uninit(struct netdev *netdev)
{
    struct netdev_offload_dpdk_data *data;

    if (!netdev_dpdk_is_ethdev(netdev)) {
        return;
    }

    data = (struct netdev_offload_dpdk_data *)
        ovsrcu_get(void *, &netdev->hw_info.offload_data);
    if (!data) {
        return;
    }

    ct_nat_miss_uninit(netdev, &data->ct_nat_miss);
    ct_zones_uninit(netdev, data);
    hairpin_uninit(netdev, &data->hairpin);
}

static int
rte_aux_tables_init(struct netdev *netdev)
{
    struct netdev_offload_dpdk_data *data;
    int ret = 0;

    if (!netdev_dpdk_is_ethdev(netdev)) {
        return 0;
    }

    data = (struct netdev_offload_dpdk_data *)
        ovsrcu_get(void *, &netdev->hw_info.offload_data);

    ret = ct_nat_miss_init(netdev, &data->ct_nat_miss);
    if (!ret) {
        ret = ct_zones_init(netdev, data);
    }
    if (!ret) {
        ret = hairpin_init(netdev, &data->hairpin);
    }
    if (ret) {
        VLOG_WARN("Cannot apply init flows for netdev %s",
                  netdev_get_name(netdev));
        rte_aux_tables_uninit(netdev);
    }

    if (netdev_dpdk_sw_meter_get(netdev, NULL, NULL)) {
        /* A SW meter was configured on that port, but it
         * is not implemented. Warn the user. */
        VLOG_WARN("%s: The configured sw-meter is not supported by dpdk "
                  "configuration.", netdev_get_name(netdev));
    }

    return ret;
}

struct offload_trace {
    char *desc;
    struct ovs_list list_node;
};

static struct ovs_mutex offload_trace_list_lock = OVS_MUTEX_INITIALIZER;
static struct ovs_list offload_trace_list OVS_GUARDED_BY(offload_trace_list_lock)
    = OVS_LIST_INITIALIZER(&offload_trace_list);

static void
offload_trace_create(struct offload_trace **ptrace, const char *desc_str)
{
    struct offload_trace *trace;

    trace = xzalloc(sizeof(struct offload_trace));
    trace->desc = xstrdup(desc_str);

    ovs_mutex_lock(&offload_trace_list_lock);
    ovs_list_push_back(&offload_trace_list, &trace->list_node);
    ovs_mutex_unlock(&offload_trace_list_lock);

    *ptrace = trace;
}

static void
offload_trace_destroy(struct offload_trace **ptrace)
{
    struct offload_trace *trace = ptrace ? *ptrace : NULL;

    if (OVS_LIKELY(!trace)) {
        return;
    }

    ovs_mutex_lock(&offload_trace_list_lock);
    ovs_list_remove(&trace->list_node);
    ovs_mutex_unlock(&offload_trace_list_lock);

    free(trace->desc);
    free(trace);
    *ptrace = NULL;
}

static void
offload_trace_list_dump(struct ds *out)
{
    struct offload_trace *trace;

    ds_put_format(out, "---\n");
    ovs_mutex_lock(&offload_trace_list_lock);
    LIST_FOR_EACH (trace, list_node, &offload_trace_list) {
        ds_put_format(out, "%s\n", trace->desc);
    }
    ovs_mutex_unlock(&offload_trace_list_lock);
    ds_put_format(out, "---\n");
}

static int
conn_build_patterns(struct netdev *netdev,
                    const struct ct_match *ct_match,
                    struct flow_patterns *patterns,
                    uint32_t ct_match_zone_id)
{
    struct rte_flow_item_port_id *port_id_spec;
    uint8_t proto = 0;

    patterns->physdev = netdev;

    port_id_spec = per_thread_xzalloc(sizeof *port_id_spec);
    port_id_spec->id = netdev_dpdk_get_port_id(patterns->physdev);
    add_flow_pattern(patterns, RTE_FLOW_ITEM_TYPE_PORT_ID, port_id_spec,
                     NULL, NULL);

    add_flow_pattern(patterns, RTE_FLOW_ITEM_TYPE_ETH, NULL, NULL, NULL);

    /* IPv4 */
    if (ct_match->key.dl_type == htons(ETH_TYPE_IP)) {
        struct rte_flow_item_ipv4 *spec, *mask;

        spec = per_thread_xzalloc(sizeof *spec);
        mask = per_thread_xzalloc(sizeof *spec);

        spec->hdr.src_addr = ct_match->key.src.addr.ipv4;
        spec->hdr.dst_addr = ct_match->key.dst.addr.ipv4;
        spec->hdr.next_proto_id = ct_match->key.nw_proto;

        mask->hdr.src_addr = OVS_BE32_MAX;
        mask->hdr.dst_addr = OVS_BE32_MAX;
        mask->hdr.next_proto_id = UINT8_MAX;

        add_flow_pattern(patterns, RTE_FLOW_ITEM_TYPE_IPV4, spec, mask, NULL);

        /* Save proto for L4 protocol setup. */
        proto = spec->hdr.next_proto_id;
    }

    /* IPv6 */
    if (ct_match->key.dl_type == htons(ETH_TYPE_IPV6)) {
        struct rte_flow_item_ipv6 *spec, *mask;

        spec = per_thread_xzalloc(sizeof *spec);
        mask = per_thread_xzalloc(sizeof *mask);

        memcpy(spec->hdr.src_addr, &ct_match->key.src.addr.ipv6,
               sizeof spec->hdr.src_addr);
        memcpy(spec->hdr.dst_addr, &ct_match->key.dst.addr.ipv6,
               sizeof spec->hdr.dst_addr);
        spec->hdr.proto = ct_match->key.nw_proto;

        memset(&mask->hdr.src_addr, 0xff, sizeof mask->hdr.src_addr);
        memset(&mask->hdr.dst_addr, 0xff, sizeof mask->hdr.dst_addr);
        mask->hdr.proto = UINT8_MAX;

        add_flow_pattern(patterns, RTE_FLOW_ITEM_TYPE_IPV6, spec, mask, NULL);

        /* Save proto for L4 protocol setup. */
        proto = spec->hdr.proto;
    }

    if (proto == IPPROTO_TCP) {
        struct rte_flow_item_tcp *spec, *mask;

        spec = per_thread_xzalloc(sizeof *spec);
        mask = per_thread_xzalloc(sizeof *mask);

        spec->hdr.src_port  = ct_match->key.src.port;
        spec->hdr.dst_port  = ct_match->key.dst.port;

        mask->hdr.src_port  = OVS_BE16_MAX;
        mask->hdr.dst_port  = OVS_BE16_MAX;
        /* Any segment syn, rst or fin must miss and go to SW to trigger
         * a connection state change. */
        mask->hdr.tcp_flags = (TCP_SYN | TCP_RST | TCP_FIN) & 0xff;

        add_flow_pattern(patterns, RTE_FLOW_ITEM_TYPE_TCP, spec, mask, NULL);
    } else if (proto == IPPROTO_UDP) {
        struct rte_flow_item_udp *spec, *mask;

        spec = per_thread_xzalloc(sizeof *spec);
        mask = per_thread_xzalloc(sizeof *mask);

        spec->hdr.src_port  = ct_match->key.src.port;
        spec->hdr.dst_port  = ct_match->key.dst.port;

        mask->hdr.src_port  = OVS_BE16_MAX;
        mask->hdr.dst_port  = OVS_BE16_MAX;

        add_flow_pattern(patterns, RTE_FLOW_ITEM_TYPE_UDP, spec, mask, NULL);
    } else {
        VLOG_ERR_RL(&rl, "Unsupported L4 protocol: 0x02%" PRIx8, proto);
        return -1;
    }

    if (add_pattern_match_reg_field(patterns, REG_FIELD_CT_ZONE,
                                    ct_match_zone_id,
                                    dpdk_offload_api->reg_fields()[REG_FIELD_CT_ZONE].mask)) {
        VLOG_ERR_RL(&rl, "Failed to add the CT zone %"PRIu16" register match",
                    ct_match->key.zone);
        return -1;
    }

    add_flow_pattern(patterns, RTE_FLOW_ITEM_TYPE_END, NULL, NULL, NULL);
    return 0;
}

static void
add_full_set_action(struct flow_actions *actions,
                    enum rte_flow_action_type type,
                    const void *value, size_t size)
{
    struct action_set_data *asd;

    asd = per_thread_xzalloc(sizeof *asd);
    memcpy(asd->value, value, size);
    memset(asd->mask, 0xFF, size);
    asd->size = size;
    add_flow_action(actions, type, asd);
}

static int
conn_build_actions(struct ct_flow_offload_item ct_offload[1],
                   struct flow_actions *actions,
                   struct act_resources *act_resources)
{
    struct rte_flow_action_set_meta *set_meta;
    struct rte_flow_action *ia;
    size_t size;

    add_flow_action(actions, RTE_FLOW_ACTION_TYPE_INDIRECT, NULL);
    actions->shared_count_action_pos = actions->cnt - 1;

    /* NAT */
    if (ct_offload->nat.mod_flags) {
        if (ct_offload->ct_match.key.dl_type == htons(ETH_TYPE_IP)) {
            /* IPv4 */
            size = sizeof ct_offload->nat.key.src.addr.ipv4;
            if (ct_offload->nat.mod_flags & NAT_ACTION_SRC) {
                add_full_set_action(actions,
                    RTE_FLOW_ACTION_TYPE_SET_IPV4_SRC,
                    &ct_offload->nat.key.src.addr.ipv4, size);
            }
            if (ct_offload->nat.mod_flags & NAT_ACTION_DST) {
                add_full_set_action(actions,
                    RTE_FLOW_ACTION_TYPE_SET_IPV4_DST,
                    &ct_offload->nat.key.dst.addr.ipv4, size);
            }
        } else {
            /* IPv6 */
            size = sizeof ct_offload->nat.key.src.addr.ipv6;
            if (ct_offload->nat.mod_flags & NAT_ACTION_SRC) {
                add_full_set_action(actions,
                    RTE_FLOW_ACTION_TYPE_SET_IPV6_SRC,
                    &ct_offload->nat.key.src.addr.ipv6, size);
            }
            if (ct_offload->nat.mod_flags & NAT_ACTION_DST) {
                add_full_set_action(actions,
                    RTE_FLOW_ACTION_TYPE_SET_IPV6_DST,
                    &ct_offload->nat.key.dst.addr.ipv6, size);
            }
        }
        /* TCP | UDP */
        if (ct_offload->nat.mod_flags & NAT_ACTION_SRC_PORT ||
            ct_offload->nat.mod_flags & NAT_ACTION_DST_PORT) {
            size = sizeof ct_offload->nat.key.src.port;
            if (ct_offload->nat.mod_flags & NAT_ACTION_SRC_PORT) {
                add_full_set_action(actions,
                    RTE_FLOW_ACTION_TYPE_SET_TP_SRC,
                    &ct_offload->nat.key.src.port, size);
            }
            if (ct_offload->nat.mod_flags & NAT_ACTION_DST_PORT) {
                add_full_set_action(actions,
                    RTE_FLOW_ACTION_TYPE_SET_TP_DST,
                    &ct_offload->nat.key.dst.port, size);
            }
        }
    }

    /* CT MARK */
    add_action_set_reg_field(actions, REG_FIELD_CT_MARK,
                             ct_offload->mark_key,
                             dpdk_offload_api->reg_fields()[REG_FIELD_CT_MARK].mask);

    /* CT LABEL */
    add_action_set_reg_field(actions, REG_FIELD_CT_LABEL_ID,
                             act_resources->ct_action_label_id,
                             dpdk_offload_api->reg_fields()[REG_FIELD_CT_LABEL_ID].mask);

    /* CT STATE */
    add_action_set_reg_field(actions, REG_FIELD_CT_STATE,
                             ct_offload->ct_state, 0xFF);

    /* Shared counter. */
    ia = &actions->actions[actions->shared_count_action_pos];
    ovs_assert(ia->type == RTE_FLOW_ACTION_TYPE_INDIRECT &&
               ia->conf == NULL);
    ia->conf = act_resources->shared_count_ctx->act_hdl;

    set_meta = per_thread_xzalloc(sizeof *set_meta);
    set_meta->data = act_resources->ct_miss_ctx_id;
    add_flow_action(actions, OVS_RTE_FLOW_ACTION_TYPE(CT_INFO), set_meta);

    /* Last CT action is to go to Post-CT. */
    add_jump_action(actions, POSTCT_TABLE_ID);
    add_flow_action(actions, RTE_FLOW_ACTION_TYPE_END, NULL);

    return 0;
}

static int
dpdk_offload_rte_insert_conn_dir(struct netdev *netdev,
                                 struct ct_flow_offload_item ct_offload[1],
                                 struct conn_item *ci,
                                 struct act_resources *act_resources)
{
    struct flow_patterns patterns = {
        .items = NULL,
        .cnt = 0,
        .s_tnl = DS_EMPTY_INITIALIZER,
    };
    struct flow_actions actions = {
        .actions = NULL,
        .cnt = 0,
        .s_tnl = DS_EMPTY_INITIALIZER,
    };
    struct rte_flow_error error;
    int ret;

    ret = conn_build_patterns(netdev, &ct_offload->ct_match, &patterns,
                              act_resources->ct_match_zone_id);
    if (ret) {
        return ret;
    }

    ret = conn_build_actions(ct_offload, &actions, act_resources);
    if (ret) {
        goto free_patterns;
    }

    ret = create_ct_conn(netdev, &patterns, &actions, &error, act_resources, ci);
    if (!ret) {
        ci->valid = true;
    }

    free_flow_actions(&actions, true);
free_patterns:
    free_flow_patterns(&patterns);

    return ret;
}

static int
dpdk_offload_rte_destroy_conn(struct netdev *netdevs[CT_DIR_NUM], struct conn *conn,
                              struct rte_flow_error *error);

static int
dpdk_offload_rte_insert_conns(struct netdev *netdevs[CT_DIR_NUM], struct batch *conns)
{
    struct rte_flow_error error;
    struct conn *conn;

    netdev_offload_dpdk_get_conns_res(netdevs, conns, true, true, false);

    BATCH_FOREACH_POP (conn, conns) {
        struct ct_flow_offload_item items[CT_DIR_NUM];
        struct ct_offload_handle *coh;
        int ret;

        coh = conntrack_offload_get(conn);
        conntrack_offload_build_add_items(NULL, conn, items, 0);

        for (int dir = 0; dir < CT_DIR_NUM; dir++) {
            ret = dpdk_offload_rte_insert_conn_dir(netdevs[dir], &items[dir],
                                                   &coh->dir[dir].conn_item,
                                                   &coh->act_resources);
            if (ret) {
                break;
            }
        }

        if (ret) {
            dpdk_offload_rte_destroy_conn(netdevs, conn, &error);
            netdev_offload_dpdk_free_conn_res(&coh->act_resources);
            VLOG_ERR_RL(&rl, "%s/%s: Failed to insert connection %p offload.",
                        netdev_get_name(netdevs[CT_DIR_INIT]),
                        netdev_get_name(netdevs[CT_DIR_REP]), conn);
        } else {
            batch_add(conns, conn);
        }
    }

    return -!batch_size(conns);
}

void
netdev_offload_dpdk_get_conns_res(struct netdev *netdevs[CT_DIR_NUM], struct batch *conns,
                                  bool with_miss_ctx, bool with_shared_counters,
                                  bool with_shared_ct_actions)
{
    struct netdev *netdev = netdevs[CT_DIR_INIT];
    struct conn *conn;

    /* Allocation failures result in the connection being removed from the batch.
     * The insertion of those connections will not be attempted, while some others might
     * still succeed. */
    BATCH_FOREACH_POP (conn, conns) {
        struct ct_flow_offload_item items[CT_DIR_NUM];
        struct ct_flow_offload_item *ct_offload;
        struct flows_counter_key counter_id_key;
        struct act_resources *act_resources;
        struct ovs_shared_ctx *ctx;

        conntrack_offload_build_add_items(NULL, conn, items, 0);

        ct_offload = &items[CT_DIR_INIT];
        act_resources = &(conntrack_offload_get(conn)->act_resources);
        act_resources->flow_id = INVALID_FLOW_MARK;

        /* CT ZONE */
        if (get_zone_id(ct_offload->ct_match.key.zone,
                        &act_resources->ct_match_zone_id)) {
            VLOG_ERR_RL(&rl, "Unable to find the offload zone id mapped to the "
                        "CT zone %" PRIu16, ct_offload->ct_match.key.zone);
            goto err;
        }

        /* CT LABEL */
        if (netdev_offload_dpdk_ct_labels_mapping) {
            if (get_label_id(&ct_offload->label_key,
                             &act_resources->ct_action_label_ctx_id)) {
                VLOG_ERR_RL(&rl, "Failed to generate a label ID");
                goto err;
            }
            act_resources->ct_action_label_id = act_resources->ct_action_label_ctx_id;
        } else {
            act_resources->ct_action_label_id = ct_offload->label_key.u32[0];
        }

        /* Shared CT actions. */
        if (with_shared_ct_actions) {
            for (int i = 0; i < CT_DIR_NUM; i++) {
                bool nat = items[i].nat.mod_flags;

                if (nat) {
                    continue;
                }
                ctx = get_shared_ct_actions_ctx(netdev, act_resources->ct_action_label_id,
                                                ct_offload->mark_key, items[i].ct_state);
                if (!ctx) {
                    goto err;
                }
                act_resources->shared_ct_actions_ctx[i] = ctx;
            }
        }

        if (with_miss_ctx) {
            struct ct_miss_ctx miss_ctx;

            memset(&miss_ctx, 0, sizeof miss_ctx);
            miss_ctx.zone = ct_offload->ct_match.key.zone;
            miss_ctx.mark = ct_offload->mark_key;
            memcpy(&miss_ctx.label, &ct_offload->label_key, sizeof miss_ctx.label);
            miss_ctx.state = ct_offload->ct_state;

            if (get_ct_ctx_id(&miss_ctx, &act_resources->ct_miss_ctx_id)) {
                VLOG_ERR("Could not get a CT context ID");
                goto err;
            }
        }

        /* Shared counter. */
        if (with_shared_counters) {
            memset(&counter_id_key, 0, sizeof counter_id_key);
            counter_id_key.ptr_key = ct_offload->ctid_key;

            ctx = get_shared_count_ctx(netdev, &counter_id_key,
                                       OVS_SHARED_CT_COUNT, true);
            if (!ctx) {
                goto err;
            }
            act_resources->shared_count_ctx = ctx;
        }

        put_table_ctx(act_resources->self_table_ctx);
        act_resources->self_table_ctx = NULL;

        batch_add(conns, conn);
        continue;

        /* Error: do not keep this conn within the batch, it cannot be inserted. */
err:
        VLOG_ERR_RL(&rl, "%s/%s: Failed to reserve resources to offload conn %p",
                    netdev_get_name(netdevs[CT_DIR_INIT]),
                    netdev_get_name(netdevs[CT_DIR_REP]), conn);
        netdev_offload_dpdk_free_conn_res(act_resources);
    }
}

void
netdev_offload_dpdk_free_conn_res(struct act_resources *act_resources)
{
    put_action_resources(act_resources);
    memset(act_resources, 0, sizeof *act_resources);
}

static void
conn_unlink(struct conn *conn)
{
    struct ct_offload_handle *coh = conntrack_offload_get(conn);

    ovs_list_remove(&coh->list_node);
    for (int dir = 0; dir < CT_DIR_NUM; dir++) {
        netdev_close(coh->dir[dir].netdev);
    }
}

static void
conn_link(struct netdev *netdevs[CT_DIR_NUM], struct conn *conn)
{
    struct ct_offload_handle *coh = conntrack_offload_get(conn);
    unsigned int tid = netdev_offload_thread_id();

    coh->insertion_tid = tid;
    ovs_spin_init(&coh->lock);
    ovs_list_push_back(&per_threads[tid].conn_list, &coh->list_node);
    coh->conn = conn;
    for (int dir = 0; dir < CT_DIR_NUM; dir++) {
        coh->dir[dir].netdev = netdev_ref(netdevs[dir]);
    }
    ct_offload_handle_active_set(coh, CT_OFFLOAD_BOTH, true);
}

int
netdev_offload_dpdk_conns_add(struct netdev *netdevs[CT_DIR_NUM], struct batch *conns)
{
    struct netdev_offload_dpdk_data *data[CT_DIR_NUM];
    unsigned int tid = netdev_offload_thread_id();
    struct conn *conn;
    int ret;

    if (!netdev_dpdk_is_ethdev(netdevs[CT_DIR_INIT]) ||
        !netdev_dpdk_is_ethdev(netdevs[CT_DIR_REP])) {
        return ENOTSUP;
    }

    /* Connection offloads can only be inserted on a single eswitch. */
    if (netdev_dpdk_get_esw_mgr_port_id(netdevs[CT_DIR_INIT]) !=
        netdev_dpdk_get_esw_mgr_port_id(netdevs[CT_DIR_REP])) {
        VLOG_DBG_RL(&rl, "%s/%s: Offloading connections crossing eswitch is not supported",
                    netdev_get_name(netdevs[CT_DIR_INIT]),
                    netdev_get_name(netdevs[CT_DIR_REP]));
        return ENOTSUP;
    }

    data[CT_DIR_INIT] = (struct netdev_offload_dpdk_data *)
        ovsrcu_get(void *, &netdevs[CT_DIR_INIT]->hw_info.offload_data);
    data[CT_DIR_REP] = (struct netdev_offload_dpdk_data *)
        ovsrcu_get(void *, &netdevs[CT_DIR_REP]->hw_info.offload_data);

    per_thread_init();
    ret = dpdk_offload_api->insert_conns(netdevs, conns);

    BATCH_FOREACH (conn, conns) {
        conn_link(netdevs, conn);
        data[CT_DIR_INIT]->conn_counters[tid]++;
        data[CT_DIR_REP]->conn_counters[tid]++;
        if (netdev_is_dpdk_offload_trace_enabled()) {
            struct ct_offload_handle *coh = conntrack_offload_get(conn);

            for (int dir = 0; dir < CT_DIR_NUM; dir++) {
                struct conn_item *ci = &coh->dir[dir].conn_item;
                struct ds dbg_ds = DS_EMPTY_INITIALIZER;

                if (!ci->valid) {
                    continue;
                }

                ds_put_format(&dbg_ds, "%s ct flow: ", netdev_get_name(netdevs[dir]));
                conntrack_offload_dump(&dbg_ds, conn, dir,
                                       coh->act_resources.ct_match_zone_id,
                                       coh->act_resources.ct_action_label_id,
                                       coh->act_resources.ct_miss_ctx_id);
                offload_trace_create(&ci->trace, ds_cstr(&dbg_ds));
                ds_destroy(&dbg_ds);
            }
        }
    }

    return ret;
}

int
netdev_offload_dpdk_conns_del(struct netdev *netdevs[CT_DIR_NUM], struct batch *conns)
{
    unsigned int tid = netdev_offload_thread_id();
    struct conn *conn;
    int err = 0;

    BATCH_FOREACH (conn, conns) {
        struct ct_offload_handle *coh;
        struct rte_flow_error error;
        int ret = -1;

        coh = conntrack_offload_get(conn);
        /* In case a connection offload is not active there is no point in
         * deleting it
         */
        if (!ct_offload_handle_active(coh, CT_OFFLOAD_BOTH)) {
            continue;
        }

        ovs_spin_lock(&coh->lock);
        ct_offload_handle_active_set(coh, CT_OFFLOAD_BOTH, false);

        ret = dpdk_offload_api->destroy_conn(netdevs, conn, &error);
        if (ret) {
            err |= ret;
            VLOG_ERR_RL(&rl, "%s/%s: failed conn 0x%"PRIxPTR" destroy: %s",
                        netdev_get_name(netdevs[CT_DIR_INIT]),
                        netdev_get_name(netdevs[CT_DIR_REP]),
                        (intptr_t) conn, error.message);
        }

        if (ret == 0) {
            for (int i = 0; i < CT_DIR_NUM; i++) {
                struct conn_item *ci = &coh->dir[i].conn_item;
                struct netdev *netdev = coh->dir[i].netdev;
                struct netdev_offload_dpdk_data *data;

                data = (struct netdev_offload_dpdk_data *)
                    ovsrcu_get(void *, &netdev->hw_info.offload_data);

                ci->flow = NULL;
                ci->valid = false;
                offload_trace_destroy(&ci->trace);

                data->conn_counters[tid]--;
                VLOG_DBG_RL(&rl, "%s: removed conn 0x%"PRIxPTR,
                            netdev_get_name(netdev), (intptr_t) conn);
            }
            conn_unlink(conn);
        }

        ovs_spin_unlock(&coh->lock);
    }

    return err;
}

int
netdev_offload_dpdk_conn_stats(struct netdev *netdevs[CT_DIR_NUM],
                               struct conn *conn,
                               struct dpif_flow_stats *stats,
                               struct dpif_flow_attrs *attrs,
                               long long int now)
{
    struct ct_offload_handle *coh;
    struct rte_flow_error error;
    int ret = ENODEV;

    coh = conntrack_offload_get(conn);
    if (!ct_offload_handle_active(coh, CT_OFFLOAD_BOTH)) {
        return ENODATA;
    }

    if (ovs_spin_trylock(&coh->lock)) {
        return ENODATA;
    }

    /* Check again the status of the offloads after locking. */
    if (!ct_offload_handle_active(coh, CT_OFFLOAD_BOTH)) {
        ret = ENODATA;
        goto unlock;
    }

    ret = dpdk_offload_api->query_conn(netdevs, conn, stats, attrs, &error, now);
    if (ret) {
        VLOG_ERR_RL(&rl, "%s/%s: failed to query statistics on conn 0x%"PRIxPTR": %s",
                    netdev_get_name(netdevs[CT_DIR_INIT]),
                    netdev_get_name(netdevs[CT_DIR_REP]),
                    (intptr_t) conn, error.message);
    }

unlock:
    ovs_spin_unlock(&coh->lock);

    return ret;
}

int
netdev_offload_dpdk_sw_meter_drops(const struct netdev *netdev,
                                   struct rte_flow_query_count *query)
{
    struct netdev_offload_dpdk_data *data;
    struct dpdk_offload_handle *doh;
    struct rte_flow_error error;

    data = (struct netdev_offload_dpdk_data *)
        ovsrcu_get(void *, &netdev->hw_info.offload_data);

    if (!data) {
        return -1;
    }

    doh = &data->sw_meter_match.doh;
    return dpdk_offload_api->query_count(CONST_CAST(struct netdev *, netdev),
                                doh, query, &error);
}

static int
dpdk_offload_rte_destroy(struct netdev *netdev,
                         struct dpdk_offload_handle *doh,
                         struct rte_flow_error *error,
                         bool esw_port_id)
{
    int ret;

    ret = netdev_dpdk_rte_flow_destroy(netdev, doh->rte_flow, error,
                                       esw_port_id);
    if (ret == 0) {
        dpdk_offload_counter_dec(netdev);
    }

    return ret;
}

static int
dpdk_offload_rte_destroy_conn(struct netdev *netdevs[CT_DIR_NUM], struct conn *conn,
                              struct rte_flow_error *error)
{
    struct ct_offload_handle *coh;
    int rets[CT_DIR_NUM];

    memset(rets, 0, sizeof rets);

    coh = conntrack_offload_get(conn);
    for (int dir = 0; dir < CT_DIR_NUM; dir++) {
        if (!coh->dir[dir].conn_item.valid) {
            continue;
        }
        if (!netdevs[dir]) {
            continue;
        }
        rets[dir] = netdev_dpdk_rte_flow_destroy(netdevs[dir], coh->dir[dir].conn_item.rte_flow,
                                                 error, true);
        if (rets[dir] == 0) {
            dpdk_offload_counter_dec(netdevs[dir]);
        }
        if (coh->dir[dir].conn_item.ct_ct_nat) {
            struct rte_flow_error ct_ct_nat_error;

            netdev_dpdk_rte_flow_destroy(netdevs[dir], coh->dir[dir].conn_item.ct_ct_nat,
                                         &ct_ct_nat_error, true);
        }
    }
    netdev_offload_dpdk_free_conn_res(&coh->act_resources);
    return rets[CT_DIR_INIT] | rets[CT_DIR_REP];
}

static int
dpdk_offload_rte_query_count(struct netdev *netdev,
                             struct dpdk_offload_handle *doh,
                             struct rte_flow_query_count *query,
                             struct rte_flow_error *error)
{
    return netdev_dpdk_rte_flow_query_count(netdev, doh->rte_flow, query,
                                            error);
}

static int
dpdk_offload_rte_shared_create(struct ovs_shared_ctx *ctx,
                               struct ovs_shared_arg *arg,
                               struct rte_flow_error *error)
{
    const struct rte_flow_action *action = arg->action;
    struct netdev *netdev = arg->netdev;

    ovs_assert(ctx->act_hdl == NULL);
    ovs_assert(arg->type != OVS_SHARED_CT_ACTIONS);

    ctx->act_hdl = netdev_dpdk_indirect_action_create(netdev, ctx->port_id, action, error);
    if (!ctx->act_hdl) {
        VLOG_DBG("%s: netdev_dpdk_indirect_action_create failed: %d (%s)",
                 netdev_get_name(netdev), error->type, error->message);
        return -1;
    }

    return 0;
}

static int
dpdk_offload_rte_shared_destroy(struct ovs_shared_ctx *ctx,
                                struct rte_flow_error *error)
{
    int ret;

    if (!ctx->act_hdl) {
        return 0;
    }

    ret = netdev_dpdk_indirect_action_destroy(ctx->port_id, ctx->act_hdl, error);
    if (ret) {
        return ret;
    }

    ctx->act_hdl = NULL;

    return 0;
}

static int
dpdk_offload_rte_shared_query(struct ovs_shared_ctx *ctx, void *data,
                              struct rte_flow_error *error)
{
    return netdev_dpdk_indirect_action_query(ctx->port_id, ctx->act_hdl, data,
                                             error);
}

static int
dpdk_offload_rte_query_conn(struct netdev *netdevs[CT_DIR_NUM],
                            struct conn *conn,
                            struct dpif_flow_stats *stats,
                            struct dpif_flow_attrs *attrs,
                            struct rte_flow_error *error,
                            long long int now)
{
    struct rte_flow_query_count query = { .reset = 1 };
    struct ct_offload_handle *coh;
    struct ovs_shared_ctx *ctx;
    int ret;

    coh = conntrack_offload_get(conn);

    if (attrs) {
        attrs->dp_extra_info = NULL;
        attrs->offloaded = true;
        attrs->dp_layer = "dpdk";
    }

    for (int dir = 0; dir < CT_DIR_NUM; dir++) {
        if (!coh->act_resources.shared_count_ctx) {
            struct conn_item *ci = &coh->dir[dir].conn_item;

            ret = netdev_dpdk_rte_flow_query_count(netdevs[dir], ci->rte_flow, &query, error);
        } else {
            ctx = coh->act_resources.shared_count_ctx;
            ret = dpdk_offload_rte_shared_query(ctx, &query, error);
        }

        if (ret) {
            VLOG_DBG_RL(&rl, "%s: Failed to query conn: %p",
                        netdev_get_name(netdevs[dir]), conn);
            /* Try 2nd netdev. */
        }
    }

    if (ret == 0) {
        netdev_offload_dpdk_update_stats(&coh->stats, attrs, &query);
        if (query.hits_set && query.hits) {
            coh->stats.used = now;
        }

        if (stats) {
            memcpy(stats, &coh->stats, sizeof *stats);
        }
    }

    return ret;
}

static int
dpdk_offload_rte_packet_hw_hash(struct netdev *netdev OVS_UNUSED,
                                struct dp_packet *packet OVS_UNUSED,
                                uint32_t seed OVS_UNUSED,
                                uint32_t *hash OVS_UNUSED)
{
    return -1;
}

static int
dpdk_offload_rte_packet_hw_entropy(struct netdev *netdev OVS_UNUSED,
                                   struct dp_packet *packet OVS_UNUSED,
                                   uint16_t *hash OVS_UNUSED)
{
    return -1;
}

static void
dpdk_offload_rte_upkeep(struct netdev *netdev OVS_UNUSED,
                        bool quiescing OVS_UNUSED)
{
    unsigned int tid = netdev_offload_thread_id();
    long long int now = time_last_usec() / 1000;

    id_refmap_upkeep(ct_ctx_irfm, tid, now);
}

void
netdev_offload_dpdk_dump_traces(struct unixctl_conn *conn, int argc OVS_UNUSED,
                                 const char *argv[] OVS_UNUSED, void *aux OVS_UNUSED)
{
    struct ds out = DS_EMPTY_INITIALIZER;

    if (!netdev_is_dpdk_offload_trace_enabled()) {
        unixctl_command_reply(conn, "Tracing not enabled\n");
        return;
    }

    offload_trace_list_dump(&out);
    unixctl_command_reply(conn, ds_cstr(&out));
    ds_destroy(&out);
}

unsigned int
dpdk_offload_thread_nb(void)
{
    if (!dpdk_offload_api) {
        /* Safe static max */
        return MAX_OFFLOAD_THREAD_NB + 1;
    }
    return dpdk_offload_api->thread_nb();
}

static unsigned int
dpdk_offload_rte_thread_nb(void)
{
    return netdev_offload_thread_nb();
}

int
netdev_offload_dpdk_aux_tables_init(struct netdev *netdev)
{
    const struct netdev_flow_api *flow_api =
        ovsrcu_get(const struct netdev_flow_api *, &netdev->flow_api);
    int rv;

    if (!dpdk_offload_api || !flow_api) {
        return -1;
    }
    rv = dpdk_offload_api->aux_tables_init(netdev);
    VLOG_DBG("%s: aux_tables_init, rv=%d", netdev_get_name(netdev), rv);
    return rv;
}

void
netdev_offload_dpdk_aux_tables_uninit(struct netdev *netdev)
{
    if (!dpdk_offload_api) {
        return;
    }
    VLOG_DBG("%s: aux_tables_uninit", netdev_get_name(netdev));
    dpdk_offload_api->aux_tables_uninit(netdev);
}

static int
dpdk_offload_rte_create(struct netdev *netdev,
                        const struct rte_flow_attr *attr,
                        struct rte_flow_item *items,
                        struct rte_flow_action *actions,
                        struct dpdk_offload_handle *doh,
                        struct rte_flow_error *error)
{
    struct rte_flow_action *a;
    struct rte_flow_item *it;

    for (it = items; it->type != RTE_FLOW_ITEM_TYPE_END; it++) {
        if (it->type == OVS_RTE_FLOW_ITEM_TYPE(FLOW_INFO)) {
            it->type = RTE_FLOW_ITEM_TYPE_MARK;
        } else if (it->type == OVS_RTE_FLOW_ITEM_TYPE(HASH)) {
            error->type = RTE_FLOW_ERROR_TYPE_ITEM;
            error->message = "Could not create hash item";
            return -1;
        }
    }

    for (a = actions; a->type != RTE_FLOW_ACTION_TYPE_END; a++) {
        int act_type = a->type;

        if (act_type == OVS_RTE_FLOW_ACTION_TYPE(FLOW_INFO)) {
            a->type = RTE_FLOW_ACTION_TYPE_MARK;
        } else if (act_type == OVS_RTE_FLOW_ACTION_TYPE(CT_INFO)) {
            struct rte_flow_action_set_meta *set_meta;

            a->type = RTE_FLOW_ACTION_TYPE_SET_META;
            set_meta = CONST_CAST(struct rte_flow_action_set_meta *, a->conf);
            set_meta->data = reg_field_spec(REG_FIELD_CT_CTX, set_meta->data);
            set_meta->mask = reg_field_mask(REG_FIELD_CT_CTX, UINT32_MAX);
        } else if (act_type == OVS_RTE_FLOW_ACTION_TYPE(PRE_CT_END)) {
            /* This is only a dummy action used to split pre and post CT. It
             * should never be actually used.
             */
            OVS_NOT_REACHED();
        } else if (act_type == OVS_RTE_FLOW_ACTION_TYPE(HASH)) {
            error->type = RTE_FLOW_ERROR_TYPE_ACTION;
            error->message = "Could not create hash action";
            return -1;
        } else if (IS_REWRITE_ACTION(act_type)) {
            struct action_set_data *asd = (struct action_set_data *) a->conf;

            /* DPDK does not support partially masked set actions. In such
             * case, fail the offload.
             */
            if (!is_all_ones(asd->mask, asd->size)) {
                VLOG_DBG_RL(&rl, "Partial mask is not supported");
                error->type = RTE_FLOW_ERROR_TYPE_ACTION;
                error->message = "Could not create partial set action";
                return -1;
            }
            if (act_type == OVS_RTE_FLOW_ACTION_TYPE(SET_UDP_DST) ||
                act_type == OVS_RTE_FLOW_ACTION_TYPE(SET_TCP_DST)) {
                a->type = RTE_FLOW_ACTION_TYPE_SET_TP_DST;
            }
            if (act_type == OVS_RTE_FLOW_ACTION_TYPE(SET_UDP_SRC) ||
                act_type == OVS_RTE_FLOW_ACTION_TYPE(SET_TCP_SRC)) {
                a->type = RTE_FLOW_ACTION_TYPE_SET_TP_SRC;
            }
        } else if (act_type == RTE_FLOW_ACTION_TYPE_SET_IPV4_DSCP ||
                   act_type == RTE_FLOW_ACTION_TYPE_SET_IPV6_DSCP) {
            error->type = RTE_FLOW_ERROR_TYPE_ACTION;
            error->message = "Could not create set dscp action";
            return -1;
        } else if (act_type == OVS_RTE_FLOW_ACTION_TYPE(SFLOW)) {
            a->type = RTE_FLOW_ACTION_TYPE_SET_META;
        } else if (act_type == OVS_RTE_FLOW_ACTION_TYPE(USERSPACE)) {
            a->type = RTE_FLOW_ACTION_TYPE_SAMPLE;
        }  else if (act_type == OVS_RTE_FLOW_ACTION_TYPE(NESTED)) {
            a->type = RTE_FLOW_ACTION_TYPE_SAMPLE;
        }
    }

    doh->rte_flow = netdev_dpdk_rte_flow_create(netdev, attr, items, actions,
                                                error);
    if (doh->rte_flow != NULL) {
        dpdk_offload_counter_inc(netdev);
    }
    return doh->rte_flow == NULL ? -1 : 0;
}

int
create_rte_flow(struct netdev *netdev,
                const struct rte_flow_attr *attr,
                struct flow_patterns *flow_patterns,
                struct flow_actions *flow_actions,
                struct dpdk_offload_handle *doh,
                struct rte_flow_error *error)
{
    struct rte_flow_action *actions = flow_actions->actions;
    struct rte_flow_item *items = flow_patterns->items;
    struct ds s_extra = DS_EMPTY_INITIALIZER;
    struct ds s = DS_EMPTY_INITIALIZER;
    char *extra_str;
    int rv;

    rv = dpdk_offload_api->create(netdev, attr, items, actions, doh, error);
    if (rv != 0) {
        enum vlog_level level = VLL_WARN;

        ovs_assert(error->message != NULL);
        if (error->type == RTE_FLOW_ERROR_TYPE_ACTION) {
            level = VLL_DBG;
        }
        VLOG_RL(&rl, level, "%s: rte_flow creation failed: %d (%s).",
                netdev_get_name(netdev), error->type, error->message);
        if (!vlog_should_drop(&this_module, level, &rl)) {
            dump_flow(&s, &s_extra, attr, flow_patterns, flow_actions);
            extra_str = ds_cstr(&s_extra);
            VLOG_RL(&rl, level, "%s: Failed flow: %s  flow create %d %s",
                    netdev_get_name(netdev), extra_str,
                    attr->transfer
                    ? netdev_dpdk_get_esw_mgr_port_id(netdev)
                    : netdev_dpdk_get_port_id(netdev), ds_cstr(&s));
        }
    } else {
        struct ds dbg_ds = DS_EMPTY_INITIALIZER;
        bool dbg_rl = VLOG_DROP_DBG(&rl);

        doh->valid = true;
        if (netdev_is_dpdk_offload_trace_enabled() || !dbg_rl) {
            dump_flow(&s, &s_extra, attr, flow_patterns, flow_actions);
            extra_str = ds_cstr(&s_extra);
            ds_put_format(&dbg_ds, "%s: %s  flow create %d user_id 0x%"PRIxPTR" %s",
                          netdev_get_name(netdev), extra_str,
                          attr->transfer
                          ? netdev_dpdk_get_esw_mgr_port_id(netdev)
                          : netdev_dpdk_get_port_id(netdev),
                          (intptr_t) doh->rte_flow, ds_cstr(&s));
            if (!dbg_rl) {
                VLOG_DBG_RL(&rl, "%s", ds_cstr(&dbg_ds));
            }

            if (netdev_is_dpdk_offload_trace_enabled()) {
                offload_trace_create(&doh->trace, ds_cstr(&dbg_ds));
            }
            ds_destroy(&dbg_ds);
        }
    }

    ds_destroy(&s);
    ds_destroy(&s_extra);
    return rv;
}

int
netdev_offload_dpdk_destroy_flow(struct netdev *netdev,
                                 struct dpdk_offload_handle *doh,
                                 const ovs_u128 *ufid, bool is_esw)
{
    struct uuid ufid0 = UUID_ZERO;
    struct rte_flow_error error;
    int ret;

    ret = dpdk_offload_api->destroy(netdev, doh, &error, is_esw);
    if (!ret) {
        VLOG_DBG_RL(&rl, "%s: flow destroy %d user_id rule 0x%"PRIxPTR" ufid "
                    UUID_FMT, netdev_get_name(netdev),
                    is_esw ? netdev_dpdk_get_esw_mgr_port_id(netdev)
                           : netdev_dpdk_get_port_id(netdev),
                    (intptr_t) doh->rte_flow,
                    UUID_ARGS(ufid ? (struct uuid *) ufid : &ufid0));
        /* In case of modification, this handle could be re-used.
         * Clear up relevant fields to make sure it is usable. */
        doh->rte_flow = NULL;
        doh->valid = false;

        offload_trace_destroy(&doh->trace);
    } else {
        VLOG_ERR("Failed: %s: flow destroy %d user_id rule 0x%"PRIxPTR" ufid "
                 UUID_FMT " %s (%u)", netdev_get_name(netdev),
                 is_esw ? netdev_dpdk_get_esw_mgr_port_id(netdev)
                        : netdev_dpdk_get_port_id(netdev),
                 (intptr_t) doh->rte_flow,
                 UUID_ARGS(ufid ? (struct uuid *) ufid : &ufid0),
                 error.message, error.type);
        return -1;
    }

    return ret;
}

int
netdev_offload_dpdk_hw_miss_packet_recover_sflow(struct netdev *netdev,
                                                 struct dp_packet *packet,
                                                 int sflow_id,
                                                 struct dpif_sflow_attr *sflow_attr)
{
    struct sflow_ctx sflow_ctx;

    /* Since sFlow does not work with CT, offloaded sampled packets
     * cannot have mark. If a packet without a mark reaches SW it
     * is either a sampled packet if a cookie is found or a datapath one.
     */
    if (find_sflow_ctx(sflow_id, &sflow_ctx)) {
        log_packet_err(netdev, packet, "sFlow id not found");
        return 0;
    }

    memcpy(sflow_attr->userdata, &sflow_ctx.cookie,
           sflow_ctx.sflow_attr.userdata_len);

    if (!is_all_zeros(&sflow_ctx.sflow_tnl, sizeof sflow_ctx.sflow_tnl)) {
        memcpy(sflow_attr->tunnel, &sflow_ctx.sflow_tnl,
               sizeof *sflow_attr->tunnel);
    } else {
        sflow_attr->tunnel = NULL;
    }

    sflow_attr->sflow = sflow_ctx.sflow_attr.sflow;
    sflow_attr->sflow_len = sflow_ctx.sflow_attr.sflow_len;
    sflow_attr->userdata_len = sflow_ctx.sflow_attr.userdata_len;
    return EIO;
}

int
netdev_offload_dpdk_hw_miss_packet_recover_ext(struct netdev *netdev,
                                               struct dp_packet *packet,
                                               struct dpdk_offload_recovery_info *info,
                                               uint8_t *skip_actions)
{
    struct flow_miss_ctx flow_miss_ctx;
    struct ct_miss_ctx ct_miss_ctx;

    if (info->flow_miss_id) {
        if (find_flow_miss_ctx(info->flow_miss_id, &flow_miss_ctx)) {
            log_packet_err(netdev, packet, "flow miss ctx id not found");
            return 0;
        }
        *skip_actions = flow_miss_ctx.skip_actions;
        packet->md.recirc_id = flow_miss_ctx.recirc_id;
        if (flow_miss_ctx.has_dp_hash) {
            /* OVS will not match on dp-hash unless it has a non-zero value.
             * The 4 LSBs are matched, but zero value is valid. In order to
             * have a non-zero value in the dp-hash of the packet, set the 5th
             * bit.
             */
            packet->md.dp_hash = info->dp_hash | 0x10;
        }

        if (flow_miss_ctx.vport != ODPP_NONE) {
            struct netdev *vport_netdev = netdev_ports_get_short(flow_miss_ctx.vport,
                                                                 netdev->dpif_type);

            if (!vport_netdev) {
                return -1;
            }
            if (is_all_zeros(&flow_miss_ctx.tnl, sizeof flow_miss_ctx.tnl)) {
                if (vport_netdev->netdev_class->pop_header) {
                    parse_tcp_flags(packet, NULL, NULL, NULL);
                    if (!vport_netdev->netdev_class->pop_header(vport_netdev, packet, false)) {
                        return -1;
                    }
                }
            } else {
                memcpy(&packet->md.tunnel, &flow_miss_ctx.tnl,
                       sizeof packet->md.tunnel);
            }
            packet->md.in_port.odp_port = flow_miss_ctx.vport;
        }
    }

    if (info->reg_tags.ct_state) {
        char tmp[50];

        packet->md.ct_state = info->reg_tags.ct_state;
        packet->md.ct_zone = info->reg_tags.ct_zone;
        packet->md.ct_mark = info->reg_tags.ct_mark;
        if (netdev_offload_dpdk_ct_labels_mapping) {
            ovs_u128 ct_label;

            if (find_ct_label(info->reg_tags.ct_label, &ct_label)) {
                snprintf(tmp, sizeof tmp, "label id %" PRIu32" not found", info->reg_tags.ct_label);
                log_packet_err(netdev, packet, tmp);
                return 0;
            }
            packet->md.ct_label = ct_label;
        } else {
            packet->md.ct_label.u32[0] = info->reg_tags.ct_label;
        }
    } else if (info->ct_miss_id) {
        if (find_ct_miss_ctx(info->ct_miss_id, &ct_miss_ctx)) {
            log_packet_err(netdev, packet, "ct miss ctx id not found");
            return 0;
        }
        packet->md.ct_state = ct_miss_ctx.state;
        packet->md.ct_zone = ct_miss_ctx.zone;
        packet->md.ct_mark = ct_miss_ctx.mark;
        packet->md.ct_label = ct_miss_ctx.label;
    }

    return 0;
}

int
parse_ct_label_match(struct match *match, struct act_resources *act_resources,
                     struct flow_patterns *patterns, struct flow *consumed_masks)
{
    uint32_t value, mask;

    if (netdev_offload_dpdk_ct_labels_mapping) {
        ovs_u128 tmp_u128;

        tmp_u128.u64.lo = match->flow.ct_label.u64.lo &
                          match->wc.masks.ct_label.u64.lo;
        tmp_u128.u64.hi = match->flow.ct_label.u64.hi &
                          match->wc.masks.ct_label.u64.hi;
        if (get_label_id(&tmp_u128, &act_resources->ct_match_label_id)) {
            return -1;
        }
        value = act_resources->ct_match_label_id;
        mask = dpdk_offload_api->reg_fields()[REG_FIELD_CT_LABEL_ID].mask;
    } else {
        if (match->wc.masks.ct_label.u32[1] ||
            match->wc.masks.ct_label.u32[2] ||
            match->wc.masks.ct_label.u32[3]) {
            return -1;
        }

        value = match->flow.ct_label.u32[0];
        mask = match->wc.masks.ct_label.u32[0];
    }

    if ((is_all_zeros(&match->flow.ct_label, sizeof match->flow.ct_label) &&
         match->flow.recirc_id == 0) ||
        !add_pattern_match_reg_field(patterns, REG_FIELD_CT_LABEL_ID,
                                     value, mask)) {
        memset(&consumed_masks->ct_label, 0,
               sizeof consumed_masks->ct_label);
    }

    return 0;
}

struct dpdk_offload_api dpdk_offload_api_rte = {
    .create = dpdk_offload_rte_create,
    .destroy = dpdk_offload_rte_destroy,
    .query_count = dpdk_offload_rte_query_count,
    .shared_create = dpdk_offload_rte_shared_create,
    .shared_destroy = dpdk_offload_rte_shared_destroy,
    .shared_query = dpdk_offload_rte_shared_query,
    .get_packet_recover_info = rte_get_packet_recovery_info,
    .insert_conns = dpdk_offload_rte_insert_conns,
    .destroy_conn = dpdk_offload_rte_destroy_conn,
    .query_conn = dpdk_offload_rte_query_conn,
    .reg_fields = rte_get_reg_fields,
    .update_stats = netdev_offload_dpdk_update_stats,
    .upkeep = dpdk_offload_rte_upkeep,
    .aux_tables_init = rte_aux_tables_init,
    .aux_tables_uninit = rte_aux_tables_uninit,
    .packet_hw_hash = dpdk_offload_rte_packet_hw_hash,
    .packet_hw_entropy = dpdk_offload_rte_packet_hw_entropy,
    .thread_nb = dpdk_offload_rte_thread_nb,
};

void
offload_data_init_ext(void)
{
    ct_ctx_init();
    flow_miss_ctx_init();
    label_id_init();
    sflow_id_init();
    shared_age_init();
    shared_count_init();
    shared_ct_actions_init();
    table_id_init();
    tnl_id_init();
    zone_id_init();
}

void
offload_provider_api_init(void)
{
    static struct ovsthread_once init_once = OVSTHREAD_ONCE_INITIALIZER;

    if (ovsthread_once_start(&init_once)) {
        if (ovs_doca_enabled()) {
            dpdk_offload_api = &dpdk_offload_api_doca;
        } else {
            dpdk_offload_api = &dpdk_offload_api_rte;
        }

        ovsthread_once_done(&init_once);
    }
}

void
netdev_dpdk_offload_init_flow_api_ext(struct netdev *netdev)
{
    struct esw_members_aux aux = {
        .esw_netdev = netdev,
        .op = init_esw_members_op,
    };

    /* If the netdev is an ESW manager, init its members too. */
    if (netdev_dpdk_is_esw_mgr(netdev)) {
        netdev_ports_traverse(netdev->dpif_type, esw_members_cb, &aux);
    }

    netdev_offload_dpdk_ct_labels_mapping = netdev_is_ct_labels_mapping_enabled();
    netdev_offload_dpdk_disable_zone_tables = netdev_is_zone_tables_disabled();
}

void
netdev_offload_dpdk_uninit_flow_api_ext(struct netdev *netdev)
{
    struct esw_members_aux aux = {
        .esw_netdev = netdev,
        .op = uninit_esw_members_op,
    };

    /* If the netdev is an ESW manager, uninit its members too. */
    if (netdev_dpdk_is_esw_mgr(netdev)) {
        netdev_ports_traverse(netdev->dpif_type, esw_members_cb, &aux);
    }
}

int
netdev_offload_dpdk_flow_flush_esw_mgr_members(struct netdev *netdev)
{
    struct esw_members_aux aux = {
        .esw_netdev = netdev,
        .ret = 0,
        .op = flush_esw_members_op,
    };

    /* If the flushed netdev is an ESW manager, flush its members too. */
    if (netdev_dpdk_is_esw_mgr(netdev)) {
               netdev_ports_traverse(netdev->dpif_type, esw_members_cb, &aux);
    }

    return aux.ret;
}

void
ufid_to_rte_flow_set(struct ufid_to_rte_flow_data *data,
                     struct netdev *netdev,
                     struct netdev *physdev,
                     struct act_resources *act_resources)
{
    if (data->netdev) {
        ovs_assert(data->netdev == netdev);
    } else {
        data->netdev = netdev_ref(netdev);
    }
    if (data->physdev) {
        ovs_assert(data->physdev == physdev);
    } else {
        data->physdev = netdev != physdev ? netdev_ref(physdev) : physdev;
    }
    memcpy(&data->act_resources, act_resources, sizeof data->act_resources);
}

int
netdev_offload_dpdk_insert_tun_header_with_vlan(
    struct raw_encap_data *raw_encap_data,
    const struct ovs_action_push_vlan *vlan)
{
    struct vlan_eth_header *veh;

    /* Add conf.data sizeof(VLAN_HEADER_LEN) and make sure
     * there is enough headroom.
     */
    raw_encap_data->conf.data -= VLAN_HEADER_LEN;
    if (raw_encap_data->conf.data < raw_encap_data->headroom) {
        return -EOPNOTSUPP;
    }
    raw_encap_data->conf.size += VLAN_HEADER_LEN;

    veh = ALIGNED_CAST(struct vlan_eth_header *, raw_encap_data->conf.data);
    memmove(veh, (char *) veh + VLAN_HEADER_LEN, 2 * ETH_ADDR_LEN);
    veh->veth_type = vlan->vlan_tpid;
    veh->veth_tci = vlan->vlan_tci & htons(~VLAN_CFI);

    return 0;
}
