/*
 * Copyright (c) 2023-2024 NVIDIA CORPORATION & AFFILIATES. All rights reserved.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at:
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

#include <config.h>

#include "batch.h"
#include "id-fpool.h"
#include "netdev-offload.h"
#include "netdev-offload-private.h"
#include "netdev-offload-provider.h"
#include "netdev-provider.h"
#include "netdev-vport.h"
#include "openvswitch/vlog.h"
#include "ovs-doca.h"

VLOG_DEFINE_THIS_MODULE(netdev_offload_ext);

static bool e2e_cache_enabled = false;
static atomic_bool dpdk_offload_trace_enabled = false;
static struct id_fpool *flow_mark_pool;
static uint32_t e2e_cache_size = 0;
bool netdev_offload_ct_on_ct_nat = false;
static bool ct_labels_mapping = false;
static bool disable_zone_tables = false;

int
netdev_conns_add(struct netdev *netdevs[CT_DIR_NUM], struct batch *conns)
{
    const struct netdev_flow_api *flow_apis[CT_DIR_NUM] = { NULL, NULL };

    for (int dir = 0; dir < CT_DIR_NUM; dir++) {
        if (!netdevs[dir]) {
            return ENODEV;
        }
        flow_apis[dir] = ovsrcu_get(const struct netdev_flow_api *,
                                    &netdevs[dir]->flow_api);
        if (!flow_apis[dir]) {
            return ENODEV;
        }
    }

    /* Do NULL-check before doing diff-check, so that an EINVAL truly
     * signals that a connection offload was attempted on two very different ports
     * which is a logical error that should not have happened.
     * The 'flow_api' is set NULL during the normal lifecycle of a port, and it is
     * expected that at some point we will read differing values as one port is
     * destroyed while the other is not.
     */
    if (flow_apis[CT_DIR_INIT] != flow_apis[CT_DIR_REP]) {
        return EINVAL;
    }

    if (!flow_apis[CT_DIR_INIT]->conns_add) {
        return EOPNOTSUPP;
    }

    if (!ovs_doca_initialized()) {
        netdev_offload_upkeep(netdevs[CT_DIR_INIT], false);
        netdev_offload_upkeep(netdevs[CT_DIR_REP], false);
    }

    return flow_apis[CT_DIR_INIT]->conns_add(netdevs, conns);
}

int
netdev_conns_del(struct netdev *netdevs[CT_DIR_NUM], struct batch *conns)
{
    const struct netdev_flow_api *flow_apis[CT_DIR_NUM] = { NULL, NULL };

    if (!netdevs[CT_DIR_INIT] && !netdevs[CT_DIR_REP]) {
        /* No port remains, assume deletion to be successful. */
        return 0;
    }

    for (int dir = 0; dir < CT_DIR_NUM; dir++) {
        if (netdevs[dir]) {
            flow_apis[dir] = ovsrcu_get(const struct netdev_flow_api *,
                                        &netdevs[dir]->flow_api);
        }
    }

    if (flow_apis[CT_DIR_INIT] && flow_apis[CT_DIR_REP] &&
        flow_apis[CT_DIR_INIT] != flow_apis[CT_DIR_REP]) {
        return EINVAL;
    }

    if (!flow_apis[CT_DIR_INIT] && flow_apis[CT_DIR_REP]) {
        flow_apis[CT_DIR_INIT] = flow_apis[CT_DIR_REP];
    }

    if (!flow_apis[CT_DIR_INIT]) {
        return ENODEV;
    }

    if (!flow_apis[CT_DIR_INIT]->conns_del) {
        return EOPNOTSUPP;
    }

    if (!ovs_doca_initialized()) {
        for (int dir = 0; dir < CT_DIR_NUM; dir++) {
            if (netdevs[dir]) {
                netdev_offload_upkeep(netdevs[dir], false);
            }
        }
    }

    return flow_apis[CT_DIR_INIT]->conns_del(netdevs, conns);
}

int
netdev_conn_stats(struct netdev *netdevs[CT_DIR_NUM], struct conn *conn,
                  struct dpif_flow_stats *stats, struct dpif_flow_attrs *attrs, long long int now)
{
    const struct netdev_flow_api *flow_api = NULL;

    if (!netdevs[CT_DIR_INIT] && !netdevs[CT_DIR_REP]) {
        /* No port remains, connection is already removed. */
        return ENODEV;
    }

    for (int dir = 0; dir < CT_DIR_NUM; dir++) {
        /* Full connection offload already enforced that
         * both netdevs use the same offload API. Read either
         * one of the two. If one netdev is NULL or in the process
         * of being destroyed, skip to the other. */
        if (netdevs[dir]) {
            flow_api = ovsrcu_get(const struct netdev_flow_api *,
                                  &netdevs[dir]->flow_api);
            if (flow_api) {
              break;
            }
        }
    }

    if (!flow_api) {
        return ENODEV;
    }

    if (!flow_api->conn_stats) {
        return EOPNOTSUPP;
    }

    return flow_api->conn_stats(netdevs, conn, stats, attrs, now);
}

void
netdev_offload_upkeep(struct netdev *netdev, bool quiescing)
{
    const struct netdev_flow_api *flow_api =
        ovsrcu_get(const struct netdev_flow_api *, &netdev->flow_api);

    if (flow_api && flow_api->upkeep) {
        flow_api->upkeep(netdev, quiescing);
    }
}

static struct ovs_list mark_release_lists[MAX_OFFLOAD_THREAD_NB];

struct mark_release_item {
    struct ovs_list node;
    long long int timestamp;
    uint32_t mark;
};

static void
mark_delayed_release(uint32_t mark)
{
    struct mark_release_item *item = xzalloc(sizeof *item);
    unsigned int tid = netdev_offload_thread_id();

    item->mark = mark;
    item->timestamp = time_msec();
    ovs_list_push_back(&mark_release_lists[tid], &item->node);
    VLOG_DBG("%s: mark=%d, timestamp=%llu", __func__, item->mark,
             item->timestamp);
}

#define DELAYED_RELEASE_TIMEOUT_MS 1000
/* This timeout is to reserve the flow mark ID for a while after the flow is
 * deleted, so that ID will not be allocated to another flow while packets
 * matching the deleted flow are still in HW queues. We assume this timeout
 * is enough for such packets to already be processed.
 */

static void
do_mark_delayed_release(unsigned int tid)
{
    struct ovs_list *mark_release_list = &mark_release_lists[tid];
    struct mark_release_item *item;
    struct ovs_list *list;
    long long int now;

    now = time_msec();
    while (!ovs_list_is_empty(mark_release_list)) {
        list = ovs_list_front(mark_release_list);
        item = CONTAINER_OF(list, struct mark_release_item, node);
        if (now < item->timestamp + DELAYED_RELEASE_TIMEOUT_MS) {
            break;
        }
        VLOG_DBG("%s: mark=%d, timestamp=%llu, now=%llu", __func__, item->mark,
                 item->timestamp, now);
        id_fpool_free_id(flow_mark_pool, tid, item->mark);
        ovs_list_remove(list);
        free(item);
    }
}

uint32_t
netdev_offload_flow_mark_alloc(void)
{
    static struct ovsthread_once init_once = OVSTHREAD_ONCE_INITIALIZER;
    unsigned int tid = netdev_offload_thread_id();
    uint32_t mark;

    if (ovsthread_once_start(&init_once)) {
        size_t i;

        /* Haven't initiated yet, do it here */
        flow_mark_pool = id_fpool_create(MAX_OFFLOAD_THREAD_NB, MIN_FLOW_MARK, NB_FLOW_MARK);
        for (i = 0; i < MAX_OFFLOAD_THREAD_NB; i++) {
            ovs_list_init(&mark_release_lists[i]);
        }

        ovsthread_once_done(&init_once);
    }

    do_mark_delayed_release(tid);
    if (id_fpool_new_id(flow_mark_pool, tid, &mark)) {
        return mark;
    }

    return INVALID_FLOW_MARK;
}

void
netdev_offload_flow_mark_free(uint32_t mark)
{
    mark_delayed_release(mark);
}

int
netdev_offload_get_stats(struct netdev *netdev,
                         struct netdev_offload_stats *stats)
{
    const struct netdev_flow_api *flow_api =
        ovsrcu_get(const struct netdev_flow_api *, &netdev->flow_api);

    return (flow_api && flow_api->get_stats)
           ? flow_api->get_stats(netdev, stats)
           : EOPNOTSUPP;
}

int
netdev_ct_counter_query(struct netdev *netdev,
                        uintptr_t counter,
                        long long now,
                        long long prev_now,
                        struct dpif_flow_stats *stats)
{
    const struct netdev_flow_api *flow_api =
        ovsrcu_get(const struct netdev_flow_api *, &netdev->flow_api);

    return (flow_api && flow_api->ct_counter_query)
           ? flow_api->ct_counter_query(netdev, counter, now, prev_now, stats)
           : EOPNOTSUPP;
}

bool
netdev_is_e2e_cache_enabled(void)
{
    return e2e_cache_enabled;
}

uint32_t
netdev_get_e2e_cache_size(void)
{
    return e2e_cache_size;
}

bool
netdev_is_ct_labels_mapping_enabled(void)
{
    return ct_labels_mapping;
}

bool
netdev_is_zone_tables_disabled(void)
{
    return disable_zone_tables;
}

bool
netdev_is_dpdk_offload_trace_enabled(void)
{
    bool val;

    atomic_read_relaxed(&dpdk_offload_trace_enabled, &val);
    return val;
}

bool
netdev_is_flow_counter_key_zero(const struct flows_counter_key *k)
{
    size_t i;

    for (i = 0; i < ARRAY_SIZE(k->ufid_key); i++) {
        if (k->ufid_key[i].u64.hi != 0 || k->ufid_key[i].u64.lo != 0) {
            return false;
        }
    }
    return true;
}

char*
netdev_flow_counter_key_to_string(const struct flows_counter_key *k,
                                  char *string, size_t sz)
{
    char *ptr = string;
    size_t i;

    if (OVS_UNLIKELY(!sz)) {
        return string;
    }
    sz--;

    if (OVS_LIKELY(sz)) {
        *ptr = '[';
        ptr++;
        sz--;
    } else {
        goto out;
    }
    for (i = 0; i < ARRAY_SIZE(k->ufid_key); i++) {
        const ovs_u128 *ufid = &k->ufid_key[i];
        int n;

        if (ufid->u64.hi == 0 && ufid->u64.lo == 0) {
            break;
        }
        if (OVS_UNLIKELY(sz < 35)) {
            goto out;
        }
        if (i > 0) {
            *ptr = ',';
            ptr++;
            sz--;
        }
        n = sprintf(ptr, "0x%"PRIx64"%"PRIx64, ufid->u64.hi, ufid->u64.lo);
        if (OVS_UNLIKELY(n <= 0)) {
            goto out;
        }
        ptr += (unsigned int) n;
        sz -= (unsigned int) n;
    }
    if (OVS_LIKELY(sz)) {
        *ptr = ']';
        ptr++;
    }
out:
    *ptr = '\0';
    return string;
}

void
netdev_set_flow_api_enabled_ext(const struct smap *ovs_other_config)
{
    if (smap_get_bool(ovs_other_config, "dpdk-offload-trace", false)) {
        static struct ovsthread_once once = OVSTHREAD_ONCE_INITIALIZER;

        if (ovsthread_once_start(&once)) {
            if (!netdev_is_dpdk_offload_trace_enabled()) {
                atomic_store_relaxed(&dpdk_offload_trace_enabled, true);
                VLOG_INFO("DPDK offload trace enabled.");
            }
            ovsthread_once_done(&once);
        }
    } else {
            atomic_store_relaxed(&dpdk_offload_trace_enabled, false);
    }

    if (smap_get_bool(ovs_other_config, "e2e-enable", false)) {
        static struct ovsthread_once once_e2e = OVSTHREAD_ONCE_INITIALIZER;

        if (ovsthread_once_start(&once_e2e)) {
            e2e_cache_enabled = true;
            if (e2e_cache_enabled) {
                e2e_cache_size = smap_get_int(ovs_other_config, "e2e-size",
                                              32000);
                VLOG_INFO("E2E cache size is %"PRIu32, e2e_cache_size);
            }
            ovsthread_once_done(&once_e2e);
        }
    }

    {
        bool req_conf = smap_get_bool(ovs_other_config,
                                      "ct-action-on-nat-conns", false);

        if (req_conf && smap_get_bool(ovs_other_config, "doca-init", false)) {
            VLOG_WARN_ONCE("ct-action-on-nat-conns is not supported by OVS-DOCA.");
        } else if (netdev_offload_ct_on_ct_nat != req_conf) {
            netdev_offload_ct_on_ct_nat = req_conf;
            VLOG_INFO("offloads CT on NAT connections: %s",
                      netdev_offload_ct_on_ct_nat ? "enabled" : "disabled");
        }
    }

    if (smap_get_bool(ovs_other_config, "ct-labels-mapping", false)) {
        static struct ovsthread_once once = OVSTHREAD_ONCE_INITIALIZER;

        if (ovsthread_once_start(&once)) {
            ct_labels_mapping = true;
            VLOG_INFO("CT offloads: labels mapping enabled");
            ovsthread_once_done(&once);
        }
    }

    if (smap_get_bool(ovs_other_config, "disable-zone-tables", false)) {
        static struct ovsthread_once once = OVSTHREAD_ONCE_INITIALIZER;

        if (ovsthread_once_start(&once)) {
            if (!smap_get_bool(ovs_other_config, "doca-init", false)) {
                disable_zone_tables = true;
                VLOG_INFO("CT offloads: zone tables disabled");
            } else {
                VLOG_WARN("disable-zone-tables flag is ignored with doca");
            }
            ovsthread_once_done(&once);
        }
    }
}

int
netdev_packet_hw_hash(struct netdev *netdev,
                      struct dp_packet *packet,
                      uint32_t seed,
                      uint32_t *hash)
{
    const struct netdev_flow_api *flow_api =
        ovsrcu_get(const struct netdev_flow_api *, &netdev->flow_api);

    return (flow_api && flow_api->packet_hw_hash)
        ? flow_api->packet_hw_hash(netdev, packet, seed, hash)
        : EOPNOTSUPP;
}

int
netdev_packet_hw_entropy(struct netdev *netdev,
                         struct dp_packet *packet,
                         uint16_t *entropy)
{
    const struct netdev_flow_api *flow_api =
        ovsrcu_get(const struct netdev_flow_api *, &netdev->flow_api);

    return (flow_api && flow_api->packet_hw_entropy)
        ? flow_api->packet_hw_entropy(netdev, packet, entropy)
        : EOPNOTSUPP;
}

/* Same as 'netdev_ports_get', but do not keep a reference on the
 * netdev returned. This is used for short accesses (within an RCU grace period)
 * when we already know any such reference will remain valid the whole time. */
struct netdev *
netdev_ports_get_short(odp_port_t port_no, const char *dpif_type)
{
    struct port_to_netdev_data *node;
    struct netdev *ret = NULL;

    ovs_rwlock_rdlock(&port_to_netdev_rwlock);
    node = netdev_ports_lookup(port_no, dpif_type);
    if (node) {
        bool visible;

        atomic_read_explicit(&node->visible, &visible, memory_order_acquire);
        if (visible) {
            ret = node->netdev;
        }
    }
    ovs_rwlock_unlock(&port_to_netdev_rwlock);

    return ret;
}
