/*
 * Copyright (c) 2022-2024 NVIDIA CORPORATION & AFFILIATES. All rights reserved.
 *
 * This software product is a proprietary product of NVIDIA CORPORATION &
 * AFFILIATES (the "Company") and all right, title, and interest in and to the
 * software product, including all associated intellectual property rights, are
 * and shall remain exclusively with the Company.
 *
 * This software product is governed by the End User License Agreement
 * provided with the software product.
 *
 */

#ifndef OVS_DOCA_H_
#define OVS_DOCA_H_

#include <config.h>

#include "dp-packet.h"
#include "dpif-netdev.h"
#include "netdev-offload.h"
#include "ovs-atomic.h"
#include "smap.h"
#include "vswitch-idl.h"

#ifdef DPDK_NETDEV

#include <rte_pci.h>

#else

#define PCI_PRI_STR_SIZE sizeof("XXXXXXXX:XX:XX.X")
#define RTE_MAX_ETHPORTS 1

#endif /* DPDK_NETDEV */

#if DOCA_OFFLOAD
#include <doca_flow.h>

/* DOCA requires everything upfront. As a WA we define max number of ESW. */
#define OVS_DOCA_MAX_ESW    2

#define OVS_DOCA_QUEUE_DEPTH 32
#define OVS_DOCA_CT_QUEUE_DEPTH 512

/*
 * Shared counter IDs allocation scheme.
 *
 *                                 ovs_doca_max_shared_counters
 * --------------------------------------------------------------------------------------------->
 *
 *                ovs_doca_max_shared_counters_per_esw
 * ---------------------------------------------------------------------->
 *                                   ESW_0                                            ESW_N
 * ,----------------------------------+----------------------------------.     ,-------+---------.
 * |                CT                |               METER              |     |       |         |
 * |                                  | ,------------------------------. |     |       |         |
 * |                                  | |    GREEN,RED,GREEN,RED,...   | | ... |  CT   |  METER  |
 * |                                  | `------------------------------' |     |       |         |
 * |                                  |                                  |     |       |         |
 * `----------------------------------+----------------------------------'     `-------+---------'
 *   ovs_doca_max_ct_counters_per_esw | OVS_DOCA_MAX_METER_COUNTERS_PER_ESW
 * ---------------------------------->|--------------------------------->
 *                                    |
 *                                    v
 *                                    OVS_DOCA_METER_COUNTERS_PER_ESW_BASE_ID
 */

/* Using shared counters we need 2 per meter */
#define OVS_DOCA_MAX_METER_COLORS 2 /* green and red */
#define OVS_DOCA_MAX_FLOW_METERS_PER_ESW MAX_METERS
#define OVS_DOCA_MAX_SW_METERS_PER_ESW RTE_MAX_ETHPORTS
/* The first shared meter ID is not actually used, because DOCA does not
 * support a shared_meter_id 0. As all meter binding is shifted by one,
 * the last meter ID is the one becoming unusable from our level.
 * As sw-meter use the DPDK port-id, increase the range by one so that
 * all ports can configure a sw-meter. */
#define OVS_DOCA_MAX_METERS_PER_ESW \
    (OVS_DOCA_MAX_FLOW_METERS_PER_ESW + OVS_DOCA_MAX_SW_METERS_PER_ESW + 1)
#define OVS_DOCA_MAX_METERS (OVS_DOCA_MAX_METERS_PER_ESW * OVS_DOCA_MAX_ESW)
#define OVS_DOCA_MAX_METER_COUNTERS_PER_ESW \
    (OVS_DOCA_MAX_FLOW_METERS_PER_ESW * OVS_DOCA_MAX_METER_COLORS)
#define OVS_DOCA_MAX_METER_COUNTERS (OVS_DOCA_MAX_METER_COUNTERS_PER_ESW * \
                                     OVS_DOCA_MAX_ESW)
/* Estimated maximum number of megaflows */
#define OVS_DOCA_MAX_MEGAFLOWS_COUNTERS (1 << 16)
#define OVS_DOCA_FOREACH_ACTIVE_ESWITCH(ID) \
    for (ID = ovs_doca_eswitch_next_active(0); \
         ID < OVS_DOCA_MAX_ESW; \
         ID = ovs_doca_eswitch_next_active(ID + 1))
#define AUX_QUEUE NETDEV_OFFLOAD_THREAD_MAIN
#define FIRST_QUEUE (AUX_QUEUE + 1)
#define ENTRY_PROCESS_TIMEOUT_US 1000

extern bool ovs_doca_async;
extern bool ovs_doca_flow_ct_enabled;
extern uint32_t ctl_pipe_size;
extern uint32_t ctl_pipe_infra_size;
extern unsigned int doca_congestion_threshold;
extern atomic_bool ovs_doca_eswitch_active_ids[OVS_DOCA_MAX_ESW];

/* sw-meter range starts immediately after the flow meter's. */
#define OVS_DOCA_SW_METER_BASE_ID OVS_DOCA_MAX_FLOW_METERS_PER_ESW

/* sFlow uses one mirror object */
#define OVS_DOCA_MAX_MIRRORS_PER_ESW 1
#define OVS_DOCA_MAX_MIRRORS (OVS_DOCA_MAX_MIRRORS_PER_ESW * OVS_DOCA_MAX_ESW)

unsigned int ovs_doca_max_ct_conns(void);

/* Connections are offloaded with one hardware rule per direction.
 * The netdev-offload layer manages offloads rule-wise, so a
 * connection is handled in two parts. This discrepancy can be
 * misleading.
 * This macro expresses the number of hardware rules required
 * to handle the number of CT connections supported by ovs-doca.
 */
static inline unsigned int
ovs_doca_max_ct_rules(void)
{
    return ovs_doca_max_ct_conns() * 2;
}

/* Using shared counters we need 1 per connection */
static inline unsigned int
ovs_doca_max_ct_counters_per_esw(void)
{
    return ovs_doca_flow_ct_enabled ? 0 : ovs_doca_max_ct_conns();
}

/* Total shared counters */
static inline unsigned int
ovs_doca_max_shared_counters_per_esw(void)
{
    return ovs_doca_max_ct_counters_per_esw() +
           OVS_DOCA_MAX_METER_COUNTERS_PER_ESW;
}

static inline unsigned int
ovs_doca_max_shared_counters(void)
{
    return ovs_doca_max_shared_counters_per_esw() * OVS_DOCA_MAX_ESW;
}

static inline int
ovs_doca_eswitch_next_active(int next)
{
    bool active;

    while (next < OVS_DOCA_MAX_ESW) {
        atomic_read_explicit(&ovs_doca_eswitch_active_ids[next], &active,
                             memory_order_consume);
        if (active) {
            return next;
        }
        next++;
    }
    return next;
}

void
ovs_doca_pipe_process_cb(struct doca_flow_pipe *pipe,
                         enum doca_flow_pipe_status status,
                         enum doca_flow_pipe_op op, void *user_ctx);

unsigned int
ovs_doca_n_offload_queues(void);

uint32_t
ovs_doca_get_post_meter_counter_id(uint32_t meter_id,
                                   enum doca_flow_meter_color meter_color);
uint32_t
ovs_doca_meter_id(uint32_t dp_meter_id, uint32_t esw_id);

/* Translate a DPDK port ID into a meter ID in the extended datapath meter
 * ID range. This ID is valid for use with netdev-dpdk meter functions and
 * then to translate it into a DOCA shared meter ID. */
static inline uint32_t
ovs_doca_sw_meter_dp_ext_id(int dpdk_port_id)
{
    return OVS_DOCA_SW_METER_BASE_ID + dpdk_port_id;
}

/* Translate a (DPDK port ID, ESW ID) tuple into a DOCA shared meter ID.This ID
 * is valid within the DOCA provider to reference a shared meter resource. */
static inline uint32_t
ovs_doca_sw_meter_id(int dpdk_port_id, uint32_t esw_id)
{
    return ovs_doca_meter_id(ovs_doca_sw_meter_dp_ext_id(dpdk_port_id),
                             esw_id);
}

#endif /* DOCA_OFFLOAD */

enum ovs_doca_rss_type {
    OVS_DOCA_RSS_IPV4_TCP,
    OVS_DOCA_RSS_IPV4_UDP,
    OVS_DOCA_RSS_IPV4_ICMP,
    OVS_DOCA_RSS_IPV4_ESP,
    OVS_DOCA_RSS_IPV4_OTHER,
    OVS_DOCA_RSS_IPV6_TCP,
    OVS_DOCA_RSS_IPV6_UDP,
    OVS_DOCA_RSS_IPV6_ICMP,
    OVS_DOCA_RSS_IPV6_ESP,
    OVS_DOCA_RSS_IPV6_OTHER,
    OVS_DOCA_RSS_OTHER,
};
/* Must be the last enum type. */
#define OVS_DOCA_RSS_NUM_ENTRIES (OVS_DOCA_RSS_OTHER + 1)

/* For slow path, doca egress the packets by its own pipe, hidden from the
 * user. For hairpin offload, it uses the user's one. Then, we need to use the
 * same mapping as doca, which is simply the port_id.
 */
static inline uint32_t
ovs_doca_port_id_pkt_meta(uint16_t port_id)
{
    return port_id;
}

struct doca_dev;
struct doca_flow_pipe;
struct doca_flow_pipe_entry;
struct doca_flow_port;
struct ofputil_meter_config;
struct rte_eth_stats;
struct rte_mtr_error;
struct rte_mtr_stats;

struct ovs_doca_port_queue {
    PADDED_MEMBERS(CACHE_LINE_SIZE,
        struct dp_packet_batch batch;
        atomic_uint64_t n_packets;
        atomic_uint64_t n_bytes;
    );
};

struct ovs_doca_tx_stats {
    PADDED_MEMBERS(CACHE_LINE_SIZE,
        atomic_uint64_t n_packets;
        atomic_uint64_t n_bytes;
    );
};

struct ovs_doca_esw_ctx {
    char pci_addr[PCI_PRI_STR_SIZE];
    struct doca_flow_pipe *egress_pipe;
    struct doca_flow_pipe *rss_pipe;
    struct doca_dev *dev;
    int cmd_fd;
    uint16_t port_id;
    struct ovs_doca_port_queue *port_queues[RTE_MAX_ETHPORTS];
};

struct ovs_doca_netdev_data {
    const char *devargs;
    struct doca_flow_port *port;
    uint16_t port_id;
    uint16_t esw_mgr_port_id;
    struct ovs_doca_esw_ctx *esw_ctx;
    struct doca_flow_pipe_entry *egress_entry;
    struct doca_flow_pipe_entry *rss_entries[OVS_DOCA_RSS_NUM_ENTRIES];
};

/* Whether DOCA support is compiled-in and user configuration
 * requested DOCA to be enabled. DOCA might not yet be initialized,
 * but the user expects a DOCA execution at some point. */
bool
ovs_doca_enabled(void);

/* Both 'ovs_doca_enabled() == true' and DOCA has successfully initialized. */
bool
ovs_doca_initialized(void);

int
ovs_doca_init(const struct smap *ovs_other_config);

int
ovs_doca_port_start(struct netdev *netdev, struct ovs_doca_netdev_data *dev_data);

int
ovs_doca_port_stop(struct netdev *netdev, struct ovs_doca_netdev_data *dev_data);

void
ovs_doca_set_properties(struct ovs_doca_netdev_data *dev_data,
                        uint16_t port_id,
                        const char *devargs);

void
ovs_doca_status(const struct ovsrec_open_vswitch *cfg);

void
print_doca_version(void);

const char *
ovs_doca_get_version(void);

int
ovs_doca_create_meter(uint32_t meter_id,
                      struct ofputil_meter_config *config,
                      struct rte_mtr_error *error);

int
ovs_doca_delete_meter(uint16_t port_id, uint32_t meter_id,
                      struct rte_mtr_error *error);

int
ovs_doca_mtr_stats_read(uint16_t port_id,
                        uint32_t mtr_id,
                        struct rte_mtr_stats *stats,
                        struct rte_mtr_error *error);

int
ovs_doca_dev_probe(struct ovs_doca_netdev_data *dev_data,
                   const char *devargs,
                   bool already_valid);

void
ovs_doca_dev_close(struct ovs_doca_netdev_data *dev_data);

int
ovs_doca_rx_burst(struct ovs_doca_netdev_data *dev_data,
                  uint16_t queue_id,
                  struct dp_packet_batch *batch);

extern struct ovs_mutex mgmt_queue_lock;

void ovs_doca_mgmt_queue_lock(void)
    OVS_ACQUIRES(mgmt_queue_lock);
void ovs_doca_mgmt_queue_unlock(void)
    OVS_RELEASES(mgmt_queue_lock);

int
ovs_doca_eth_stats_get(const struct netdev *netdev,
                       struct ovs_doca_netdev_data *dev_data,
                       struct rte_eth_stats *rte_stats);

int
ovs_doca_get_custom_stats(const struct netdev *netdev,
                          const struct ovs_doca_netdev_data *dev_data,
                          const struct ovs_doca_tx_stats *tx_stats,
                          struct netdev_custom_stats *custom_stats);

#endif
