/*
 * Copyright (c) 2024 NVIDIA CORPORATION & AFFILIATES. All rights reserved.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at:
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */


#include <config.h>

#include "cmap.h"
#include "hash.h"
#include "openvswitch/vlog.h"
#include "ovs-atomic.h"
#include "refmap.h"
#include "timeval.h"

VLOG_DEFINE_THIS_MODULE(refmap);
static struct vlog_rate_limit rl = VLOG_RATE_LIMIT_INIT(600, 600);

struct refmap {
    struct cmap map;
    struct ovs_mutex map_lock;
    size_t key_size;
    size_t value_size;
    refmap_value_init value_init;
    refmap_value_uninit value_uninit;
    refmap_value_format value_format;
    char *name;
};

struct refmap_node {
    struct ovsrcu_gc_node rcu_node;
    /* CMAP related: */
    struct cmap_node map_node;
    uint32_t hash;
    /* Content: */
    struct ovs_refcount refcount;
    /* Key, then Value follows. */
};

struct refmap *
refmap_create(const char *name,
              size_t key_size,
              size_t value_size,
              refmap_value_init value_init,
              refmap_value_uninit value_uninit,
              refmap_value_format value_format,
              bool disable_map_shrink)
{
    struct refmap *rfm;

    rfm = xzalloc(sizeof *rfm);
    rfm->name = xstrdup(name);
    rfm->key_size = key_size;
    rfm->value_size = value_size;
    rfm->value_init = value_init;
    rfm->value_uninit = value_uninit;
    rfm->value_format = value_format;

    ovs_mutex_init(&rfm->map_lock);
    cmap_init(&rfm->map);

    if (disable_map_shrink) {
        cmap_set_min_load(&rfm->map, 0.0);
    }

    return rfm;
}

static void
refmap_node_free(struct refmap_node *node)
{
    free(node);
}

void
refmap_destroy(struct refmap *rfm)
{
    if (rfm == NULL) {
        return;
    }

    ovs_mutex_lock(&rfm->map_lock);
    if (!cmap_is_empty(&rfm->map)) {
        struct refmap_node *node;

        VLOG_WARN("%s: %s called with elements remaining in the map",
                  rfm->name, __func__);
        CMAP_FOR_EACH (node, map_node, &rfm->map) {
            /* No need to remove the node from the CMAP, it will
             * be destroyed immediately. */
            ovsrcu_gc(refmap_node_free, node, rcu_node);
        }
    }
    cmap_destroy(&rfm->map);
    ovs_mutex_unlock(&rfm->map_lock);

    ovs_mutex_destroy(&rfm->map_lock);
    free(rfm->name);
    free(rfm);
}

static size_t
refmap_key_size(struct refmap *rfm)
{
    return ROUND_UP(rfm->key_size, 8);
}

static void *
refmap_node_key(struct refmap_node *node)
{
    return node + 1;
}

static void *
refmap_node_value(struct refmap *rfm, struct refmap_node *node)
{
    return ((char *) refmap_node_key(node)) + refmap_key_size(rfm);
}

static size_t
refmap_node_total_size(struct refmap *rfm)
{
    return sizeof(struct refmap_node) +
           refmap_key_size(rfm) + rfm->value_size;
}

static struct refmap_node *
refmap_node_from_value(struct refmap *rfm, void *value)
{
    size_t offset = sizeof(struct refmap_node) + refmap_key_size(rfm);

    if ((uintptr_t) value < offset) {
        return NULL;
    }
    return (void *) (((char *) value) - offset);
}

void
refmap_for_each(struct refmap *rfm, void (*cb)(void *value, void *key, void *arg), void *arg)
{
    struct refmap_node *node;

    CMAP_FOR_EACH (node, map_node, &rfm->map) {
        if (ovs_refcount_read(&node->refcount) != 0) {
            cb(refmap_node_value(rfm, node), refmap_node_key(node), arg);
        }
    }
}

void *
refmap_find(struct refmap *rfm, void *key)
{
    struct refmap_node *node;
    uint32_t hash;

    hash = hash_bytes(key, rfm->key_size, 0);
    CMAP_FOR_EACH_WITH_HASH (node, map_node, hash, &rfm->map) {
        if (!memcmp(key, refmap_node_key(node), rfm->key_size) &&
            ovs_refcount_read(&node->refcount) != 0) {
            return refmap_node_value(rfm, node);
        }
    }

    return NULL;
}

void *
refmap_ref(struct refmap *rfm, void *key, void *arg)
{
    struct refmap_node *node;
    void *value;
    bool error;

    error = false;

    value = refmap_find(rfm, key);
    if (value) {
        node = refmap_node_from_value(rfm, value);
        ovs_assert(NULL != node);
        if (ovs_refcount_try_ref_rcu(&node->refcount)) {
            goto log_value;
        }
    }

    node = xzalloc(refmap_node_total_size(rfm));
    node->hash = hash_bytes(key, rfm->key_size, 0);
    ovs_refcount_init(&node->refcount);
    memcpy(refmap_node_key(node), key, rfm->key_size);
    value = refmap_node_value(rfm, node);

    ovs_mutex_lock(&rfm->map_lock);
    if (rfm->value_init(value, arg) == 0) {
        cmap_insert(&rfm->map, &node->map_node, node->hash);
    } else {
        value = NULL;
        error = true;
    }
    ovs_mutex_unlock(&rfm->map_lock);

log_value:
    if (OVS_UNLIKELY(!VLOG_DROP_DBG((&rl)))) {
        struct ds s = DS_EMPTY_INITIALIZER;

        if (rfm->value_format) {
            ds_put_cstr(&s, ", '");
            rfm->value_format(&s, key, value, arg);
            ds_put_cstr(&s, "'");
        }
        VLOG_DBG("%s: %p ref, refcnt=%d%s", rfm->name,
                 value, ovs_refcount_read(&node->refcount),
                 ds_cstr(&s));
        ds_destroy(&s);
    }

    if (error) {
        free(node);
        return NULL;
    }

    return value;
}

bool
refmap_ref_value(struct refmap *rfm, void *value, bool safe)
{
    struct refmap_node *node;

    node = refmap_node_from_value(rfm, value);
    ovs_assert(NULL != node);
    if (safe) {
        ovs_refcount_ref(&node->refcount);
    } else if (!ovs_refcount_try_ref_rcu(&node->refcount)) {
        return false;
    }

    if (OVS_UNLIKELY(!VLOG_DROP_DBG((&rl)))) {
        VLOG_DBG("%s: %p ref, refcnt=%d", rfm->name,
                 value, ovs_refcount_read(&node->refcount));
    }

    return true;
}

bool
refmap_unref(struct refmap *rfm, void *value)
{
    struct refmap_node *node;

    if (value == NULL) {
        return false;
    }

    node = refmap_node_from_value(rfm, value);

    ovs_assert(NULL != node);

    if (OVS_UNLIKELY(!VLOG_DROP_DBG((&rl)))) {
        struct ds s = DS_EMPTY_INITIALIZER;

        VLOG_DBG("%s: %p deref, refcnt=%d", rfm->name,
                 value, ovs_refcount_read(&node->refcount));
        ds_destroy(&s);
    }

    if (ovs_refcount_unref(&node->refcount) == 1) {
        ovs_mutex_lock(&rfm->map_lock);
        rfm->value_uninit(refmap_node_value(rfm, node));
        cmap_remove(&rfm->map, &node->map_node, node->hash);
        ovs_mutex_unlock(&rfm->map_lock);
        ovsrcu_gc(refmap_node_free, node, rcu_node);
        return true;
    }
    return false;
}

void *
refmap_key_from_value(struct refmap *rfm, void *value)
{
    return refmap_node_key(refmap_node_from_value(rfm, value));
}
