/*
 * Copyright (c) 2023-2024 NVIDIA CORPORATION & AFFILIATES. All rights reserved.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at:
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

#include <config.h>

#include <doca_flow.h>
#include <doca_flow_ct.h>
#include <rte_flow.h>
#include <sys/types.h>

#include "conntrack-offload.h"
#include "coverage.h"
#include "doca-pipe-group.h"
#include "dp-packet.h"
#include "dpif-netdev-ext.h"
#include "dpdk-offload-doca.h"
#include "dpdk-offload-provider.h"
#include "id-fpool.h"
#include "openflow/openflow.h"
#include "openvswitch/vlog.h"
#include "offload-metadata.h"
#include "openvswitch/list.h"
#include "openvswitch/match.h"
#include "openvswitch/vlog.h"
#include "ovs-doca.h"
#include "netdev-dpdk.h"
#include "netdev-offload-dpdk-ext.h"
#include "netdev-vport.h"
#include "refmap.h"
#include "timeval.h"
#include "util.h"

/*
 * DOCA offload implementation for DPDK provider.
 *
 * Connection Tracking
 * ===================
 *
 * The CT offload implementation over basic pipes is designed as such:
 *
 * +---------------------------------------------------------------------------------------------+
 * | Control pipes                                                                               |
 * |                                                                                             |
 * |                 ,-[ CT Zone X ]-----.                                                       |
 * |                 |  ,-[ CT Zone Y ]-----.                                                    |
 * |                 |  |  ,-[ CT Zone Z ]-----.                                                 |
 * |                 |  |  |                   |                                                 |
 * |                 |  |  |  ,---------.      |                                                 |
 * |                 |  |  |  |ct_zone=Z+------+--------------------------------------.          |
 * |                 |  |  |  `---------'hit   |                                      |          |
 * |                 |  |  |                   | +----------------------------+       |          |
 * |                 |  |  |                   | | Basic pipes                |       |          |
 * |                 |  |  |                   | |                            |       |          |
 * |                 |  |  |                   | |  ,-[ CT IPv4 x UDP ]-.     |       |          |
 * |                 |  |  |    ,----------.   | |  |                   |     |       |          |
 * |                 |  |  |    |IPv4 + UDP+---+--->|  ,-[ CT IPv4 x TCP ]-.  |       v          |
 * |  ,-[ Pre-CT ]-. |  |  |    `----------'hit| |  |  |                   |  | ,-[ Post-CT ]--. |
 * |  |            +------>|    ,----------.   | |  |  | ,-------------.   |--->|              | |
 * |  |            | |  |  |    |IPv4 + TCP+---+------>| | CT entries  +---+--->|              | |
 * |  `------------' |  |  |    `----------'hit| |  |  | `-------------'hit|  | `--------------' |
 * |                 |  |  |       ,---------. | |  |  |   ,---------.     |  |                  |
 * |                 |  |  |       |Catch-all| | |  |  |   |Catch-all|     |  |                  |
 * |                 `--|  |       `----+----' | |  `--|   `----+----'     |  |                  |
 * |                    `--|            |      | |     `--------+----------'  |                  |
 * |                       `------------+------' |           |  |             |                  |
 * |                                    |        +-----------|--|-------------+                  |
 * |                                    v                    |  |                                |
 * |                             ,-[ Miss pipe ]-------------v--v-------.                        |
 * |                             |         < Execute miss handling >    |                        |
 * |                             `--------------------------------------'                        |
 * +---------------------------------------------------------------------------------------------+
 *
 * This model is replicated once per eswitch.
 *
 * A megaflow that contains a 'ct()' action is split
 * into its 'pre-CT' and 'post-CT' part. The pre-CT is inserted
 * into the eswitch root pipe, and contains the megaflow original
 * match.
 *
 * On match, to execute CT, the packet is sent to the 'CT-zone' pipes,
 * one pipe per CT zone ID. If the ct_zone value is already set on the packet
 * and the value matches that of the current CT-zone pipe, then CT is known
 * to have already been executed. The packet is thus immediately forwarded to
 * post-CT. Post-CT contains the rest of the original megaflow that was not
 * used in pre-CT.
 *
 * If this ct_zone match fails, then either CT was never executed, or
 * it was executed in a different CT zone. If it matches the currently
 * supported CT (network x protocol) tuple, then its ct_zone is set and
 * it is forwarded to the corresponding CT pipe. If no (network x protocol)
 * tuple matches, then CT is not supported for this flow and the packet
 * goes to software.
 *
 * The CT pipe is a basic pipe with a single action type, which writes to
 *
 *  * The packet registers used for CT metadata.
 *  * The packet 5-tuple.
 *
 * For plain CT, the 5-tuple is overwritten with its own values.
 * For NAT, the translations are written instead where relevant.
 *
 * In both cases, all fields are written anyway.
 * This way, the number of template used by the CT pipe is minimal.
 * During performance tests, no impact was measured due to the
 * superfluous writes.
 *
 * If a CT entry matches the packet, the CT pipe action is executed
 * and the packet is then forwarded to post-CT. Otherwise, the packet
 * goes to the sw-meter pipe to execute miss handling.
 *
 * IPv6 Connection Tracking Implementation
 * =======================================
 *
 * By default IPv6 connection offloading is disabled.
 * Set 'other_config:hw-offload-ct-ipv6-enabled=true' to enable.
 *
 * The diagram below shows how IPv6 connection tracking is implemented in the
 * hardware datapath. IPv6 CT rules are too large to fit into single steering
 * hardware objects (STE) and must be split.
 *
 * Note:
 * This is HW specific.
 * In BF3 and above, jumbo STE is supported and the rules can match in a
 * single STE. However, as OVS is HW agnostic, and to support < BF3 cards,
 * this split is done.
 *
 * The IPv6 tuple is divided into a prefix and a suffix, each inserted into
 * their respective pipe. A packet has to match both prefix a suffix rules
 * for a complete IPv6 5-tuple match and continue into the common post-CT pipe.
 *
 * This implementation complements and integrates with the IPv4 model.
 * The distinction is made during the CT-zone stage, matching on L3 protocol
 * to steer the packet toward the relevant basic pipe. Beside splitting match
 * into two, the same logic applies as for the IPv4 implementation.
 *
 *               +-----------------------------------------------+
 *               |IPv6 CT basic pipes                            |
 *               |                         +-[CT suffix TCP]-+   |
 *               |                         |                 |   |
 *               |                         +-----------------+   |
 *               |                         |prefix_id +      |   |
 *               |                         |IPv6.dst +       +-+ |
 *               |                         |TCP ports        | | |
 *               | +[CT prefix pipe]+      +-----------------+ | |
 *+[CT Zone X]+  | |                |      |                 | | |
 *|           |  | |                | +--->|                 | | |
 *|           |  | |                | |    |                 | | |
 *|           |  | |+-------------+ | |    |  +------------+ | | |
 *|+---------+|  | ||IPv6.src+TCP +-+-+ +--+--+ Miss fwd   | | | | +-[Post CT]+
 *||IPv6+UDP ++--+>|+-------------+ |   |  |  +------------+ | | | |          |
 *|+---------+|  | ||IPv6.src+UDP +-+-+ |  |                 | | | |          |
 *||IPv6+TCP ++--+>|+-------------+ | | |  +-----------------+ | | |          |
 *|+---------+|  | |                | | |                      +-+>|          |
 *|           |  | |                | | |  +-[CT suffix UDP]-+ | | |          |
 *|           |  | |                | +-+->|                 | | | |          |
 *|           |  | |+-------------+ |   |  |                 | | | |          |
 *|           |  | ||  Miss fwd   | |   |  +-----------------+ | | |          |
 *|           |  | |+-------+-----+ |   |  |Prefix_id +      | | | |          |
 *|           |  | |        |       |   |  |IPv6.dst +       +-+ | |          |
 *+-----------+  | |        |       |   |  |UDP ports        |   | |          |
 *               | +--------+-------+   |  +-----------------+   | +----------+
 *               |          |           |  |  +------------+ |   |
 *               |          |           | ++--+ Miss fwd   | |   |
 *               |          |           | ||  +------------+ |   |
 *               |          |           | ||                 |   |
 *               |          |           | |+-----------------+   |
 *               +----------+-----------+-+----------------------+
 *                          |           | |
 *                          |    +------v-v---------+
 *                          +--->|     Miss Pipe    |
 *                               +------------------+
 *
 * Meter action post processing
 * ============================
 *
 * +-[pre-CT]------+
 * |               |
 * | +-[CT-zones]--+--+
 * | |                |
 * | | +-[post-CT]----+-+         +-[normal-tables]-+        +-[POST-METER]-------+
 * | | |                |         |                 |        |                    |
 * | | |                |         |                 |        |  +--------------+  |
 * | | |                |         |       ...       |    +---+->|action=meter2 +--+--+
 * | | |                |         |                 |    |   |  +--------------+  |  |
 * | | |                |         | +------------+  |    |   |        ...         |  |
 * | | |                +-------->| |action=meter+--+----+   |  +--------------+  |  |
 * | | |                |         | +------------+  |        |  |action=meterN |<-+--+
 * +-+ |                |         |                 |        |  +-------+------+  |
 *   | |                |         |       ...       |        |          |         |      +------+
 *   +-+                |         |                 |        |          +---------+----->| port |
 *     |                |         |                 |        |                    |      +------+
 *     +----------------+         +-----------------+        +--------------------+
 *
 * Single meter action forwards to the POSTMETER_TABLE where matching on
 * red/green color is done and green packets get forwarded to the real
 * destination of the original flow.
 *
 * If there are more then one meter action, then there is a loop over
 * POSTMETER_TABLE to perform second meter action, then third meter action and
 * so on. Each iteration drops red packets and forwards green packets to the
 * next meter in POSTMETER_TABLE. When last meter action is reached, the next
 * match in POSTMETER_TABLE forwards green packets to the real destination.
 *
 * Multiple split flow
 * ===================
 *
 * If an original flow is too big to fit in a single STE, split the matches of
 * such flow to up to 10 different flows, in which each split flow takes some of
 * the original flow's matches. each match is mapped to an id "prefix_id" which
 * is then matched in the following split flow.
 *
 * Notes:
 * If an original flow has a tunnel header match, the split is done forcefully
 * over the outer header first then the inner.
 *
 * +-[split_depth(0)]---+    +-[split_depth(1)]---+       +-[split_depth(n)]---+
 * |                    |    |                    |       |                    |
 * |match:              |    |match:              |       |match:              |
 * | M(0)               |    | M(1)               |       | M(n)               |
 * |                    |    | prefix_id(1)       |       | prefix_id(n-1)     |
 * |                    |    |                    |       |                    |
 * |actions:            +--->|actions:            | . . . |actions:            |
 * | set prefix_id(1)   |    | set prefix_id(2)   |       | orig_actions       |
 * | jump split_depth(1)|    | jump split_depth(2)|       |                    |
 * |                    |    |                    |       |                    |
 * +--------------------+    +--------------------+       +--------------------+
 *
 * The diagrams were initially drawn with https://asciiflow.com/ and edited in VIM.
 * The resulting extended ASCII chars should however be avoided.
 *
 * Miss handling
 * =============
 *
 * The hardware datapath is magnitudes faster than its software
 * counterpart and is capable of flooding it in some occasion.
 * The exception path sending packets from hardware to software
 * datapath implements a flood protection in the form of a per-port
 * software rate-limit.
 *
 * It is implemented on each eswitch as follows:
 *
 *     .-[ Control ]---.     .-[ Basic ]-----.
 *     |               |     |               |
 *     `-------+-------'     `-------+-------'
 *             |miss                 |miss
 *             `----------+----------'
 *                        |
 *   .-[ pre-miss ]-------V------------------------.
 *   |    .--------------------------------------. |
 *   |    | match eth_type 0x8809 | fwd to kernel| |
 *   |    | match eth_type 0x88cc | fwd to kernel| |
 *   |    `--------------------------------------' |
 *   |      .-------------.                        |
 *   |  miss|  Catch-all  +--,                     |
 *   |      `-------------'  |                     |
 *   `-----------------------|---------------------'
 *   .-[ sw-meter ]----------v----.  .-[ post-sw-meter ]-----------------.
 *   |     Match      Action      |  |                                   |
 *   |    .---------------------. |  |    .----------------------------. |
 *   |    |Port 0   | Meter     +-------->| Port 0 + RED | count, drop | |
 *   |    `---------------------' |  |    `----------------------------' |
 *   |    .---------------------. |  |    .----------------------------. |
 *   |    |Port N   | Meter     +-------->| Port N + RED | count, drop | |
 *   |    `---------------------' |  |    `----------------------------' |
 *   |    .------------------.    |  |    .-------------.                |
 *   |miss|    Catch-all     |    |  |miss|  Catch-all  +--,             |
 *   |    `------------+-----'    |  |    `-------------'  |             |
 *   `-----------------|----------'  `---------------------|-------------'
 *   .-[ meta-push ]---v-----------------------------------v-------------.
 *   |                  .------------------------------.                 |
 *   |                  |    Catch-all  | push vxlan   +--,              |
 *   |                  `------------------------------'  |              |
 *   `----------------------------------------------------|--------------'
 *   .-[ meta-copy ]--------------------------------------v--------------.
 *   |                  .------------------------------.                 |
 *   |                  |    Catch-all  | copy tags    +--,              |
 *   |                  `------------------------------'  |              |
 *   `----------------------------------------------------|--------------'
 *   .-[ miss ]-----------------------------------.       |
 *   |                                            |<------'
 *   |     < Empty pipe defaults to RSS >         |
 *   `--------------------------------------------'
 *
 * All pipe types that are expected to miss (control and basic) are configured
 * with a miss rule forwarding to the kernel pipe which will match on LACP
 * or LLDP ether types, and all other packets will be
 * forwarded to the sw-meter pipe.
 *
 * The sw-meter pipe is itself configured with a miss directive steering
 * toward the meta-push pipe. If a software meter is configured on a port,
 * a metering rule is inserted in the sw-meter pipe matching that specific
 * port. This rule avoids the miss config and apply the requested rate
 * limiting, before forwarding the packet to the post-sw-meter pipe.
 *
 * There, if the color RED is set on the packet, it matches a rule
 * that counts and drop that packet. Otherwise, if the packet was marked
 * GREEN, no rule matches and the catch-all miss directive applies,
 * forwarding the packet back to the meta-push pipe.
 *
 * The meta-push pipe itself is configured with a catch all rule to push a VXLAN
 * header to the packet which then is forwarded toward the meta-copy pipe.
 *
 * The meta-copy pipe contains a catch all rule which copies tag registers that are
 * not propagated to SW otherwise into the VXLAN header which then sends the packet
 * onward to the miss pipe.
 * The miss pipe remains empty, which following DOCA API, triggers the execution of
 * the RSS config on that port, moving the packet to the RX queues.
 *
 * Sample handling
 * ===============
 *
 * Packet sampling is not active by default and doesn't have any overhead. Only
 * when "sflow" option is added to a bridge the sampling pipeline becomes
 * active. It is implemented on each eswitch as follows:
 *
 * +---------------------------+
 * |                           |
 * |         root pipe         |                    Normal
 * |                           +---------------->  datapath
 * |                           |                  processing
 * |                           |
 * |            ...            |   set NONSAMPLE_FLOW_MARK
 * |                           |<----------------------------+
 * |                           |                             | 99%
 * | +-----------------------+ |                 +-----------+-----------+
 * | |                       | |                 |                       |
 * | | prio0: FLOW_INFO == 0 +-+---------------->|      sample pipe      |
 * | |                       | |                 |                       |
 * | +-----------------------+ |                 +-----------+-----------+
 * |                           |                             |
 * +---------------------------+     set NONSAMPLE_FLOW_MARK | 1%
 *                      ^                                    v
 *                      |                           +-----------------+
 *                      |                original   |                 |
 *                      +---------------------------+   mirror pipe   |
 *                                                  |                 |
 *                                                  +--------+--------+
 *                                                           |
 *                                                           | clone
 *                                                           v
 *               +---------------+                  +------------------+
 *               |               |                  |                  |
 *               | misspath pipe |<-----------------+ post-mirror pipe |
 *               |               |       set        |                  |
 *               +---------------+ SAMPLE_FLOW_MARK +------------------+
 *
 * The sample, mirror and post-mirror pipes are created at eswitch
 * initialization. They're not active however, until the first rule that has
 * "sample" action is offloaded. When it happens, a prio0 entry is added to root
 * pipe that forwards all packets w/o FLOW_INFO metadata to the sample pipe.
 * Then the packets are forwarded back to the root pipe and they continue normal
 * processing - they miss prio0 entry because they have FLOW_INFO set.
 *
 * Some percentage of packets is mirrored to the slow path where they will be
 * marked with a "sample" flag in the HW recovery handler. Such packets will
 * continue datapath processing in software and will be dropped eventually when
 * "sample" action is reached.
 *
 * The rest of the "normal" offloads in the root pipe have their original
 * priority lowered to make prio0 reserved for this special rule that does not
 * exist by default. This is why when sampling is disabled it doesn't have any
 * overhead.
 */

COVERAGE_DEFINE(doca_queue_empty);
COVERAGE_DEFINE(doca_queue_none_processed);
COVERAGE_DEFINE(doca_resize_block);

#define BE16_MAX ((OVS_FORCE doca_be16_t) 0xffff)
#define BE32_MAX ((OVS_FORCE doca_be32_t) 0xffffffff)

/* This is hardware-specific.
 * Current NIC must abide by this limit.
 *
 * Need a feature to query it from the actual HW.
 *
 * The value reflects the max number of bytes we can match on
 * in single rule where each field in the match contributes to
 * the total matched bytes according to his size.
 */
#define MAX_FIELD_BYTES 32

#define MAX_OFFLOAD_QUEUE_NB MAX_OFFLOAD_THREAD_NB
#define MAX_GENEVE_OPT 1

#define SHARED_MTR_N_IDS (OVS_DOCA_MAX_MEGAFLOWS_COUNTERS * \
                          DPDK_OFFLOAD_MAX_METERS_PER_FLOW)
#define MIN_SHARED_MTR_FLOW_ID 1

#define META_PUSH_VXLAN_UDP_PORT 4789
#define META_PUSH_VNI 0xdead

/* Calculate the bit offset of a field from the end of its containing type. */
#define REV_BIT_OFFSETOF(STRUCT, FIELD) \
    (8 * (sizeof(STRUCT) - offsetof(STRUCT, FIELD) - MEMBER_SIZEOF(STRUCT, FIELD)))

VLOG_DEFINE_THIS_MODULE(dpdk_offload_doca);
static struct vlog_rate_limit rl = VLOG_RATE_LIMIT_INIT(600, 600);

#define SPLIT_FIELD(field, type, proto_type, proto_field) \
                    {type, proto_type, \
                     offsetof(struct doca_flow_header_format, field), \
                     MEMBER_SIZEOF(struct doca_flow_header_format, field), \
                     offsetof(struct doca_flow_header_format, proto_field), \
                     MEMBER_SIZEOF(struct doca_flow_header_format, proto_field)}

enum split_field_type {
    FIELD_TYPE_INVALID,
    FIELD_TYPE_NONE,
    FIELD_TYPE_SRC,
    FIELD_TYPE_DST,
};

struct split_field {
    int type;
    int proto_type;
    size_t offset;
    size_t size;
    size_t proto_offset;
    size_t proto_size;
};

enum split_l2_field_names {
    FIELD_ETH_MAC_SRC,
    FIELD_ETH_MAC_DST,
    FIELD_ETH_TYPE,
    FIELD_ETH_VLAN_TCI,
    NUM_L2_FIELDS,
};

enum split_l3_field_names {
    FIELD_L3_TYPE,
    FIELD_IP4_SRC,
    FIELD_IP4_DST,
    FIELD_IP4_VER_IHL,
    FIELD_IP4_DSCP_ECN,
    FIELD_IP4_NXT_PROTO,
    FIELD_IP4_TTL,
    FIELD_IP6_SRC,
    FIELD_IP6_DST,
    FIELD_IP6_DSCP_ECN,
    FIELD_IP6_NXT_PROTO,
    FIELD_IP6_HOP_LIMIT,
    NUM_L3_FIELDS,
};

enum split_l4_field_names {
    FIELD_L4_TYPE,
    FIELD_UDP_SRC,
    FIELD_UDP_DSR,
    FIELD_TCP_SRC,
    FIELD_TCP_DST,
    FIELD_TCP_FLAGS,
    FIELD_ICMP_TYPE,
    FIELD_ICMP_CODE,
    NUM_L4_FIELDS,
};

enum split_field_layer {
    L2_HEADERS,
    L3_HEADERS,
    L4_HEADERS,
};

static struct split_field split_fields[][NUM_L3_FIELDS] = {
    [L2_HEADERS] = {
        SPLIT_FIELD(eth.src_mac, FIELD_TYPE_SRC, 0, l2_valid_headers),
        SPLIT_FIELD(eth.dst_mac, FIELD_TYPE_DST, 0, l2_valid_headers),
        SPLIT_FIELD(eth.type, FIELD_TYPE_NONE, 0, l2_valid_headers),
        SPLIT_FIELD(eth_vlan[0].tci, FIELD_TYPE_NONE, 0, l2_valid_headers),
    },
    [L3_HEADERS] = {
        SPLIT_FIELD(ip4.src_ip, FIELD_TYPE_SRC, DOCA_FLOW_L3_TYPE_IP4, l3_type),
        SPLIT_FIELD(ip4.dst_ip, FIELD_TYPE_DST, DOCA_FLOW_L3_TYPE_IP4, l3_type),
        SPLIT_FIELD(ip4.version_ihl, FIELD_TYPE_NONE, DOCA_FLOW_L3_TYPE_IP4, l3_type),
        SPLIT_FIELD(ip4.dscp_ecn, FIELD_TYPE_NONE, DOCA_FLOW_L3_TYPE_IP4, l3_type),
        SPLIT_FIELD(ip4.next_proto, FIELD_TYPE_NONE, DOCA_FLOW_L3_TYPE_IP4, l3_type),
        SPLIT_FIELD(ip4.ttl, FIELD_TYPE_NONE, DOCA_FLOW_L3_TYPE_IP4, l3_type),
        SPLIT_FIELD(ip6.src_ip, FIELD_TYPE_SRC, DOCA_FLOW_L3_TYPE_IP6, l3_type),
        SPLIT_FIELD(ip6.dst_ip, FIELD_TYPE_DST, DOCA_FLOW_L3_TYPE_IP6, l3_type),
        SPLIT_FIELD(ip6.traffic_class, FIELD_TYPE_NONE, DOCA_FLOW_L3_TYPE_IP6, l3_type),
        SPLIT_FIELD(ip6.next_proto, FIELD_TYPE_NONE, DOCA_FLOW_L3_TYPE_IP6, l3_type),
        SPLIT_FIELD(ip6.hop_limit, FIELD_TYPE_NONE, DOCA_FLOW_L3_TYPE_IP6, l3_type),
    },
    [L4_HEADERS] = {
        SPLIT_FIELD(udp.l4_port.src_port, FIELD_TYPE_SRC,
                    DOCA_FLOW_L4_TYPE_EXT_UDP, l4_type_ext),
        SPLIT_FIELD(udp.l4_port.dst_port, FIELD_TYPE_DST,
                    DOCA_FLOW_L4_TYPE_EXT_UDP, l4_type_ext),
        SPLIT_FIELD(tcp.l4_port.src_port, FIELD_TYPE_SRC,
                    DOCA_FLOW_L4_TYPE_EXT_TCP, l4_type_ext),
        SPLIT_FIELD(tcp.l4_port.dst_port, FIELD_TYPE_DST,
                    DOCA_FLOW_L4_TYPE_EXT_TCP, l4_type_ext),
        SPLIT_FIELD(tcp.flags, FIELD_TYPE_NONE, DOCA_FLOW_L4_TYPE_EXT_TCP, l4_type_ext),
        SPLIT_FIELD(icmp.type, FIELD_TYPE_NONE, DOCA_FLOW_L4_TYPE_EXT_ICMP, l4_type_ext),
        SPLIT_FIELD(icmp.code, FIELD_TYPE_NONE, DOCA_FLOW_L4_TYPE_EXT_ICMP, l4_type_ext),
        SPLIT_FIELD(icmp.type, FIELD_TYPE_NONE, DOCA_FLOW_L4_TYPE_EXT_ICMP6, l4_type_ext),
        SPLIT_FIELD(icmp.code, FIELD_TYPE_NONE, DOCA_FLOW_L4_TYPE_EXT_ICMP6, l4_type_ext),
    },
};

enum ct_fwd_type {
    CT_FWD_POST,
    CT_FWD_MISS,
    NUM_CT_FWD,
};

enum ct_nw_type {
    CT_NW_IP4, /* CT on IPv4 networks. */
    CT_NW_IP6, /* CT on IPv6 networks. */
    NUM_CT_NW,
};

enum ct_tp_type {
    CT_TP_UDP, /* CT on UDP datagrams. */
    CT_TP_TCP, /* CT on TCP streams. */
    NUM_CT_TP,
};

/* DOCA-CT templates are defined as:
 * [0]: Plain CT (IPv4 or IPv6)
 * [1]: CT-NAT IPv4
 * [2]: CT-NAT IPv6
 *
 * If Plain-CT, action template [0] is always used.
 * If CT-NAT, action template [1] + [NW] is used.
 */
#define CT_ACTIONS_IDX(NW, NAT) (!!(NAT) * (!!(NAT) + (NW)))
#define NUM_CT_ACTIONS (1 + NUM_CT_NW)

enum hash_pipe_type {
    HASH_TYPE_IPV4_UDP,
    HASH_TYPE_IPV4_TCP,
    HASH_TYPE_IPV4_L3,
    HASH_TYPE_IPV6_UDP_SUF,
    HASH_TYPE_IPV6_TCP_SUF,
    HASH_TYPE_IPV6_L3_SUF,
    HASH_TYPE_IPV6_UDP_PRE,
    HASH_TYPE_IPV6_TCP_PRE,
    HASH_TYPE_IPV6_L3_PRE,
    NUM_HASH_PIPE_TYPE,
};
/* The SUF types must be in lower positions than the PRE ones, as this is the
 * order of initialization, and release in the opposite order.
 */
BUILD_ASSERT_DECL(HASH_TYPE_IPV6_UDP_SUF < HASH_TYPE_IPV6_UDP_PRE);
BUILD_ASSERT_DECL(HASH_TYPE_IPV6_TCP_SUF < HASH_TYPE_IPV6_TCP_PRE);
BUILD_ASSERT_DECL(HASH_TYPE_IPV6_L3_SUF < HASH_TYPE_IPV6_L3_PRE);

struct doca_basic_pipe_ctx {
    struct doca_flow_pipe *pipe;
    struct doca_pipe_group_ctx *fwd_group_ctx;
    struct doca_pipe_group_ctx *miss_group_ctx;
};

/* +--------+   +-------------+
 * |IPv4-UDP|-->|HASH-IPv4-UDP|
 * |        |   +-------------+
 * |        |   +-------------+
 * |IPv4-TCP|-->|HASH-IPv4-TCP|
 * +---+----+   +-------------+
 *     | miss   +-------------+
 *     +------->|HASH-IPv4-L3 |
 *              +-------------+
 * +--------+   +-----------------+   +-----------------+
 * |IPv6-UDP|-->|HASH-IPv6-UDP_PRE|-->|HASH-IPv6-UDP_SUF|
 * |        |   +-----------------+   +-----------------+
 * |        |   +-----------------+   +-----------------+
 * |IPv6-TCP|-->|HASH-IPv6-TCP_PRE|-->|HASH-IPv6-TCP_SUF|
 * +---+----+   +-----------------+   +-----------------+
 *     | miss   +-----------------+   +-----------------+
 *     +------->|HASH-IPv6-L3_PRE |-->|HASH-IPv6-L3_SUF |
 *              +-----------------+   +-----------------+
 * OVS always matches on ether type.
 * We only need to know TCP/UDP, or miss to simple L3.
 * With IPv6, full IP match cannot fit in a single STE, thus a split is done.
 * In PRE, the SRC IP/port are hashed. In SUF, the hash result from the PRE
 * and the DST IP/port are hashed, to create the final hash.
 */
enum hash_tp_type {
    HASH_TP_UDP,
    HASH_TP_TCP,
    NUM_HASH_TP,
};

enum hash_nw_type {
    HASH_NW_IP4, /* CT on IPv4 networks. */
    HASH_NW_IP6, /* CT on IPv6 networks. */
    NUM_HASH_NW,
};

struct doca_meta_transfer_ctx {
    struct doca_basic_pipe_ctx meta_push_pipe_ctx;
    struct doca_basic_pipe_ctx meta_copy_pipe_ctx;
    struct fixed_rule meta_push_hdr;
    struct fixed_rule meta_copy_to_hdr;
    struct netdev *netdev;
};

struct doca_hash_pipe_ctx {
    struct {
        struct doca_flow_pipe *pipe;
        struct doca_flow_pipe_entry *entry;
    } hashes[NUM_HASH_PIPE_TYPE];
    struct {
        struct doca_flow_pipe *pipe;
        struct doca_flow_pipe_entry *tcpudp[NUM_HASH_TP];
    } classifier[NUM_HASH_NW];
    struct doca_pipe_group_ctx *post_hash_group_ctx;
    struct netdev *netdev;
};

struct doca_async_entry {
    struct netdev *netdev; /* If set, port that posted this entry. */
    struct doca_flow_pipe_entry **pentry; /* Storage for this entry. */
    unsigned int index; /* Index of this entry within the aync_state array. */
};

struct gnv_opt_parser {
    struct ovsthread_once once;
    struct doca_flow_parser *parser;
};

struct ct_zone_ctx {
    OVSRCU_TYPE(struct doca_pipe_group_ctx *) plain_ctx;
    OVSRCU_TYPE(struct doca_pipe_group_ctx *) nat_ctx;
    struct fixed_rule entries[2][CT_ZONE_FLOWS_NUM];
};

enum pre_miss_types {
    SEND_TO_KERNEL_LACP,
    SEND_TO_KERNEL_LLDP,
    NUM_SEND_TO_KERNEL,
};

static uint16_t pre_miss_mapping[NUM_SEND_TO_KERNEL] = {
    [SEND_TO_KERNEL_LACP] = ETH_TYPE_LACP,
    [SEND_TO_KERNEL_LLDP] = ETH_TYPE_LLDP,
};

struct doca_pre_miss_ctx {
    struct doca_basic_pipe_ctx pre_miss_pipe_ctx;
    struct fixed_rule pre_miss_rules[NUM_SEND_TO_KERNEL];
    struct netdev *netdev;
};

struct doca_sample_ctx {
    struct doca_flow_pipe_entry *root_sample_entry;
    struct doca_pipe_group_mask_ctx *root_sample_entry_mctx;
    struct ovs_refcount root_sample_entry_refcount;
    struct doca_pipe_group_ctx *sample_pipe_ctx;
    struct doca_flow_pipe_entry *sample_pipe_entry;
    struct doca_pipe_group_mask_ctx *sample_pipe_entry_mctx;
    struct doca_flow_pipe_entry *sample_pipe_miss_entry;
    struct doca_pipe_group_mask_ctx *sample_pipe_miss_entry_mctx;
    struct doca_basic_pipe_ctx mirror_pipe_ctx;
    struct doca_flow_pipe_entry *mirror_pipe_entry;
    struct doca_pipe_group_ctx *postmirror_pipe_ctx;
    struct doca_flow_pipe_entry *postmirror_pipe_entry;
    struct doca_pipe_group_mask_ctx *postmirror_pipe_entry_mctx;
    struct doca_flow_pipe_entry *postmirror_lacp_pipe_entry;
    struct doca_pipe_group_mask_ctx *postmirror_lacp_pipe_entry_mctx;
    uint32_t ratio;
};

struct doca_offload_queue {
    PADDED_MEMBERS(CACHE_LINE_SIZE,
        unsigned int n_waiting_entries;
    );
};

struct doca_offload_esw_ctx {
    struct doca_flow_port *esw_port;
    struct doca_basic_pipe_ctx post_sw_meter_pipe_ctx;
    struct doca_basic_pipe_ctx sw_meter_pipe_ctx;
    struct doca_meta_transfer_ctx meta_transfer_ctx;
    struct doca_pre_miss_ctx pre_miss_ctx;
    struct doca_pipe_group_ctx *root_pipe_group_ctx;
    struct gnv_opt_parser gnv_opt_parser;
    struct {
        struct doca_flow_pipe *pipe;
        struct doca_pipe_group_ctx *post_ct_ctx;
    } ct;
    struct doca_basic_pipe_ctx ct_pipes[NUM_CT_NW][NUM_CT_TP];
    struct ovs_mutex ct_zones_mutex;
    struct ct_zone_ctx ct_zones[NUM_CT_ZONES];
    struct doca_basic_pipe_ctx ct_ip6_prefix;
    struct id_fpool *shared_counter_id_pool;
    struct id_fpool *shared_ct_counter_id_pool;
    struct id_fpool *shared_mtr_flow_id_pool;
    uint32_t esw_id;
    struct ovs_refcount pipe_resizing;
    struct ovs_list resized_pipe_lists[MAX_OFFLOAD_QUEUE_NB];
    struct doca_pipe_group_ctx *post_meter_pipe_group_ctx;
    struct doca_hash_pipe_ctx *hash_pipe_ctx;
    struct doca_sample_ctx sample_ctx;
    struct ovs_rwlock port_rwlock;
    struct doca_offload_queue *offload_queues;
    struct ovs_list destroy_pipe_lists[MAX_OFFLOAD_QUEUE_NB];
};

static doca_error_t
doca_offload_complete_queue_esw(struct doca_offload_esw_ctx *esw, unsigned int qid, bool sync)
    OVS_EXCLUDED(mgmt_queue_lock);

static bool
doca_flow_ct_shared_actions_create(struct doca_offload_esw_ctx *esw,
                                   struct shared_ct_actions_key *key,
                                   uint32_t *id, struct rte_flow_error *error);

static void
doca_flow_ct_shared_actions_destroy(struct doca_offload_esw_ctx *esw,
                                    uint32_t id);

static struct doca_flow_pipe *
doca_offload_get_misspath_pipe(struct doca_offload_esw_ctx *esw_ctx)
{
    /* Get the start of the miss-path handling. */
    return esw_ctx->pre_miss_ctx.pre_miss_pipe_ctx.pipe;
}

void
doca_offload_esw_ctx_pipe_resizing_ref(struct doca_offload_esw_ctx *esw_ctx)
{
    ovs_refcount_ref(&esw_ctx->pipe_resizing);
}

void
doca_offload_esw_ctx_pipe_resizing_unref(struct doca_offload_esw_ctx *esw_ctx)
{
    ovs_refcount_unref(&esw_ctx->pipe_resizing);
}

void
doca_offload_esw_ctx_resized_list_add(struct doca_offload_esw_ctx *esw_ctx,
                                      struct ovs_list *resized_list_node)
{
    unsigned int tid = netdev_offload_thread_id();

    ovs_list_push_back(&esw_ctx->resized_pipe_lists[tid], resized_list_node);
}

struct meter_info {
    struct ovs_list list_node;
    uint32_t id;
    uint32_t flow_id;
};

struct doca_act_vars {
    uint32_t flow_id;
    uint32_t mtr_flow_id;
    struct ovs_list next_meters;
    const struct hash_data *hash_data;
    uint32_t sample_ratio;
};

struct doca_meter_ctx {
    struct ovs_list list_node;
    uint32_t post_meter_flow_id;
    struct doca_flow_pipe_entry *post_meter_red_entry;
    struct doca_pipe_group_mask_ctx *post_meter_red_entry_mctx;
    struct doca_flow_pipe_entry *post_meter_green_entry;
    struct doca_pipe_group_mask_ctx *post_meter_green_entry_mctx;
};

static int
destroy_dpdk_offload_handle(struct doca_offload_esw_ctx *esw,
                            struct netdev *netdev,
                            struct dpdk_offload_handle *doh,
                            unsigned int queue_id,
                            struct rte_flow_error *error);
static struct id_fpool *esw_id_pool;

static int
shared_mtr_flow_id_alloc(struct doca_offload_esw_ctx *esw_ctx,
                         struct rte_flow_error *error);

static void
shared_mtr_flow_id_free(struct doca_offload_esw_ctx *esw_ctx, uint32_t id);

static int
doca_ct_zone_init(struct netdev *netdev, struct doca_offload_esw_ctx *esw_ctx,
                  uint32_t zone_group);

static inline enum ct_nw_type
l3_to_nw_type(enum doca_flow_l3_type l3_type)
{
    switch (l3_type) {
    case DOCA_FLOW_L3_TYPE_IP4: return CT_NW_IP4;
    case DOCA_FLOW_L3_TYPE_IP6: return CT_NW_IP6;
    case DOCA_FLOW_L3_TYPE_NONE: return NUM_CT_NW;
    };
    return NUM_CT_NW;
}

static inline enum ct_tp_type
l4_to_tp_type(enum doca_flow_l4_type_ext l4_type)
{
    switch (l4_type) {
    case DOCA_FLOW_L4_TYPE_EXT_TCP: return CT_TP_TCP;
    case DOCA_FLOW_L4_TYPE_EXT_UDP: return CT_TP_UDP;
    case DOCA_FLOW_L4_TYPE_EXT_ICMP:
    case DOCA_FLOW_L4_TYPE_EXT_ICMP6:
    case DOCA_FLOW_L4_TYPE_EXT_TRANSPORT:
    case DOCA_FLOW_L4_TYPE_EXT_NONE: return NUM_CT_TP;
    }
    return NUM_CT_TP;
}

#define MIN_SPLIT_PREFIX_ID 1
#define MAX_SPLIT_PREFIX_ID (reg_fields[REG_FIELD_SCRATCH].mask - 1)
#define NUM_SPLIT_PREFIX_ID (MAX_SPLIT_PREFIX_ID - MIN_SPLIT_PREFIX_ID + 1)
static struct refmap *split_prefix_rfm;
static struct id_fpool *split_prefix_id_pool = NULL;

OVS_ASSERT_PACKED(struct doca_split_prefix_key,
    struct doca_flow_match spec;
    struct doca_flow_match mask;
    union {
        struct doca_pipe_group_ctx *prefix_group_ctx;
        struct doca_flow_pipe *prefix_pipe;
    };
);

struct doca_split_prefix_ctx {
    struct netdev *netdev;
    struct dpdk_offload_handle doh;
    uint32_t id;
};

struct doca_split_prefix_arg {
    struct netdev *netdev;
    struct doca_flow_match *spec;
    struct doca_flow_match *mask;
    bool prefix_is_group;
    union {
        struct doca_pipe_group_ctx *prefix_group_ctx;
        struct doca_flow_pipe *prefix_pipe;
    };
    struct doca_flow_pipe *suffix_pipe;
    uint32_t set_flow_info_id;
    uint32_t prio;
};

doca_error_t
doca_offload_add_entry(unsigned int qid,
                       struct doca_flow_pipe *pipe,
                       const struct doca_flow_match *match,
                       const struct doca_flow_actions *actions,
                       const struct doca_flow_monitor *monitor,
                       const struct doca_flow_fwd *fwd,
                       uint32_t flags,
                       struct doca_offload_esw_ctx *esw,
                       struct doca_flow_pipe_entry **pentry)
    OVS_EXCLUDED(mgmt_queue_lock)
{
    doca_error_t err;

    if (qid == AUX_QUEUE) {
        ovs_doca_mgmt_queue_lock();
        err = doca_flow_pipe_add_entry(qid, pipe, match, actions, monitor,
                                       fwd, flags, esw, pentry);
        ovs_doca_mgmt_queue_unlock();
    } else {
        err = doca_flow_pipe_add_entry(qid, pipe, match, actions, monitor,
                                       fwd, flags, esw, pentry);
    }
    if (err == DOCA_SUCCESS) {
        if (esw) {
            esw->offload_queues[qid].n_waiting_entries++;
        } else {
            VLOG_DBG("added entry %p without an eswitch handle", *pentry);
        }
    }
    return err;
}

doca_error_t
doca_offload_remove_entry(struct doca_offload_esw_ctx *esw,
                          unsigned int qid, uint32_t flags,
                          struct doca_flow_pipe_entry *entry)
    OVS_EXCLUDED(mgmt_queue_lock)
{
    doca_error_t err;

    if (qid == AUX_QUEUE) {
        ovs_doca_mgmt_queue_lock();
        err = doca_flow_pipe_remove_entry(qid, flags, entry);
        ovs_doca_mgmt_queue_unlock();
    } else {
        err = doca_flow_pipe_remove_entry(qid, flags, entry);
    }
    if (err == DOCA_SUCCESS) {
        esw->offload_queues[qid].n_waiting_entries++;
        if (qid == AUX_QUEUE) {
            /* Ignore potential errors here, as even if the queue completion
             * failed, the entry removal would still be issued. The caller
             * requires knowing so. */
            doca_offload_complete_queue_esw(esw, qid, true);
        }
    }
    return err;
}

static void
doca_hash_pipe_ctx_uninit(struct doca_offload_esw_ctx *esw,
                          struct doca_hash_pipe_ctx *ctx)
{
    unsigned int queue_id = netdev_offload_thread_id();
    enum hash_nw_type nw_type;
    int i;

    if (ctx == NULL) {
        return;
    }

    /* Destroy classifiers. */
    for (nw_type = 0; nw_type < NUM_HASH_NW; nw_type++) {
        for (i = 0; i < NUM_HASH_TP; i++) {
            if (ctx->classifier[nw_type].tcpudp[i]) {
                doca_offload_remove_entry(esw, queue_id, DOCA_FLOW_NO_WAIT,
                                          ctx->classifier[nw_type].tcpudp[i]);
                dpdk_offload_counter_dec(ctx->netdev);
            }
        }
        if (ctx->classifier[nw_type].pipe) {
            doca_flow_pipe_destroy(ctx->classifier[nw_type].pipe);
        }
    }

    /* Remove entries first then complete the queue,
     * before starting pipe destroy. */

    for (i = 0; i < NUM_HASH_PIPE_TYPE; i++) {
        if (ctx->hashes[i].entry) {
            doca_offload_remove_entry(esw, queue_id, DOCA_FLOW_NO_WAIT,
                                      ctx->hashes[i].entry);
            dpdk_offload_counter_dec(ctx->netdev);
        }
    }

    doca_offload_complete_queue_esw(esw, queue_id, true);

    /* Release hash pipes in the reverse order of their dependency.
     * Ordering is described and enforced at the `hash_pipe_type` definition.
     */
    for (i = NUM_HASH_PIPE_TYPE - 1; i >= 0; i--) {
        if (ctx->hashes[i].pipe) {
            doca_flow_pipe_destroy(ctx->hashes[i].pipe);
        }
    }

    doca_pipe_group_ctx_unref(ctx->post_hash_group_ctx);

    free(ctx);
}

static void
doca_basic_pipe_ctx_uninit(struct doca_basic_pipe_ctx *ctx)
{
    doca_flow_pipe_destroy(ctx->pipe);
    ctx->pipe = NULL;
    doca_pipe_group_ctx_unref(ctx->miss_group_ctx);
    ctx->miss_group_ctx = NULL;
    doca_pipe_group_ctx_unref(ctx->fwd_group_ctx);
    ctx->fwd_group_ctx = NULL;
}

static struct reg_field reg_fields[] = {
    [REG_FIELD_FLOW_INFO] = {
        .type = REG_TYPE_META,
        .index = 0,
        .offset = 12,
        .mask = 0x000FFFFF,
    },
    [REG_FIELD_META_PUSH] = {
        .type = REG_TYPE_META,
        .index = 0,
        .offset = 0,
        .mask = 0x00000001,
    },
    /* Not in use. */
    [REG_FIELD_CT_CTX] = {
        .type = REG_TYPE_META,
        .index = 0,
        .offset = 1,
        .mask = 0x000007FF,
    },
    [REG_FIELD_DP_HASH] = {
        .type = REG_TYPE_META,
        .index = 0,
        .offset = 8,
        .mask = 0x0000000F,
    },
    [REG_FIELD_TUN_INFO] = {
        .type = REG_TYPE_TAG,
        .index = 0,
        .offset = 16,
        .mask = 0x0000FFFF,
    },
    /* Scratch info:
     * - prefix id for split flows.
     * - dp-hash ipv6 split hash result.
     * - dp-hash seed.
     */
    [REG_FIELD_SCRATCH] = {
        .type = REG_TYPE_TAG,
        .index = 1,
        .offset = 0,
        .mask = 0xFFFFFFFF,
    },
    /* sFlow context is not used for DOCA. */
    [REG_FIELD_SFLOW_CTX] = {
        .type = REG_TYPE_META,
        .index = 0,
        .offset = 0,
        .mask = 0x0000FFFF,
    },
    [REG_FIELD_RECIRC] = {
        .type = REG_TYPE_TAG,
        .index = 0,
        .offset = 0,
        .mask = MATCH_RECIRC_ID_MASK,
    },
    [REG_FIELD_CT_MARK] = {
        .type = REG_TYPE_TAG,
        .index = 2,
        .offset = 0,
        .mask = 0xFFFFFFFF,
    },
    [REG_FIELD_CT_LABEL_ID] = {
        .type = REG_TYPE_TAG,
        .index = 3,
        .offset = 0,
        .mask = 0xFFFFFFFF,
    },
    [REG_FIELD_CT_STATE] = {
        .type = REG_TYPE_TAG,
        .index = 4,
        .offset = 24,
        .mask = 0x000000FF,
    },
    [REG_FIELD_CT_ZONE] = {
        .type = REG_TYPE_TAG,
        .index = 4,
        .offset = 0,
        .mask = 0x0000FFFF,
    },
    [REG_FIELD_TAG0] = {
        .type = REG_TYPE_TAG,
        .index = 0,
        .offset = 0,
        .mask = 0xFFFFFFFF,
    },
};

/* Register usage:
 * +------+--------+--------+--------+--------+--------+--------+--------+--------+
 * |      |  31-28 |  27-24 |  23-20 |  19-16 |  15-12 |  11-8  |   7-4  |   3-0  |
 * |      +-----------------+-----------------+-----------------+-----------------+
 * |      |      31-24      |      23-16      |      15-8       |       7-0       |
 * |      +-----------------+-----------------+-----------------+-----------------+
 * |      |                                 31-0                                  |
 * +------+-----------------+--------------------------+--------+--------+------+-+
 * | meta |               FLOW_INFO                    |dp-hash |      FREE     |M|
 * |      | - Pre/post CT: flow-id. In case            |        |               |E|
 * |      |   of CT miss, it has a                     |        |               |T|
 * |      |   flow_miss_ctx to recover.                |        |               |A|
 * |      | - Post-hash: flow-id.                      |        |               | |
 * |      | - Meters: meter-flow-id. As a              |        |               |P|
 * |      |   temp id in the post-meter loop.          |        |               |U|
 * |      |   Last iteration restores to the           |        |               |S|
 * |      |   flow-id.                                 |        |               |H|
 * +------+-----------------+-----------------+--------+--------+--------+------+-+
 * | tag0 |           TUNNEL_INFO             |             recirc_id             |
 * |      | - recirc_id(0) matches on the     |                                   |
 * |      |   packet's fields, and decap. The |                                   |
 * |      |   tunnel match is mapped.         |                                   |
 * |      |   Following recircs map on this   |                                   |
 * |      |   ID.                             |                                   |
 * |      |        ---- TAG0 should be cleared before leaving FDB----             |
 * +------+-----------------+-----------------+-----------------+-----------------+
 * | tag1 |                                SCRATCH                                |
 * |      | - Prefix id for split flows.                                          |
 * |      | - dp-hash ipv6 split hash result.                                     |
 * |      | - dp-hash seed.                                                       |
 * |      |     ---- TAG1 is mapped to regc_3 which is used also by dpdk. ----    |
 * |      |     ---- Can be used only temorarily (SCRATCH). ----                  |
 * +------+-----------------+-----------------+-----------------+-----------------+
 * | tag2 |                               CT_MARK                                 |
 * +------+-----------------+-----------------+-----------------+-----------------+
 * | tag3 |                      CT_LABEL. 32 low bits or mapped.                 |
 * +------+-----------------+-----------------+-----------------+-----------------+
 * | tag4 |     CT_STATE    |                 |               CT_ZONE             |
 * +------+-----------------+-----------------+-----------------+-----------------+
 */

static struct reg_field *
dpdk_offload_doca_get_reg_fields(void)
{
    return reg_fields;
}

static void
doca_set_reg_val(struct doca_flow_meta *dmeta, uint8_t reg_field_id,
                 uint32_t val)
{
    struct reg_field *reg_field = &reg_fields[reg_field_id];

    if (reg_field->type == REG_TYPE_TAG) {
        dmeta->u32[reg_field->index] |= reg_field_spec(reg_field_id, val);
    } else {
        dmeta->pkt_meta |= reg_field_spec(reg_field_id, val);
    }
}

static void
doca_set_reg_mask(struct doca_flow_meta *dmeta, uint8_t reg_field_id)
{
    doca_set_reg_val(dmeta, reg_field_id, UINT32_MAX);
}

void
doca_offload_set_reg_template(struct doca_flow_meta *dmeta, enum dpdk_reg_id id)
{
    struct reg_field *reg_field = &reg_fields[id];
    uint32_t *u32;

    /* This function does not respect the logical layout of the reg-fields.
     * It sets the whole register used by 'id' to 0xff, as is currently
     * required by DOCA-flow for pipe templates definitions.
     */
    u32 = (reg_field->type == REG_TYPE_TAG) ? &dmeta->u32[reg_field->index]
                                            : &dmeta->pkt_meta;
    memset(u32, 0xff, sizeof *u32);
}

void
doca_offload_set_reg_mask(struct doca_flow_meta *dmeta, enum dpdk_reg_id id)
{
    doca_set_reg_val(dmeta, id, UINT32_MAX);
}

static void
doca_set_reg_val_mask(struct doca_flow_meta *dmeta,
                      struct doca_flow_meta *dmeta_mask,
                      uint8_t reg_field_id,
                      uint32_t val)
{
    doca_set_reg_val(dmeta, reg_field_id, val);
    doca_set_reg_mask(dmeta_mask, reg_field_id);
}

static uint32_t
doca_get_reg_bit_offset(uint8_t reg_field_id)
{
    struct reg_field *reg_field = &reg_fields[reg_field_id];
    uint32_t offset;

    offset = reg_field->offset;

    if (reg_field->type == REG_TYPE_TAG) {
        offset += offsetof(struct doca_flow_meta, u32[reg_field->index]) * 8;
    } else {
        offset += offsetof(struct doca_flow_meta, pkt_meta) * 8;
    }

    return offset;
}

static uint32_t
doca_get_reg_width(uint8_t reg_field_id)
{
    struct reg_field *reg_field = &reg_fields[reg_field_id];

    return ffsl(~((uint64_t) reg_field->mask)) - 1;
}

static void
doca_spec_mask(void *dspec__, void *dmask__, size_t size)
{
    char *dspec = dspec__;
    char *dmask = dmask__;
    int i;

    if (!dspec || !dmask) {
        return;
    }

    for (i = 0; i < size; i++) {
        dspec[i] &= dmask[i];
    }
}

static void
doca_copy_mask_field__(void *dspec, void *dmask,
                       const void *spec, const void *mask,
                       size_t size)
{
    memcpy(dspec, spec, size);
    memcpy(dmask, mask, size);
    doca_spec_mask(dspec, dmask, size);
}

static void
doca_translate_gre_key_item(const struct rte_flow_item *item,
                            struct doca_flow_match *doca_spec,
                            struct doca_flow_match *doca_mask)
{
    const rte_be32_t *key_spec, *key_mask;

    doca_spec->tun.type = DOCA_FLOW_TUN_GRE;
    doca_mask->tun.type = DOCA_FLOW_TUN_GRE;

    key_spec = item->spec;
    key_mask = item->mask;

    if (item->spec) {
        doca_spec->tun.gre_key = (OVS_FORCE doca_be32_t) *key_spec;
    }
    if (item->mask) {
        doca_mask->tun.gre_key = (OVS_FORCE doca_be32_t) *key_mask;
    }

    doca_spec_mask(&doca_spec->tun.gre_key, &doca_mask->tun.gre_key,
                   sizeof doca_spec->tun.gre_key);
}

static void
doca_translate_gre_item(const struct rte_flow_item *item,
                        struct doca_flow_match *doca_spec,
                        struct doca_flow_match *doca_mask)
{
    const struct rte_gre_hdr *greh_spec, *greh_mask;

    doca_spec->tun.type = DOCA_FLOW_TUN_GRE;
    doca_mask->tun.type = DOCA_FLOW_TUN_GRE;

    greh_spec = (struct rte_gre_hdr *) item->spec;
    greh_mask = (struct rte_gre_hdr *) item->mask;

    if (item->spec) {
        doca_spec->tun.key_present = greh_spec->k;
    }
    if (item->mask) {
        doca_mask->tun.key_present = greh_mask->k;
    }

    doca_spec_mask(&doca_spec->tun.key_present, &doca_mask->tun.key_present,
                   sizeof doca_spec->tun.key_present);
}

static void
doca_translate_geneve_item(const struct rte_flow_item *item,
                           struct doca_flow_match *doca_spec,
                           struct doca_flow_match *doca_mask)
{
    const struct rte_flow_item_geneve *gnv_spec = item->spec;
    const struct rte_flow_item_geneve *gnv_mask = item->mask;

    if (!item->spec || !item->mask) {
        return;
    }

    doca_spec->tun.type = DOCA_FLOW_TUN_GENEVE;
    doca_spec->tun.geneve.vni =
        (OVS_FORCE doca_be32_t) get_unaligned_be32(ALIGNED_CAST(ovs_be32 *, gnv_spec->vni));

    doca_mask->tun.type = DOCA_FLOW_TUN_GENEVE;
    doca_mask->tun.geneve.vni =
        (OVS_FORCE doca_be32_t) get_unaligned_be32(ALIGNED_CAST(ovs_be32 *, gnv_mask->vni));

    doca_spec_mask(&doca_spec->tun.geneve.vni, &doca_mask->tun.geneve.vni,
                   sizeof doca_spec->tun.geneve.vni);
}

static int
doca_init_geneve_opt_parser(struct netdev *netdev,
                            const struct rte_flow_item *item)
{
    struct doca_flow_parser_geneve_opt_cfg opt_cfg[MAX_GENEVE_OPT];
    const struct rte_flow_item_geneve_opt *geneve_opt_spec;
    struct doca_offload_esw_ctx *esw_ctx;
    int ret;

    esw_ctx = doca_offload_esw_ctx_get(netdev);
    if (!esw_ctx) {
        VLOG_ERR("%s: Failed to create geneve_opt parser - esw_ctx is NULL",
                 netdev_get_name(netdev));
        return -1;
    }

    if (!ovsthread_once_start(&esw_ctx->gnv_opt_parser.once)) {
        return 0;
    }
    geneve_opt_spec = item->spec;

    memset(&opt_cfg[0], 0, sizeof(opt_cfg[0]));
    opt_cfg[0].match_on_class_mode =
        DOCA_FLOW_PARSER_GENEVE_OPT_MODE_FIXED;
    opt_cfg[0].option_len = geneve_opt_spec->option_len;
    opt_cfg[0].option_class = (OVS_FORCE doca_be16_t) geneve_opt_spec->option_class;
    opt_cfg[0].option_type = geneve_opt_spec->option_type;
    BUILD_ASSERT_DECL(MEMBER_SIZEOF(struct doca_flow_parser_geneve_opt_cfg,
                                    data_mask[0]) ==
                      MEMBER_SIZEOF(struct rte_flow_item_geneve_opt, data[0]));
    memset(&opt_cfg[0].data_mask[0], UINT32_MAX,
           sizeof(opt_cfg[0].data_mask[0]) * geneve_opt_spec->option_len);

    ret = doca_flow_parser_geneve_opt_create(esw_ctx->esw_port, opt_cfg,
                                             MAX_GENEVE_OPT,
                                             &esw_ctx->gnv_opt_parser.parser);
    if (ret) {
        VLOG_DBG_RL(&rl, "%s: Create geneve_opt parser failed - doca call failure "
                    "rc %d, (%s)",netdev_get_name(netdev), ret,
                    doca_error_get_descr(ret));
        ovsthread_once_reset(&esw_ctx->gnv_opt_parser.once);
        return -1;
    }
    ovsthread_once_done(&esw_ctx->gnv_opt_parser.once);
    return 0;
}

static void
doca_translate_geneve_opt_item(const struct rte_flow_item *item,
                               struct doca_flow_match *doca_spec,
                               struct doca_flow_match *doca_mask)
{
    union doca_flow_geneve_option *doca_opt_spec, *doca_opt_mask;
    const struct rte_flow_item_geneve_opt *geneve_opt_spec;
    const struct rte_flow_item_geneve_opt *geneve_opt_mask;

    geneve_opt_spec = item->spec;
    geneve_opt_mask = item->mask;
    doca_opt_spec = &doca_spec->tun.geneve_options[0];
    doca_opt_mask = &doca_mask->tun.geneve_options[0];

    doca_opt_spec->length = geneve_opt_spec->option_len;
    doca_opt_spec->class_id = (OVS_FORCE doca_be16_t) geneve_opt_spec->option_class;
    doca_opt_spec->type = geneve_opt_spec->option_type;
    doca_opt_mask->length = geneve_opt_mask->option_len;
    doca_opt_mask->class_id = (OVS_FORCE doca_be16_t) geneve_opt_mask->option_class;
    doca_opt_mask->type = geneve_opt_mask->option_type;
    doca_spec_mask(&doca_opt_spec->length, &doca_opt_mask->length,
                   sizeof doca_opt_spec->length);
    doca_spec_mask(&doca_opt_spec->class_id, &doca_opt_mask->class_id,
                   sizeof doca_opt_spec->class_id);
    doca_spec_mask(&doca_opt_spec->type, &doca_opt_mask->type,
                   sizeof doca_opt_spec->type);

    doca_opt_mask->length = doca_opt_spec->length;
    doca_opt_mask->type =  doca_opt_spec->type;

    /* doca_flow represents the geneve option header as an array of a union of
     * 32 bits, the array's first element is the type/class/len and this
     * option's data starts from the next element in the array up to option_len
     */
    doca_opt_spec++;
    doca_opt_mask++;
    BUILD_ASSERT_DECL(sizeof(doca_opt_spec->data) ==
                      sizeof(geneve_opt_spec->data[0]));
    memcpy(&doca_opt_spec->data, &geneve_opt_spec->data[0],
           sizeof(doca_opt_spec->data) * geneve_opt_spec->option_len);
    memcpy(&doca_opt_mask->data, &geneve_opt_mask->data[0],
           sizeof(doca_opt_mask->data) * geneve_opt_spec->option_len);
    doca_spec_mask(&doca_opt_spec->data, &doca_opt_mask->data,
                   sizeof(doca_opt_spec->data) * geneve_opt_spec->option_len);
}

static void
doca_translate_vxlan_item(const struct rte_flow_item *item,
                          struct doca_flow_match *doca_spec,
                          struct doca_flow_match *doca_mask)
{
    const struct vxlan_data *vxlan_spec = item->spec;
    const struct vxlan_data *vxlan_mask = item->mask;
    ovs_be32 spec_vni, mask_vni;

    doca_spec->tun.type = DOCA_FLOW_TUN_VXLAN;
    if (item->spec) {
        spec_vni = get_unaligned_be32(ALIGNED_CAST(ovs_be32 *,
                    vxlan_spec->conf.vni));
        doca_spec->tun.vxlan_tun_id = (OVS_FORCE doca_be32_t) spec_vni;
        if (vxlan_spec->gbp_id) {
            doca_spec->tun.vxlan_type = DOCA_FLOW_TUN_EXT_VXLAN_GBP;
            doca_spec->tun.vxlan_group_policy_id = htons(vxlan_spec->gbp_id);
        }
    }

    doca_mask->tun.type = DOCA_FLOW_TUN_VXLAN;
    if (item->mask) {
        mask_vni = get_unaligned_be32(ALIGNED_CAST(ovs_be32 *,
                    vxlan_mask->conf.vni));
        doca_mask->tun.vxlan_tun_id = (OVS_FORCE doca_be32_t) mask_vni;
        if (vxlan_spec->gbp_id) {
            doca_mask->tun.vxlan_type = DOCA_FLOW_TUN_EXT_VXLAN_GBP;
            doca_mask->tun.vxlan_group_policy_id = htons(vxlan_spec->gbp_id);
            memset(&doca_mask->tun.vxlan_group_policy_id, 0xFF,
                   sizeof doca_mask->tun.vxlan_group_policy_id);
        }
    }

    doca_spec_mask(&doca_spec->tun.vxlan_tun_id, &doca_mask->tun.vxlan_tun_id,
                   sizeof doca_spec->tun.vxlan_tun_id);
}

static int
doca_translate_items(struct netdev *netdev,
                     const struct rte_flow_attr *attr,
                     const struct rte_flow_item *items,
                     struct doca_flow_match *doca_spec,
                     struct doca_flow_match *doca_mask)
{
    struct doca_flow_header_format *doca_hdr_spec, *doca_hdr_mask;

#define doca_copy_mask_field(dfield, field)                   \
    if (spec && mask) {                                       \
        doca_copy_mask_field__(&doca_hdr_spec->dfield,        \
                               &doca_hdr_mask->dfield,        \
                               &spec->field,                  \
                               &mask->field,                  \
                               sizeof doca_hdr_spec->dfield); \
    }

    /* Start by filling out outer header match and
     * switch to inner in case we encounter a tnl proto.
     */
    doca_hdr_spec = &doca_spec->outer;
    doca_hdr_mask = &doca_mask->outer;

    for (; items->type != RTE_FLOW_ITEM_TYPE_END; items++) {
        int item_type = items->type;

        if (item_type == RTE_FLOW_ITEM_TYPE_PORT_ID) {
            const struct rte_flow_item_port_id *spec = items->spec;

            /* Only recirc_id 0 (group_id == 0) may hold flows
             * from different source ports since it's the root table.
             * For every other recirc_id we have a table per port and
             * therefore we can skip matching on port id for those
             * tables.
             */
            if (attr->group > 0) {
                continue;
            }

            doca_spec->parser_meta.port_meta = spec->id;
            doca_mask->parser_meta.port_meta = 0xFFFFFFFF;
        } else if (item_type == RTE_FLOW_ITEM_TYPE_ETH) {
            const struct rte_flow_item_eth *spec = items->spec;
            const struct rte_flow_item_eth *mask = items->mask;

            if (!spec || !mask) {
                continue;
            }

            doca_copy_mask_field(eth.src_mac, src);
            doca_copy_mask_field(eth.dst_mac, dst);
            doca_copy_mask_field(eth.type, type);

            if (mask->has_vlan) {
                if (doca_hdr_spec == &doca_spec->outer) {
                    doca_spec->parser_meta.outer_l2_type =
                        spec->has_vlan
                        ? DOCA_FLOW_L2_META_SINGLE_VLAN
                        : DOCA_FLOW_L2_META_NO_VLAN;
                    doca_mask->parser_meta.outer_l2_type = UINT32_MAX;
                } else {
                    doca_spec->parser_meta.inner_l2_type =
                        spec->has_vlan
                        ? DOCA_FLOW_L2_META_SINGLE_VLAN
                        : DOCA_FLOW_L2_META_NO_VLAN;
                    doca_mask->parser_meta.inner_l2_type = UINT32_MAX;
                }
            }
        } else if (item_type == RTE_FLOW_ITEM_TYPE_VLAN) {
            const struct rte_flow_item_vlan *spec = items->spec;
            const struct rte_flow_item_vlan *mask = items->mask;

            /* HW supports match on one Ethertype, the Ethertype following the
             * last VLAN tag of the packet (see PRM). DOCA API has only that one.
             * Add a match on it as part of the doca eth header.
             */
            doca_copy_mask_field(eth.type, inner_type);
            doca_copy_mask_field(eth_vlan[0].tci, tci);
            doca_hdr_spec->l2_valid_headers = DOCA_FLOW_L2_VALID_HEADER_VLAN_0;
            doca_hdr_mask->l2_valid_headers = DOCA_FLOW_L2_VALID_HEADER_VLAN_0;
        /* L3 */
        } else if (item_type == RTE_FLOW_ITEM_TYPE_IPV4) {
            const struct rte_flow_item_ipv4 *spec = items->spec;
            const struct rte_flow_item_ipv4 *mask = items->mask;

            doca_hdr_spec->l3_type = DOCA_FLOW_L3_TYPE_IP4;
            doca_hdr_mask->l3_type = DOCA_FLOW_L3_TYPE_IP4;

            doca_copy_mask_field(ip4.next_proto, hdr.next_proto_id);
            doca_copy_mask_field(ip4.src_ip, hdr.src_addr);
            doca_copy_mask_field(ip4.dst_ip, hdr.dst_addr);
            doca_copy_mask_field(ip4.ttl, hdr.time_to_live);
            doca_copy_mask_field(ip4.dscp_ecn, hdr.type_of_service);

            /* DOCA API does not provide distinguishment between first/later
             * frags. Reject both.
             */
            if (items->last || (spec && spec->hdr.fragment_offset)) {
                return -1;
            }

            if (doca_hdr_spec == &doca_spec->outer) {
                doca_mask->parser_meta.outer_ip_fragmented = UINT8_MAX;
            } else {
                doca_mask->parser_meta.inner_ip_fragmented = UINT8_MAX;
                doca_mask->parser_meta.outer_ip_fragmented = 0;
            }
        } else if (item_type == RTE_FLOW_ITEM_TYPE_IPV6) {
            const struct rte_flow_item_ipv6 *spec = items->spec;
            const struct rte_flow_item_ipv6 *mask = items->mask;
            uint32_t vtc_spec, vtc_mask;

            doca_hdr_spec->l3_type = DOCA_FLOW_L3_TYPE_IP6;
            doca_hdr_mask->l3_type = DOCA_FLOW_L3_TYPE_IP6;

            doca_copy_mask_field(ip6.src_ip, hdr.src_addr);
            doca_copy_mask_field(ip6.dst_ip, hdr.dst_addr);
            doca_copy_mask_field(ip6.next_proto, hdr.proto);
            vtc_spec = ntohl(spec->hdr.vtc_flow) >> RTE_IPV6_HDR_TC_SHIFT;
            vtc_mask = ntohl(mask->hdr.vtc_flow) >> RTE_IPV6_HDR_TC_SHIFT;
            doca_hdr_spec->ip6.traffic_class = vtc_spec & vtc_mask;
            doca_hdr_mask->ip6.traffic_class = vtc_mask;
            doca_copy_mask_field(ip6.hop_limit, hdr.hop_limits);

            if (doca_hdr_spec == &doca_spec->outer) {
                doca_mask->parser_meta.outer_ip_fragmented = UINT8_MAX;
            } else {
                doca_mask->parser_meta.inner_ip_fragmented = UINT8_MAX;
                doca_mask->parser_meta.outer_ip_fragmented = 0;
            }
        /* L4 */
        } else if (item_type == RTE_FLOW_ITEM_TYPE_UDP) {
            const struct rte_flow_item_udp *spec = items->spec;
            const struct rte_flow_item_udp *mask = items->mask;

            doca_hdr_spec->l4_type_ext = DOCA_FLOW_L4_TYPE_EXT_UDP;
            doca_hdr_mask->l4_type_ext = DOCA_FLOW_L4_TYPE_EXT_UDP;

            doca_copy_mask_field(udp.l4_port.src_port, hdr.src_port);
            doca_copy_mask_field(udp.l4_port.dst_port, hdr.dst_port);
        } else if (item_type ==  RTE_FLOW_ITEM_TYPE_TCP) {
            const struct rte_flow_item_tcp *spec = items->spec;
            const struct rte_flow_item_tcp *mask = items->mask;

            doca_hdr_spec->l4_type_ext = DOCA_FLOW_L4_TYPE_EXT_TCP;
            doca_hdr_mask->l4_type_ext = DOCA_FLOW_L4_TYPE_EXT_TCP;

            doca_copy_mask_field(tcp.l4_port.src_port, hdr.src_port);
            doca_copy_mask_field(tcp.l4_port.dst_port, hdr.dst_port);
            doca_copy_mask_field(tcp.flags, hdr.tcp_flags);
        } else if (item_type == RTE_FLOW_ITEM_TYPE_VXLAN) {
            doca_translate_vxlan_item(items, doca_spec, doca_mask);

            doca_hdr_spec = &doca_spec->inner;
            doca_hdr_mask = &doca_mask->inner;
        } else if (item_type == RTE_FLOW_ITEM_TYPE_GRE) {
            doca_translate_gre_item(items, doca_spec, doca_mask);

            doca_hdr_spec = &doca_spec->inner;
            doca_hdr_mask = &doca_mask->inner;
        } else if (item_type == RTE_FLOW_ITEM_TYPE_GRE_KEY) {
            doca_translate_gre_key_item(items, doca_spec, doca_mask);

            doca_hdr_spec = &doca_spec->inner;
            doca_hdr_mask = &doca_mask->inner;
        } else if (item_type == RTE_FLOW_ITEM_TYPE_GENEVE) {
            doca_translate_geneve_item(items, doca_spec, doca_mask);

            doca_hdr_spec = &doca_spec->inner;
            doca_hdr_mask = &doca_mask->inner;
        } else if (item_type == RTE_FLOW_ITEM_TYPE_GENEVE_OPT) {
            if (doca_init_geneve_opt_parser(netdev, items)) {
                return -1;
            }
            doca_translate_geneve_opt_item(items, doca_spec, doca_mask);
        } else if (item_type == RTE_FLOW_ITEM_TYPE_ICMP) {
            const struct rte_flow_item_icmp *spec = items->spec;
            const struct rte_flow_item_icmp *mask = items->mask;

            doca_hdr_spec->l4_type_ext  = DOCA_FLOW_L4_TYPE_EXT_ICMP;
            doca_hdr_mask->l4_type_ext = DOCA_FLOW_L4_TYPE_EXT_ICMP;

            doca_copy_mask_field(icmp.type, hdr.icmp_type);
            doca_copy_mask_field(icmp.code, hdr.icmp_code);
        } else if (item_type == RTE_FLOW_ITEM_TYPE_ICMP6) {
            const struct rte_flow_item_icmp6 *spec = items->spec;
            const struct rte_flow_item_icmp6 *mask = items->mask;

            doca_hdr_spec->l4_type_ext = DOCA_FLOW_L4_TYPE_EXT_ICMP6;
            doca_hdr_mask->l4_type_ext = DOCA_FLOW_L4_TYPE_EXT_ICMP6;

            doca_copy_mask_field(icmp.code, code);
            doca_copy_mask_field(icmp.type, type);
        } else if (item_type == RTE_FLOW_ITEM_TYPE_TAG) {
            const struct rte_flow_item_tag *spec = items->spec;
            const struct rte_flow_item_tag *mask = items->mask;

            if (spec && mask) {
                doca_spec->meta.u32[spec->index] |= spec->data & mask->data;
                doca_mask->meta.u32[spec->index] |= mask->data;
            }
        } else if (item_type == OVS_RTE_FLOW_ITEM_TYPE(FLOW_INFO)) {
            uint32_t reg_offset = reg_fields[REG_FIELD_FLOW_INFO].offset;
            uint32_t reg_mask = reg_fields[REG_FIELD_FLOW_INFO].mask;
            const struct rte_flow_item_mark *spec = items->spec;

            if (spec) {
                doca_spec->meta.pkt_meta |= (spec->id & reg_mask) << reg_offset;
                doca_mask->meta.pkt_meta |= reg_mask << reg_offset;
            }
        } else if (item_type == OVS_RTE_FLOW_ITEM_TYPE(HASH)) {
            struct reg_field *reg_field = &reg_fields[REG_FIELD_DP_HASH];
            const struct rte_flow_item_mark *hash_spec = items->spec;
            const struct rte_flow_item_mark *hash_mask = items->mask;

            /* In case of non-IPv4, the first flow with the hash function is
             * not offloaded, so there is no point to offload this flow as it
             * will never be hit.
             */
            if (doca_mask->outer.l3_type != DOCA_FLOW_L3_TYPE_IP4 &&
                doca_mask->outer.l3_type != DOCA_FLOW_L3_TYPE_IP6) {
                return -1;
            }
            if (!hash_spec || !hash_mask || hash_mask->id & ~reg_field->mask) {
                /* Can't support larger mask. */
                return -1;
            }

            doca_spec->meta.pkt_meta |=
                (hash_spec->id & reg_field->mask) << reg_field->offset;
            doca_mask->meta.pkt_meta |=
                (hash_mask->id & reg_field->mask) << reg_field->offset;
        } else {
            VLOG_DBG_RL(&rl, "item %d is not supported", item_type);
            return -1;
        }
    }

    return 0;
}

static int
doca_translate_geneve_encap(const struct genevehdr *geneve,
                            struct doca_flow_actions *dacts,
                            struct doca_flow_actions *dacts_masks)
{
    struct doca_flow_encap_action *encap_mask = &dacts_masks->encap_cfg.encap;
    struct doca_flow_encap_action *encap = &dacts->encap_cfg.encap;

    encap->tun.type = DOCA_FLOW_TUN_GENEVE;
    encap_mask->tun.type = encap->tun.type;
    encap->tun.geneve.ver_opt_len = geneve->opt_len;
    encap->tun.geneve.ver_opt_len |= geneve->ver << 6;
    encap_mask->tun.geneve.ver_opt_len = 0xFF;
    encap->tun.geneve.o_c = geneve->critical << 6;
    encap->tun.geneve.o_c |= geneve->oam << 7;
    encap_mask->tun.geneve.o_c = 0xFF;
    encap->tun.geneve.next_proto = (OVS_FORCE doca_be16_t) geneve->proto_type;
    encap_mask->tun.geneve.next_proto = 0xFFFF;
    encap->tun.geneve.vni = (OVS_FORCE doca_be32_t) get_16aligned_be32(&geneve->vni);
    encap_mask->tun.geneve.vni = UINT32_MAX;

    if (geneve->options[0].length) {
        encap->tun.geneve_options[0].class_id =
            (OVS_FORCE doca_be16_t) geneve->options[0].opt_class;
        encap_mask->tun.geneve_options[0].class_id = 0xFFFF;
        encap->tun.geneve_options[0].type = geneve->options[0].type;
        encap_mask->tun.geneve_options[0].type = 0xFF;
        encap->tun.geneve_options[0].length = geneve->options[0].length;
        encap_mask->tun.geneve_options[0].length = 0xFF;

        /* doca_flow represents the geneve option header as an array of a union
         * of 32 bits, the array's first element is the type/class/len and this
         * option's data starts from the next element in the array up to option_len
         */
        BUILD_ASSERT_DECL(sizeof(encap->tun.geneve_options[1].data) ==
                          sizeof(geneve->options[1]));
        memcpy(&encap->tun.geneve_options[1].data, &geneve->options[1],
               sizeof(encap->tun.geneve_options[1].data) * geneve->options[0].length);
        memset(&encap_mask->tun.geneve_options[1].data, 0xFF,
               sizeof(encap->tun.geneve_options[1].data) * geneve->options[0].length);
    }

    return 0;
}

static int
doca_translate_gre_encap(const struct gre_base_hdr *gre,
                         struct doca_flow_actions *dacts,
                         struct doca_flow_actions *dacts_masks)
{
    struct doca_flow_encap_action *encap_mask = &dacts_masks->encap_cfg.encap;
    struct doca_flow_encap_action *encap = &dacts->encap_cfg.encap;

    encap->tun.type = DOCA_FLOW_TUN_GRE;
    encap_mask->tun.type = encap->tun.type;
    encap->tun.protocol = (OVS_FORCE doca_be16_t) gre->protocol;
    encap_mask->tun.protocol = 0xFFFF;
    encap->tun.key_present = !!(gre->flags & htons(GRE_KEY));
    encap_mask->tun.key_present = encap->tun.key_present;

    if (encap->tun.key_present) {
        const void *gre_key = gre + 1;

        memcpy(&encap->tun.gre_key, gre_key, sizeof encap->tun.gre_key);
        encap_mask->tun.gre_key = UINT32_MAX;
    }

    return 0;
}

static int
doca_translate_vxlan_encap(const struct vxlanhdr *vxlan,
                           struct doca_flow_actions *dacts,
                           struct doca_flow_actions *dacts_masks)
{
    struct doca_flow_encap_action *encap_mask = &dacts_masks->encap_cfg.encap;
    struct doca_flow_encap_action *encap = &dacts->encap_cfg.encap;
    ovs_be32 vx_flags = get_16aligned_be32(&vxlan->vx_flags);

    encap->tun.type = DOCA_FLOW_TUN_VXLAN;
    encap_mask->tun.type = encap->tun.type;
    if (vx_flags & htonl(VXLAN_HF_GBP)) {
        /* DOCA doesn't exposes flags in the API, so currently it's not
         * possible to offload non-default flags setting.
         */
        if (vxlan->vx_gbp.flags) {
            return -1;
        }
        encap->tun.vxlan_type = DOCA_FLOW_TUN_EXT_VXLAN_GBP;
        encap_mask->tun.vxlan_type = encap->tun.vxlan_type;
        encap->tun.vxlan_group_policy_id = vxlan->vx_gbp.policy_id;
        memset(&encap_mask->tun.vxlan_group_policy_id, 0xFF,
               sizeof encap_mask->tun.vxlan_group_policy_id);
    }

    memcpy(&encap->tun.vxlan_tun_id, &vxlan->vx_vni, sizeof vxlan->vx_vni);
    /* Explicitly set the vxlan_tun_id mask as UINT32_MAX to indicate for doca-flow
     * that this field is changeable and should not be ignored.
     */
    encap_mask->tun.vxlan_tun_id = UINT32_MAX;

    return 0;
}

static int
doca_translate_raw_encap(const struct rte_flow_action *action,
                         struct doca_flow_actions *dacts,
                         struct doca_flow_actions *dacts_masks)
{
    struct doca_flow_header_format *outer_mask = &dacts_masks->encap_cfg.encap.outer;
    struct doca_flow_header_format *outer = &dacts->encap_cfg.encap.outer;
    const struct raw_encap_data *data = action->conf;
    struct ovs_16aligned_ip6_hdr *ip6;
    struct vlan_header *vlan;
    struct udp_header *udp;
    struct eth_header *eth;
    struct ip_header *ip;
    ovs_be16 proto;
    void *l4;

    /* L2 */
    eth = find_raw_encap_spec(data, RTE_FLOW_ITEM_TYPE_ETH);
    if (!eth) {
        return -1;
    }

    dacts->encap_type = DOCA_FLOW_RESOURCE_TYPE_NON_SHARED;
    dacts_masks->encap_type = dacts->encap_type;
    dacts->encap_cfg.is_l2 = true;
    dacts_masks->encap_cfg.is_l2 = dacts->encap_cfg.is_l2;

    memcpy(&outer->eth.src_mac, &eth->eth_src, sizeof outer->eth.src_mac);
    memset(&outer_mask->eth.src_mac, 0xFF, sizeof outer_mask->eth.src_mac);
    memcpy(&outer->eth.dst_mac, &eth->eth_dst, sizeof outer->eth.dst_mac);
    memset(&outer_mask->eth.dst_mac, 0xFF, sizeof outer_mask->eth.dst_mac);

    outer->eth.type = (OVS_FORCE doca_be16_t) eth->eth_type;
    outer_mask->eth.type = 0xFFFF;

    proto = eth->eth_type;
    if (proto == htons(ETH_TYPE_VLAN_8021Q)) {
        vlan = ALIGNED_CAST(struct vlan_header *, (uint8_t *) (eth + 1));

        outer->eth_vlan[0].tci = (OVS_FORCE doca_be16_t) vlan->vlan_tci;
        outer_mask->eth_vlan[0].tci = 0xFFFF;
        outer->l2_valid_headers = DOCA_FLOW_L2_VALID_HEADER_VLAN_0;
        outer_mask->l2_valid_headers = outer->l2_valid_headers;

        proto = vlan->vlan_next_type;
    }

    /* L3 */
    if (proto == htons(ETH_TYPE_IP)) {
        ip = find_raw_encap_spec(data, RTE_FLOW_ITEM_TYPE_IPV4);
        if (!ip) {
            return -1;
        }

        outer->l3_type = DOCA_FLOW_L3_TYPE_IP4;
        outer_mask->l3_type = outer->l3_type;
        outer->ip4.src_ip = (OVS_FORCE doca_be32_t) get_16aligned_be32(&ip->ip_src);
        outer_mask->ip4.src_ip = 0xFFFFFFFF;
        outer->ip4.dst_ip = (OVS_FORCE doca_be32_t) get_16aligned_be32(&ip->ip_dst);
        outer_mask->ip4.dst_ip = 0xFFFFFFFF;
        outer->ip4.ttl = ip->ip_ttl;
        outer_mask->ip4.ttl = 0xFF;
        outer->ip4.dscp_ecn = ip->ip_tos;
        outer_mask->ip4.dscp_ecn = 0xFF;
        l4 = ip + 1;
    } else if (proto == htons(ETH_TYPE_IPV6)) {
        ip6 = find_raw_encap_spec(data, RTE_FLOW_ITEM_TYPE_IPV6);
        if (!ip6) {
            return -1;
        }

        outer->l3_type = DOCA_FLOW_L3_TYPE_IP6;
        outer_mask->l3_type = outer->l3_type;
        memcpy(&outer->ip6.src_ip, &ip6->ip6_src, sizeof ip6->ip6_src);
        memset(&outer_mask->ip6.src_ip, 0xFF, sizeof outer_mask->ip6.src_ip);
        memcpy(&outer->ip6.dst_ip, &ip6->ip6_dst, sizeof ip6->ip6_dst);
        memset(&outer_mask->ip6.dst_ip, 0xFF, sizeof outer_mask->ip6.dst_ip);
        outer->ip6.hop_limit = ip6->ip6_hlim;
        outer_mask->ip6.hop_limit = 0xFF;
        outer->ip6.traffic_class = ntohl(get_16aligned_be32(&ip6->ip6_flow)) >> 20;
        outer_mask->ip6.traffic_class = 0xFF;
        l4 = ip6 + 1;
    } else {
        return -1;
    }

    /* Tunnel */
    if (data->tnl_type == OVS_VPORT_TYPE_GRE
        || data->tnl_type == OVS_VPORT_TYPE_IP6GRE) {
        return doca_translate_gre_encap(l4, dacts, dacts_masks);
    }

    udp = l4;
    /* L4 */
    outer->l4_type_ext = DOCA_FLOW_L4_TYPE_EXT_UDP;
    outer_mask->l4_type_ext = outer_mask->l4_type_ext;
    outer->udp.l4_port.src_port = (OVS_FORCE doca_be16_t) udp->udp_src;
    outer_mask->udp.l4_port.src_port = 0xFFFF;
    outer->udp.l4_port.dst_port = (OVS_FORCE doca_be16_t) udp->udp_dst;
    outer_mask->udp.l4_port.dst_port = 0xFFFF;

    if (data->tnl_type == OVS_VPORT_TYPE_GENEVE) {
        return doca_translate_geneve_encap((void *) (udp + 1), dacts, dacts_masks);
    }

    if (data->tnl_type == OVS_VPORT_TYPE_VXLAN) {
        return doca_translate_vxlan_encap((void *) (udp + 1), dacts, dacts_masks);
    }

    return -1;
}

static int
doca_hash_pipe_init(struct netdev *netdev,
                    unsigned int queue_id,
                    struct doca_hash_pipe_ctx *hash_pipe_ctx,
                    enum hash_pipe_type type)
{
    struct doca_flow_pipe *next_pipe = doca_pipe_group_get_pipe(hash_pipe_ctx->post_hash_group_ctx);
    struct doca_flow_match hash_matches[NUM_HASH_PIPE_TYPE] = {
        [HASH_TYPE_IPV4_UDP] = {
            .outer.l3_type = DOCA_FLOW_L3_TYPE_IP4,
            .outer.ip4.src_ip = BE32_MAX,
            .outer.ip4.dst_ip = BE32_MAX,
            .outer.l4_type_ext = DOCA_FLOW_L4_TYPE_EXT_UDP,
            .outer.udp.l4_port.src_port = BE16_MAX,
            .outer.udp.l4_port.dst_port = BE16_MAX,
        },
        [HASH_TYPE_IPV4_TCP] = {
            .outer.l3_type = DOCA_FLOW_L3_TYPE_IP4,
            .outer.ip4.src_ip = BE32_MAX,
            .outer.ip4.dst_ip = BE32_MAX,
            .outer.l4_type_ext = DOCA_FLOW_L4_TYPE_EXT_TCP,
            .outer.tcp.l4_port.src_port = BE16_MAX,
            .outer.tcp.l4_port.dst_port = BE16_MAX,
        },
        [HASH_TYPE_IPV4_L3] = {
            .outer.l3_type = DOCA_FLOW_L3_TYPE_IP4,
            .outer.ip4.src_ip = BE32_MAX,
            .outer.ip4.dst_ip = BE32_MAX,
        },
        [HASH_TYPE_IPV6_UDP_SUF] = {
            .outer.l3_type = DOCA_FLOW_L3_TYPE_IP6,
            .outer.ip6.dst_ip[0] = BE32_MAX,
            .outer.ip6.dst_ip[1] = BE32_MAX,
            .outer.ip6.dst_ip[2] = BE32_MAX,
            .outer.ip6.dst_ip[3] = BE32_MAX,
            .outer.l4_type_ext = DOCA_FLOW_L4_TYPE_EXT_UDP,
            .outer.udp.l4_port.dst_port = BE16_MAX,
        },
        [HASH_TYPE_IPV6_TCP_SUF] = {
            .outer.l3_type = DOCA_FLOW_L3_TYPE_IP6,
            .outer.ip6.dst_ip[0] = BE32_MAX,
            .outer.ip6.dst_ip[1] = BE32_MAX,
            .outer.ip6.dst_ip[2] = BE32_MAX,
            .outer.ip6.dst_ip[3] = BE32_MAX,
            .outer.l4_type_ext = DOCA_FLOW_L4_TYPE_EXT_TCP,
            .outer.tcp.l4_port.dst_port = BE16_MAX,
        },
        [HASH_TYPE_IPV6_L3_SUF] = {
            .outer.l3_type = DOCA_FLOW_L3_TYPE_IP6,
            .outer.ip6.dst_ip[0] = BE32_MAX,
            .outer.ip6.dst_ip[1] = BE32_MAX,
            .outer.ip6.dst_ip[2] = BE32_MAX,
            .outer.ip6.dst_ip[3] = BE32_MAX,
        },
        [HASH_TYPE_IPV6_UDP_PRE] = {
            .outer.l3_type = DOCA_FLOW_L3_TYPE_IP6,
            .outer.ip6.src_ip[0] = BE32_MAX,
            .outer.ip6.src_ip[1] = BE32_MAX,
            .outer.ip6.src_ip[2] = BE32_MAX,
            .outer.ip6.src_ip[3] = BE32_MAX,
            .outer.l4_type_ext = DOCA_FLOW_L4_TYPE_EXT_UDP,
            .outer.udp.l4_port.src_port = BE16_MAX,
        },
        [HASH_TYPE_IPV6_TCP_PRE] = {
            .outer.l3_type = DOCA_FLOW_L3_TYPE_IP6,
            .outer.ip6.src_ip[0] = BE32_MAX,
            .outer.ip6.src_ip[1] = BE32_MAX,
            .outer.ip6.src_ip[2] = BE32_MAX,
            .outer.ip6.src_ip[3] = BE32_MAX,
            .outer.l4_type_ext = DOCA_FLOW_L4_TYPE_EXT_TCP,
            .outer.tcp.l4_port.src_port = BE16_MAX,
        },
        [HASH_TYPE_IPV6_L3_PRE] = {
            .outer.l3_type = DOCA_FLOW_L3_TYPE_IP6,
            .outer.ip6.src_ip[0] = BE32_MAX,
            .outer.ip6.src_ip[1] = BE32_MAX,
            .outer.ip6.src_ip[2] = BE32_MAX,
            .outer.ip6.src_ip[3] = BE32_MAX,
        },
    };
    struct doca_flow_actions actions, *actions_arr[1];
    struct doca_flow_action_descs *descs_arr[1];
    struct doca_offload_esw_ctx *esw_ctx;
    struct doca_flow_pipe_entry **pentry;
    struct doca_flow_action_descs descs;
    struct doca_flow_action_desc desc;
    struct doca_flow_fwd fwd, miss;
    struct doca_flow_pipe_cfg *cfg;
    struct doca_flow_pipe **ppipe;
    char pipe_name[50];
    int ret;

    esw_ctx = doca_offload_esw_ctx_get(netdev);

    ppipe = &hash_pipe_ctx->hashes[type].pipe;
    pentry = &hash_pipe_ctx->hashes[type].entry;

    snprintf(pipe_name, sizeof pipe_name, "OVS_HASH_PIPE_type_%u", type);

    memset(&fwd, 0, sizeof(fwd));
    memset(&miss, 0, sizeof(miss));
    memset(&descs, 0, sizeof(descs));
    memset(&actions, 0, sizeof(actions));
    memset(&desc, 0, sizeof desc);

    ret = doca_flow_pipe_cfg_create(&cfg, esw_ctx->esw_port);
    if (ret) {
        VLOG_ERR("%s: Could not create doca_flow_pipe_cfg", netdev_get_name(netdev));
        return ret;
    }
    if (doca_flow_pipe_cfg_set_name(cfg, pipe_name) ||
        doca_flow_pipe_cfg_set_type(cfg, DOCA_FLOW_PIPE_HASH) ||
        doca_flow_pipe_cfg_set_enable_strict_matching(cfg, true) ||
        doca_flow_pipe_cfg_set_nr_entries(cfg, 1)) {
        VLOG_ERR("%s: Could not set doca_flow_pipe_cfg", netdev_get_name(netdev));
        ret = -1;
        goto out;
    }
    descs_arr[0] = &descs;
    descs.desc_array = &desc;
    descs.nb_action_desc = 1;
    actions_arr[0] = &actions;

    desc.type = DOCA_FLOW_ACTION_COPY;
    desc.field_op.src.field_string = "parser_meta.hash.result";
    desc.field_op.src.bit_offset = 0;
    desc.field_op.dst.field_string = "meta.data";
    desc.field_op.dst.bit_offset = doca_get_reg_bit_offset(REG_FIELD_DP_HASH);
    desc.field_op.width = doca_get_reg_width(REG_FIELD_DP_HASH);

    if (type >= HASH_TYPE_IPV6_UDP_PRE && type <= HASH_TYPE_IPV6_L3_PRE) {
        uint32_t offset = HASH_TYPE_IPV6_UDP_SUF - HASH_TYPE_IPV6_UDP_PRE;

        next_pipe = hash_pipe_ctx->hashes[type + offset].pipe;

        desc.field_op.dst.bit_offset =
            doca_get_reg_bit_offset(REG_FIELD_SCRATCH);
        desc.field_op.width = doca_get_reg_width(REG_FIELD_SCRATCH);
    }

    /* For SUF pipes, it's the hash result of the PRE pipes. For others it's
     * the seed.
     */
    doca_set_reg_mask(&hash_matches[type].meta, REG_FIELD_SCRATCH);

    fwd.type = DOCA_FLOW_FWD_PIPE;
    fwd.next_pipe = next_pipe;
    miss.type = DOCA_FLOW_FWD_DROP;

    if (doca_flow_pipe_cfg_set_match(cfg, NULL, &hash_matches[type]) ||
        doca_flow_pipe_cfg_set_actions(cfg, actions_arr, NULL, descs_arr, 1)) {
        VLOG_ERR("%s: Could not set doca_flow_pipe_cfg", netdev_get_name(netdev));
        ret = -1;
        goto out;
    }
    ret = doca_flow_pipe_create(cfg, &fwd, &miss, ppipe);
    if (ret) {
        VLOG_ERR("Failed to create hash pipe: %d (%s)", ret,
                 doca_error_get_descr(ret));
        goto out;
    }

    ret = doca_flow_pipe_hash_add_entry(queue_id, *ppipe, 0, NULL, NULL, NULL,
                                        DOCA_FLOW_NO_WAIT, esw_ctx, pentry);
    if (ret) {
        VLOG_ERR("Failed to create hash pipe entry. Error: %d (%s)", ret,
                 doca_error_get_descr(ret));
        goto out;
    }
    esw_ctx->offload_queues[queue_id].n_waiting_entries++;
    dpdk_offload_counter_inc(netdev);

    ret = doca_offload_complete_queue(netdev, queue_id, true);
    if (ret) {
        VLOG_ERR("Failed to process hash pipe entry. Error: %d (%s)", ret,
                 doca_error_get_descr(ret));
    }
out:
    doca_flow_pipe_cfg_destroy(cfg);
    return ret;
}

static struct doca_hash_pipe_ctx *
doca_hash_pipe_ctx_init(struct netdev *netdev)
{
    unsigned int queue_id = netdev_offload_thread_id();
    struct doca_hash_pipe_ctx *hash_pipe_ctx;
    struct doca_flow_pipe_entry **pentry;
    struct doca_offload_esw_ctx *esw_ctx;
    struct doca_flow_pipe_cfg *cfg;
    struct doca_flow_match spec;
    struct doca_flow_match mask;
    struct doca_flow_fwd miss;
    enum hash_nw_type nw_type;
    struct doca_flow_fwd fwd;
    char pipe_name[50];
    doca_error_t err;
    int type;

    esw_ctx = doca_offload_esw_ctx_get(netdev);
    if (esw_ctx == NULL) {
        return NULL;
    }

    hash_pipe_ctx = xzalloc(sizeof *hash_pipe_ctx);
    hash_pipe_ctx->netdev = netdev;

    hash_pipe_ctx->post_hash_group_ctx =
        doca_pipe_group_ctx_ref(netdev, POSTHASH_TABLE_ID);
    if (hash_pipe_ctx->post_hash_group_ctx == NULL) {
        VLOG_ERR("%s: Failed to create post-hash group", netdev_get_name(netdev));
        goto pipe_ctx_uninit;
    }

    for (type = 0; type < NUM_HASH_PIPE_TYPE; type++) {
        int ret;

        ret = doca_hash_pipe_init(netdev, queue_id, hash_pipe_ctx, type);
        if (ret) {
            VLOG_ERR("%s: Failed to create hash pipe ctx",
                     netdev_get_name(netdev));
            goto pipe_ctx_uninit;
        }
    }

    /* Classifier pipe. */
    for (nw_type = 0; nw_type < NUM_HASH_NW; nw_type++) {
        struct doca_flow_pipe **ppipe;

        memset(&mask, 0, sizeof mask);
        memset(&spec, 0, sizeof spec);
        memset(&fwd, 0, sizeof fwd);
        fwd.type = DOCA_FLOW_FWD_PIPE;
        memset(&miss, 0, sizeof miss);
        miss.type = DOCA_FLOW_FWD_PIPE;

        if (nw_type == HASH_NW_IP4) {
            mask.outer.l3_type = DOCA_FLOW_L3_TYPE_IP4;
            mask.outer.ip4.next_proto = 0xFF;
            miss.next_pipe = hash_pipe_ctx->hashes[HASH_TYPE_IPV4_L3].pipe;
            snprintf(pipe_name, sizeof pipe_name, "OVS_HASH_IPv4_CLASSIFIER_PIPE");
        } else {
            mask.outer.l3_type = DOCA_FLOW_L3_TYPE_IP6;
            mask.outer.ip6.next_proto = 0xFF;
            miss.next_pipe = hash_pipe_ctx->hashes[HASH_TYPE_IPV6_L3_PRE].pipe;
            snprintf(pipe_name, sizeof pipe_name, "OVS_HASH_IPv6_CLASSIFIER_PIPE");
        }

        ppipe = &hash_pipe_ctx->classifier[nw_type].pipe;

        if (doca_flow_pipe_cfg_create(&cfg, esw_ctx->esw_port)) {
            VLOG_ERR("%s: Could not create doca_flow_pipe_cfg", netdev_get_name(netdev));
            goto pipe_ctx_uninit;
        }
        if (doca_flow_pipe_cfg_set_name(cfg, pipe_name) ||
            doca_flow_pipe_cfg_set_type(cfg, DOCA_FLOW_PIPE_BASIC) ||
            doca_flow_pipe_cfg_set_enable_strict_matching(cfg, true) ||
            doca_flow_pipe_cfg_set_nr_entries(cfg, 2) ||
            doca_flow_pipe_cfg_set_match(cfg, &mask, &mask)) {
            VLOG_ERR("%s: Could not set doca_flow_pipe_cfg", netdev_get_name(netdev));
            goto err;
        }

        err = doca_flow_pipe_create(cfg, &fwd, &miss, ppipe);
        if (err) {
            VLOG_ERR("%s: Failed to create hash classifier pipe: %d (%s)",
                     netdev_get_name(netdev), err, doca_error_get_descr(err));
            goto err;
        }

        /* TCP/UDP entries. */
        pentry = &hash_pipe_ctx->classifier[nw_type].tcpudp[HASH_TP_UDP];
        if (nw_type == HASH_NW_IP4) {
            spec.outer.l3_type = DOCA_FLOW_L3_TYPE_IP4;
            spec.outer.ip4.next_proto = IPPROTO_UDP;
            fwd.next_pipe = hash_pipe_ctx->hashes[HASH_TYPE_IPV4_UDP].pipe;
        } else {
            spec.outer.l3_type = DOCA_FLOW_L3_TYPE_IP6;
            spec.outer.ip6.next_proto = IPPROTO_UDP;
            fwd.next_pipe = hash_pipe_ctx->hashes[HASH_TYPE_IPV6_UDP_PRE].pipe;
        }
        err = doca_offload_add_entry(queue_id, *ppipe, &spec, NULL, NULL,
                                     &fwd, DOCA_FLOW_NO_WAIT, esw_ctx, pentry);
        if (err) {
            VLOG_ERR("%s: Failed to create UDP classifier entry: %d (%s)",
                     netdev_get_name(netdev), err, doca_error_get_descr(err));
            goto err;
        }
        dpdk_offload_counter_inc(netdev);

        pentry = &hash_pipe_ctx->classifier[nw_type].tcpudp[HASH_TP_TCP];
        if (nw_type == HASH_NW_IP4) {
            spec.outer.l3_type = DOCA_FLOW_L3_TYPE_IP4;
            spec.outer.ip4.next_proto = IPPROTO_TCP;
            fwd.next_pipe = hash_pipe_ctx->hashes[HASH_TYPE_IPV4_TCP].pipe;
        } else {
            spec.outer.l3_type = DOCA_FLOW_L3_TYPE_IP6;
            spec.outer.ip6.next_proto = IPPROTO_TCP;
            fwd.next_pipe = hash_pipe_ctx->hashes[HASH_TYPE_IPV6_TCP_PRE].pipe;
        }
        err = doca_offload_add_entry(queue_id, *ppipe, &spec, NULL, NULL,
                                     &fwd, DOCA_FLOW_NO_WAIT, esw_ctx, pentry);
        if (err) {
            VLOG_ERR("%s: Failed to create TCP classifier entry: %d (%s)",
                     netdev_get_name(netdev), err, doca_error_get_descr(err));
            goto err;
        }
        dpdk_offload_counter_inc(netdev);

        err = doca_offload_complete_queue(netdev, queue_id, true);
        if (err) {
            VLOG_ERR("%s: Failed to poll classifier completion: queue %u. "
                     "Error: %d (%s)", netdev_get_name(netdev), queue_id, err,
                     doca_error_get_descr(err));
            goto err;
        }
        doca_flow_pipe_cfg_destroy(cfg);
    }

    return hash_pipe_ctx;
err:
    doca_flow_pipe_cfg_destroy(cfg);
pipe_ctx_uninit:
    doca_hash_pipe_ctx_uninit(esw_ctx, hash_pipe_ctx);
    return NULL;
}

static struct doca_flow_pipe *
get_group_root(struct doca_hash_pipe_ctx *hash_pipe_ctx,
               struct doca_pipe_group_ctx *next_group_ctx,
               struct doca_flow_match *spec,
               struct doca_flow_actions *dacts,
               bool has_dp_hash)
{
    struct doca_flow_header_format *hf;

    if (!has_dp_hash) {
        return doca_pipe_group_get_pipe(next_group_ctx);
    }

    if (dacts->encap_type == DOCA_FLOW_RESOURCE_TYPE_NON_SHARED) {
        if (dacts->encap_cfg.encap.outer.l3_type == DOCA_FLOW_L3_TYPE_IP4) {
            if (dacts->encap_cfg.encap.tun.type == DOCA_FLOW_TUN_VXLAN ||
                dacts->encap_cfg.encap.tun.type == DOCA_FLOW_TUN_GENEVE) {
                return hash_pipe_ctx->hashes[HASH_TYPE_IPV4_UDP].pipe;
            }
            return hash_pipe_ctx->hashes[HASH_TYPE_IPV4_L3].pipe;
        }

        if (dacts->encap_cfg.encap.outer.l3_type == DOCA_FLOW_L3_TYPE_IP6) {
            if (dacts->encap_cfg.encap.tun.type == DOCA_FLOW_TUN_VXLAN ||
                dacts->encap_cfg.encap.tun.type == DOCA_FLOW_TUN_GENEVE) {
                return hash_pipe_ctx->hashes[HASH_TYPE_IPV6_UDP_PRE].pipe;
            }
            return hash_pipe_ctx->hashes[HASH_TYPE_IPV6_L3_PRE].pipe;
        }
        return NULL;
    }

    hf = dacts->decap_cfg.is_l2 ? &spec->inner : &spec->outer;

    if (hf->l3_type == DOCA_FLOW_L3_TYPE_IP4) {
        if (hf->l4_type_ext == DOCA_FLOW_L4_TYPE_EXT_TCP) {
            return hash_pipe_ctx->hashes[HASH_TYPE_IPV4_TCP].pipe;
        } else if (hf->l4_type_ext == DOCA_FLOW_L4_TYPE_EXT_UDP) {
            return hash_pipe_ctx->hashes[HASH_TYPE_IPV4_UDP].pipe;
        }
        return hash_pipe_ctx->classifier[HASH_NW_IP4].pipe;
    }

    if (hf->l3_type == DOCA_FLOW_L3_TYPE_IP6) {
        if (hf->l4_type_ext == DOCA_FLOW_L4_TYPE_EXT_TCP) {
            return hash_pipe_ctx->hashes[HASH_TYPE_IPV6_TCP_PRE].pipe;
        } else if (hf->l4_type_ext == DOCA_FLOW_L4_TYPE_EXT_UDP) {
            return hash_pipe_ctx->hashes[HASH_TYPE_IPV6_UDP_PRE].pipe;
        }
        return hash_pipe_ctx->classifier[HASH_NW_IP6].pipe;
    }

    return NULL;
}

static int
doca_translate_actions(struct netdev *netdev,
                       struct doca_flow_match *spec,
                       const struct rte_flow_action *actions,
                       struct doca_flow_actions *dacts,
                       struct doca_flow_actions *dacts_masks,
                       struct doca_flow_fwd *fwd,
                       struct doca_flow_monitor *monitor,
                       struct doca_flow_handle_resources *flow_res,
                       struct doca_act_vars *dact_vars)
{
    struct doca_flow_header_format *outer_masks = &dacts_masks->outer;
    struct doca_flow_header_format *outer = &dacts->outer;
    struct doca_offload_esw_ctx *esw_ctx;
    bool vlan_act_push = false;

    esw_ctx = doca_offload_esw_ctx_get(netdev);
    for (; actions->type != RTE_FLOW_ACTION_TYPE_END; actions++) {
        int act_type = actions->type;

        if (act_type == RTE_FLOW_ACTION_TYPE_DROP) {
            fwd->type = DOCA_FLOW_FWD_DROP;
        } else if (act_type == RTE_FLOW_ACTION_TYPE_SET_MAC_SRC) {
            const struct action_set_data *asd = actions->conf;

            memcpy(&outer->eth.src_mac, asd->value, asd->size);
            memcpy(&outer_masks->eth.src_mac, asd->mask, asd->size);
        } else if (act_type == RTE_FLOW_ACTION_TYPE_SET_MAC_DST) {
            const struct action_set_data *asd = actions->conf;

            memcpy(&outer->eth.dst_mac, asd->value, asd->size);
            memcpy(&outer_masks->eth.dst_mac, asd->mask, asd->size);
        } else if (act_type == RTE_FLOW_ACTION_TYPE_OF_SET_VLAN_VID) {
            const struct rte_flow_action_of_set_vlan_vid *rte_vlan_vid;

            rte_vlan_vid = actions->conf;
            /* If preceeded by vlan push action, this is a new
             * vlan tag. Otherwise, perfrom vlan modification.
             */
            if (vlan_act_push) {
                dacts->push.type = DOCA_FLOW_PUSH_ACTION_VLAN;
                dacts_masks->push.type = DOCA_FLOW_PUSH_ACTION_VLAN;
                dacts->push.vlan.tci = (OVS_FORCE doca_be16_t) rte_vlan_vid->vlan_vid;
                memset(&dacts_masks->push.vlan, 0xFF, sizeof dacts_masks->push.vlan);
                dacts->has_push = true;
                dacts_masks->has_push = true;
            } else {
                outer->eth_vlan[0].tci = (OVS_FORCE doca_be16_t) rte_vlan_vid->vlan_vid;
                outer->l2_valid_headers = DOCA_FLOW_L2_VALID_HEADER_VLAN_0;
                memset(&outer_masks->eth_vlan[0].tci, 0xFF,
                       sizeof outer_masks->eth_vlan[0].tci);
                outer_masks->l2_valid_headers =
                    DOCA_FLOW_L2_VALID_HEADER_VLAN_0;
            }
        } else if (act_type == RTE_FLOW_ACTION_TYPE_SET_IPV4_SRC) {
            const struct action_set_data *asd = actions->conf;

            outer->l3_type = DOCA_FLOW_L3_TYPE_IP4;
            memcpy(&outer->ip4.src_ip, asd->value, asd->size);
            memcpy(&outer_masks->ip4.src_ip, asd->mask, asd->size);
        } else if (act_type == RTE_FLOW_ACTION_TYPE_SET_IPV4_DST) {
            const struct action_set_data *asd = actions->conf;

            outer->l3_type = DOCA_FLOW_L3_TYPE_IP4;
            memcpy(&outer->ip4.dst_ip, asd->value, asd->size);
            memcpy(&outer_masks->ip4.dst_ip, asd->mask, asd->size);
        } else if (act_type == RTE_FLOW_ACTION_TYPE_SET_IPV4_TTL) {
            const struct action_set_data *asd = actions->conf;

            outer->l3_type = DOCA_FLOW_L3_TYPE_IP4;
            memcpy(&outer->ip4.ttl, asd->value, asd->size);
            memcpy(&outer_masks->ip4.ttl, asd->mask, asd->size);
        } else if (act_type == RTE_FLOW_ACTION_TYPE_SET_IPV4_DSCP) {
            const struct action_set_data *asd = actions->conf;

            outer->l3_type = DOCA_FLOW_L3_TYPE_IP4;
            memcpy(&outer->ip4.dscp_ecn, asd->value, asd->size);
            memcpy(&outer_masks->ip4.dscp_ecn, asd->mask, asd->size);
        } else if (act_type == RTE_FLOW_ACTION_TYPE_SET_IPV6_HOP) {
            const struct action_set_data *asd = actions->conf;

            outer->l3_type = DOCA_FLOW_L3_TYPE_IP6;
            memcpy(&outer->ip6.hop_limit, asd->value, asd->size);
            memcpy(&outer_masks->ip6.hop_limit, asd->mask, asd->size);
        } else if (act_type == RTE_FLOW_ACTION_TYPE_SET_IPV6_SRC) {
            const struct action_set_data *asd = actions->conf;

            outer->l3_type = DOCA_FLOW_L3_TYPE_IP6;
            memcpy(&outer->ip6.src_ip, asd->value, asd->size);
            memcpy(&outer_masks->ip6.src_ip, asd->mask, asd->size);
        } else if (act_type == RTE_FLOW_ACTION_TYPE_SET_IPV6_DST) {
            const struct action_set_data *asd = actions->conf;

            outer->l3_type = DOCA_FLOW_L3_TYPE_IP6;
            memcpy(&outer->ip6.dst_ip, asd->value, asd->size);
            memcpy(&outer_masks->ip6.dst_ip, asd->mask, asd->size);
        } else if (act_type == RTE_FLOW_ACTION_TYPE_SET_IPV6_DSCP) {
            const struct action_set_data *asd = actions->conf;

            outer->l3_type = DOCA_FLOW_L3_TYPE_IP6;
            memcpy(&outer->ip6.traffic_class, asd->value, asd->size);
            memcpy(&outer_masks->ip6.traffic_class, asd->mask, asd->size);
        } else if (act_type == OVS_RTE_FLOW_ACTION_TYPE(SET_UDP_SRC)) {
            const struct action_set_data *asd = actions->conf;

            outer->l4_type_ext = DOCA_FLOW_L4_TYPE_EXT_UDP;
            memcpy(&outer->udp.l4_port.src_port, asd->value, asd->size);
            memcpy(&outer_masks->udp.l4_port.src_port, asd->mask, asd->size);
        } else if (act_type == OVS_RTE_FLOW_ACTION_TYPE(SET_UDP_DST)) {
            const struct action_set_data *asd = actions->conf;

            outer->l4_type_ext = DOCA_FLOW_L4_TYPE_EXT_UDP;
            memcpy(&outer->udp.l4_port.dst_port, asd->value, asd->size);
            memcpy(&outer_masks->udp.l4_port.dst_port, asd->mask, asd->size);
        } else if (act_type == OVS_RTE_FLOW_ACTION_TYPE(SET_TCP_SRC)) {
            const struct action_set_data *asd = actions->conf;

            outer->l4_type_ext = DOCA_FLOW_L4_TYPE_EXT_TCP;
            memcpy(&outer->tcp.l4_port.src_port, asd->value, asd->size);
            memcpy(&outer_masks->tcp.l4_port.src_port, asd->mask, asd->size);
        } else if (act_type == OVS_RTE_FLOW_ACTION_TYPE(SET_TCP_DST)) {
            const struct action_set_data *asd = actions->conf;

            outer->l4_type_ext = DOCA_FLOW_L4_TYPE_EXT_TCP;
            memcpy(&outer->tcp.l4_port.dst_port, asd->value, asd->size);
            memcpy(&outer_masks->tcp.l4_port.dst_port, asd->mask, asd->size);
        } else if (act_type == RTE_FLOW_ACTION_TYPE_REPRESENTED_PORT) {
            const struct rte_flow_action_ethdev *ethdev = actions->conf;

            /* DOCA supports hairpin only for WIRE ports. In case of another
             * port, the rule is wrongly offloaded and packets will be dropped.
             * As a WA for this, deny non-wire hairpin offload. */
            if (netdev_dpdk_get_port_id(netdev) == ethdev->port_id &&
                netdev_dpdk_get_port_dir(netdev) != NETDEV_DPDK_PORT_DIR_RX) {
                return -1;
            }

            fwd->type = DOCA_FLOW_FWD_PORT;
            fwd->port_id = ethdev->port_id;
        } else if (act_type == RTE_FLOW_ACTION_TYPE_NVGRE_DECAP ||
                   act_type == RTE_FLOW_ACTION_TYPE_VXLAN_DECAP ||
                   act_type == RTE_FLOW_ACTION_TYPE_RAW_DECAP) {
            dacts->decap_type = DOCA_FLOW_RESOURCE_TYPE_NON_SHARED;
            dacts->decap_cfg.is_l2 = true;
            dacts_masks->decap_type = DOCA_FLOW_RESOURCE_TYPE_NON_SHARED;
            dacts_masks->decap_cfg.is_l2 = true;
        } else if (act_type == RTE_FLOW_ACTION_TYPE_COUNT) {
            monitor->counter_type = DOCA_FLOW_RESOURCE_TYPE_NON_SHARED;
        } else if (act_type == RTE_FLOW_ACTION_TYPE_JUMP) {
            const struct rte_flow_action_jump *jump = actions->conf;
            struct doca_pipe_group_ctx *next_group_ctx;

            if (doca_pipe_group_is_ct_zone_group_id(jump->group)
                && doca_ct_zone_init(netdev, esw_ctx, jump->group)) {
                return -1;
            }

            fwd->type = DOCA_FLOW_FWD_PIPE;

            if (jump->group == MISS_TABLE_ID) {
                fwd->next_pipe = doca_offload_get_misspath_pipe(esw_ctx);
            } else {
                next_group_ctx = doca_pipe_group_ctx_ref(netdev, jump->group);
                if (!next_group_ctx) {
                    return -1;
                }

                fwd->next_pipe = get_group_root(esw_ctx->hash_pipe_ctx,
                                                next_group_ctx, spec, dacts,
                                                !!dact_vars->hash_data);
                if (!fwd->next_pipe) {
                    doca_pipe_group_ctx_unref(next_group_ctx);
                    return -1;
                }
                flow_res->next_group_ctx = next_group_ctx;
            }
        } else if (act_type == RTE_FLOW_ACTION_TYPE_RAW_ENCAP) {
            if (doca_translate_raw_encap(actions, dacts, dacts_masks)) {
                return -1;
            }
        } else if (act_type == OVS_RTE_FLOW_ACTION_TYPE(FLOW_INFO)) {
            const struct rte_flow_action_mark *mark = actions->conf;

            if (!dact_vars->mtr_flow_id) {
                doca_set_reg_val_mask(&dacts->meta, &dacts_masks->meta,
                                      REG_FIELD_FLOW_INFO, mark->id);
                dact_vars->flow_id = mark->id;
            }
        } else if (act_type == RTE_FLOW_ACTION_TYPE_SET_TAG) {
            const struct rte_flow_action_set_tag *set_tag = actions->conf;
            uint8_t index = set_tag->index;

            dacts->meta.u32[index] |= set_tag->data;
            dacts_masks->meta.u32[index] |= set_tag->mask;
        } else if (act_type == RTE_FLOW_ACTION_TYPE_OF_POP_VLAN) {
            /* Current support is for a single VLAN tag */
            if (dacts->pop) {
                return -1;
            }
            dacts->pop = true;
            dacts_masks->pop = true;
        } else if (act_type == RTE_FLOW_ACTION_TYPE_OF_PUSH_VLAN) {
            if (vlan_act_push) {
                return -1;
            }
            vlan_act_push = true;
        } else if (act_type == RTE_FLOW_ACTION_TYPE_OF_SET_VLAN_PCP) {
            continue;
        } else if (act_type == OVS_RTE_FLOW_ACTION_TYPE(HASH)) {
            const struct hash_data *hash_data = actions->conf;

            if (!dact_vars->mtr_flow_id) {
                doca_set_reg_val_mask(&dacts->meta, &dacts_masks->meta,
                                      REG_FIELD_FLOW_INFO, hash_data->flow_id);
                doca_set_reg_val_mask(&dacts->meta, &dacts_masks->meta,
                                      REG_FIELD_SCRATCH, hash_data->seed);
            }
            dact_vars->hash_data = hash_data;
        } else if (act_type == RTE_FLOW_ACTION_TYPE_METER) {
            const struct meter_data *mtr_data = actions->conf;
            uint32_t doca_mtr_id, flow_id;

            flow_id = shared_mtr_flow_id_alloc(esw_ctx, NULL);
            if (flow_id == -1) {
                return -1;
            }

            /* id is determine by both the upper layer id, and the esw_id. */
            doca_mtr_id = ovs_doca_meter_id(mtr_data->conf.mtr_id,
                                            esw_ctx->esw_id);

            if (monitor->meter_type == DOCA_FLOW_RESOURCE_TYPE_NONE) {
                /* first meter in a multi-meter action */
                monitor->meter_type = DOCA_FLOW_RESOURCE_TYPE_SHARED;
                monitor->shared_meter.shared_meter_id = doca_mtr_id;
                dact_vars->mtr_flow_id = flow_id;
                dact_vars->flow_id = mtr_data->flow_id;
                doca_set_reg_val_mask(&dacts->meta, &dacts_masks->meta,
                                      REG_FIELD_FLOW_INFO, flow_id);
            } else {
                struct meter_info *mtr_info = xzalloc(sizeof *mtr_info);

                mtr_info->id = doca_mtr_id;
                mtr_info->flow_id = flow_id;
                ovs_list_push_back(&dact_vars->next_meters,
                                   &mtr_info->list_node);
            }
        } else if (act_type == RTE_FLOW_ACTION_TYPE_SAMPLE ||
                   /* When sFlow sampling ratio is 1 the ovs action is
                    * translated into OVS_ACTION_ATTR_USERSPACE.
                    */
                   act_type == OVS_RTE_FLOW_ACTION_TYPE(USERSPACE)) {
            const struct rte_flow_action_sample *sample = actions->conf;

            dact_vars->sample_ratio = sample->ratio;
        } else if (act_type == OVS_RTE_FLOW_ACTION_TYPE(SFLOW)) {
            /* The sflow_id action parameter is not used with DOCA, but the
             * action still needs to be processed.
             */
            continue;
        } else if (act_type == RTE_FLOW_ACTION_TYPE_VOID) {
            continue;
        } else {
            return -1;
        }
    }

    outer_masks->l2_valid_headers = outer->l2_valid_headers;
    outer_masks->l3_type = outer->l3_type;
    outer_masks->l4_type_ext = outer->l4_type_ext;

    return 0;
}

static uint16_t
doca_offload_ct_queue_id(unsigned int tid)
{
    if (ovs_doca_flow_ct_enabled) {
        /* Threads that interact with DOCA-CT are either
         * polling threads (updates) or auxiliary threads (queries).
         * Queries are made with the queue ID used for entry creation.
         * Regular offload threads do not interact and do not have a queue
         * configured.
         *
         * Example with 2 offload threads, 3 polling threads:
         *
         *       [6|7|8]  DOCA-CT thread IDs
         *  x x x | | |
         * [0|1|2|3|4|5]  netdev-offload thread IDs
         *  | | | `-`-`-o polling
         *  | `-`-------o offload
         *  `-----------o main
         *
         * In this example,
         *    ovs_doca_n_offload_queues() is 6 (2 offload, 3 PMDs, 1 main)
         *    netdev_offload_thread_nb() is 2
         */
        ovs_assert(tid > netdev_offload_thread_nb());
        return ovs_doca_n_offload_queues() +
               (tid - netdev_offload_thread_nb() - 1);
    }
    return tid;
}

void
doca_offload_entry_process(struct doca_flow_pipe_entry *entry,
                           uint16_t qid,
                           enum doca_flow_entry_status status,
                           enum doca_flow_entry_op op,
                           void *aux)
{
    static const char *status_desc[] = {
        [DOCA_FLOW_ENTRY_STATUS_IN_PROCESS] = "in-process",
        [DOCA_FLOW_ENTRY_STATUS_SUCCESS] = "success",
        [DOCA_FLOW_ENTRY_STATUS_ERROR] = "failure",
    };
    static const char *op_desc[] = {
        [DOCA_FLOW_ENTRY_OP_ADD] = "add",
        [DOCA_FLOW_ENTRY_OP_DEL] = "del",
        [DOCA_FLOW_ENTRY_OP_UPD] = "mod",
        [DOCA_FLOW_ENTRY_OP_AGED] = "age",
    };
    bool error = status == DOCA_FLOW_ENTRY_STATUS_ERROR;
    unsigned int tid = netdev_offload_thread_id();
    struct doca_offload_esw_ctx *esw = aux;

    VLOG_RL(&rl, error ? VLL_ERR : VLL_DBG,
            "%s: [tid:%u][qid:%" PRIu16 "] %s aux=%p entry %p %s",
            __func__, tid, qid, op_desc[op], aux, entry, status_desc[status]);

    if (esw && status != DOCA_FLOW_ENTRY_STATUS_IN_PROCESS) {
        esw->offload_queues[qid].n_waiting_entries--;
    }
}

/* Complete the queue 'qid' on the netdev's eswitch until 'min_room'
 * entries are available. If 'min_room' is 0, complete any available.
 */
static doca_error_t
doca_offload_complete_queue_n_protected(struct doca_offload_esw_ctx *esw,
                                        unsigned int qid, uint32_t min_room)
    OVS_REQ_RDLOCK(esw->port_rwlock)
{
    struct doca_offload_queue *queue;
    long long int timeout_ms;
    unsigned int n_waiting;
    bool pipe_resizing;
    doca_error_t err;
    bool is_doca_ct;
    uint32_t room;
    int retries;

    is_doca_ct = qid >= ovs_doca_n_offload_queues();
    queue = &esw->offload_queues[qid];
    n_waiting = queue->n_waiting_entries;
    pipe_resizing = !is_doca_ct && ovs_refcount_read(&esw->pipe_resizing) > 1;

    if (!pipe_resizing && n_waiting == 0) {
        COVERAGE_INC(doca_queue_empty);
        return DOCA_SUCCESS;
    }

    if (qid == AUX_QUEUE) {
        /* 10 seconds timeout when synchronizing the AUX_QUEUE. */
        timeout_ms = time_msec() + 10 * 1000;
    } else {
        timeout_ms = 0;
    }

    retries = 100;
    do {
        unsigned int n_processed;

        if (is_doca_ct) {
            err = doca_flow_ct_entries_process(esw->esw_port, qid,
                                               min_room, min_room,
                                               &room);
            /* If this assertion fails, DOCA has increased the size of their
             * internal CT queue, which breaks synchronous polling.
             * OVS_DOCA_CT_QUEUE_DEPTH must be updated. */
            if (min_room == OVS_DOCA_CT_QUEUE_DEPTH) {
                ovs_assert(room <= OVS_DOCA_CT_QUEUE_DEPTH);
            }
        } else {
            /* Use 'max_processed_entries' == 0 to always attempt processing
             * the full length of the queue. */
            err = doca_flow_entries_process(esw->esw_port, qid,
                                            ENTRY_PROCESS_TIMEOUT_US, 0);
        }
        if (err) {
            VLOG_WARN_RL(&rl, "eswitch %" PRIu32 ": Failed to process entries "
                         "in queue %u: %s", esw->esw_id, qid,
                         doca_error_get_descr(err));
            return err;
        }

        n_processed = n_waiting - queue->n_waiting_entries;
        if (min_room && n_processed == 0) {
            COVERAGE_INC(doca_queue_none_processed);
        }
        n_waiting = queue->n_waiting_entries;

        if (!is_doca_ct) {
            if (pipe_resizing && qid == FIRST_QUEUE) {
                ovs_doca_mgmt_queue_lock();
                err = doca_offload_complete_queue_n_protected(esw, AUX_QUEUE, min_room);
                ovs_doca_mgmt_queue_unlock();
            }

            doca_pipe_group_ctx_resize_end_list(&esw->resized_pipe_lists[qid]);

            room = OVS_DOCA_QUEUE_DEPTH - n_waiting;
            pipe_resizing = ovs_refcount_read(&esw->pipe_resizing) > 1;
            if (pipe_resizing && n_processed == 0) {
                if (retries-- <= 0) {
                    COVERAGE_INC(doca_resize_block);
                    break;
                }
            }

            if (timeout_ms && time_msec() > timeout_ms) {
                ovs_abort(0, "Timeout reached trying to complete queue %u: %u remaining entries",
                          qid, n_waiting);
            }
        }
    } while (err == DOCA_SUCCESS && min_room &&
             ((room < min_room) || pipe_resizing));

    return err;
}

static doca_error_t
doca_offload_complete_queue_n(struct doca_offload_esw_ctx *esw,
                              unsigned int qid, uint32_t min_room)
    OVS_EXCLUDED(mgmt_queue_lock)
{
    doca_error_t err;

    ovs_rwlock_rdlock(&esw->port_rwlock);
    if (qid == AUX_QUEUE) {
        ovs_doca_mgmt_queue_lock();
        err = doca_offload_complete_queue_n_protected(esw, qid, min_room);
        ovs_doca_mgmt_queue_unlock();
    } else {
        err = doca_offload_complete_queue_n_protected(esw, qid, min_room);
    }
    ovs_rwlock_unlock(&esw->port_rwlock);

    return err;
}

static doca_error_t
doca_offload_complete_queue_esw(struct doca_offload_esw_ctx *esw, unsigned int qid, bool sync)
    OVS_EXCLUDED(mgmt_queue_lock)
{
    uint32_t min_room = 0;

    if (sync) {
        bool is_doca_ct = (qid >= ovs_doca_n_offload_queues());

        min_room = is_doca_ct ? OVS_DOCA_CT_QUEUE_DEPTH
                              : OVS_DOCA_QUEUE_DEPTH;
    }

    return doca_offload_complete_queue_n(esw, qid, min_room);
}

doca_error_t
doca_offload_complete_queue(struct netdev *netdev, unsigned int qid, bool sync)
    OVS_EXCLUDED(mgmt_queue_lock)
{
    struct doca_offload_esw_ctx *esw;

    if (netdev == NULL || !netdev_dpdk_is_ethdev(netdev)) {
        return DOCA_SUCCESS;
    }

    esw = doca_offload_esw_ctx_get(netdev);
    /* vports won't take an eswitch reference. */
    if (esw == NULL) {
        return DOCA_SUCCESS;
    }

    return doca_offload_complete_queue_esw(esw, qid, sync);
}

static void
doca_offload_resolve_pipe_resize(struct doca_offload_esw_ctx *esw)
    OVS_REQ_WRLOCK(esw->port_rwlock)
{
    /* Use a 10 seconds timeout. */
    long long int timeout_time_ms = time_msec() + (10 * 1000);
    bool pipe_resizing;
    unsigned int qid;

    pipe_resizing = ovs_refcount_read(&esw->pipe_resizing) > 1;
    while (pipe_resizing) {
        if (time_msec() >= timeout_time_ms) {
            VLOG_ERR("Pipe resize resolution took more than 10 seconds!!");
            break;
        }
        for (qid = 0; qid < ovs_doca_n_offload_queues(); qid++) {
            doca_offload_complete_queue_n_protected(esw, qid, 0);
        }
        pipe_resizing = ovs_refcount_read(&esw->pipe_resizing) > 1;
    }
}

static void
dpdk_offload_doca_upkeep(struct netdev *netdev, bool quiescing)
{
    unsigned int qid = netdev_offload_thread_id();
    struct doca_offload_esw_ctx *esw;
    struct destroy_pipe_elem *elem;
    bool is_doca_ct;
    bool is_pmd;

    if (!netdev_dpdk_is_esw_mgr(netdev)) {
        return;
    }

    /* Safe to get an eswitch context after verify netdev is an 'ethdev'. */
    esw = doca_offload_esw_ctx_get(netdev);

    if (!esw) {
        return;
    }

    doca_offload_complete_queue(netdev, qid, quiescing);

    is_doca_ct = qid >= ovs_doca_n_offload_queues();
    is_pmd = qid > netdev_offload_thread_nb() && !is_doca_ct;

    if (ovs_doca_flow_ct_enabled && is_pmd) {
        /* If DOCA-CT has been initialized, poll the CT queues as well from
         * threads in the polling threads range. */
        doca_offload_complete_queue(netdev, doca_offload_ct_queue_id(qid), quiescing);
    }

    if (!ovs_list_is_empty(&esw->destroy_pipe_lists[qid])) {
        /* If this assert triggers, it means multiple hw-offload thread have been
         * destroying pipes. In such case, the implementation of destroy_pipe_lists
         * must be reworked. */
        ovs_assert(qid == FIRST_QUEUE);
        ovs_rwlock_wrlock(&esw->port_rwlock);
        /* While under lock, ensure no pipe resize is currently occurring.
         * This is a necessary condition before destroying any pipe. */
        doca_offload_resolve_pipe_resize(esw);
        LIST_FOR_EACH_POP (elem, node, &esw->destroy_pipe_lists[qid]) {
            doca_flow_pipe_destroy(elem->pipe);
            free(elem);
        }
        ovs_rwlock_unlock(&esw->port_rwlock);
    }
}

static int
create_doca_basic_flow_entry(struct netdev *netdev,
                             unsigned int queue_id,
                             struct doca_flow_pipe *pipe,
                             struct doca_flow_match *spec,
                             struct doca_flow_actions *actions,
                             struct doca_flow_monitor *monitor,
                             struct doca_flow_fwd *fwd,
                             struct doca_flow_pipe_entry **pentry,
                             struct rte_flow_error *error,
                             bool force_sync)
    OVS_EXCLUDED(mgmt_queue_lock)
{
    enum doca_flow_flags_type doca_flags;
    struct doca_offload_esw_ctx *esw;
    bool wait = !force_sync;
    doca_error_t err;

    esw = doca_offload_esw_ctx_get(netdev);
    doca_flags = wait ? DOCA_FLOW_WAIT_FOR_BATCH : DOCA_FLOW_NO_WAIT;

    err = doca_offload_add_entry(queue_id, pipe, spec, actions, monitor, fwd,
                                 doca_flags, esw, pentry);
    if (err) {
        VLOG_WARN_RL(&rl, "%s: Failed to create basic pipe entry. Error: %d (%s)",
                     netdev_get_name(netdev), err, doca_error_get_descr(err));
        error->type = RTE_FLOW_ERROR_TYPE_HANDLE;
        error->message = doca_error_get_descr(err);
        return -1;
    }

    if (force_sync) {
        err = doca_offload_complete_queue(netdev, queue_id, true);
        if (err != DOCA_SUCCESS) {
            error->type = RTE_FLOW_ERROR_TYPE_HANDLE;
            error->message = doca_error_get_descr(err);
            return -1;
        }
    }

    dpdk_offload_counter_inc(netdev);

    return 0;
}

static struct doca_flow_pipe_entry *
create_doca_pipe_group_entry(struct netdev *netdev,
                             unsigned int queue_id,
                             struct doca_pipe_group_ctx *self_group_ctx,
                             uint32_t prio,
                             struct doca_flow_match *spec,
                             struct doca_flow_match *mask,
                             struct doca_flow_actions *actions,
                             struct doca_flow_actions *actions_masks,
                             struct doca_flow_action_descs *dacts_descs,
                             struct doca_flow_monitor *monitor,
                             struct doca_flow_fwd *fwd,
                             struct doca_pipe_group_mask_ctx **mask_ctx,
                             struct rte_flow_error *error)
    OVS_EXCLUDED(mgmt_queue_lock)
{
    struct doca_flow_pipe_entry *entry;
    doca_error_t err;

    err = doca_pipe_group_add_entry(netdev, queue_id, prio, self_group_ctx,
                                    spec, mask, actions, actions_masks,
                                    dacts_descs, monitor, fwd, mask_ctx, &entry);
    if (err) {
        if ((doca_error_t) err != DOCA_ERROR_TOO_BIG) {
            VLOG_WARN_RL(&rl, "%s: Failed to create group entry. Error: %d (%s)",
                         netdev_get_name(netdev), err, doca_error_get_descr(err));
        }
        error->type = (enum rte_flow_error_type) err;
        error->message = doca_error_get_descr(err);
        return NULL;
    }

    dpdk_offload_counter_inc(netdev);
    ovs_assert(entry);

    return entry;
}

static struct doca_flow_pipe *
doca_get_ct_pipe(struct doca_offload_esw_ctx *ctx,
                 struct doca_flow_match *spec)
{
    enum ct_nw_type nw_type;
    enum ct_tp_type tp_type;

    if (ctx == NULL) {
        return NULL;
    }

    if (ctx->ct.pipe) {
        return ctx->ct.pipe;
    }

    nw_type = l3_to_nw_type(spec->outer.l3_type);
    if (nw_type >= NUM_CT_NW) {
        VLOG_DBG_RL(&rl, "Unsupported CT network type.");
        return NULL;
    }

    if (nw_type == CT_NW_IP6) {
        return ctx->ct_ip6_prefix.pipe;
    }

    tp_type = l4_to_tp_type(spec->outer.l4_type_ext);
    if (tp_type >= NUM_CT_TP) {
        VLOG_DBG_RL(&rl, "Unsupported CT protocol type.");
        return NULL;
    }

    return ctx->ct_pipes[nw_type][tp_type].pipe;
}

static int
doca_split_prefix_ctx_init(void *ctx_, void *arg_)
{
    unsigned int tid = netdev_offload_thread_id();
    struct doca_flow_actions dacts, dacts_mask;
    struct doca_split_prefix_ctx *ctx = ctx_;
    struct doca_split_prefix_arg *arg = arg_;
    struct doca_flow_handle *hndl;
    struct rte_flow_error error;
    struct doca_flow_fwd fwd;
    unsigned int queue_id;
    uint32_t id;

    memset(&dacts, 0, sizeof dacts);
    memset(&dacts_mask, 0, sizeof dacts_mask);
    memset(&fwd, 0, sizeof fwd);

    if (!id_fpool_new_id(split_prefix_id_pool, tid, &id)) {
        return -1;
    }
    ctx->id = id;

    doca_set_reg_val_mask(&dacts.meta, &dacts_mask.meta,
                          REG_FIELD_SCRATCH, id);

    fwd.type = DOCA_FLOW_FWD_PIPE;
    fwd.next_pipe = arg->suffix_pipe;
    queue_id = netdev_offload_thread_id();

    if (!arg->prefix_is_group) {
        if (create_doca_basic_flow_entry(arg->netdev, queue_id,
                                         arg->prefix_pipe, arg->spec, &dacts,
                                         NULL, &fwd, &ctx->doh.dfh.flow, &error, false)) {
            VLOG_WARN_RL(&rl, "%s: Failed to create basic split prefix entry:"
                         " Error %d (%s)", netdev_get_name(arg->netdev),
                         error.type, error.message);
            return -1;
        }
    } else {
        if (arg->set_flow_info_id) {
            /* Set flow info id to pkt meta.
             * If packet misses on suffix table the flow info
             * is already stored for proper recovery.
             */
            doca_set_reg_val_mask(&dacts.meta, &dacts_mask.meta,
                                  REG_FIELD_FLOW_INFO, arg->set_flow_info_id);
        }
        hndl = &ctx->doh.dfh;
        hndl->flow = create_doca_pipe_group_entry(arg->netdev, queue_id,
                                                  arg->prefix_group_ctx, arg->prio,
                                                  arg->spec, arg->mask,
                                                  &dacts, &dacts_mask,
                                                  NULL, NULL, &fwd,
                                                  &hndl->flow_res.mask_ctx,
                                                  &error);
        if (!hndl->flow) {
            return -1;
        }
    }

    ctx->netdev = arg->netdev;

    return 0;
}

static void
doca_split_prefix_ctx_uninit(void *ctx_)
{
    unsigned int tid = netdev_offload_thread_id();
    struct doca_split_prefix_ctx *ctx = ctx_;
    struct doca_offload_esw_ctx *esw;
    unsigned int queue_id = tid;
    struct rte_flow_error error;

    esw = doca_offload_esw_ctx_get(ctx->netdev);
    id_fpool_free_id(split_prefix_id_pool, tid, ctx->id);
    destroy_dpdk_offload_handle(esw, ctx->netdev, &ctx->doh, queue_id, &error);
}

static struct ds *
dump_split_prefix_id(struct ds *s, void *key_, void *ctx_, void *arg OVS_UNUSED)
{
    struct doca_split_prefix_key *key = key_;
    struct doca_split_prefix_ctx *ctx = ctx_;
    struct doca_flow_handle *hndl;

    if (ctx) {
        hndl = &ctx->doh.dfh;
        ds_put_format(s, "prefix_flow=%p, ", hndl->flow);
    }
    ds_put_format(s, "prefix_pipe=%p, ", key->prefix_pipe);

    return s;
}

static void
split_prefix_id_init(void)
{
    static struct ovsthread_once init_once = OVSTHREAD_ONCE_INITIALIZER;

    if (ovsthread_once_start(&init_once)) {
        split_prefix_id_pool = id_fpool_create(MAX_OFFLOAD_THREAD_NB, MIN_SPLIT_PREFIX_ID,
                                               NUM_SPLIT_PREFIX_ID);
        split_prefix_rfm = refmap_create("prefix-id", sizeof(struct doca_split_prefix_key),
                                         sizeof(struct doca_split_prefix_ctx),
                                         doca_split_prefix_ctx_init, doca_split_prefix_ctx_uninit,
                                         dump_split_prefix_id, false);
        ovsthread_once_done(&init_once);
    }
}

static struct doca_split_prefix_ctx *
split_prefix_ctx_ref(struct doca_split_prefix_key *key,
                     struct doca_split_prefix_arg *args,
                     uint32_t *prefix_id)
{
    struct doca_split_prefix_ctx *ctx;

    split_prefix_id_init();
    ctx = refmap_ref(split_prefix_rfm, key, args);
    if (ctx) {
        *prefix_id = ctx->id;
    }
    return ctx;
}

static void
split_prefix_ctx_unref(struct doca_split_prefix_ctx *ctx)
{
    refmap_unref(split_prefix_rfm, ctx);
}

static void
reset_split_header_proto(enum split_field_layer layer, uint32_t proto,
                         struct doca_flow_header_format *smask,
                         void *sspec_proto, void *smask_proto, size_t proto_size)
{
    switch (layer) {
    case L2_HEADERS:
        if (!is_all_zeros(&smask->eth, sizeof smask->eth)) {
            return;
        }
        break;
    case L3_HEADERS:
        if (proto == DOCA_FLOW_L3_TYPE_IP4) {
            if (!is_all_zeros(&smask->ip4, sizeof smask->ip4)) {
                return;
            }
        } else if (!is_all_zeros(&smask->ip6, sizeof smask->ip6)) {
            return;
        }
        break;
    case L4_HEADERS:
        if (proto == DOCA_FLOW_L4_TYPE_EXT_TCP) {
            if (!is_all_zeros(&smask->tcp, sizeof smask->tcp)) {
                return;
            }
        } else if (proto == DOCA_FLOW_L4_TYPE_EXT_UDP) {
            if (!is_all_zeros(&smask->udp, sizeof smask->udp)) {
                return;
            }
        } else if (proto == DOCA_FLOW_L4_TYPE_EXT_ICMP ||
                   proto == DOCA_FLOW_L4_TYPE_EXT_ICMP6) {
            if (!is_all_zeros(&smask->icmp, sizeof smask->icmp)) {
                return;
            }
        }
        break;
    }
    memset(sspec_proto, 0, proto_size);
    memset(smask_proto, 0, proto_size);
}

static void
split_match_header_fields(struct doca_flow_header_format *spec_,
                          struct doca_flow_header_format *mask_,
                          struct doca_flow_header_format *pspec_,
                          struct doca_flow_header_format *pmask_,
                          struct doca_flow_header_format *sspec_,
                          struct doca_flow_header_format *smask_,
                          enum split_field_type avoid_type,
                          size_t *match_bytes)
{
    enum split_field_layer layer;
    struct split_field *field;
    char *pspec, *pmask;
    char *sspec, *smask;
    char *spec, *mask;
    int i, num_fields;
    uint32_t proto;

    spec = (void *) spec_;
    mask = (void *) mask_;
    pspec = (void *) pspec_;
    pmask = (void *) pmask_;
    sspec = (void *) sspec_;
    smask = (void *) smask_;

    for (layer = L2_HEADERS;
         layer <= L4_HEADERS && *match_bytes < MAX_FIELD_BYTES; layer++) {
        switch (layer) {
        case L2_HEADERS:
            proto = spec_->l2_valid_headers;
            num_fields = NUM_L2_FIELDS;
            break;
        case L3_HEADERS:
            proto = spec_->l3_type;
            num_fields = NUM_L3_FIELDS;
            break;
        case L4_HEADERS:
            proto = spec_->l4_type_ext;
            num_fields = NUM_L4_FIELDS;
            break;
        }

        for (i = 0; i < num_fields && *match_bytes < MAX_FIELD_BYTES; i++) {
            field = &split_fields[layer][i];
            if (field->type != avoid_type &&
                field->proto_type == proto &&
                !is_all_zeros(mask + field->offset, field->size) &&
                *match_bytes + field->size <= MAX_FIELD_BYTES) {
                memcpy(pspec + field->offset, spec + field->offset, field->size);
                memcpy(pmask + field->offset, mask + field->offset, field->size);
                memset(sspec + field->offset, 0, field->size);
                memset(smask + field->offset, 0, field->size);
                *match_bytes += field->size;
                memcpy(pspec + field->proto_offset, &field->proto_type, field->proto_size);
                memcpy(pmask + field->proto_offset, &field->proto_type, field->proto_size);
            }
        }
        field = &split_fields[layer][0];
        reset_split_header_proto(layer, proto, smask_, sspec + field->proto_offset,
                                 smask + field->proto_offset, field->proto_size);
    }
}

static int
split_tunnel_header_fields(struct doca_flow_match *spec,
                           struct doca_flow_match *mask,
                           struct doca_flow_match *pspec,
                           struct doca_flow_match *pmask,
                           struct doca_flow_match *sspec,
                           struct doca_flow_match *smask,
                           size_t *match_bytes)
{
    pspec->tun.type = spec->tun.type;
    pmask->tun.type = spec->tun.type;
    memset(&sspec->tun, 0, sizeof smask->tun);
    memset(&smask->tun, 0, sizeof smask->tun);

    /* Assume the tunnel match will fit into the prefix header */
    if (spec->tun.type == DOCA_FLOW_TUN_GENEVE) {
        pspec->tun.geneve.vni = spec->tun.geneve.vni;
        pmask->tun.geneve.vni = mask->tun.geneve.vni;
        *match_bytes += sizeof spec->tun.geneve.vni;

        memcpy(pspec->tun.geneve_options, spec->tun.geneve_options,
               sizeof pspec->tun.geneve_options);
        memcpy(pmask->tun.geneve_options, mask->tun.geneve_options,
               sizeof pmask->tun.geneve_options);
        *match_bytes += spec->tun.geneve_options->length * 4;
    } else if (spec->tun.type == DOCA_FLOW_TUN_VXLAN) {
        pspec->tun.vxlan_tun_id = spec->tun.vxlan_tun_id;
        pmask->tun.vxlan_tun_id = mask->tun.vxlan_tun_id;
        *match_bytes += sizeof spec->tun.vxlan_tun_id;
        if (spec->tun.vxlan_type == DOCA_FLOW_TUN_EXT_VXLAN_GBP) {
            pspec->tun.vxlan_type = spec->tun.vxlan_type;
            pmask->tun.vxlan_type = mask->tun.vxlan_type;
            *match_bytes += sizeof spec->tun.vxlan_type;
            pspec->tun.vxlan_group_policy_id = spec->tun.vxlan_group_policy_id;
            pmask->tun.vxlan_group_policy_id = mask->tun.vxlan_group_policy_id;
            *match_bytes += sizeof spec->tun.vxlan_group_policy_id;
        }
    } else if (spec->tun.type == DOCA_FLOW_TUN_GRE) {
        if (mask->tun.key_present && spec->tun.key_present) {
            pspec->tun.key_present = true;
            pmask->tun.key_present = true;
            pspec->tun.gre_key = spec->tun.gre_key;
            pmask->tun.gre_key = mask->tun.gre_key;
            *match_bytes += sizeof spec->tun.gre_key;
        }
    } else {
        return -1;
    }

    return 0;
}

static int
split_doca_flow_match(struct netdev *netdev,
                      struct doca_flow_match *spec,
                      struct doca_flow_match *mask,
                      struct doca_flow_match *pspec,
                      struct doca_flow_match *pmask,
                      struct doca_flow_match *sspec,
                      struct doca_flow_match *smask,
                      uint8_t depth)
{
    enum split_field_type avoid_type = FIELD_TYPE_INVALID;
    bool has_tunnel, is_n2h;
    size_t match_bytes = 0;
    int i;

    memset(pspec, 0, sizeof *pspec);
    memset(pmask, 0, sizeof *pmask);

    has_tunnel = spec->tun.type != DOCA_FLOW_TUN_NONE;
    is_n2h = netdev_dpdk_is_esw_mgr(netdev);

    memcpy(sspec, spec, sizeof *spec);
    memcpy(smask, mask, sizeof *mask);

    /* Collect meta matches to prefix first and assume we
     * always have enough room in prefix for all meta.
     * Port meta match always exists and takes 4 bytes and
     * we set it for both prefix and suffix to utilize the
     * doca flow direction optimization and reduce number
     * of flows.
     */
    pspec->parser_meta.port_meta = spec->parser_meta.port_meta;
    pmask->parser_meta.port_meta = mask->parser_meta.port_meta;
    match_bytes += sizeof spec->parser_meta.port_meta;

    if (mask->meta.pkt_meta) {
        pspec->meta.pkt_meta = spec->meta.pkt_meta;
        pmask->meta.pkt_meta = mask->meta.pkt_meta;
        sspec->meta.pkt_meta = 0;
        smask->meta.pkt_meta = 0;
        match_bytes += sizeof spec->meta.pkt_meta;
    }

    for (i = 0; i < ARRAY_SIZE(spec->meta.u32); i++) {
        if (mask->meta.u32[i]) {
            pspec->meta.u32[i] = spec->meta.u32[i];
            pmask->meta.u32[i] = mask->meta.u32[i];
            sspec->meta.u32[i] = 0;
            smask->meta.u32[i] = 0;
            match_bytes += sizeof spec->meta.u32[i];
        }
    }

    if (has_tunnel) {
        if (split_tunnel_header_fields(spec, mask, pspec, pmask,
                                       sspec, smask, &match_bytes)) {
            return -1;
        }

        /* Tunnel outer header is common among flows so collect
         * them into the prefix flow.
         */
        split_match_header_fields(&spec->outer, &mask->outer, &pspec->outer,
                                  &pmask->outer, &sspec->outer, &smask->outer,
                                  FIELD_TYPE_INVALID, &match_bytes);
        if (depth && is_all_zeros(&mask->outer, sizeof mask->outer)) {
            pspec->inner.l3_type = spec->inner.l3_type;
            pmask->inner.l3_type = mask->inner.l3_type;
            pspec->inner.l4_type_ext = spec->inner.l4_type_ext;
            pmask->inner.l4_type_ext = mask->inner.l4_type_ext;
            avoid_type = depth == 0 ? FIELD_TYPE_SRC : FIELD_TYPE_INVALID;
            split_match_header_fields(&spec->inner, &mask->inner, &pspec->inner,
                                      &pmask->inner, &sspec->inner, &smask->inner,
                                      avoid_type, &match_bytes);
        }

        /* Set tun type for decap action */
        sspec->tun.type = spec->tun.type;
        smask->tun.type = spec->tun.type;
    } else {
        /* In case there's no tunnel and it is the the first split,
         * collect fields to prefix based on the flow's direction.
         * is_n2h will indicate the flow is network 2 host flow and
         * therefore, dst addresses are more likely to be common among
         * these flows and we have a better chance to create a prefix
         * flow that will be used by multiple n2h split flows.
         * The other option is that the flow is host 2 network
         * and in such case, the src addresses are taken into the prefix flow
         * as they are the common addresses among these flows.
         */
        if (depth == 0) {
            avoid_type = is_n2h ? FIELD_TYPE_SRC : FIELD_TYPE_DST;
        }
        split_match_header_fields(&spec->outer, &mask->outer, &pspec->outer,
                                  &pmask->outer, &sspec->outer, &smask->outer,
                                  avoid_type, &match_bytes);
    }

    return 0;
}

static struct doca_flow_pipe_entry *
create_split_doca_flow_entry(struct netdev *netdev,
                             unsigned int queue_id,
                             uint32_t prio,
                             struct doca_pipe_group_ctx *group_ctx,
                             struct doca_flow_match *spec,
                             struct doca_flow_match *mask,
                             struct doca_flow_actions *actions,
                             struct doca_flow_actions *actions_masks,
                             struct doca_flow_action_descs *dacts_descs,
                             struct doca_flow_monitor *monitor,
                             struct doca_flow_fwd *fwd,
                             struct doca_flow_handle_resources *flow_res,
                             uint8_t depth,
                             struct rte_flow_error *error)
{
    struct doca_pipe_group_ctx *suffix_group;
    struct doca_split_prefix_key prefix_key;
    struct doca_split_prefix_ctx *pctx;
    struct doca_flow_match suffix_spec;
    struct doca_flow_match suffix_mask;
    struct doca_flow_pipe_entry *entry;
    struct doca_split_prefix_arg args;
    uint32_t prefix_id;

    if (unlikely(depth == MAX_SPLIT_DEPTH)) {
        VLOG_DBG_RL(&rl, "Exceeded max split depth %d", MAX_SPLIT_DEPTH);
        return NULL;
    }

    memset(&suffix_spec, 0, sizeof(suffix_spec));
    memset(&suffix_mask, 0, sizeof(suffix_mask));
    /* Split the original match to prefix and suffix keys */
    if (split_doca_flow_match(netdev, spec, mask,
                              &prefix_key.spec, &prefix_key.mask,
                              &suffix_spec, &suffix_mask, depth)) {
        error->type = RTE_FLOW_ERROR_TYPE_UNSPECIFIED;
        error->message = "Could not split flow";
        return NULL;
    }

    suffix_group = doca_pipe_group_ctx_ref(netdev, SPLIT_DEPTH_TABLE_ID(depth));
    if (!suffix_group) {
        error->type = RTE_FLOW_ERROR_TYPE_UNSPECIFIED;
        error->message = "Could not get suffix group for split flow";
        return NULL;
    }

    args.netdev = netdev;
    args.spec = &prefix_key.spec;
    args.mask = &prefix_key.mask;
    args.prefix_is_group = true;
    args.prefix_group_ctx = group_ctx;
    args.suffix_pipe = doca_pipe_group_get_pipe(suffix_group);
    args.set_flow_info_id = doca_pipe_group_get_miss_flow_id(group_ctx);
    args.prio = prio;

    prefix_key.prefix_group_ctx = args.prefix_group_ctx;

    /* Get prefix flow id. */
    pctx = split_prefix_ctx_ref(&prefix_key, &args, &prefix_id);
    if (!pctx) {
        error->type = RTE_FLOW_ERROR_TYPE_UNSPECIFIED;
        error->message = "Could not get split flow prefix id";
        goto err_prefix;
    }

    if (depth > 0) {
        /* If this is not the first split attempt, suffix_spec and mask are
         * copied from the previous split suffix which included a match on the
         * previous prefix_id, mask that match before setting a match on
         * the current prefix_id
         */
        memset(&suffix_spec.meta, 0, sizeof(suffix_spec.meta));
    }

    /* Add prefix ID match to suffix rule. */
    doca_set_reg_val_mask(&suffix_spec.meta, &suffix_mask.meta,
                          REG_FIELD_SCRATCH, prefix_id);

    /* Insert suffix rule. */
    entry = create_doca_pipe_group_entry(netdev, queue_id, suffix_group, prio,
                                         &suffix_spec, &suffix_mask, actions,
                                         actions_masks, dacts_descs, monitor,
                                         fwd, &flow_res->mask_ctx, error);
    if (!entry) {
        if (depth != MAX_SPLIT_DEPTH &&
            (doca_error_t) error->type == DOCA_ERROR_TOO_BIG) {
            VLOG_DBG_RL(&rl, "%s: Split attempt %d flow entry is too big to "
                        "insert directly. Attempt another flow split.",
                        netdev_get_name(netdev), depth);
            entry = create_split_doca_flow_entry(netdev, queue_id, prio,
                                                 suffix_group, &suffix_spec,
                                                 &suffix_mask, actions,
                                                 actions_masks, dacts_descs,
                                                 monitor, fwd, flow_res,
                                                 depth + 1, error);

            if (entry) {
                goto out;
            }
        }
        error->type = RTE_FLOW_ERROR_TYPE_HANDLE;
        error->message = "Could not insert split flow suffix rule";
        goto err_suffix;
    }

out:
    flow_res->split_ctx[depth].curr_split_ctx = pctx;
    flow_res->split_ctx[depth].next_split_group_ctx = suffix_group;

    error->type = (enum rte_flow_error_type) DOCA_SUCCESS;
    return entry;

err_suffix:
    split_prefix_ctx_unref(pctx);
err_prefix:
    doca_pipe_group_ctx_unref(suffix_group);

    return NULL;
}

static struct doca_flow_handle *
create_doca_flow_handle(struct netdev *netdev,
                        unsigned int queue_id,
                        uint32_t prio,
                        uint32_t group,
                        uint32_t miss_flow_id,
                        struct doca_flow_match *spec,
                        struct doca_flow_match *mask,
                        struct doca_flow_actions *actions,
                        struct doca_flow_actions *actions_masks,
                        struct doca_flow_action_descs *dacts_descs,
                        struct doca_flow_monitor *monitor,
                        struct doca_flow_fwd *fwd,
                        struct doca_flow_handle_resources *flow_res,
                        struct dpdk_offload_handle *doh,
                        struct rte_flow_error *error)
{
    struct doca_pipe_group_ctx *group_ctx = NULL;
    struct doca_flow_handle *hndl;

    hndl = &doh->dfh;

    /* get self table pointer */
    group_ctx = doca_pipe_group_ctx_ref(netdev, group);
    if (!group_ctx) {
        error->type = RTE_FLOW_ERROR_TYPE_UNSPECIFIED;
        error->message = "Could not create group";
        goto err_pipe;
    }
    /* insert rule */
    if (!spec || spec->tun.type == DOCA_FLOW_TUN_NONE) {
        hndl->flow = create_doca_pipe_group_entry(netdev, queue_id, group_ctx,
                                                  prio, spec, mask, actions,
                                                  actions_masks, dacts_descs,
                                                  monitor, fwd, &flow_res->mask_ctx, error);
    } else {
        hndl->flow = NULL;
        error->type = (enum rte_flow_error_type) DOCA_ERROR_TOO_BIG;
    }
    if (!hndl->flow && spec && mask) {
        if ((doca_error_t) error->type == DOCA_ERROR_TOO_BIG) {
            VLOG_DBG_RL(&rl, "%s: Flow entry is too big to insert directly. "
                        "Attempt flow split.",
                        netdev_get_name(netdev));
            hndl->flow = create_split_doca_flow_entry(netdev, queue_id, prio,
                                                      group_ctx, spec, mask,
                                                      actions, actions_masks,
                                                      dacts_descs, monitor,
                                                      fwd, flow_res, 0, error);
        } else {
            error->type = RTE_FLOW_ERROR_TYPE_UNSPECIFIED;
            error->message = "Could not create offload rule";
        }
    }

    if (!hndl->flow) {
        goto err_insert;
    }

    memcpy(&hndl->flow_res, flow_res, sizeof *flow_res);
    hndl->flow_res.self_group_ctx = group_ctx;
    /* miss_flow_id will be non 0 only for pipe miss flows
     * and should be set once per pipe when the miss flow
     * is created. To avoid overwriting it when regular flows
     * are inserted to the pipe, we perform the check for
     * non 0 value.
     */
    if (miss_flow_id) {
        doca_pipe_group_set_miss_flow_id(group_ctx, miss_flow_id);
    }

    return hndl;

err_insert:
    doca_pipe_group_ctx_unref(group_ctx);
err_pipe:
    return NULL;
}

int
doca_pipe_group_ctx_add_miss(struct netdev *netdev,
                             struct doca_pipe_group_ctx *group_ctx)
{
    struct doca_pipe_group_mask_ctx *mask_ctx;
    struct doca_offload_esw_ctx *esw_ctx;
    struct doca_flow_handle *hndl;
    struct rte_flow_error error;
    struct doca_flow_fwd fwd;
    struct fixed_rule *fr;

    esw_ctx = doca_offload_esw_ctx_get(netdev);
    if (!esw_ctx) {
        VLOG_ERR("%s: ESW context is not usable, impossible to create "
                 "ctl pipe default miss rule.", netdev_get_name(netdev));
        return -1;
    }

    memset(&fwd, 0, sizeof fwd);

    fr = doca_pipe_group_get_miss_rule(group_ctx);
    hndl = &fr->doh.dfh;
    fwd.type = DOCA_FLOW_FWD_PIPE;
    fwd.next_pipe = doca_offload_get_misspath_pipe(esw_ctx);

    /* We *must not* use create_doca_flow_handle,
     * as it has to take a reference on the pipe
     * the entry is being inserted into. It would
     * create a recursive dependency.
     *
     * Similarly, no reference is taken on the next-pipe,
     * instead using a static pointer from the esw-ctx.
     */
    hndl->flow = create_doca_pipe_group_entry(netdev, AUX_QUEUE, group_ctx,
                                              DPDK_OFFLOAD_PRIORITY_MISS, NULL,
                                              NULL, NULL, NULL, NULL, NULL,
                                              &fwd,  &hndl->flow_res.mask_ctx,
                                              &error);
    if (!hndl->flow) {
        VLOG_ERR("%s: Failed to create root pipe default miss rule.",
                 netdev_get_name(netdev));
        return -1;
    }

    mask_ctx = hndl->flow_res.mask_ctx;

    /* No reference was taken on the 'next-pipe' for this fixed-rule.
     * None should be released when destroying it.
     * Force the 'flow_res' field to 0 to ensure that. */
    memset(&hndl->flow_res, 0, sizeof hndl->flow_res);

    hndl->flow_res.mask_ctx = mask_ctx;

    return 0;
}

static struct doca_flow_pipe_entry *
add_doca_post_meter_red_entry(struct netdev *netdev,
                              unsigned int queue_id,
                              uint32_t shared_meter_id,
                              uint32_t flow_id,
                              struct doca_pipe_group_mask_ctx **mask_ctx,
                              struct rte_flow_error *error)
{
    struct doca_pipe_group_ctx *post_meter_pipe_group_ctx;
    struct doca_offload_esw_ctx *esw_ctx;
    struct doca_flow_pipe_entry *entry;
    struct doca_flow_monitor monitor;
    struct doca_flow_match red_match;
    struct doca_flow_match red_mask;
    struct doca_flow_fwd fwd;

    esw_ctx = doca_offload_esw_ctx_get(netdev);
    post_meter_pipe_group_ctx = esw_ctx->post_meter_pipe_group_ctx;

    memset(&red_match, 0, sizeof(red_match));
    memset(&red_mask, 0, sizeof(red_mask));
    memset(&monitor, 0, sizeof(monitor));
    memset(&fwd, 0, sizeof fwd);

    fwd.type = DOCA_FLOW_FWD_DROP;
    monitor.counter_type = DOCA_FLOW_RESOURCE_TYPE_SHARED;
    monitor.shared_counter.shared_counter_id =
        ovs_doca_get_post_meter_counter_id(shared_meter_id,
                                           DOCA_FLOW_METER_COLOR_RED);

    doca_set_reg_val_mask(&red_match.meta, &red_mask.meta,
                          REG_FIELD_FLOW_INFO, flow_id);

    /* Insert red rule with low priority, lower than the corresponding
     * green rule, optimizing for the case when traffic stays below metered
     * rate.
     */
    entry = create_doca_pipe_group_entry(netdev, queue_id, post_meter_pipe_group_ctx,
                                         DPDK_OFFLOAD_PRIORITY_MED, &red_match,
                                         &red_mask, NULL, NULL, NULL, &monitor,
                                         &fwd, mask_ctx, error);
    if (!entry) {
        VLOG_ERR_RL(&rl, "%s: Failed to create shared meter red rule for flow ID %u",
                    netdev_get_name(netdev), flow_id);
        return NULL;
    }

    return entry;
}

static struct doca_flow_pipe_entry *
add_doca_post_meter_green_entry(struct netdev *netdev,
                                unsigned int queue_id,
                                uint32_t shared_meter_id,
                                uint32_t flow_id,
                                const struct doca_act_vars *dact_vars,
                                const struct meter_info *next_meter,
                                struct doca_flow_fwd *orig_fwd,
                                struct doca_pipe_group_mask_ctx **mask_ctx,
                                struct rte_flow_error *error)
{
    struct doca_pipe_group_ctx *post_meter_pipe_group_ctx;
    struct doca_flow_actions dacts, dacts_masks;
    struct doca_offload_esw_ctx *esw_ctx;
    struct doca_flow_fwd *fwd = orig_fwd;
    struct doca_flow_pipe_entry *entry;
    struct doca_flow_match green_match;
    struct doca_flow_match green_mask;
    struct doca_flow_fwd next_mtr_fwd;
    struct doca_flow_monitor monitor;
    uint32_t fid = dact_vars->flow_id;

    esw_ctx = doca_offload_esw_ctx_get(netdev);
    post_meter_pipe_group_ctx = esw_ctx->post_meter_pipe_group_ctx;

    memset(&next_mtr_fwd, 0, sizeof next_mtr_fwd);
    memset(&green_match, 0, sizeof(green_match));
    memset(&dacts_masks, 0, sizeof dacts_masks);
    memset(&green_mask, 0, sizeof(green_mask));
    memset(&monitor, 0, sizeof(monitor));
    memset(&dacts, 0, sizeof dacts);

    monitor.counter_type = DOCA_FLOW_RESOURCE_TYPE_SHARED;
    monitor.shared_counter.shared_counter_id =
        ovs_doca_get_post_meter_counter_id(shared_meter_id,
                                           DOCA_FLOW_METER_COLOR_GREEN);

    doca_set_reg_val_mask(&green_match.meta, &green_mask.meta,
                          REG_FIELD_FLOW_INFO, flow_id);
    green_match.parser_meta.meter_color = DOCA_FLOW_METER_COLOR_GREEN;
    green_mask.parser_meta.meter_color = 0xff;

    if (next_meter) {
        monitor.meter_type = DOCA_FLOW_RESOURCE_TYPE_SHARED;
        monitor.shared_meter.shared_meter_id = next_meter->id;
        fid = next_meter->flow_id;
        next_mtr_fwd.type = DOCA_FLOW_FWD_PIPE;
        next_mtr_fwd.next_pipe = doca_pipe_group_get_pipe(post_meter_pipe_group_ctx);
        fwd = &next_mtr_fwd;
    } else if (dact_vars->hash_data) {
        doca_set_reg_val_mask(&dacts.meta, &dacts_masks.meta, REG_FIELD_SCRATCH,
                              dact_vars->hash_data->seed);
    }

    doca_set_reg_val_mask(&dacts.meta, &dacts_masks.meta,
                          REG_FIELD_FLOW_INFO, fid);

    /* Insert green rule with high prio, which is higher than the corresponding
     * red rule, optimizing for the case when traffic stays below metered rate.
     */
    entry = create_doca_pipe_group_entry(netdev, queue_id, post_meter_pipe_group_ctx,
                                         DPDK_OFFLOAD_PRIORITY_HIGH, &green_match,
                                         &green_mask, &dacts, &dacts_masks, NULL,
                                         &monitor, fwd, mask_ctx, error);
    if (!entry) {
        VLOG_ERR_RL(&rl, "%s: Failed to create shared meter green rule for mtr"
                    " ID %u, flow ID %u", netdev_get_name(netdev),
                    shared_meter_id, flow_id);
        return NULL;
    }

    if (!next_meter) {
        /* replace original fwd with the internal meter pipe */
        memset(orig_fwd, 0, sizeof *orig_fwd);
        orig_fwd->type = DOCA_FLOW_FWD_PIPE;
        orig_fwd->next_pipe = doca_pipe_group_get_pipe(post_meter_pipe_group_ctx);
    }

    return entry;
}

static int
destroy_meter_hierarchy(struct netdev *netdev, unsigned int queue_id,
                        struct doca_flow_handle_resources *flow_res,
                        struct rte_flow_error *error)
{
    struct doca_offload_esw_ctx *esw_ctx = doca_offload_esw_ctx_get(netdev);
    struct doca_meter_ctx *mtr_ctx;
    int err;

    if (!flow_res->meters_ctx) {
        return 0;
    }

    LIST_FOR_EACH_SAFE (mtr_ctx, list_node, flow_res->meters_ctx) {
        /* RED entry */
        if (mtr_ctx->post_meter_red_entry) {
            err = doca_offload_remove_entry(esw_ctx, queue_id, DOCA_FLOW_NO_WAIT,
                                            mtr_ctx->post_meter_red_entry);
            if (err) {
                if (error) {
                    error->type = RTE_FLOW_ERROR_TYPE_HANDLE;
                    error->message = doca_error_get_descr(err);
                }
                return -1;
            }
            doca_pipe_group_mask_ctx_unref(mtr_ctx->post_meter_red_entry_mctx);
            dpdk_offload_counter_dec(netdev);
        }

        /* GREEN entry */
        if (mtr_ctx->post_meter_green_entry) {
            err = doca_offload_remove_entry(esw_ctx, queue_id, DOCA_FLOW_NO_WAIT,
                                            mtr_ctx->post_meter_green_entry);
            if (err) {
                if (error) {
                    error->type = RTE_FLOW_ERROR_TYPE_HANDLE;
                    error->message = doca_error_get_descr(err);
                }
                return -1;
            }
            doca_pipe_group_mask_ctx_unref(mtr_ctx->post_meter_green_entry_mctx);
            dpdk_offload_counter_dec(netdev);
        }

        /* Meter internal flow ID */
        if (mtr_ctx->post_meter_flow_id) {
            shared_mtr_flow_id_free(esw_ctx, mtr_ctx->post_meter_flow_id);
        }

        ovs_list_remove(&mtr_ctx->list_node);
        free(mtr_ctx);
    }

    free(flow_res->meters_ctx);
    flow_res->meters_ctx = NULL;

    return 0;
}

static void
destroy_post_hash_entry(struct netdev *netdev, unsigned int queue_id,
                        struct doca_flow_handle_resources *flow_res)
{
    struct doca_offload_esw_ctx *esw;

    if (!flow_res->post_hash_entry) {
        return;
    }

    esw = doca_offload_esw_ctx_get(netdev);
    doca_offload_remove_entry(esw, queue_id, DOCA_FLOW_NO_WAIT,
                              flow_res->post_hash_entry);
    doca_pipe_group_mask_ctx_unref(flow_res->post_hash_mask_ctx);
    dpdk_offload_counter_dec(netdev);
    flow_res->post_hash_entry = NULL;
    flow_res->post_hash_mask_ctx = NULL;
}

static struct doca_flow_pipe_entry *
create_post_hash_entry(struct netdev *netdev, unsigned int queue_id,
                       uint32_t flow_id,
                       struct doca_pipe_group_ctx *next_group_ctx,
                       struct doca_pipe_group_mask_ctx **mask_ctx,
                       struct rte_flow_error *error)
{
    struct doca_offload_esw_ctx *esw_ctx = doca_offload_esw_ctx_get(netdev);
    struct doca_pipe_group_ctx *post_hash_group_ctx;
    struct doca_flow_pipe_entry *entry;
    struct doca_flow_match spec;
    struct doca_flow_match mask;
    struct doca_flow_fwd fwd;

    memset(&mask, 0, sizeof mask);
    memset(&spec, 0, sizeof spec);
    memset(&fwd, 0, sizeof fwd);

    post_hash_group_ctx = esw_ctx->hash_pipe_ctx->post_hash_group_ctx;

    doca_set_reg_val_mask(&spec.meta, &mask.meta, REG_FIELD_FLOW_INFO, flow_id);
    fwd.type = DOCA_FLOW_FWD_PIPE;
    fwd.next_pipe = doca_pipe_group_get_pipe(next_group_ctx);

    entry = create_doca_pipe_group_entry(netdev, queue_id, post_hash_group_ctx, 0,
                                         &spec, &mask, NULL, NULL, NULL, NULL,
                                         &fwd, mask_ctx, error);
    return entry;
}

static int
create_meter_hierarchy(struct netdev *netdev, unsigned int queue_id,
                       uint32_t shared_meter_id,
                       struct doca_act_vars *dact_vars,
                       struct doca_flow_handle_resources *flow_res,
                       struct doca_flow_fwd *fwd, struct rte_flow_error *error)
{
    struct doca_offload_esw_ctx *esw = doca_offload_esw_ctx_get(netdev);
    struct meter_info *next_mtr = NULL;
    struct doca_meter_ctx *mtr_ctx;
    struct meter_info *cur_mtr;

    if (!flow_res->meters_ctx) {
        flow_res->meters_ctx = xzalloc(sizeof *flow_res->meters_ctx);
        ovs_list_init(flow_res->meters_ctx);
    }

    /* Meter entries are added in reverse order to have the full hierarchy
     * already in place at time when the first packet arrives. The entries are
     * "chained" by the means of mark and match on meter_info->flow_id.
     */
    LIST_FOR_EACH_REVERSE (cur_mtr, list_node, &dact_vars->next_meters) {
        mtr_ctx = xzalloc(sizeof *mtr_ctx);

        /* RED entry */
        mtr_ctx->post_meter_red_entry =
            add_doca_post_meter_red_entry(netdev, queue_id, cur_mtr->id,
                                          cur_mtr->flow_id,
                                          &mtr_ctx->post_meter_red_entry_mctx, error);
        if (!mtr_ctx->post_meter_red_entry) {
            if (error) {
                error->type = RTE_FLOW_ERROR_TYPE_ACTION;
                error->message = "Could not create red post multi-meter rule";
            }
            goto err;
        }

        /* GREEN entry */
        mtr_ctx->post_meter_green_entry =
            add_doca_post_meter_green_entry(netdev, queue_id, cur_mtr->id,
                                            cur_mtr->flow_id, dact_vars,
                                            next_mtr, fwd,
                                            &mtr_ctx->post_meter_green_entry_mctx,
                                            error);
        if (!mtr_ctx->post_meter_green_entry) {
            if (error) {
                error->type = RTE_FLOW_ERROR_TYPE_ACTION;
                error->message = "Could not create green post multi-meter rule";
            }
            doca_offload_remove_entry(esw, queue_id, DOCA_FLOW_NO_WAIT,
                                      mtr_ctx->post_meter_red_entry);
            doca_pipe_group_mask_ctx_unref(mtr_ctx->post_meter_red_entry_mctx);
            goto err;
        }

        mtr_ctx->post_meter_flow_id = cur_mtr->flow_id;
        ovs_list_push_front(flow_res->meters_ctx, &mtr_ctx->list_node);
        next_mtr = cur_mtr;
    }

    mtr_ctx = xzalloc(sizeof *mtr_ctx);

    /* First RED entry */
    mtr_ctx->post_meter_red_entry =
        add_doca_post_meter_red_entry(netdev, queue_id, shared_meter_id,
                                      dact_vars->mtr_flow_id,
                                      &mtr_ctx->post_meter_red_entry_mctx,
                                      error);
    if (!mtr_ctx->post_meter_red_entry) {
        if (error) {
            error->type = RTE_FLOW_ERROR_TYPE_ACTION;
            error->message = "Could not create red post meter rule";
        }
        goto err;
    }

    /* First GREEN entry */
    mtr_ctx->post_meter_green_entry =
        add_doca_post_meter_green_entry(netdev, queue_id, shared_meter_id,
                                        dact_vars->mtr_flow_id, dact_vars,
                                        next_mtr, fwd,
                                        &mtr_ctx->post_meter_green_entry_mctx,
                                        error);
    if (!mtr_ctx->post_meter_green_entry) {
        if (error) {
            error->type = RTE_FLOW_ERROR_TYPE_ACTION;
            error->message = "Could not create green post meter rule";
        }
        doca_offload_remove_entry(esw, queue_id, DOCA_FLOW_NO_WAIT,
                                  mtr_ctx->post_meter_red_entry);
        doca_pipe_group_mask_ctx_unref(mtr_ctx->post_meter_red_entry_mctx);
        goto err;
    }

    mtr_ctx->post_meter_flow_id = dact_vars->mtr_flow_id;
    ovs_list_push_front(flow_res->meters_ctx, &mtr_ctx->list_node);

    return 0;

err:
    free(mtr_ctx);
    destroy_meter_hierarchy(netdev, queue_id, flow_res, NULL);
    if (error) {
        ovs_assert(error->message);
        ovs_assert(error->type);
    }
    return -1;
}

/* Random mask HW register is limited to 16 bits. Because it's a mask, it
 * doesn't support arbitrary sampling ratios. Only a fixed number of ratios that
 * correspond to each bit in the mask are supported - starting from the maximum
 * supported ratio of 50%, down to the minimal supported ratio of 0.0015258789%.
 * The percentage value is halved on each step:
 *
 *     50, 25, 12.5, ... , 0.0015258789
 *
 *      1:  50.000000% -> 0x1
 *      2:  25.000000% -> 0x3
 *      3:  12.500000% -> 0x7
 *      4:   6.250000% -> 0xf
 *      5:   3.125000% -> 0x1f
 *      6:   1.562500% -> 0x3f
 *      7:   0.781250% -> 0x7f
 *      8:   0.390625% -> 0xff
 *      9:   0.195312% -> 0x1ff
 *     10:   0.097656% -> 0x3ff
 *     11:   0.048828% -> 0x7ff
 *     12:   0.024414% -> 0xfff
 *     13:   0.012207% -> 0x1fff
 *     14:   0.006104% -> 0x3fff
 *     15:   0.003052% -> 0x7fff
 *     16:   0.001526% -> 0xffff
 *
 * It finds the nearest supported percentage which is not greater than the
 * requested ratio and calculates corresponding mask value.
 */
static uint16_t
sample_ratio_to_random_mask(uint32_t ratio)
{
    double next_highest_supported_percentage = 50;
    double percentage = 1.0 / ratio * 100;
    uint16_t mask;
    int i;

    for (i = 1; i <= UINT16_WIDTH; ++i) {
        if (percentage >= next_highest_supported_percentage) {
            break;
        }

        next_highest_supported_percentage /= 2;
    }

    mask = (1 << i) - 1;

    VLOG_INFO("Using random mask 0x%04x for sampling %g%% of traffic"
              " (%g%% requested)", mask, next_highest_supported_percentage,
              percentage);

    return mask;
}

static void
doca_sample_pipe_del_entry(struct netdev *netdev)
{
    struct doca_offload_esw_ctx *ctx = doca_offload_esw_ctx_get(netdev);

    if (ctx->sample_ctx.sample_pipe_entry) {
        doca_offload_remove_entry(ctx, AUX_QUEUE, DOCA_FLOW_NO_WAIT,
                                  ctx->sample_ctx.sample_pipe_entry);
        doca_pipe_group_mask_ctx_unref(ctx->sample_ctx.sample_pipe_entry_mctx);
        dpdk_offload_counter_dec(netdev);
        ctx->sample_ctx.sample_pipe_entry = NULL;
        ctx->sample_ctx.ratio = 0;
    }
}

static int
doca_sample_pipe_add_entry(struct netdev *netdev, uint32_t ratio)
{
    struct doca_offload_esw_ctx *ctx = doca_offload_esw_ctx_get(netdev);
    struct doca_pipe_group_mask_ctx **entry_mctx;
    struct doca_flow_monitor monitor = {
        .counter_type = DOCA_FLOW_RESOURCE_TYPE_NON_SHARED,
    };
    struct doca_flow_pipe_entry *entry;
    struct doca_flow_actions act_mask;
    struct doca_flow_actions act;
    struct doca_flow_match spec;
    struct doca_flow_match mask;
    struct doca_flow_fwd fwd = {
        .type = DOCA_FLOW_FWD_PIPE,
        .next_pipe = ctx->sample_ctx.mirror_pipe_ctx.pipe,
    };
    struct rte_flow_error error;

    memset(&act_mask, 0, sizeof act_mask);
    memset(&spec, 0, sizeof spec);
    memset(&mask, 0, sizeof mask);
    memset(&act, 0, sizeof act);

    /* For ratio == 1 the match mask doesn't need to be set, effectively meaning
     * "matchall".
     */
    if (ratio != 1) {
        mask.parser_meta.random = sample_ratio_to_random_mask(ratio);
    }
    doca_set_reg_val_mask(&act.meta, &act_mask.meta, REG_FIELD_FLOW_INFO,
                          NONSAMPLE_FLOW_MARK);
    entry_mctx = &ctx->sample_ctx.sample_pipe_entry_mctx;

    /* Main sampling entry, it matches with a probability lower than the
     * requested sampling ratio and forwards packets to the sample-mirror pipe.
     * Everything else, that is above the requested sampling ratio probability,
     * is captured by the second lower priority match-all entry and is forwarded
     * back to the root pipe.
     */
    entry = create_doca_pipe_group_entry(netdev, AUX_QUEUE, ctx->sample_ctx.sample_pipe_ctx,
                                         DPDK_OFFLOAD_PRIORITY_HIGH, &spec,
                                         &mask, &act, &act_mask, NULL, &monitor,
                                         &fwd, entry_mctx, &error);
    if (!entry) {
            VLOG_ERR_RL(&rl, "%s: Failed to create sample entry.",
                        netdev_get_name(netdev));
            return -1;
    }
    ctx->sample_ctx.sample_pipe_entry = entry;
    dpdk_offload_counter_inc(netdev);

    if (ctx->sample_ctx.sample_pipe_miss_entry) {
        goto out;
    }

    entry_mctx = &ctx->sample_ctx.sample_pipe_miss_entry_mctx;
    fwd.next_pipe = doca_pipe_group_get_pipe(ctx->root_pipe_group_ctx);

    entry = create_doca_pipe_group_entry(netdev, AUX_QUEUE, ctx->sample_ctx.sample_pipe_ctx,
                                         DPDK_OFFLOAD_PRIORITY_LOW, &spec,
                                         &spec, &act, &act_mask, NULL, NULL, &fwd,
                                         entry_mctx, &error);
    if (!entry) {
            VLOG_ERR_RL(&rl, "%s: Failed to create sample miss entry.",
                        netdev_get_name(netdev));
            doca_sample_pipe_del_entry(netdev);
            return -1;
    }
    ctx->sample_ctx.sample_pipe_miss_entry = entry;
    dpdk_offload_counter_inc(netdev);

out:
    ctx->sample_ctx.ratio = ratio;
    doca_offload_complete_queue(netdev, AUX_QUEUE, true);
    return 0;
}

static int
doca_sample_root_add_entry(struct netdev *netdev, struct rte_flow_error *error)
{
    struct doca_offload_esw_ctx *ctx = doca_offload_esw_ctx_get(netdev);
    struct doca_pipe_group_mask_ctx **entry_mctx;
    struct doca_flow_pipe_entry *entry;
    struct doca_flow_match spec;
    struct doca_flow_match mask;
    struct doca_flow_fwd fwd = {
        .type = DOCA_FLOW_FWD_PIPE,
        .next_pipe = doca_pipe_group_get_pipe(ctx->sample_ctx.sample_pipe_ctx),
    };

    memset(&spec, 0, sizeof spec);
    memset(&mask, 0, sizeof mask);

    entry_mctx = &ctx->sample_ctx.root_sample_entry_mctx;
    doca_set_reg_mask(&mask.meta, REG_FIELD_FLOW_INFO);

    entry = create_doca_pipe_group_entry(netdev, AUX_QUEUE, ctx->root_pipe_group_ctx,
                                         DPDK_OFFLOAD_PRIORITY_DOCA_SAMPLE, &spec,
                                         &mask, NULL, NULL, NULL, NULL, &fwd,
                                         entry_mctx, error);
    if (!entry) {
            VLOG_ERR_RL(&rl, "%s: Failed to add sampling entry to root pipe: "
                        "%s.", netdev_get_name(netdev), error->message);
            return -1;
    }
    ctx->sample_ctx.root_sample_entry = entry;
    ovs_refcount_init(&ctx->sample_ctx.root_sample_entry_refcount);
    doca_offload_complete_queue(netdev, AUX_QUEUE, true);
    VLOG_DBG("%s: Added sampling entry to root pipe.", netdev_get_name(netdev));

    return 0;
}

static int
dpdk_offload_doca_create(struct netdev *netdev,
                         const struct rte_flow_attr *attr,
                         struct rte_flow_item *items,
                         struct rte_flow_action *actions,
                         struct dpdk_offload_handle *doh,
                         struct rte_flow_error *error)
{
    struct doca_offload_esw_ctx *esw_ctx = doca_offload_esw_ctx_get(netdev);
    unsigned int tid = netdev_offload_thread_id();
    struct doca_flow_actions dacts, dacts_masks;
    struct doca_flow_handle_resources flow_res;
    struct doca_flow_action_descs dacts_descs;
    struct doca_flow_action_desc desc_array;
    struct doca_flow_monitor monitor;
    struct doca_act_vars dact_vars;
    struct doca_flow_handle *hndl;
    struct doca_flow_match mask;
    struct doca_flow_match spec;
    struct meter_info *next_mtr;
    unsigned int queue_id = tid;
    uint32_t miss_flow_id = 0;
    struct doca_flow_fwd fwd;

    memset(error, 0, sizeof *error);
    memset(&dact_vars, 0, sizeof dact_vars);
    ovs_list_init(&dact_vars.next_meters);

    /* If it's a post ct rule, check for eswitch ct offload support */
    if (attr->group == POSTCT_TABLE_ID &&
        !(esw_ctx->shared_ct_counter_id_pool || ovs_doca_flow_ct_enabled)) {
        error->type = RTE_FLOW_ERROR_TYPE_UNSPECIFIED;
        error->message = "CT offload disabled on this eswitch";
        goto err;
    }

    memset(&dacts_masks, 0, sizeof dacts_masks);
    memset(&dacts_descs, 0, sizeof dacts_descs);
    memset(&desc_array, 0, sizeof desc_array);
    memset(&flow_res, 0, sizeof flow_res);
    memset(&monitor, 0, sizeof monitor);
    memset(&dacts, 0, sizeof dacts);
    memset(&mask, 0, sizeof mask);
    memset(&spec, 0, sizeof spec);
    memset(&fwd, 0, sizeof fwd);

    if (doca_translate_items(netdev, attr, items, &spec, &mask)) {
        error->type = RTE_FLOW_ERROR_TYPE_ITEM;
        error->message = "Could not create items";
        goto err;
    }

    /* parse actions */
    dacts_descs.desc_array = &desc_array;
    dacts_descs.nb_action_desc = 1;
    if (doca_translate_actions(netdev, &spec, actions, &dacts, &dacts_masks,
                               &fwd, &monitor, &flow_res, &dact_vars)) {
        error->type = RTE_FLOW_ERROR_TYPE_ACTION;
        error->message = "Could not create actions";
        goto err;
    }

    /* Detect a miss flow which sets flow id for the
     * Ctl pipe and store it for split flows usage.
     */
    if (attr->priority == DPDK_OFFLOAD_PRIORITY_MISS && dact_vars.flow_id) {
        miss_flow_id = dact_vars.flow_id;
    }

    /* Create post-hash entry */
    if (dact_vars.hash_data && flow_res.next_group_ctx) {
        struct doca_pipe_group_ctx *next_group_ctx = flow_res.next_group_ctx;
        uint32_t flow_id = dact_vars.hash_data->flow_id;

        flow_res.post_hash_entry =
            create_post_hash_entry(netdev, queue_id, flow_id, next_group_ctx,
                                   &flow_res.post_hash_mask_ctx, error);
        if (flow_res.post_hash_entry == NULL) {
            goto err;
        }
    }

    if (monitor.meter_type == DOCA_FLOW_RESOURCE_TYPE_SHARED) {
        if (create_meter_hierarchy(netdev, queue_id, monitor.shared_meter.shared_meter_id,
                                   &dact_vars, &flow_res, &fwd, error)) {
            goto err;
        }
    }

    if (dact_vars.sample_ratio) {
        if (!esw_ctx->sample_ctx.sample_pipe_entry) {
            if (doca_sample_pipe_add_entry(netdev, dact_vars.sample_ratio)) {
                error->type = RTE_FLOW_ERROR_TYPE_ACTION;
                error->message = "Could not offload sample action";
                goto err;
            }
        }
        if (!esw_ctx->sample_ctx.root_sample_entry) {
            if (dact_vars.sample_ratio != esw_ctx->sample_ctx.ratio) {
                doca_sample_pipe_del_entry(netdev);
                if (doca_sample_pipe_add_entry(netdev, dact_vars.sample_ratio)) {
                    error->type = RTE_FLOW_ERROR_TYPE_ACTION;
                    error->message = "Could not offload sample action";
                    goto err;
                }
            }

            if (doca_sample_root_add_entry(netdev, error)) {
                goto err;
            }
        }
        if (dact_vars.sample_ratio != esw_ctx->sample_ctx.ratio) {
            error->type = RTE_FLOW_ERROR_TYPE_ACTION;
            error->message = "Could not offload sample action";
            VLOG_WARN_RL(&rl, "%s: Cannot add sFlow ratio %u while another "
                         " ratio %u is active", netdev_get_name(netdev),
                         dact_vars.sample_ratio, esw_ctx->sample_ctx.ratio);
            goto err;
        }
    }

    hndl = create_doca_flow_handle(netdev, queue_id, attr->priority, attr->group,
                                   miss_flow_id, &spec, &mask, &dacts,
                                   &dacts_masks, &dacts_descs, &monitor, &fwd,
                                   &flow_res, doh, error);
    if (!hndl) {
        destroy_post_hash_entry(netdev, queue_id, &flow_res);
        doca_pipe_group_ctx_unref(flow_res.next_group_ctx);
        if (monitor.meter_type == DOCA_FLOW_RESOURCE_TYPE_SHARED) {
            destroy_meter_hierarchy(netdev, queue_id, &flow_res, NULL);
        }
        goto err;
    }

    if (dact_vars.sample_ratio) {
        hndl->flow_res.sampled = true;
        ovs_refcount_ref(&esw_ctx->sample_ctx.root_sample_entry_refcount);
    }

    LIST_FOR_EACH_POP (next_mtr, list_node, &dact_vars.next_meters) {
        free(next_mtr);
    }

    return 0;

err:
    LIST_FOR_EACH_POP (next_mtr, list_node, &dact_vars.next_meters) {
        free(next_mtr);
    }
    ovs_assert(error->message);
    ovs_assert(error->type);
    doh->rte_flow = NULL;
    return -1;
}

static int
destroy_dpdk_offload_handle(struct doca_offload_esw_ctx *esw,
                            struct netdev *netdev,
                            struct dpdk_offload_handle *doh,
                            unsigned int queue_id,
                            struct rte_flow_error *error)
{
    doca_error_t err;

    /* Deletion is always synchronous.
     *
     * If async deletion is implemented, aux-table uninit calls deleting
     * entries will use the offload queues in conflict with offload threads
     * polling them during upkeep. It should result in a crash or
     * in a lockup of the queues. */
    err = doca_offload_remove_entry(esw, queue_id, DOCA_FLOW_NO_WAIT, doh->dfh.flow);
    if (err) {
        if (error) {
            error->type = RTE_FLOW_ERROR_TYPE_HANDLE;
            error->message = doca_error_get_descr(err);
        }
        return -1;
    }

    doca_pipe_group_mask_ctx_unref(doh->dfh.flow_res.mask_ctx);

    if (doh->dfh.flow_res.meters_ctx) {
        err = destroy_meter_hierarchy(netdev, queue_id, &doh->dfh.flow_res,
                                      error);
        if (err) {
            return -1;
        }
    }

    if (doh->dfh.flow_res.split_ctx[0].curr_split_ctx) {
        for (int i = MAX_SPLIT_DEPTH - 1; i >= 0; i--) {
            split_prefix_ctx_unref(doh->dfh.flow_res.split_ctx[i].curr_split_ctx);
            doca_pipe_group_ctx_unref(doh->dfh.flow_res.split_ctx[i].next_split_group_ctx);
        }
    }

    destroy_post_hash_entry(netdev, queue_id, &doh->dfh.flow_res);

    if (doh->dfh.flow_res.sampled) {
        struct doca_offload_esw_ctx *esw_ctx = doca_offload_esw_ctx_get(netdev);
        /* root_sample_entry_refcount is initialized to 1 when entry is added.
         * There is also additional ovs_refcount_ref() for every offloaded flow
         * with sample action. So when there is only a single flow, the count is
         * 2. When this flow is destroyed, there will be no other sample action
         * flow, so the root entry needs to be removed and the counter value to
         * check for this condition is 2.
         */
        if (ovs_refcount_unref(&esw_ctx->sample_ctx.root_sample_entry_refcount) == 2) {
            doca_offload_remove_entry(esw_ctx, AUX_QUEUE, DOCA_FLOW_NO_WAIT,
                                      esw_ctx->sample_ctx.root_sample_entry);
            dpdk_offload_counter_dec(netdev);
            esw_ctx->sample_ctx.root_sample_entry = NULL;
            doca_pipe_group_mask_ctx_unref(esw_ctx->sample_ctx.root_sample_entry_mctx);
            VLOG_DBG("%s: Removed sampling entry from root pipe.",
                     netdev_get_name(netdev));
        }
    }

    /* Netdev can only be NULL during aux tables uninit. */
    if (netdev) {
        dpdk_offload_counter_dec(netdev);
    }

    doca_pipe_group_ctx_unref(doh->dfh.flow_res.next_group_ctx);
    doca_pipe_group_ctx_unref(doh->dfh.flow_res.self_group_ctx);

    return 0;
}

static int
dpdk_offload_doca_destroy(struct netdev *netdev,
                          struct dpdk_offload_handle *doh,
                          struct rte_flow_error *error,
                          bool esw_port_id OVS_UNUSED)
{
    struct doca_offload_esw_ctx *esw = doca_offload_esw_ctx_get(netdev);
    unsigned int queue_id = netdev_offload_thread_id();

    return destroy_dpdk_offload_handle(esw, netdev, doh, queue_id, error);
}

static int
doca_query_count(struct netdev *netdev,
                 struct doca_flow_pipe_entry *entry,
                 struct rte_flow_query_count *query,
                 struct rte_flow_error *error)
{
    struct ovs_rwlock *port_rwlock = doca_offload_rwlock_get(netdev);
    struct doca_flow_resource_query stats;
    doca_error_t err;

    memset(query, 0, sizeof *query);
    memset(&stats, 0, sizeof stats);

    if (entry == NULL) {
        /* The async entry has not yet been completed,
         * it cannot have done anything yet. */
        return 0;
    }

    err = doca_offload_query_entry(port_rwlock, entry, &stats);
    if (err) {
        VLOG_WARN_RL(&rl, "%s: Failed to query entry: %p. Error %d (%s)",
                     netdev_get_name(netdev), entry, err,
                     doca_error_get_descr(err));
        error->type = RTE_FLOW_ERROR_TYPE_UNSPECIFIED;
        error->message = doca_error_get_descr(err);
        return -1;
    }

    query->hits = stats.counter.total_pkts;
    query->bytes = stats.counter.total_bytes;

    return 0;
}

static int
dpdk_offload_doca_query_count(struct netdev *netdev,
                              struct dpdk_offload_handle *doh,
                              struct rte_flow_query_count *query,
                              struct rte_flow_error *error)
{
    return doca_query_count(netdev, doh->dfh.flow, query, error);
}

static int
dpdk_offload_doca_shared_create(struct ovs_shared_ctx *ctx,
                                struct ovs_shared_arg *arg,
                                struct rte_flow_error *error)
{
    const struct rte_flow_action *action = arg->action;
    unsigned int tid = netdev_offload_thread_id();
    struct doca_offload_esw_ctx *esw_ctx;
    struct netdev *netdev = arg->netdev;
    uint32_t id;

    esw_ctx = doca_offload_esw_ctx_get(netdev);

    if (action && action->type != RTE_FLOW_ACTION_TYPE_COUNT) {
        return -1;
    }

    error->type = RTE_FLOW_ERROR_TYPE_UNSPECIFIED;
    error->message = NULL;

    switch (ctx->res_type) {
    case OVS_SHARED_COUNT:
        if (!id_fpool_new_id(esw_ctx->shared_counter_id_pool, tid, &id)) {
            error->message = "Flow counter exhausted: no free ID";
        }
        if (error->message) {
            /* Flow-related message are not expected to be great in number,
             * no need to rate-limit errors. */
            VLOG_ERR("%s", error->message);
            return -1;
        }
        break;
    case OVS_SHARED_CT_COUNT:
        if (!esw_ctx->shared_ct_counter_id_pool) {
            error->message = "Shared CT counters not enabled on this eswitch";
        } else if (!id_fpool_new_id(esw_ctx->shared_ct_counter_id_pool,
                                    tid, &id)) {
            error->message = "CT counter exhausted: no free ID";
        }
        if (error->message) {
            /* CT-related message can be in high-volume, do not DoS our
             * syslog with errors. */
            VLOG_ERR_RL(&rl, "%s", error->message);
            return -1;
        }
        break;
    case OVS_SHARED_CT_ACTIONS:
        /* This is only compatible with DOCA-CT. */
        if (!esw_ctx->ct.pipe) {
            error->message = "Shared CT actions requires DOCA-CT";
            return -1;
        }
        if (!doca_flow_ct_shared_actions_create(esw_ctx, arg->ct_actions_key, &id, error)) {
            return -1;
        }
        break;
    case OVS_SHARED_UNDEFINED:
        error->message = "Unsupported shared resource type requested";
        return -1;
    }

    ctx->res_id = id;
    if (action) {
        ctx->act_type = action->type;
    }

    return 0;
}

static int
dpdk_offload_doca_shared_destroy(struct ovs_shared_ctx *ctx,
                                 struct rte_flow_error *error OVS_UNUSED)
{
    struct doca_offload_esw_ctx *esw_ctx = doca_offload_esw_ctx_get(ctx->netdev);
    unsigned int tid = netdev_offload_thread_id();

    switch (ctx->res_type) {
    case OVS_SHARED_COUNT:
        id_fpool_free_id(esw_ctx->shared_counter_id_pool, tid, ctx->res_id);
        break;
    case OVS_SHARED_CT_COUNT:
        if (esw_ctx->shared_ct_counter_id_pool) {
            id_fpool_free_id(esw_ctx->shared_ct_counter_id_pool, tid,
                             ctx->res_id);
        }
        break;
    case OVS_SHARED_CT_ACTIONS:
        if (ovs_doca_flow_ct_enabled) {
            doca_flow_ct_shared_actions_destroy(esw_ctx, ctx->res_id);
        }
        break;
    case OVS_SHARED_UNDEFINED:
        VLOG_ERR("Unsupported shared resource type deletion");
        return -1;
    }

    return 0;
}

static int
dpdk_offload_doca_shared_query(struct ovs_shared_ctx *ctx,
                               void *data,
                               struct rte_flow_error *error)
{
    struct doca_flow_resource_query query_results;
    struct rte_flow_query_count *query;
    doca_error_t ret;
    uint32_t cnt_id;

    /* Only shared counter supported at the moment */
    if (ctx->act_type != RTE_FLOW_ACTION_TYPE_COUNT) {
        return -1;
    }

    query = (struct rte_flow_query_count *) data;
    memset(query, 0, sizeof *query);
    memset(&query_results, 0, sizeof query_results);

    cnt_id = ctx->res_id;
    ret = doca_flow_shared_resources_query(DOCA_FLOW_SHARED_RESOURCE_COUNTER,
                                           &cnt_id, &query_results, 1);
    if (ret != DOCA_SUCCESS) {
        VLOG_ERR("Failed to query shared counter id 0x%.8x: %s",
                 ctx->res_id, doca_error_get_descr(ret));
        error->type = RTE_FLOW_ERROR_TYPE_UNSPECIFIED;
        error->message = doca_error_get_descr(ret);
        return -1;
    }

    query->hits = query_results.counter.total_pkts;
    query->bytes = query_results.counter.total_bytes;

    return 0;
}

static int
ipv4_packet_hw_hash(struct doca_hash_pipe_ctx *hash_pipe_ctx,
                    struct dp_packet *packet,
                    uint32_t seed,
                    uint32_t *hash)
{
    struct doca_flow_match field_values;
    struct doca_flow_pipe *pipe;
    struct udp_header *udp;
    struct tcp_header *tcp;
    struct ip_header *ip;

    ip = dp_packet_l3(packet);

    memset(&field_values.meta, 0, sizeof field_values.meta);
    doca_set_reg_val(&field_values.meta, REG_FIELD_SCRATCH, seed);
    if (ip->ip_proto == IPPROTO_UDP) {
        pipe = hash_pipe_ctx->hashes[HASH_TYPE_IPV4_UDP].pipe;
        udp = (struct udp_header *) (ip + 1);
        field_values.outer.udp.l4_port.src_port = (OVS_FORCE doca_be16_t) udp->udp_src;
        field_values.outer.udp.l4_port.dst_port = (OVS_FORCE doca_be16_t) udp->udp_dst;
        field_values.outer.l4_type_ext = DOCA_FLOW_L4_TYPE_EXT_UDP;
    } else if (ip->ip_proto == IPPROTO_TCP) {
        pipe = hash_pipe_ctx->hashes[HASH_TYPE_IPV4_TCP].pipe;
        tcp = (struct tcp_header *) (ip + 1);
        field_values.outer.tcp.l4_port.src_port = (OVS_FORCE doca_be16_t) tcp->tcp_src;
        field_values.outer.tcp.l4_port.dst_port = (OVS_FORCE doca_be16_t) tcp->tcp_dst;
        field_values.outer.l4_type_ext = DOCA_FLOW_L4_TYPE_EXT_TCP;
    } else {
        pipe = hash_pipe_ctx->hashes[HASH_TYPE_IPV4_L3].pipe;
    }

    field_values.outer.l3_type = DOCA_FLOW_L3_TYPE_IP4;
    field_values.outer.ip4.src_ip = (OVS_FORCE doca_be32_t) get_16aligned_be32(&ip->ip_src);
    field_values.outer.ip4.dst_ip = (OVS_FORCE doca_be32_t) get_16aligned_be32(&ip->ip_dst);

    return doca_flow_pipe_calc_hash(pipe, &field_values, hash);
}

static int
ipv6_packet_hw_hash(struct doca_hash_pipe_ctx *hash_pipe_ctx,
                    struct dp_packet *packet,
                    uint32_t seed,
                    uint32_t *hash)
{
    struct doca_flow_match field_values;
    struct ovs_16aligned_ip6_hdr *ip6;
    struct doca_flow_pipe *pre, *suf;
    struct udp_header *udp;
    struct tcp_header *tcp;
    int rv;

    ip6 = dp_packet_l3(packet);

    memset(&field_values.meta, 0, sizeof field_values.meta);
    doca_set_reg_val(&field_values.meta, REG_FIELD_SCRATCH, seed);
    if (ip6->ip6_nxt == IPPROTO_UDP) {
        pre = hash_pipe_ctx->hashes[HASH_TYPE_IPV6_UDP_PRE].pipe;
        suf = hash_pipe_ctx->hashes[HASH_TYPE_IPV6_UDP_SUF].pipe;
        udp = (struct udp_header *) (ip6 + 1);
        field_values.outer.udp.l4_port.src_port = (OVS_FORCE doca_be16_t) udp->udp_src;
        field_values.outer.udp.l4_port.dst_port = (OVS_FORCE doca_be16_t) udp->udp_dst;
        field_values.outer.l4_type_ext = DOCA_FLOW_L4_TYPE_EXT_UDP;
    } else if (ip6->ip6_nxt == IPPROTO_TCP) {
        pre = hash_pipe_ctx->hashes[HASH_TYPE_IPV6_TCP_PRE].pipe;
        suf = hash_pipe_ctx->hashes[HASH_TYPE_IPV6_TCP_SUF].pipe;
        tcp = (struct tcp_header *) (ip6 + 1);
        field_values.outer.tcp.l4_port.src_port = (OVS_FORCE doca_be16_t) tcp->tcp_src;
        field_values.outer.tcp.l4_port.dst_port = (OVS_FORCE doca_be16_t) tcp->tcp_dst;
        field_values.outer.l4_type_ext = DOCA_FLOW_L4_TYPE_EXT_TCP;
    } else {
        pre = hash_pipe_ctx->hashes[HASH_TYPE_IPV6_L3_PRE].pipe;
        suf = hash_pipe_ctx->hashes[HASH_TYPE_IPV6_L3_SUF].pipe;
    }

    field_values.outer.l3_type = DOCA_FLOW_L3_TYPE_IP6;
    memcpy(&field_values.outer.ip6.src_ip, &ip6->ip6_src, sizeof ip6->ip6_src);
    memcpy(&field_values.outer.ip6.dst_ip, &ip6->ip6_dst, sizeof ip6->ip6_dst);

    rv = doca_flow_pipe_calc_hash(pre, &field_values, hash);

    if (rv) {
        return rv;
    }

    memset(&field_values.meta, 0, sizeof field_values.meta);
    doca_set_reg_val(&field_values.meta, REG_FIELD_SCRATCH, *hash);

    return doca_flow_pipe_calc_hash(suf, &field_values, hash);
}

static int
dpdk_offload_doca_packet_hw_hash(struct netdev *netdev,
                                 struct dp_packet *packet,
                                 uint32_t seed,
                                 uint32_t *hash)
{
    struct doca_hash_pipe_ctx *hash_pipe_ctx;
    struct doca_offload_esw_ctx *esw_ctx;
    ovs_be16 eth_proto;
    int rv = -1;

    esw_ctx = doca_offload_esw_ctx_get(netdev);
    if (!esw_ctx) {
        return -1;
    }
    hash_pipe_ctx = esw_ctx->hash_pipe_ctx;

    parse_tcp_flags(packet, &eth_proto, NULL, NULL);
    if (eth_proto == htons(ETH_TYPE_IP)) {
        rv = ipv4_packet_hw_hash(hash_pipe_ctx, packet, seed, hash);
    } else if (eth_proto == htons(ETH_TYPE_IPV6)) {
        rv = ipv6_packet_hw_hash(hash_pipe_ctx, packet, seed, hash);
    } else {
        /* Only IPv4/IPv6 are supported. */
        return -1;
    }

    if (rv) {
        *hash = 0;
    } else {
        *hash &= 0x0000FFFF;
        *hash |= 0xd0ca0000;
    }
    return rv;
}

/* Examines the VNI (VXLAN Network Identifier) and destination MAC for
 * specific values to determine whether the packet contains a VXLAN header
 * used for metadata.
 */
static bool
dpdk_offload_doca_has_meta_header(struct dp_packet *packet)
{
    uint32_t val;

    if (!dpdk_offload_get_reg_field(packet, REG_FIELD_META_PUSH, &val)) {
        return false;
    }

    return !!val;
}

/* Parse the hardware register values embedded within the encapsulated
 * IP header. Extract the source (src) and destination (dst) IP addresses,
 * which hold registers of index 0 and 1. Additionally, it decapsulates
 * the VXLAN header that was used to encapsulate and convey these metadata.
 */
static bool
dpdk_offload_doca_packet_parse_meta(struct dp_packet *packet, struct reg_tags *reg_tags)
{
    struct vxlan_meta_header *meta_hdr;

    if (!dpdk_offload_doca_has_meta_header(packet)) {
        return false;
    }

    meta_hdr = dp_packet_data(packet);

    /* Parse reg tags */
    reg_tags->ct_state = meta_hdr->eth_dst.reg_field_ct_state;
    reg_tags->ct_zone = ntohs(meta_hdr->eth_dst.reg_field_ct_zone);
    reg_tags->ct_mark = ntohl(get_16aligned_be32(&meta_hdr->reg_field_ct_mark));
    reg_tags->ct_label = ntohl(get_16aligned_be32(&meta_hdr->reg_field_ct_label_id));

    /* Drop vxlan header used for storing metadata */
    dp_packet_pull(packet, sizeof *meta_hdr);

    return true;
}

static bool
dpdk_offload_doca_get_pkt_recover_info(struct dp_packet *p,
                                       struct dpdk_offload_recovery_info *info)
{
    memset(info, 0, sizeof *info);

    if (!dpdk_offload_doca_packet_parse_meta(p, &info->reg_tags)) {
        return false;
    }

    if (dpdk_offload_get_reg_field(p, REG_FIELD_FLOW_INFO, &info->flow_miss_id)) {
        if (info->flow_miss_id == SAMPLE_FLOW_MARK) {
            p->md.sample = true;
            memset(info, 0, sizeof *info);
            return true;
        }
        if (info->flow_miss_id == NONSAMPLE_FLOW_MARK) {
            memset(info, 0, sizeof *info);
            return true;
        }
        dp_packet_set_flow_mark(p, info->flow_miss_id);
        dpdk_offload_get_reg_field(p, REG_FIELD_DP_HASH, &info->dp_hash);
    }

    return true;
}

static void
dpdk_offload_doca_update_stats(struct dpif_flow_stats *stats,
                               struct dpif_flow_attrs *attrs,
                               struct rte_flow_query_count *query)
{
    if (attrs) {
        attrs->dp_layer = "doca";
    }

    if (!stats || !query) {
        return;
    }

    if (stats->n_packets != query->hits) {
        query->hits_set = 1;
        query->bytes_set = 1;
    }

    stats->n_packets = query->hits;
    stats->n_bytes = query->bytes;
}

static int
doca_offload_query_conn_dirs(struct netdev *netdevs[CT_DIR_NUM],
                             struct conn *conn,
                             struct dpif_flow_stats *stats,
                             struct dpif_flow_attrs *attrs,
                             struct rte_flow_error *error,
                             long long int now)
{
    struct rte_flow_query_count query = { .reset = 1 };
    struct ct_offload_handle *coh;
    int ret;

    if (attrs) {
        attrs->dp_extra_info = NULL;
        attrs->offloaded = true;
        attrs->dp_layer = "doca";
    }

    coh = conntrack_offload_get(conn);
    for (int dir = 0; dir < CT_DIR_NUM; dir++) {
        struct conn_item *ci = &coh->dir[dir].conn_item;
        struct ovs_shared_ctx *ctx;

        if (!coh->act_resources.shared_count_ctx) {
            ret = doca_query_count(netdevs[dir], ci->doca_flow, &query, error);
        } else {
            ctx = coh->act_resources.shared_count_ctx;
            ret = dpdk_offload_doca_shared_query(ctx, &query, error);
        }

        if (ret == 0) {
            break;
        }

        VLOG_DBG_RL(&rl, "%s: Failed to query conn %p",
                    netdev_get_name(netdevs[dir]), conn);
    }

    if (ret == 0) {
        dpdk_offload_doca_update_stats(&coh->stats, attrs, &query);
        if (query.hits_set && query.hits) {
            coh->stats.used = now;
        }

        if (stats) {
            memcpy(stats, &coh->stats, sizeof *stats);
        }
    }

    return ret;
}

static int
doca_offload_query_conn_doca_ct(struct netdev *netdevs[CT_DIR_NUM],
                                struct conn *conn,
                                struct dpif_flow_stats *stats,
                                struct dpif_flow_attrs *attrs,
                                struct rte_flow_error *error,
                                long long int now)
{
    struct doca_flow_resource_query queries[CT_DIR_NUM];
    struct doca_offload_esw_ctx *esw;
    struct ct_offload_handle *coh;
    struct netdev *netdev;
    struct conn_item *ci;
    uint64_t last_hit_s;
    unsigned int qid;
    doca_error_t err;

    if (attrs) {
        attrs->dp_extra_info = NULL;
        attrs->offloaded = true;
        attrs->dp_layer = "doca";
    }

    netdev = netdevs[CT_DIR_INIT] ? netdevs[CT_DIR_INIT]
                                  : netdevs[CT_DIR_REP];
    if (!netdev) {
        error->type = RTE_FLOW_ERROR_TYPE_UNSPECIFIED;
        error->message = "Missing netdev";
        return -ENODEV;
    }

    coh = conntrack_offload_get(conn);
    ci = &coh->dir[0].conn_item;
    qid = doca_offload_ct_queue_id(coh->insertion_tid);
    esw = doca_offload_esw_ctx_get(netdev);

    if (ci->doca_flow == NULL) {
        error->type = RTE_FLOW_ERROR_TYPE_HANDLE;
        error->message = "missing entry handle";
        return -1;
    }

    err = doca_flow_ct_query_entry(qid, esw->ct.pipe,
                                   DOCA_FLOW_CT_ENTRY_FLAGS_NO_WAIT,
                                   ci->doca_flow, &queries[CT_DIR_INIT],
                                   &queries[CT_DIR_REP], &last_hit_s);
    if (DOCA_IS_ERROR(err)) {
        VLOG_DBG_RL(&rl, "%s/%s: Failed to query conn %p",
                    netdev_get_name(netdevs[CT_DIR_INIT]),
                    netdev_get_name(netdevs[CT_DIR_REP]),
                    conn);
        error->type = RTE_FLOW_ERROR_TYPE_UNSPECIFIED;
        error->message = doca_error_get_descr(err);
        return -1;
    }

    if (coh->stats.n_packets != queries[CT_DIR_INIT].counter.total_pkts) {
        coh->stats.used = now;
    }
    coh->stats.n_packets = queries[CT_DIR_INIT].counter.total_pkts;
    coh->stats.n_bytes = queries[CT_DIR_INIT].counter.total_bytes;

    if (stats) {
        memcpy(stats, &coh->stats, sizeof *stats);
    }

    return 0;
}

static int
dpdk_offload_doca_query_conn(struct netdev *netdevs[CT_DIR_NUM],
                             struct conn *conn,
                             struct dpif_flow_stats *stats,
                             struct dpif_flow_attrs *attrs,
                             struct rte_flow_error *error,
                             long long int now)
{
    if (ovs_doca_flow_ct_enabled) {
        return doca_offload_query_conn_doca_ct(netdevs, conn, stats, attrs, error, now);
    } else {
        return doca_offload_query_conn_dirs(netdevs, conn, stats, attrs, error, now);
    }
}

void
doca_offload_fixed_rule_uninit(struct doca_offload_esw_ctx *esw,
                               struct netdev *netdev, struct fixed_rule *fr)
{
    if (!fr->doh.dfh.flow) {
        return;
    }

    destroy_dpdk_offload_handle(esw, netdev, &fr->doh, AUX_QUEUE, NULL);
    fr->doh.dfh.flow = NULL;
    /* Ignore potential errors here, as even if the queue completion
     * failed, the entry removal would still be issued. The caller
     * requires knowing so. */
    doca_offload_complete_queue_esw(esw, AUX_QUEUE, true);
}

static int
doca_create_ct_zone_revisit_rule(struct netdev *netdev, uint32_t group,
                                 uint16_t zone, int nat,
                                 struct dpdk_offload_handle *doh)
{
    struct doca_flow_handle_resources flow_res;
    struct doca_pipe_group_ctx *next_group_ctx;
    uint32_t ct_state_spec, ct_state_mask;
    struct doca_flow_handle *hndl;
    struct rte_flow_error error;
    struct doca_flow_match mask;
    struct doca_flow_match spec;
    struct doca_flow_fwd fwd;

    memset(&flow_res, 0, sizeof flow_res);
    memset(&mask, 0, sizeof mask);
    memset(&spec, 0, sizeof spec);
    memset(&fwd, 0, sizeof fwd);

    /* If the zone is the same, and already visited ct/ct-nat, skip
     * ct/ct-nat and jump directly to post-ct.
     */
    ct_state_spec = OVS_CS_F_TRACKED;
    if (nat) {
        ct_state_spec |= OVS_CS_F_NAT_MASK;
    }
    ct_state_mask = ct_state_spec;

    doca_set_reg_val_mask(&spec.meta, &mask.meta, REG_FIELD_CT_ZONE, zone);
    doca_set_reg_val(&spec.meta, REG_FIELD_CT_STATE, ct_state_spec);
    doca_set_reg_val(&mask.meta, REG_FIELD_CT_STATE, ct_state_mask);

    next_group_ctx = doca_pipe_group_ctx_ref(netdev, POSTCT_TABLE_ID);
    if (!next_group_ctx) {
        return -1;
    }

    fwd.type = DOCA_FLOW_FWD_PIPE;
    fwd.next_pipe = doca_pipe_group_get_pipe(next_group_ctx);
    flow_res.next_group_ctx = next_group_ctx;

    hndl = create_doca_flow_handle(netdev, AUX_QUEUE,
                                   DPDK_OFFLOAD_PRIORITY_HIGH, group, 0, &spec,
                                   &mask, NULL, NULL, NULL, NULL, &fwd,
                                   &flow_res, doh, &error);
    if (!hndl) {
        doca_pipe_group_ctx_unref(next_group_ctx);
        return -1;
    }
    return 0;
}

static int
doca_create_ct_zone_uphold_rule(struct netdev *netdev,
                                struct doca_offload_esw_ctx *ctx,
                                uint32_t group,
                                uint16_t zone,
                                enum ct_zone_cls_flow_type type,
                                struct dpdk_offload_handle *doh)
{
    struct doca_flow_actions dacts, dacts_masks;
    struct doca_flow_handle_resources flow_res;
    struct doca_flow_handle *hndl;
    struct doca_flow_match spec;
    struct doca_flow_match mask;
    struct rte_flow_error error;
    struct doca_flow_fwd fwd;

    memset(&dacts_masks, 0, sizeof dacts_masks);
    memset(&flow_res, 0, sizeof flow_res);
    memset(&dacts, 0, sizeof dacts);
    memset(&fwd, 0, sizeof fwd);
    memset(&spec, 0, sizeof spec);
    memset(&mask, 0, sizeof mask);

    if (type == CT_ZONE_FLOW_UPHOLD_IP6_TCP ||
        type == CT_ZONE_FLOW_UPHOLD_IP6_UDP) {
        spec.outer.l3_type = DOCA_FLOW_L3_TYPE_IP6;
        mask.outer.l3_type = DOCA_FLOW_L3_TYPE_IP6;
    } else {
        spec.outer.l3_type = DOCA_FLOW_L3_TYPE_IP4;
        mask.outer.l3_type = DOCA_FLOW_L3_TYPE_IP4;
    }

    if (type == CT_ZONE_FLOW_UPHOLD_IP4_TCP ||
        type == CT_ZONE_FLOW_UPHOLD_IP6_TCP) {
        spec.outer.l4_type_ext = DOCA_FLOW_L4_TYPE_EXT_TCP;
        mask.outer.l4_type_ext = DOCA_FLOW_L4_TYPE_EXT_TCP;
        /* Ensure that none of SYN | RST | FIN flag is set in
         * packets going to CT: they must miss and go to SW. */
        mask.outer.tcp.flags = TCP_SYN | TCP_RST | TCP_FIN;
    } else {
        spec.outer.l4_type_ext = DOCA_FLOW_L4_TYPE_EXT_UDP;
        mask.outer.l4_type_ext = DOCA_FLOW_L4_TYPE_EXT_UDP;
    }

    doca_set_reg_val_mask(&dacts.meta, &dacts_masks.meta, REG_FIELD_CT_ZONE, zone);

    fwd.type = DOCA_FLOW_FWD_PIPE;
    fwd.next_pipe = doca_get_ct_pipe(ctx, &spec);
    flow_res.next_group_ctx = NULL;

    hndl = create_doca_flow_handle(netdev, AUX_QUEUE,
                                   DPDK_OFFLOAD_PRIORITY_MED, group, 0, &spec,
                                   &mask, &dacts, &dacts_masks, NULL, NULL,
                                   &fwd, &flow_res, doh, &error);
    if (!hndl) {
        return -1;
    }
    return 0;
}

static void
doca_ct_zone_rules_uninit(struct doca_offload_esw_ctx *esw_ctx, uint32_t zone)
    OVS_REQUIRES(esw_ctx->ct_zones_mutex)
{
    struct fixed_rule *fr;
    int nat, i;

    if (netdev_is_zone_tables_disabled()) {
        VLOG_ERR("Disabling ct zones is not supported with doca");
        return;
    }

    for (nat = 0; nat < 2; nat++) {
        for (i = 0; i < CT_ZONE_FLOWS_NUM; i++) {
            fr = &esw_ctx->ct_zones[zone].entries[nat][i];
            doca_offload_fixed_rule_uninit(esw_ctx, NULL, fr);
        }
    }
}

static void
doca_ct_zone_ctx_uninit(struct doca_offload_esw_ctx *esw_ctx, uint32_t zone)
    OVS_REQUIRES(esw_ctx->ct_zones_mutex)
{
    struct doca_pipe_group_ctx *ct_zone_nat_ctx;
    struct doca_pipe_group_ctx *ct_zone_ctx;

    ct_zone_nat_ctx = (struct doca_pipe_group_ctx *)
        ovsrcu_get(struct doca_pipe_group_ctx *,
                   &esw_ctx->ct_zones[zone].nat_ctx);
    ct_zone_ctx = (struct doca_pipe_group_ctx *)
        ovsrcu_get(struct doca_pipe_group_ctx *,
                   &esw_ctx->ct_zones[zone].plain_ctx);

    doca_pipe_group_ctx_unref(ct_zone_ctx);
    doca_pipe_group_ctx_unref(ct_zone_nat_ctx);
    ovsrcu_set(&esw_ctx->ct_zones[zone].plain_ctx, NULL);
    ovsrcu_set(&esw_ctx->ct_zones[zone].nat_ctx, NULL);
}

static int
doca_ct_zone_rules_init(struct netdev *netdev,
                        struct doca_offload_esw_ctx *esw_ctx, uint32_t zone)
    OVS_REQUIRES(esw_ctx->ct_zones_mutex)
{
    struct fixed_rule *fr;
    uint32_t base_group;
    int nat;

    for (nat = 0; nat < 2; nat++) {
        base_group = nat ? CTNAT_TABLE_ID : CT_TABLE_ID;
        /* If the zone is already set, then CT for this zone has already
         * been executed: skip to post-ct. */

        fr = &esw_ctx->ct_zones[zone].entries[nat][CT_ZONE_FLOW_REVISIT];
        if (doca_create_ct_zone_revisit_rule(netdev, base_group + zone,
                                             zone, nat, &fr->doh)) {
            goto err;
        }

        /* Otherwise, set the zone and go to CT/CT-NAT. */

        fr = &esw_ctx->ct_zones[zone].entries[nat][CT_ZONE_FLOW_UPHOLD_IP4_UDP];
        if (doca_create_ct_zone_uphold_rule(netdev, esw_ctx,
                                            base_group + zone, zone,
                                            CT_ZONE_FLOW_UPHOLD_IP4_UDP,
                                            &fr->doh)) {
            goto err;
        }

        fr = &esw_ctx->ct_zones[zone].entries[nat][CT_ZONE_FLOW_UPHOLD_IP4_TCP];
        if (doca_create_ct_zone_uphold_rule(netdev, esw_ctx,
                                            base_group + zone,
                                            zone,
                                            CT_ZONE_FLOW_UPHOLD_IP4_TCP,
                                            &fr->doh)) {
            goto err;
        }

        if (!conntrack_offload_ipv6_is_enabled()) {
            continue;
        }

        fr = &esw_ctx->ct_zones[zone].entries[nat][CT_ZONE_FLOW_UPHOLD_IP6_UDP];
        if (doca_create_ct_zone_uphold_rule(netdev, esw_ctx,
                                            base_group + zone,
                                            zone,
                                            CT_ZONE_FLOW_UPHOLD_IP6_UDP,
                                            &fr->doh)) {
            goto err;
        }

        fr = &esw_ctx->ct_zones[zone].entries[nat][CT_ZONE_FLOW_UPHOLD_IP6_TCP];
        if (doca_create_ct_zone_uphold_rule(netdev, esw_ctx,
                                            base_group + zone,
                                            zone,
                                            CT_ZONE_FLOW_UPHOLD_IP6_TCP,
                                            &fr->doh)) {
            goto err;
        }
    }

    return 0;

err:
    doca_ct_zone_rules_uninit(esw_ctx, zone);
    doca_ct_zone_ctx_uninit(esw_ctx, zone);
    return -1;
}

static int
doca_ct_zone_ctx_init(struct netdev *netdev, struct doca_offload_esw_ctx *esw_ctx,
                      uint32_t zone)
    OVS_REQUIRES(esw_ctx->ct_zones_mutex)
{
    struct ct_zone_ctx *ct_zone_ctx = &esw_ctx->ct_zones[zone];

    ovsrcu_set(&ct_zone_ctx->plain_ctx,
               doca_pipe_group_ctx_ref(netdev, CT_TABLE_ID + zone));
    if (!ovsrcu_get(struct doca_pipe_group_ctx *,
                    &esw_ctx->ct_zones[zone].plain_ctx)) {
        return -1;
    }

    ovsrcu_set(&ct_zone_ctx->nat_ctx,
               doca_pipe_group_ctx_ref(netdev, CTNAT_TABLE_ID + zone));
    if (!ovsrcu_get(struct doca_pipe_group_ctx *,
                    &esw_ctx->ct_zones[zone].nat_ctx)) {
        doca_pipe_group_ctx_unref(ovsrcu_get(struct doca_pipe_group_ctx *,
                                  &esw_ctx->ct_zones[zone].plain_ctx));
        ovsrcu_set(&esw_ctx->ct_zones[zone].plain_ctx, NULL);
        return -1;
    }

    return 0;
}

static int
doca_ct_zone_init(struct netdev *netdev, struct doca_offload_esw_ctx *esw_ctx,
                  uint32_t zone_group)
    OVS_EXCLUDED(esw_ctx->ct_zones_mutex)
{
    uint32_t zone = zone_group & reg_fields[REG_FIELD_CT_ZONE].mask;
    int ret = 0;

    if (!doca_pipe_group_is_ct_zone_group_id(zone_group)) {
        return -1;
    }

    if (ovsrcu_get(struct doca_pipe_group_ctx *,
                   &esw_ctx->ct_zones[zone].plain_ctx) != NULL) {
        return ret;
    }

    ovs_mutex_lock(&esw_ctx->ct_zones_mutex);
    ovs_assert(!netdev_is_zone_tables_disabled());
    /* Read again after locking to resolve concurrent initialization. */
    if (ovsrcu_get(struct doca_pipe_group_ctx *,
                   &esw_ctx->ct_zones[zone].plain_ctx) != NULL) {
        goto unlock;
    }

    if (doca_ct_zone_ctx_init(netdev, esw_ctx, zone)) {
        ret = -1;
        goto unlock;
    }

    if (doca_ct_zone_rules_init(netdev, esw_ctx, zone)) {
        doca_ct_zone_ctx_uninit(esw_ctx, zone);
        ret = -1;
        goto unlock;
    }

unlock:
    ovs_mutex_unlock(&esw_ctx->ct_zones_mutex);
    return ret;
}

static void
doca_ct_zones_uninit(struct doca_offload_esw_ctx *esw_ctx)
    OVS_EXCLUDED(esw_ctx->ct_zones_mutex)
{
    uint32_t zone;

    if (netdev_is_zone_tables_disabled()) {
        VLOG_ERR("Disabling ct zones is not supported with doca");
        return;
    }

    ovs_mutex_lock(&esw_ctx->ct_zones_mutex);
    for (zone = 0; zone < NUM_CT_ZONES; zone++) {
        if (ovsrcu_get(struct doca_pipe_group_ctx *,
                       &esw_ctx->ct_zones[zone].plain_ctx) == NULL) {
            continue;
        }

        doca_ct_zone_rules_uninit(esw_ctx, zone);
        doca_ct_zone_ctx_uninit(esw_ctx, zone);
    }
    ovs_mutex_unlock(&esw_ctx->ct_zones_mutex);
}

static void
doca_ct_pipe_destroy(struct doca_basic_pipe_ctx *pipe_ctx)
{
    doca_basic_pipe_ctx_uninit(pipe_ctx);
}

static void
doca_ct_pipes_destroy(struct doca_offload_esw_ctx *ctx)
{
    struct doca_basic_pipe_ctx *pipe_ctx;
    int i, j;

    doca_ct_pipe_destroy(&ctx->ct_ip6_prefix);

    for (i = 0; i < NUM_CT_NW; i++) {
        for (j = 0; j < NUM_CT_TP; j++) {
            pipe_ctx = &ctx->ct_pipes[i][j];
            doca_ct_pipe_destroy(pipe_ctx);
        }
    }
}

static void
doca_basic_pipe_name(struct ds *s, struct netdev *netdev,
                     enum ct_nw_type nw_type,
                     enum ct_tp_type tp_type)
{
    ds_put_format(s, "OVS_BASIC_CT_PIPE_%d",
                  netdev_dpdk_get_esw_mgr_port_id(netdev));

    switch (nw_type) {
    case CT_NW_IP4:
        ds_put_cstr(s, "_IP4");
        break;
    case CT_NW_IP6:
        ds_put_cstr(s, "_IP6_SUFFIX");
        break;
    case NUM_CT_NW:
       OVS_NOT_REACHED();
    }

    switch (tp_type) {
    case CT_TP_UDP:
        ds_put_cstr(s, "_UDP");
        break;
    case CT_TP_TCP:
        ds_put_cstr(s, "_TCP");
        break;
    case NUM_CT_TP:
       OVS_NOT_REACHED();
    }
}

static struct doca_flow_match ct_matches[NUM_CT_NW][NUM_CT_TP] = {
    [CT_NW_IP4] = {
        [CT_TP_UDP] = {
            .outer.l3_type = DOCA_FLOW_L3_TYPE_IP4,
            .outer.ip4.src_ip = BE32_MAX,
            .outer.ip4.dst_ip = BE32_MAX,
            .outer.l4_type_ext = DOCA_FLOW_L4_TYPE_EXT_UDP,
            .outer.udp.l4_port.src_port = BE16_MAX,
            .outer.udp.l4_port.dst_port = BE16_MAX,
            /* For pure functionality this match is not needed, instead use it
             * as a WA to take advantage of the lower level optimization.
             */
            .parser_meta.port_meta = UINT32_MAX,
        },
        [CT_TP_TCP] = {
            .outer.l3_type = DOCA_FLOW_L3_TYPE_IP4,
            .outer.ip4.src_ip = BE32_MAX,
            .outer.ip4.dst_ip = BE32_MAX,
            .outer.l4_type_ext = DOCA_FLOW_L4_TYPE_EXT_TCP,
            .outer.tcp.l4_port.src_port = BE16_MAX,
            .outer.tcp.l4_port.dst_port = BE16_MAX,
            /* For pure functionality this match is not needed, instead use it
             * as a WA to take advantage of the lower level optimization.
             */
            .parser_meta.port_meta = UINT32_MAX,
        },
    },
    [CT_NW_IP6] = {
        [CT_TP_UDP] = {
            .outer.l3_type = DOCA_FLOW_L3_TYPE_IP6,
            .outer.ip6.dst_ip[0] = BE32_MAX,
            .outer.ip6.dst_ip[1] = BE32_MAX,
            .outer.ip6.dst_ip[2] = BE32_MAX,
            .outer.ip6.dst_ip[3] = BE32_MAX,
            .outer.l4_type_ext = DOCA_FLOW_L4_TYPE_EXT_UDP,
            .outer.udp.l4_port.src_port = BE16_MAX,
            .outer.udp.l4_port.dst_port = BE16_MAX,
            /* For pure functionality this match is not needed, instead use it
             * as a WA to take advantage of the lower level optimization.
             */
            .parser_meta.port_meta = UINT32_MAX,
        },
        [CT_TP_TCP] = {
            .outer.l3_type = DOCA_FLOW_L3_TYPE_IP6,
            .outer.ip6.dst_ip[0] = BE32_MAX,
            .outer.ip6.dst_ip[1] = BE32_MAX,
            .outer.ip6.dst_ip[2] = BE32_MAX,
            .outer.ip6.dst_ip[3] = BE32_MAX,
            .outer.l4_type_ext = DOCA_FLOW_L4_TYPE_EXT_TCP,
            .outer.tcp.l4_port.src_port = BE16_MAX,
            .outer.tcp.l4_port.dst_port = BE16_MAX,
            /* For pure functionality this match is not needed, instead use it
             * as a WA to take advantage of the lower level optimization.
             */
            .parser_meta.port_meta = UINT32_MAX,
        },
    },
};

static int
doca_ct_ip6_prefix_pipe_init(struct netdev *netdev,
                             struct doca_offload_esw_ctx *ctx)
{
    struct doca_flow_actions *actions_masks_list;
    struct doca_flow_match match, match_mask;
    struct doca_flow_actions actions_masks;
    struct doca_flow_actions *actions_list;
    struct doca_basic_pipe_ctx *pipe_ctx;
    struct doca_flow_actions actions;
    struct doca_flow_port *doca_port;
    struct doca_flow_pipe_cfg *cfg;
    struct doca_flow_fwd miss;
    struct doca_flow_fwd fwd;
    struct ds pipe_name;
    int ret;

    pipe_ctx = &ctx->ct_ip6_prefix;

    memset(&fwd, 0, sizeof fwd);
    memset(&miss, 0, sizeof miss);
    memset(&match, 0, sizeof match);
    memset(&actions, 0, sizeof actions);
    memset(&match_mask, 0, sizeof match_mask);
    memset(&actions_masks, 0, sizeof actions_masks);

    miss.type = DOCA_FLOW_FWD_PIPE;
    miss.next_pipe = doca_offload_get_misspath_pipe(ctx);

    actions_list = &actions;
    actions_masks_list = &actions_masks;

    doca_set_reg_mask(&actions.meta, REG_FIELD_SCRATCH);
    doca_set_reg_mask(&actions_masks.meta, REG_FIELD_SCRATCH);

    /* Set ip6 ct tuple match template (zone + src_ip + l4_type) */
    doca_set_reg_mask(&match.meta, REG_FIELD_CT_ZONE);
    doca_set_reg_mask(&match_mask.meta, REG_FIELD_CT_ZONE);
    match.parser_meta.port_meta = UINT32_MAX;
    match_mask.parser_meta.port_meta = UINT32_MAX;
    match.outer.l3_type = DOCA_FLOW_L3_TYPE_IP6;
    match_mask.outer.l3_type = DOCA_FLOW_L3_TYPE_IP6;
    memset(&match.outer.ip6.src_ip, 0xFF, sizeof match.outer.ip6.src_ip);
    memset(&match_mask.outer.ip6.src_ip, 0xFF, sizeof match.outer.ip6.src_ip);
    match.outer.ip6.next_proto = UINT8_MAX;
    match_mask.outer.ip6.next_proto = UINT8_MAX;

    doca_port = netdev_dpdk_doca_port_get(netdev);

    ds_init(&pipe_name);
    ds_put_format(&pipe_name, "OVS_BASIC_CT_PIPE_%d_IP6_PREFIX",
                  netdev_dpdk_get_esw_mgr_port_id(netdev));

    ret = doca_flow_pipe_cfg_create(&cfg, doca_flow_port_switch_get(doca_port));
    if (ret) {
        VLOG_ERR("%s: Could not create doca_flow_pipe_cfg", netdev_get_name(netdev));
        return -1;
    }
    if (doca_flow_pipe_cfg_set_name(cfg, ds_cstr(&pipe_name)) ||
        doca_flow_pipe_cfg_set_type(cfg, DOCA_FLOW_PIPE_BASIC) ||
        doca_flow_pipe_cfg_set_is_root(cfg, false) ||
        doca_flow_pipe_cfg_set_nr_entries(cfg, ovs_doca_max_ct_rules()) ||
        doca_flow_pipe_cfg_set_enable_strict_matching(cfg, true) ||
        doca_flow_pipe_cfg_set_match(cfg, &match, &match_mask) ||
        doca_flow_pipe_cfg_set_actions(cfg, &actions_list, &actions_masks_list, NULL, 1)) {
        VLOG_ERR("%s: Could not set doca_flow_pipe_cfg", netdev_get_name(netdev));
        ret = -1;
        goto out;
    }

    /* Next pipe will be determined per connection based on L4 type */
    fwd.type = DOCA_FLOW_FWD_PIPE;
    fwd.next_pipe = NULL;

    ret = doca_flow_pipe_create(cfg, &fwd, &miss, &pipe_ctx->pipe);
    ds_destroy(&pipe_name);
    if (ret) {
        VLOG_ERR("%s: Failed to create ct prefix basic pipe: %d (%s)",
                 netdev_get_name(netdev), ret,
                 doca_error_get_descr(ret));
        goto out;
    }

out:
    doca_flow_pipe_cfg_destroy(cfg);
    return ret;
}

static void
build_ct_actions_templates(struct doca_flow_actions *actions,
                           struct doca_flow_actions *actions_masks,
                           enum ct_nw_type nw_type, enum ct_tp_type tp_type, bool nat)
{
    enum dpdk_reg_id regs[] = {
        REG_FIELD_CT_STATE,
        REG_FIELD_CT_MARK,
        REG_FIELD_CT_LABEL_ID,
    };

    memset(actions, 0, sizeof *actions);
    if (actions_masks) {
        memset(actions_masks, 0, sizeof *actions_masks);
    }

    for (size_t i = 0; i < ARRAY_SIZE(regs); i++) {
        /* Use 0xFFs values to set meta.u32 in the action upon pipe create
         * and have the mask in the actions_masks
         */
        doca_offload_set_reg_template(&actions->meta, regs[i]);
        if (actions_masks && actions_masks != actions) {
            doca_set_reg_mask(&actions_masks->meta, regs[i]);
        }
    }

    if (nat) {
        struct doca_flow_header_format *outer = &actions->outer;

        if (nw_type == CT_NW_IP4) {
            outer->l3_type = DOCA_FLOW_L3_TYPE_IP4;
            outer->ip4.src_ip = BE32_MAX;
            outer->ip4.dst_ip = BE32_MAX;
        } else if (nw_type == CT_NW_IP6) {
            outer->l3_type = DOCA_FLOW_L3_TYPE_IP6;
            memset(&outer->ip6.src_ip, UINT8_MAX, sizeof outer->ip6.src_ip) ;
            memset(&outer->ip6.dst_ip, UINT8_MAX, sizeof outer->ip6.dst_ip) ;
        } else {
            OVS_NOT_REACHED();
        }

        if (tp_type == CT_TP_UDP) {
            outer->l4_type_ext = DOCA_FLOW_L4_TYPE_EXT_UDP;
            outer->udp.l4_port.src_port = BE16_MAX;
            outer->udp.l4_port.dst_port = BE16_MAX;
        } else {
            outer->l4_type_ext = DOCA_FLOW_L4_TYPE_EXT_TCP;
            outer->tcp.l4_port.src_port = BE16_MAX;
            outer->tcp.l4_port.dst_port = BE16_MAX;
        }
        if (actions_masks && actions_masks != actions) {
            memcpy(&actions_masks->outer, &actions->outer, sizeof actions_masks->outer);
        }
    }
}

static int
doca_ct_pipe_init(struct netdev *netdev, struct doca_offload_esw_ctx *ctx,
                  enum ct_nw_type nw_type, enum ct_tp_type tp_type)
{
    struct doca_pipe_group_ctx *fwd_group_ctx = NULL;
    struct doca_flow_actions *actions_masks_list;
    struct doca_flow_actions actions_masks;
    struct doca_flow_actions *actions_list;
    struct doca_basic_pipe_ctx *pipe_ctx;
    struct doca_flow_match match_mask;
    struct doca_flow_actions actions;
    struct doca_flow_port *doca_port;
    struct doca_flow_monitor monitor;
    struct doca_flow_pipe_cfg *cfg;
    struct doca_flow_fwd miss;
    struct doca_flow_fwd fwd;
    struct ds pipe_name;
    int ret;

    pipe_ctx = &ctx->ct_pipes[nw_type][tp_type];

    /* Do not re-init a pipe if already done. */
    if (pipe_ctx->pipe != NULL) {
        return 0;
    }

    /* Do not initialize IPv6 pipes if not enabled. */
    if (!conntrack_offload_ipv6_is_enabled() &&
        nw_type == CT_NW_IP6) {
        return 0;
    }

    memset(&cfg, 0, sizeof cfg);
    memset(&fwd, 0, sizeof fwd);
    memset(&miss, 0, sizeof miss);
    memset(&monitor, 0, sizeof monitor);

    ds_init(&pipe_name);
    doca_basic_pipe_name(&pipe_name, netdev, nw_type, tp_type);

    actions_list = &actions;
    actions_masks_list = &actions_masks;
    /* CT offloads on basic pipes *always* modify the packet header,
     * even for plain-CT. In such case, all fields are written with
     * their own value. */
    build_ct_actions_templates(&actions, &actions_masks, nw_type, tp_type, true);

    /* Finalize the match templates. */
    doca_set_reg_mask(&ct_matches[CT_NW_IP4][CT_TP_UDP].meta, REG_FIELD_CT_ZONE);
    doca_set_reg_mask(&ct_matches[CT_NW_IP4][CT_TP_TCP].meta, REG_FIELD_CT_ZONE);
    /* Add prefix id match to IPv6 pipes */
    doca_set_reg_mask(&ct_matches[CT_NW_IP6][CT_TP_UDP].meta, REG_FIELD_SCRATCH);
    doca_set_reg_mask(&ct_matches[CT_NW_IP6][CT_TP_TCP].meta, REG_FIELD_SCRATCH);
    /* The mask is identical to the match itself. */
    match_mask = ct_matches[nw_type][tp_type];
    doca_port = netdev_dpdk_doca_port_get(netdev);

    monitor.counter_type = DOCA_FLOW_RESOURCE_TYPE_SHARED;
    monitor.shared_counter.shared_counter_id = UINT32_MAX;

    ret = doca_flow_pipe_cfg_create(&cfg, doca_flow_port_switch_get(doca_port));
    if (ret) {
        VLOG_ERR("%s: Could not create doca_flow_pipe_cfg", netdev_get_name(netdev));
        return -1;
    }
    if (doca_flow_pipe_cfg_set_name(cfg, ds_cstr(&pipe_name)) ||
        doca_flow_pipe_cfg_set_type(cfg, DOCA_FLOW_PIPE_BASIC) ||
        doca_flow_pipe_cfg_set_is_root(cfg, false) ||
        doca_flow_pipe_cfg_set_nr_entries(cfg, ovs_doca_max_ct_rules()) ||
        doca_flow_pipe_cfg_set_enable_strict_matching(cfg, true) ||
        doca_flow_pipe_cfg_set_match(cfg, &ct_matches[nw_type][tp_type], &match_mask) ||
        doca_flow_pipe_cfg_set_actions(cfg, &actions_list, &actions_masks_list, NULL, 1) ||
        doca_flow_pipe_cfg_set_monitor(cfg, &monitor)) {
        VLOG_ERR("%s: Could not set doca_flow_pipe_cfg", netdev_get_name(netdev));
        ret = -1;
        goto out;
    }

    fwd_group_ctx = doca_pipe_group_ctx_ref(netdev, POSTCT_TABLE_ID);
    if (fwd_group_ctx == NULL) {
        VLOG_ERR("%s: Failed to take a reference on post-ct table",
                 netdev_get_name(netdev));
        return -1;
    }
    fwd.type = DOCA_FLOW_FWD_PIPE;
    fwd.next_pipe = doca_pipe_group_get_pipe(fwd_group_ctx);

    miss.type = DOCA_FLOW_FWD_PIPE;
    miss.next_pipe = doca_offload_get_misspath_pipe(ctx);

    ret = doca_flow_pipe_create(cfg, &fwd, &miss, &pipe_ctx->pipe);
    if (ret) {
        VLOG_ERR("%s: Failed to create basic pipe: %d (%s)",
                 netdev_get_name(netdev), ret,
                 doca_error_get_descr(ret));
        ret = -1;
        goto out;
    }

    pipe_ctx->fwd_group_ctx = fwd_group_ctx;

out:
    doca_flow_pipe_cfg_destroy(cfg);
    ds_destroy(&pipe_name);
    if (ret) {
        doca_pipe_group_ctx_unref(fwd_group_ctx);
    }
    return ret;
}

static int
doca_ct_pipes_init(struct netdev *netdev, struct doca_offload_esw_ctx *ctx)
{
    int i, j;

    for (i = 0; i < NUM_CT_NW; i++) {
        for (j = 0; j < NUM_CT_TP; j++) {
            if (doca_ct_pipe_init(netdev, ctx, i, j)) {
                goto error;
            }
        }
    }

    if (conntrack_offload_ipv6_is_enabled()) {
        if (doca_ct_ip6_prefix_pipe_init(netdev, ctx)) {
            goto error;
        }
    }

    return 0;

error:
    /* Rollback any pipe creation. */
    doca_ct_pipes_destroy(ctx);
    return -1;
}

static void
doca_offload_ct_pipe_uninit(struct doca_offload_esw_ctx *esw)
{
    doca_pipe_group_ctx_unref(esw->ct.post_ct_ctx);
    esw->ct.post_ct_ctx = NULL;

    doca_flow_pipe_destroy(esw->ct.pipe);
    esw->ct.pipe = NULL;
}

static int
doca_offload_ct_pipe_init(struct netdev *netdev, struct doca_offload_esw_ctx *esw)
{
    struct doca_flow_actions *actions_masks_arr[NUM_CT_ACTIONS];
    struct doca_flow_actions actions_masks[NUM_CT_ACTIONS];
    struct doca_flow_actions *actions_arr[NUM_CT_ACTIONS];
    struct doca_flow_actions actions[NUM_CT_ACTIONS];
    struct doca_pipe_group_ctx *post_ct_ctx = NULL;
    doca_error_t err = DOCA_ERROR_UNKNOWN;
    struct doca_flow_pipe_cfg *cfg = NULL;
    struct doca_flow_pipe *ct_pipe = NULL;
    struct doca_flow_fwd fwd[NUM_CT_FWD];
    struct doca_flow_match match;
    struct doca_flow_port *port;
    char pipe_name[50];

    memset(&fwd, 0, sizeof fwd);
    memset(&match, 0, sizeof match);

    /* Create forward to post-ct. */
    post_ct_ctx = doca_pipe_group_ctx_ref(netdev, POSTCT_TABLE_ID);
    if (post_ct_ctx == NULL) {
        VLOG_ERR("%s: Failed to take a reference on post-ct table",
                 netdev_get_name(netdev));
        goto out;
    }
    fwd[CT_FWD_POST].type = DOCA_FLOW_FWD_PIPE;
    fwd[CT_FWD_POST].next_pipe = doca_pipe_group_get_pipe(post_ct_ctx);

    /* Create forward to miss handling. */
    fwd[CT_FWD_MISS].type = DOCA_FLOW_FWD_PIPE;
    fwd[CT_FWD_MISS].next_pipe = doca_offload_get_misspath_pipe(esw);

    port = doca_flow_port_switch_get(netdev_dpdk_doca_port_get(netdev));

    for (int nw = 0; nw < NUM_CT_NW; nw++) {
        for (int nat = 0; nat < 2; nat++) {
            size_t idx;

            idx = CT_ACTIONS_IDX(nw, nat);
            ovs_assert(idx < ARRAY_SIZE(actions));
            /* Transport type is irrelevant for the DOCA-CT pipe. */
            build_ct_actions_templates(&actions[idx], &actions_masks[idx], nw, 0, nat);
            actions[idx].action_idx = idx;
            actions_arr[idx] = &actions[idx];
            actions_masks[idx].action_idx = idx;
            actions_masks_arr[idx] = &actions_masks[idx];
        }
    }

    snprintf(pipe_name, sizeof pipe_name, "OVS_DOCA_CT_PIPE_%s", netdev_get_name(netdev));
    if (doca_flow_pipe_cfg_create(&cfg, port) ||
        doca_flow_pipe_cfg_set_name(cfg, pipe_name) ||
        doca_flow_pipe_cfg_set_match(cfg, &match, NULL) ||
        doca_flow_pipe_cfg_set_actions(cfg, actions_arr, actions_masks_arr, NULL,
                                       ARRAY_SIZE(actions)) ||
        doca_flow_pipe_cfg_set_type(cfg, DOCA_FLOW_PIPE_CT)) {
        VLOG_ERR("%s: Could not create DOCA-CT pipe configuration", netdev_get_name(netdev));
        goto out;
    }

    err = doca_flow_pipe_create(cfg, &fwd[CT_FWD_POST], &fwd[CT_FWD_MISS], &ct_pipe);
    if (DOCA_IS_ERROR(err)) {
        VLOG_ERR("%s: Failed to create CT pipe: %s",
                 netdev_get_name(netdev), doca_error_get_descr(err));
        goto out;
    }

    esw->ct.pipe = ct_pipe;
    esw->ct.post_ct_ctx = post_ct_ctx;

out:
    if (DOCA_IS_ERROR(err)) {
        doca_flow_pipe_destroy(ct_pipe);
        doca_pipe_group_ctx_unref(post_ct_ctx);
    }

    if (cfg) {
        doca_flow_pipe_cfg_destroy(cfg);
    }
    return -((int) DOCA_IS_ERROR(err));
}

/* Init the shared counter id map for the first
 * eswitch context that requests it. This is the only
 * eswitch that will support CT offload for now.
 * After DOCA adds proper support this limitation should
 * be lifted and support shared counters for every eswitch
 * will be added.
 */
static void
shared_counter_id_pool_init(struct doca_offload_esw_ctx *ctx)
{
    static struct ovsthread_once init_once = OVSTHREAD_ONCE_INITIALIZER;
    uint32_t base_id;

    if (ovsthread_once_start(&init_once)) {
        esw_id_pool = id_fpool_create(1, 0, OVS_DOCA_MAX_ESW);
        ovsthread_once_done(&init_once);
    }
    if (!esw_id_pool || !id_fpool_new_id(esw_id_pool, 0, &ctx->esw_id)) {
        VLOG_ERR("Failed to alloc a new esw id");
        return;
    }

    base_id = ovs_doca_max_shared_counters_per_esw() * ctx->esw_id;
    if (ovs_doca_max_ct_counters_per_esw() > 0) {
        ctx->shared_ct_counter_id_pool =
            id_fpool_create(MAX_OFFLOAD_THREAD_NB, base_id,
                            ovs_doca_max_ct_counters_per_esw());
    } else {
        ctx->shared_ct_counter_id_pool = NULL;
    }

    base_id += ovs_doca_max_ct_counters_per_esw();
    ctx->shared_counter_id_pool =
        id_fpool_create(MAX_OFFLOAD_THREAD_NB, base_id,
                        OVS_DOCA_MAX_METER_COUNTERS_PER_ESW);
}

static int
doca_bind_shared_cntrs(struct doca_offload_esw_ctx *ctx)
{
    struct doca_flow_shared_resource_cfg cfg =
        { .domain = DOCA_FLOW_PIPE_DOMAIN_DEFAULT };
    uint32_t base_id;
    uint32_t *ids;
    int i, ret;

    ids = xcalloc(ovs_doca_max_shared_counters_per_esw(), sizeof *ids);

    base_id = ovs_doca_max_shared_counters_per_esw() * ctx->esw_id;
    for (i = 0; i < ovs_doca_max_shared_counters_per_esw(); i++) {
        ids[i] = base_id + i;
        ret = doca_flow_shared_resource_set_cfg(DOCA_FLOW_SHARED_RESOURCE_COUNTER,
                                                ids[i], &cfg);
        if (ret != DOCA_SUCCESS) {
            VLOG_ERR("Failed to config shared counter id %d, err %d - %s",
                     ids[i], ret, doca_error_get_descr(ret));
            goto free_ids;
        }
    }

    ret = doca_flow_shared_resources_bind(DOCA_FLOW_SHARED_RESOURCE_COUNTER, ids,
                                       ovs_doca_max_shared_counters_per_esw(),
                                       ctx->esw_port);
    if (ret != DOCA_SUCCESS) {
        VLOG_ERR("Shared counters binding failed, ids %d-%d, err %d - %s",
                 ids[0], ids[ovs_doca_max_shared_counters_per_esw() - 1], ret,
                 doca_error_get_descr(ret));
        goto free_ids;
    }

free_ids:
    free(ids);
    return ret ? -1 : 0;
}

static int
shared_mtr_flow_id_alloc(struct doca_offload_esw_ctx *esw_ctx,
                         struct rte_flow_error *error)
{
    unsigned int tid = netdev_offload_thread_id();
    uint32_t id;

    if (!esw_ctx->shared_mtr_flow_id_pool) {
        if (error) {
            error->type = RTE_FLOW_ERROR_TYPE_UNSPECIFIED;
            error->message = "Could not allocate meter flow ID, id-pool is not initialized";
        }
        goto err;
    }

    if (!id_fpool_new_id(esw_ctx->shared_mtr_flow_id_pool, tid, &id)) {
        VLOG_ERR("Failed to alloc a new shared meter flow id");
        if (error) {
            error->type = RTE_FLOW_ERROR_TYPE_UNSPECIFIED;
            error->message = "Failed to alloc a new shared meter flow id";
        }
        goto err;
    }

    return id;
err:
    if (error) {
        ovs_assert(error->message);
        ovs_assert(error->type);
    }
    return -1;
}

static void
shared_mtr_flow_id_free(struct doca_offload_esw_ctx *esw_ctx, uint32_t id)
{
    unsigned int tid = netdev_offload_thread_id();

    id_fpool_free_id(esw_ctx->shared_mtr_flow_id_pool, tid, id);
}

static void
shared_mtr_flow_id_pool_init(struct doca_offload_esw_ctx *ctx)
{
    uint32_t base_id;

    base_id = SHARED_MTR_N_IDS * ctx->esw_id + MIN_SHARED_MTR_FLOW_ID;
    ctx->shared_mtr_flow_id_pool = id_fpool_create(MAX_OFFLOAD_THREAD_NB,
                                                   base_id, SHARED_MTR_N_IDS);
}

static int
doca_bind_shared_meters(struct doca_offload_esw_ctx *ctx)
{
    struct doca_flow_shared_resource_cfg dummy_cfg = {
        .domain = DOCA_FLOW_PIPE_DOMAIN_DEFAULT,
        .meter_cfg.limit_type = DOCA_FLOW_METER_LIMIT_TYPE_BYTES,
        .meter_cfg.cir = 125000,
        .meter_cfg.cbs = 12500,
    };
    uint32_t ids[OVS_DOCA_MAX_METERS_PER_ESW];
    int i, id, ret;

    /* DOCA allows meter IDs to start from 0, but it's problematic to have a
     * meter with ID 0 because in such case it will be impossible to disable
     * shared meter in doca_flow_monitor struct later, so meter with ID 0 is
     * not configured and not bound to avoid this issue.
     *
     * Total number of shared meters is OVS_DOCA_MAX_METERS_PER_ESW-1 because
     * meter ID 0 is not used.
     */
    for (i = 0, id = 1; i < OVS_DOCA_MAX_METERS_PER_ESW - 1; i++, id++) {
        ids[i] = ovs_doca_meter_id(id, ctx->esw_id);
        /* DOCA will fail to bind a shared meter if it's unconfigured, which is
         * a bug, so a dummy configuration is used as a W/A; actual meter
         * configuration will be set by the user when OVS meter is added with
         * `ovs-ofctl add-meter` command.
         */
        ret = doca_flow_shared_resource_set_cfg(DOCA_FLOW_SHARED_RESOURCE_METER,
                                                ids[i], &dummy_cfg);
        if (ret != DOCA_SUCCESS) {
            VLOG_ERR("Failed to init shared meter (id %d), err %d - %s",
                    ids[i], ret, doca_error_get_descr(ret));
            return -1;
        }
    }

    ret = doca_flow_shared_resources_bind(DOCA_FLOW_SHARED_RESOURCE_METER, ids,
                                          OVS_DOCA_MAX_METERS_PER_ESW - 1,
                                          ctx->esw_port);
    if (ret != DOCA_SUCCESS) {
        VLOG_ERR("Shared meters binding failed, ids %d-%d, err %d - %s",
                 ids[0], ids[OVS_DOCA_MAX_METERS_PER_ESW - 2], ret,
                 doca_error_get_descr(ret));
        return -1;
    }

    return 0;
}

static int
doca_bind_sample_shared_mirror(struct doca_offload_esw_ctx *ctx)
{
    struct doca_flow_mirror_target target = {
        .fwd = {
            .type = DOCA_FLOW_FWD_PIPE,
            .next_pipe = doca_pipe_group_get_pipe(ctx->sample_ctx.postmirror_pipe_ctx),
        }
    };
    struct doca_flow_shared_resource_cfg shared_res_cfg = {
        .mirror_cfg = {
            .nr_targets = 1,
            .target = &target,
        }
    };
    uint32_t mirror_id = ctx->esw_id;
    int err;

    err = doca_flow_shared_resource_set_cfg(DOCA_FLOW_SHARED_RESOURCE_MIRROR,
                                            mirror_id, &shared_res_cfg);

    if (err != DOCA_SUCCESS) {
        VLOG_ERR("Failed to init sFlow shared mirror (id %d), err %d - %s",
                 mirror_id, err, doca_error_get_descr(err));
        return -1;
    }

    err = doca_flow_shared_resources_bind(DOCA_FLOW_SHARED_RESOURCE_MIRROR,
                                          &mirror_id,
                                          OVS_DOCA_MAX_MIRRORS_PER_ESW,
                                          ctx->esw_port);
    if (err != DOCA_SUCCESS) {
        VLOG_ERR("Failed to bind sFlow shared mirror (id %d), err %d - %s",
                 mirror_id, err, doca_error_get_descr(err));
        return -1;
    }

    return 0;
}

static int
doca_post_meter_pipe_init(struct netdev *netdev,
                          struct doca_offload_esw_ctx *ctx)
{
    struct doca_pipe_group_ctx *group_ctx;

    group_ctx = doca_pipe_group_ctx_ref(netdev, POSTMETER_TABLE_ID);
    if (!group_ctx) {
        return -1;
    }
    ctx->post_meter_pipe_group_ctx = group_ctx;

    return 0;
}

static int
doca_sample_pipe_init(struct netdev *netdev, struct doca_offload_esw_ctx *ctx)
{
    struct doca_pipe_group_ctx *pipe_ctx;

    pipe_ctx = doca_pipe_group_ctx_ref(netdev, SAMPLE_TABLE_ID);
    if (!pipe_ctx) {
        return -1;
    }
    ctx->sample_ctx.sample_pipe_ctx = pipe_ctx;

    return 0;
}

static int
doca_sample_mirror_pipe_init(struct netdev *netdev,
                             struct doca_offload_esw_ctx *ctx)
{
    struct doca_basic_pipe_ctx *pipe_ctx = &ctx->sample_ctx.mirror_pipe_ctx;
    struct doca_flow_monitor mirror_monitor = {
        .shared_mirror_id = UINT32_MAX
    };
    struct doca_flow_pipe_cfg *pipe_cfg;
    struct doca_flow_match matchall;
    struct doca_flow_fwd fwd = {
        .type = DOCA_FLOW_FWD_PIPE,
        .next_pipe = NULL,
    };
    char pipe_name[50];
    int err;

    memset(&matchall, 0, sizeof matchall);

    err = doca_flow_pipe_cfg_create(&pipe_cfg,
                                    doca_flow_port_switch_get(ctx->esw_port));
    if (err) {
        VLOG_ERR("%s: Could not create doca_flow_pipe_cfg", netdev_get_name(netdev));
        return err;
    }

    snprintf(pipe_name, sizeof pipe_name, "SAMPLE_MIRROR_PIPE%" PRIu32,
             ctx->esw_id);

    if (doca_flow_pipe_cfg_set_name(pipe_cfg, pipe_name) ||
        doca_flow_pipe_cfg_set_type(pipe_cfg, DOCA_FLOW_PIPE_BASIC) ||
        doca_flow_pipe_cfg_set_is_root(pipe_cfg, false) ||
        doca_flow_pipe_cfg_set_nr_entries(pipe_cfg, 1) ||
        doca_flow_pipe_cfg_set_enable_strict_matching(pipe_cfg, true) ||
        doca_flow_pipe_cfg_set_match(pipe_cfg, &matchall, &matchall) ||
        doca_flow_pipe_cfg_set_monitor(pipe_cfg, &mirror_monitor)) {
        VLOG_ERR("%s: Could not set doca_flow_pipe_cfg", netdev_get_name(netdev));
        err = -1;
        goto out;
    }

    err = doca_flow_pipe_create(pipe_cfg, &fwd, NULL, &pipe_ctx->pipe);
    if (err) {
        VLOG_ERR("%s: Failed to create sample mirror pipe: %d (%s)",
                 netdev_get_name(netdev), err, doca_error_get_descr(err));
        goto out;
    }

    fwd.next_pipe = doca_pipe_group_get_pipe(ctx->root_pipe_group_ctx);
    mirror_monitor.shared_mirror_id = ctx->esw_id;
    err = doca_offload_add_entry(AUX_QUEUE, pipe_ctx->pipe, &matchall, NULL,
                                 &mirror_monitor, &fwd, DOCA_FLOW_NO_WAIT,
                                 ctx, &ctx->sample_ctx.mirror_pipe_entry);
    if (err) {
        VLOG_ERR("%s: Failed to add entry to sample mirror pipe: %d (%s)",
                 netdev_get_name(netdev), err, doca_error_get_descr(err));
        doca_flow_pipe_destroy(pipe_ctx->pipe);
        ctx->sample_ctx.mirror_pipe_ctx.pipe = NULL;
        goto out;
    }
    dpdk_offload_counter_inc(netdev);

    doca_offload_complete_queue(netdev, AUX_QUEUE, true);
out:
    doca_flow_pipe_cfg_destroy(pipe_cfg);
    return err;
}

static int
doca_sample_post_mirror_pipe_init(struct netdev *netdev,
                                  struct doca_offload_esw_ctx *ctx)
{
    struct doca_pipe_group_mask_ctx **entry_mctx;
    struct doca_pipe_group_ctx *pipe_ctx;
    struct doca_flow_pipe_entry *entry;
    struct doca_flow_actions act_mask;
    struct doca_flow_fwd fwd_miss = {
        .type = DOCA_FLOW_FWD_PIPE,
        .next_pipe = NULL,
    };
    struct doca_flow_fwd fwd_drop = {
        .type = DOCA_FLOW_FWD_DROP,
    };
    struct doca_flow_match lacp_spec;
    struct doca_flow_match matchall;
    struct doca_flow_actions act;
    struct rte_flow_error error;

    memset(&lacp_spec, 0, sizeof lacp_spec);
    memset(&act_mask, 0, sizeof act_mask);
    memset(&matchall, 0, sizeof matchall);
    memset(&act, 0, sizeof act);

    pipe_ctx = doca_pipe_group_ctx_ref(netdev, SAMPLE_POSTMIRROR_TABLE_ID);
    if (!pipe_ctx) {
        return -1;
    }
    ctx->sample_ctx.postmirror_pipe_ctx = pipe_ctx;

    entry_mctx = &ctx->sample_ctx.postmirror_pipe_entry_mctx;
    fwd_miss.next_pipe = doca_offload_get_misspath_pipe(ctx);
    doca_set_reg_val_mask(&act.meta, &act_mask.meta, REG_FIELD_FLOW_INFO,
                          SAMPLE_FLOW_MARK);

    entry = create_doca_pipe_group_entry(netdev, AUX_QUEUE, pipe_ctx,
                                         DPDK_OFFLOAD_PRIORITY_LOW, &matchall,
                                         &matchall, &act, &act_mask, NULL, NULL,
                                         &fwd_miss, entry_mctx, &error);
    if (!entry) {
            VLOG_ERR("%s: Failed to create sample postmirror entry: %s.",
                     netdev_get_name(netdev), error.message);
            goto err;
    }
    ctx->sample_ctx.postmirror_pipe_entry = entry;

    /* LACP packets should not be duplicated so if such packet has been sampled
     * it must be dropped.
     */
    lacp_spec.outer.eth.type = htons(ETH_TYPE_LACP);
    lacp_spec.parser_meta.outer_l2_type = DOCA_FLOW_L2_META_NO_VLAN;
    entry_mctx = &ctx->sample_ctx.postmirror_lacp_pipe_entry_mctx;
    entry = create_doca_pipe_group_entry(netdev, AUX_QUEUE, pipe_ctx,
                                         DPDK_OFFLOAD_PRIORITY_HIGH, &lacp_spec,
                                         &lacp_spec, NULL, NULL, NULL, NULL,
                                         &fwd_drop, entry_mctx, &error);
    if (!entry) {
            VLOG_ERR("%s: Failed to create sample postmirror entry: %s.",
                     netdev_get_name(netdev), error.message);
            goto err_lacp;
    }
    ctx->sample_ctx.postmirror_lacp_pipe_entry = entry;

    doca_offload_complete_queue(netdev, AUX_QUEUE, true);
    return 0;

err_lacp:
    doca_offload_remove_entry(ctx, AUX_QUEUE, DOCA_FLOW_NO_WAIT,
                              ctx->sample_ctx.postmirror_pipe_entry);
    doca_pipe_group_mask_ctx_unref(ctx->sample_ctx.postmirror_pipe_entry_mctx);
    dpdk_offload_counter_dec(netdev);
err:
    doca_pipe_group_ctx_unref(pipe_ctx);
    ctx->sample_ctx.postmirror_pipe_ctx = NULL;
    return -1;
}

static void
doca_post_sw_meter_pipe_uninit(struct doca_offload_esw_ctx *ctx)
{
    doca_basic_pipe_ctx_uninit(&ctx->post_sw_meter_pipe_ctx);
}

static int
doca_post_sw_meter_pipe_init(struct netdev *netdev,
                             struct doca_offload_esw_ctx *ctx)
{
    struct doca_basic_pipe_ctx *miss_pipe_ctx = NULL;
    struct doca_basic_pipe_ctx *pipe_ctx;
    struct doca_flow_match match, mask;
    struct doca_flow_monitor monitor;
    struct doca_flow_port *doca_port;
    struct doca_flow_pipe_cfg *cfg;
    struct doca_flow_fwd miss;
    struct doca_flow_fwd fwd;
    char pipe_name[50];
    int ret;

    ovs_assert(ctx->post_sw_meter_pipe_ctx.pipe == NULL);

    memset(&fwd, 0, sizeof fwd);
    memset(&miss, 0, sizeof miss);
    memset(&mask, 0, sizeof mask);
    memset(&match, 0, sizeof match);
    memset(&monitor, 0, sizeof monitor);

    snprintf(pipe_name, sizeof pipe_name, "OVS_POST_SW_METER_PIPE_%" PRIu32,
             ctx->esw_id);

    doca_port = netdev_dpdk_doca_port_get(netdev);

    ret = doca_flow_pipe_cfg_create(&cfg,
                                    doca_flow_port_switch_get(doca_port));
    if (ret) {
        VLOG_ERR("%s: Could not create doca_flow_pipe_cfg for %s",
                 netdev_get_name(netdev), pipe_name);
        return ret;
    }

    miss_pipe_ctx = &ctx->meta_transfer_ctx.meta_push_pipe_ctx;
    miss.type = DOCA_FLOW_FWD_PIPE;
    miss.next_pipe = miss_pipe_ctx->pipe;

    fwd.type = DOCA_FLOW_FWD_CHANGEABLE;

    match.parser_meta.port_meta = UINT32_MAX;
    match.parser_meta.meter_color = 0xff;
    memcpy(&mask, &match, sizeof mask);

    monitor.counter_type = DOCA_FLOW_RESOURCE_TYPE_NON_SHARED;

    if (doca_flow_pipe_cfg_set_name(cfg, pipe_name) ||
        doca_flow_pipe_cfg_set_type(cfg, DOCA_FLOW_PIPE_BASIC) ||
        doca_flow_pipe_cfg_set_is_root(cfg, false) ||
        doca_flow_pipe_cfg_set_nr_entries(cfg, OVS_DOCA_MAX_SW_METERS_PER_ESW) ||
        doca_flow_pipe_cfg_set_match(cfg, &mask, &mask) ||
        doca_flow_pipe_cfg_set_monitor(cfg, &monitor)) {
            VLOG_ERR("%s: Could not set doca_flow_pipe_cfg for %s",
                     netdev_get_name(netdev), pipe_name);
            goto error;
    }

    pipe_ctx = &ctx->post_sw_meter_pipe_ctx;
    ret = doca_flow_pipe_create(cfg, &fwd, &miss, &pipe_ctx->pipe);
    if (ret) {
        VLOG_ERR("%s: Failed to create basic pipe '%s': %d (%s)",
                 netdev_get_name(netdev), pipe_name, ret,
                 doca_error_get_descr(ret));
        goto error;
    }

    doca_flow_pipe_cfg_destroy(cfg);

    return 0;

error:
    doca_flow_pipe_cfg_destroy(cfg);
    return -1;
}

static void
doca_sw_meter_pipe_uninit(struct doca_offload_esw_ctx *ctx)
{
    doca_basic_pipe_ctx_uninit(&ctx->sw_meter_pipe_ctx);
}

/* The sw-meter table is a basic pipe with a default miss decision
 * to forward to the miss-pipe. The miss-pipe, being empty, executes
 * the RSS config to send the packet to SW. */
static int
doca_sw_meter_pipe_init(struct netdev *netdev,
                        struct doca_offload_esw_ctx *ctx)
{
    struct doca_basic_pipe_ctx *miss_pipe_ctx = NULL;
    struct doca_basic_pipe_ctx *pipe_ctx;
    struct doca_flow_monitor monitor;
    struct doca_flow_port *doca_port;
    struct doca_flow_pipe_cfg *cfg;
    struct doca_flow_match match;
    struct doca_flow_fwd miss;
    struct doca_flow_fwd fwd;
    char pipe_name[50];
    int ret;

    ovs_assert(ctx->sw_meter_pipe_ctx.pipe == NULL);

    memset(&fwd, 0, sizeof fwd);
    memset(&miss, 0, sizeof miss);
    memset(&match, 0, sizeof match);
    memset(&monitor, 0, sizeof monitor);

    snprintf(pipe_name, sizeof pipe_name, "OVS_SW_METER_PIPE_%" PRIu32,
             ctx->esw_id);

    doca_port = netdev_dpdk_doca_port_get(netdev);

    ret = doca_flow_pipe_cfg_create(&cfg,
                                    doca_flow_port_switch_get(doca_port));
    if (ret) {
        VLOG_ERR("%s: Could not create doca_flow_pipe_cfg for %s",
                 netdev_get_name(netdev), pipe_name);
        return ret;
    }

    miss_pipe_ctx = &ctx->meta_transfer_ctx.meta_push_pipe_ctx;
    miss.type = DOCA_FLOW_FWD_PIPE;
    miss.next_pipe = miss_pipe_ctx->pipe;

    fwd.type = DOCA_FLOW_FWD_PIPE;
    fwd.next_pipe = ctx->post_sw_meter_pipe_ctx.pipe;
    ovs_assert(ctx->post_sw_meter_pipe_ctx.pipe);

    match.parser_meta.port_meta = UINT32_MAX;
    monitor.meter_type = DOCA_FLOW_RESOURCE_TYPE_SHARED;
    monitor.shared_meter.shared_meter_id = UINT32_MAX;

    if (doca_flow_pipe_cfg_set_name(cfg, pipe_name) ||
        doca_flow_pipe_cfg_set_type(cfg, DOCA_FLOW_PIPE_BASIC) ||
        doca_flow_pipe_cfg_set_is_root(cfg, false) ||
        doca_flow_pipe_cfg_set_nr_entries(cfg, OVS_DOCA_MAX_SW_METERS_PER_ESW) ||
        doca_flow_pipe_cfg_set_match(cfg, &match, &match) ||
        doca_flow_pipe_cfg_set_monitor(cfg, &monitor)) {
            VLOG_ERR("%s: Could not set doca_flow_pipe_cfg for %s",
                     netdev_get_name(netdev), pipe_name);
            goto error;
    }

    pipe_ctx = &ctx->sw_meter_pipe_ctx;
    ret = doca_flow_pipe_create(cfg, &fwd, &miss, &pipe_ctx->pipe);
    if (ret) {
        VLOG_ERR("%s: Failed to create basic pipe '%s': %d (%s)",
                 netdev_get_name(netdev), pipe_name, ret,
                 doca_error_get_descr(ret));
        goto error;
    }

    doca_flow_pipe_cfg_destroy(cfg);

    return 0;

error:
    doca_flow_pipe_cfg_destroy(cfg);
    return -1;
}

static void
doca_pre_miss_pipe_uninit(struct doca_offload_esw_ctx *ctx)
{
    doca_basic_pipe_ctx_uninit(&ctx->pre_miss_ctx.pre_miss_pipe_ctx);
}

static int
doca_pre_miss_pipe_init(struct netdev *netdev, struct doca_offload_esw_ctx *ctx)
{
    struct doca_basic_pipe_ctx *pre_miss_pipe_ctx, *sw_meter_pipe_ctx;
    struct doca_flow_target *kernel_target;
    struct doca_flow_port *doca_port;
    struct doca_flow_fwd fwd, miss;
    struct doca_flow_pipe_cfg *cfg;
    struct doca_flow_match match;
    char pipe_name[50];
    int ret;

    memset(&match, 0, sizeof match);
    memset(&miss, 0, sizeof miss);
    memset(&fwd, 0, sizeof fwd);

    pre_miss_pipe_ctx = &ctx->pre_miss_ctx.pre_miss_pipe_ctx;
    ctx->pre_miss_ctx.netdev = netdev;

    snprintf(pipe_name, sizeof pipe_name, "OVS_PRE_MISS_PIPE_%" PRIu32, ctx->esw_id);

    doca_port = netdev_dpdk_doca_port_get(netdev);

    ret = doca_flow_pipe_cfg_create(&cfg, doca_flow_port_switch_get(doca_port));
    if (ret) {
        VLOG_ERR("%s: Could not create doca_flow_pipe_cfg for %s",
                 netdev_get_name(netdev), pipe_name);
        return ret;
    }

    sw_meter_pipe_ctx = &ctx->sw_meter_pipe_ctx;
    miss.type = DOCA_FLOW_FWD_PIPE;
    miss.next_pipe = sw_meter_pipe_ctx->pipe;

    ret = doca_flow_get_target(DOCA_FLOW_TARGET_KERNEL, &kernel_target);
    if (ret) {
        VLOG_ERR("%s: Could not get miss to kernel target",
                 netdev_get_name(netdev));
        goto out;
    }
    fwd.type = DOCA_FLOW_FWD_TARGET;
    fwd.target = kernel_target;

    match.outer.eth.type = UINT16_MAX;
    match.parser_meta.outer_l2_type = DOCA_FLOW_L2_META_NO_VLAN;

    if (doca_flow_pipe_cfg_set_name(cfg, pipe_name) ||
        doca_flow_pipe_cfg_set_type(cfg, DOCA_FLOW_PIPE_BASIC) ||
        doca_flow_pipe_cfg_set_is_root(cfg, false) ||
        doca_flow_pipe_cfg_set_nr_entries(cfg, NUM_SEND_TO_KERNEL) ||
        doca_flow_pipe_cfg_set_match(cfg, &match, &match)) {
            VLOG_ERR("%s: Could not set doca_flow_pipe_cfg for %s",
                     netdev_get_name(netdev), pipe_name);
            ret = -1;
            goto out;
    }

    ret = doca_flow_pipe_create(cfg, &fwd, &miss, &pre_miss_pipe_ctx->pipe);
    if (ret) {
        VLOG_ERR("%s: Failed to create basic pipe '%s': %d (%s)",
                 netdev_get_name(netdev), pipe_name, ret,
                 doca_error_get_descr(ret));
    }

out:
    doca_flow_pipe_cfg_destroy(cfg);
    return ret;
}

static void
doca_pre_miss_rules_uninit(struct doca_offload_esw_ctx *ctx)
{
    for (int i = 0 ; i < NUM_SEND_TO_KERNEL ; i++) {
        doca_offload_fixed_rule_uninit(ctx, ctx->pre_miss_ctx.netdev,
                                       &ctx->pre_miss_ctx.pre_miss_rules[i]);
    }
}

static int
doca_pre_miss_rules_init(struct netdev *netdev, struct doca_offload_esw_ctx *ctx)
{
    struct doca_flow_match match;
    struct rte_flow_error error;
    struct fixed_rule *fr;
    int ret;

    memset(&match, 0, sizeof match);

    for (int i = 0 ; i < NUM_SEND_TO_KERNEL ; i++) {
        fr = &ctx->pre_miss_ctx.pre_miss_rules[i];

        match.outer.eth.type = htons(pre_miss_mapping[i]);
        ret = create_doca_basic_flow_entry(netdev, AUX_QUEUE,
                                           ctx->pre_miss_ctx.pre_miss_pipe_ctx.pipe,
                                           &match, NULL, NULL, NULL, &fr->doh.dfh.flow,
                                           &error, true);
        if (ret) {
            VLOG_ERR("%s: Failed to create pre_miss %x rule: Error %d (%s)",
                     netdev_get_name(netdev), pre_miss_mapping[i],
                     error.type, error.message);
            doca_pre_miss_rules_uninit(ctx);
            break;
        }
    }

    return ret;
}

static void
doca_meta_push_pipe_uninit(struct doca_offload_esw_ctx *ctx)
{
    doca_basic_pipe_ctx_uninit(&ctx->meta_transfer_ctx.meta_push_pipe_ctx);
}

/* The meta-push table functions as a basic pipe, employing a catch-all rule
 * for encapsulating metadata before forwarding it to the miss-pipe.
 * The miss-pipe, being empty, sends the packet to SW.
 * Metadata encapsulation is accomplished through both the meta-push
 * and meta-copy pipes, where the former adds a header, and the latter coppies
 * the data into the header.
 */
static int
doca_meta_push_pipe_init(struct netdev *netdev, struct doca_offload_esw_ctx *ctx)
{
    struct doca_flow_actions actions_masks, *actions_masks_arr[1];
    struct doca_flow_actions actions, *actions_arr[1];
    struct doca_basic_pipe_ctx *pipe_ctx;
    struct doca_flow_port *doca_port;
    struct doca_flow_pipe_cfg *cfg;
    struct doca_flow_match match;
    struct doca_flow_fwd miss;
    struct doca_flow_fwd fwd;
    char pipe_name[50];
    int ret;

    ovs_assert(ctx->meta_transfer_ctx.meta_push_pipe_ctx.pipe == NULL);

    memset(&fwd, 0, sizeof fwd);
    memset(&miss, 0, sizeof miss);
    memset(&match, 0, sizeof match);
    memset(&actions, 0, sizeof actions);
    memset(&actions_masks, 0, sizeof actions);

    snprintf(pipe_name, sizeof pipe_name, "OVS_META_PUSH_PIPE_%" PRIu32, ctx->esw_id);

    doca_port = netdev_dpdk_doca_port_get(netdev);

    ret = doca_flow_pipe_cfg_create(&cfg, doca_flow_port_switch_get(doca_port));
    if (ret) {
        VLOG_ERR("%s: Could not create doca_flow_pipe_cfg for %s",
                 netdev_get_name(netdev), pipe_name);
        return ret;
    }

    fwd.type = DOCA_FLOW_FWD_PIPE;
    fwd.next_pipe = ctx->meta_transfer_ctx.meta_copy_pipe_ctx.pipe;
    ovs_assert(ctx->meta_transfer_ctx.meta_copy_pipe_ctx.pipe);

    actions_arr[0] = &actions;
    actions_masks_arr[0] = &actions_masks;
    /* Build basic outer VXLAN encap data */
    actions.encap_type = DOCA_FLOW_RESOURCE_TYPE_NON_SHARED;
    actions.encap_cfg.is_l2 = true;
    actions.encap_cfg.encap.outer.l3_type = DOCA_FLOW_L3_TYPE_IP4;
    memset(&actions.encap_cfg.encap.outer.ip4.src_ip, 0xFF,
           sizeof actions.encap_cfg.encap.outer.ip4.src_ip);
    memset(&actions.encap_cfg.encap.outer.ip4.dst_ip, 0xFF,
           sizeof actions.encap_cfg.encap.outer.ip4.dst_ip);
    memset(&actions.encap_cfg.encap.outer.eth.src_mac, 0xFF,
           sizeof actions.encap_cfg.encap.outer.eth.src_mac);
    memset(&actions.encap_cfg.encap.outer.eth.dst_mac, 0xFF,
           sizeof actions.encap_cfg.encap.outer.eth.dst_mac);
    memset(&actions.encap_cfg.encap.outer.ip4.ttl, 0xFF,
           sizeof actions.encap_cfg.encap.outer.ip4.ttl);
    memset(&actions.encap_cfg.encap.outer.udp.l4_port.dst_port, 0xFF,
           sizeof actions.encap_cfg.encap.outer.udp.l4_port.dst_port);
    memset(&actions.encap_cfg.encap.tun.vxlan_tun_id, 0xFF,
           sizeof actions.encap_cfg.encap.tun.vxlan_tun_id);
    actions.encap_cfg.encap.outer.l4_type_ext = DOCA_FLOW_L4_TYPE_EXT_UDP;
    actions.encap_cfg.encap.tun.type = DOCA_FLOW_TUN_VXLAN;

    doca_set_reg_val(&actions.meta, REG_FIELD_META_PUSH, 1);
    doca_set_reg_val(&actions_masks.meta, REG_FIELD_META_PUSH, 1);

    if (doca_flow_pipe_cfg_set_name(cfg, pipe_name) ||
        doca_flow_pipe_cfg_set_type(cfg, DOCA_FLOW_PIPE_BASIC) ||
        doca_flow_pipe_cfg_set_is_root(cfg, false) ||
        doca_flow_pipe_cfg_set_nr_entries(cfg, 1) ||
        doca_flow_pipe_cfg_set_match(cfg, &match, &match) ||
        doca_flow_pipe_cfg_set_actions(cfg, actions_arr, actions_masks_arr, NULL,
                                       ARRAY_SIZE(actions_arr))) {
        VLOG_ERR("%s: Could not set doca_flow_pipe_cfg", netdev_get_name(netdev));
        goto out;
    }

    pipe_ctx = &ctx->meta_transfer_ctx.meta_push_pipe_ctx;
    ret = doca_flow_pipe_create(cfg, &fwd, NULL, &pipe_ctx->pipe);
    if (ret) {
        VLOG_ERR("%s: Failed to create basic pipe '%s': %d (%s)",
                 netdev_get_name(netdev), pipe_name, ret,
                 doca_error_get_descr(ret));
        goto out;
    }

out:
    doca_flow_pipe_cfg_destroy(cfg);
    return ret;
}

static void
doca_meta_copy_pipe_uninit(struct doca_offload_esw_ctx *ctx)
{
    doca_basic_pipe_ctx_uninit(&ctx->meta_transfer_ctx.meta_copy_pipe_ctx);
}

static int
doca_meta_push_rule_init(struct netdev *netdev, struct doca_offload_esw_ctx *ctx)
{
    struct doca_meta_transfer_ctx *meta_trans_ctx;
    struct doca_flow_actions dacts;
    struct rte_flow_error error;
    struct doca_flow_fwd fwd;
    struct fixed_rule *fr;
    int ret = 0;

    memset(&dacts, 0, sizeof dacts);

    /* Eth */
    memset(&dacts.encap_cfg.encap.outer.eth.src_mac, 0xFF,
           sizeof dacts.encap_cfg.encap.outer.eth.src_mac);
    dacts.encap_cfg.encap.outer.eth.type = (OVS_FORCE doca_be16_t) htons(ETH_TYPE_IP);
    /* IP */
    dacts.encap_cfg.encap.outer.l3_type = DOCA_FLOW_L3_TYPE_IP4;
    dacts.encap_cfg.encap.outer.ip4.next_proto = IPPROTO_UDP;
    /* UDP */
    dacts.encap_cfg.encap.outer.l4_type_ext = DOCA_FLOW_L4_TYPE_EXT_UDP;
    dacts.encap_cfg.encap.outer.udp.l4_port.dst_port =
        (OVS_FORCE doca_be16_t) htons(META_PUSH_VXLAN_UDP_PORT);
    /* VXLAN */
    dacts.encap_cfg.encap.tun.type = DOCA_FLOW_TUN_VXLAN;
    dacts.encap_cfg.encap.tun.vxlan_tun_id = (OVS_FORCE doca_be32_t) htonl(META_PUSH_VNI << 8);

    fwd.type = DOCA_FLOW_FWD_PIPE;
    fwd.next_pipe = ctx->meta_transfer_ctx.meta_copy_pipe_ctx.pipe;

    fr = &ctx->meta_transfer_ctx.meta_push_hdr;
    meta_trans_ctx = &ctx->meta_transfer_ctx;

    ret = create_doca_basic_flow_entry(netdev, AUX_QUEUE,
                                       meta_trans_ctx->meta_push_pipe_ctx.pipe,
                                       NULL, &dacts, NULL, &fwd, &fr->doh.dfh.flow,
                                       &error, true);
    if (ret) {
        VLOG_ERR("%s: Failed to create meta-push rule: Error %d (%s)",
                 netdev_get_name(netdev),error.type, error.message);
    }

    return ret;
}

static void
doca_meta_push_rule_uninit(struct doca_offload_esw_ctx *ctx)
{
    doca_offload_fixed_rule_uninit(ctx, ctx->meta_transfer_ctx.netdev,
                                   &ctx->meta_transfer_ctx.meta_push_hdr);
}

static int
doca_meta_copy_rule_init(struct netdev *netdev, struct doca_offload_esw_ctx *ctx)
{
    struct doca_meta_transfer_ctx *meta_trans_ctx;
    struct rte_flow_error error;
    struct fixed_rule *fr;
    int ret = 0;

    fr = &ctx->meta_transfer_ctx.meta_copy_to_hdr;
    meta_trans_ctx = &ctx->meta_transfer_ctx;

    ret = create_doca_basic_flow_entry(netdev, AUX_QUEUE,
                                       meta_trans_ctx->meta_copy_pipe_ctx.pipe,
                                       NULL, NULL, NULL, NULL, &fr->doh.dfh.flow,
                                       &error, true);
    if (ret) {
        VLOG_ERR("%s: Failed to create meta-copy rule: Error %d (%s)",
                 netdev_get_name(netdev),error.type, error.message);
    }

    return ret;
}

static void
doca_meta_copy_rule_uninit(struct doca_offload_esw_ctx *ctx)
{
    doca_offload_fixed_rule_uninit(ctx, ctx->meta_transfer_ctx.netdev,
                                   &ctx->meta_transfer_ctx.meta_copy_to_hdr);
}

static void
doca_meta_transfer_uninit(struct doca_offload_esw_ctx *ctx)
{
    doca_meta_push_rule_uninit(ctx);
    doca_meta_push_pipe_uninit(ctx);
    doca_meta_copy_rule_uninit(ctx);
    doca_meta_copy_pipe_uninit(ctx);
}


static int
doca_meta_copy_pipe_init(struct netdev *netdev, struct doca_offload_esw_ctx *ctx)
{
    const uint32_t n_descs = sizeof(struct reg_tags) / sizeof(uint32_t);
    struct doca_flow_action_descs descs, *descs_arr[1];
    struct doca_flow_actions actions, *actions_arr[1];
    struct doca_flow_action_desc desc_array[n_descs];
    struct doca_flow_actions *actions_masks_arr[1];
    struct doca_pipe_group_ctx *next_group_ctx;
    struct doca_flow_actions actions_masks;
    struct doca_basic_pipe_ctx *pipe_ctx;
    struct doca_flow_port *doca_port;
    struct doca_flow_pipe_cfg *cfg;
    struct doca_flow_match match;
    int ret = 0, desc_ind = 0;
    struct doca_flow_fwd fwd;
    char pipe_name[50];

    ovs_assert(ctx->meta_transfer_ctx.meta_copy_pipe_ctx.pipe == NULL);

    memset(&fwd, 0, sizeof fwd);
    memset(&match, 0, sizeof match);
    memset(&actions, 0, sizeof actions);
    memset(&actions_masks, 0, sizeof actions_masks);
    memset(&descs, 0, sizeof descs);

    snprintf(pipe_name, sizeof pipe_name, "OVS_META_COPY_PIPE_%" PRIu32, ctx->esw_id);

    doca_port = netdev_dpdk_doca_port_get(netdev);
    ret = doca_flow_pipe_cfg_create(&cfg, doca_flow_port_switch_get(doca_port));
    if (ret) {
        VLOG_ERR("%s: Could not create doca_flow_pipe_cfg for %s",
                 netdev_get_name(netdev), pipe_name);
        return ret;
    }

    next_group_ctx = doca_pipe_group_ctx_ref(netdev, MISS_TABLE_ID);
    if (next_group_ctx == NULL) {
        VLOG_ERR("%s: Failed to take a reference on the miss pipe",
                 netdev_get_name(netdev));
        ret = -1;
        goto out;
    }
    fwd.type = DOCA_FLOW_FWD_PIPE;
    fwd.next_pipe = doca_pipe_group_get_pipe(next_group_ctx);

    actions_arr[0] = &actions;
    actions_masks_arr[0] = &actions_masks;
    descs_arr[0] = &descs;
    descs.nb_action_desc = n_descs;
    descs.desc_array = desc_array;

    /* Copy REG_FIELD_CT_STATE */
    desc_array[desc_ind].type = DOCA_FLOW_ACTION_COPY;
    desc_array[desc_ind].field_op.src.field_string = "meta.data";
    desc_array[desc_ind].field_op.src.bit_offset = doca_get_reg_bit_offset(REG_FIELD_CT_STATE);
    desc_array[desc_ind].field_op.dst.field_string = "outer.eth.dst_mac";
    desc_array[desc_ind].field_op.dst.bit_offset =
        REV_BIT_OFFSETOF(union meta_eth_dst, reg_field_ct_state);
    desc_array[desc_ind].field_op.width = doca_get_reg_width(REG_FIELD_CT_STATE);
    desc_ind++;

    /* Copy REG_FIELD_CT_ZONE */
    desc_array[desc_ind].type = DOCA_FLOW_ACTION_COPY;
    desc_array[desc_ind].field_op.src.field_string = "meta.data";
    desc_array[desc_ind].field_op.src.bit_offset = doca_get_reg_bit_offset(REG_FIELD_CT_ZONE);
    desc_array[desc_ind].field_op.dst.field_string = "outer.eth.dst_mac";
    desc_array[desc_ind].field_op.dst.bit_offset =
        REV_BIT_OFFSETOF(union meta_eth_dst, reg_field_ct_zone);
    desc_array[desc_ind].field_op.width = doca_get_reg_width(REG_FIELD_CT_ZONE);
    desc_ind++;

    /* Copy REG_FIELD_CT_MARK */
    desc_array[desc_ind].type = DOCA_FLOW_ACTION_COPY;
    desc_array[desc_ind].field_op.src.field_string = "meta.data";
    desc_array[desc_ind].field_op.src.bit_offset = doca_get_reg_bit_offset(REG_FIELD_CT_MARK);
    desc_array[desc_ind].field_op.dst.field_string = "outer.ipv4.dst_ip";
    desc_array[desc_ind].field_op.dst.bit_offset = 0;
    desc_array[desc_ind].field_op.width = doca_get_reg_width(REG_FIELD_CT_MARK);
    desc_ind++;

    /* Copy REG_FIELD_CT_LABEL_ID */
    desc_array[desc_ind].type = DOCA_FLOW_ACTION_COPY;
    desc_array[desc_ind].field_op.src.field_string = "meta.data";
    desc_array[desc_ind].field_op.src.bit_offset = doca_get_reg_bit_offset(REG_FIELD_CT_LABEL_ID);
    desc_array[desc_ind].field_op.dst.field_string = "outer.ipv4.src_ip";
    desc_array[desc_ind].field_op.dst.bit_offset = 0;
    desc_array[desc_ind].field_op.width = doca_get_reg_width(REG_FIELD_CT_LABEL_ID);
    desc_ind++;

    ovs_assert(desc_ind == n_descs);

    doca_set_reg_val_mask(&actions.meta, &actions_masks.meta, REG_FIELD_TAG0, 0);

    if (doca_flow_pipe_cfg_set_name(cfg, pipe_name) ||
        doca_flow_pipe_cfg_set_type(cfg, DOCA_FLOW_PIPE_BASIC) ||
        doca_flow_pipe_cfg_set_is_root(cfg, false) ||
        doca_flow_pipe_cfg_set_nr_entries(cfg, 1) ||
        doca_flow_pipe_cfg_set_match(cfg, &match, &match) ||
        doca_flow_pipe_cfg_set_actions(cfg, actions_arr, actions_masks_arr, descs_arr,
                                       ARRAY_SIZE(actions_arr))) {
        VLOG_ERR("%s: Could not set doca_flow_pipe_cfg", netdev_get_name(netdev));
        ret = -1;
        goto out;
    }

    pipe_ctx = &ctx->meta_transfer_ctx.meta_copy_pipe_ctx;
    ret = doca_flow_pipe_create(cfg, &fwd, NULL, &pipe_ctx->pipe);
    if (ret) {
        VLOG_ERR("%s: Failed to create basic pipe '%s': %d (%s)",
                 netdev_get_name(netdev), pipe_name, ret,
                 doca_error_get_descr(ret));
        goto out;
    }

out:
    if (!ret) {
        pipe_ctx->fwd_group_ctx = next_group_ctx;
    } else {
        doca_pipe_group_ctx_unref(next_group_ctx);
    }
    doca_flow_pipe_cfg_destroy(cfg);
    return ret;
}

static int
doca_meta_transfer_init(struct netdev *netdev, struct doca_offload_esw_ctx *ctx)
{
    ctx->meta_transfer_ctx.netdev = netdev;
    /* The meta-push pipe depends on the meta-copy pipe. */
    if (doca_meta_copy_pipe_init(netdev, ctx)) {
        return -1;
    }

    if (doca_meta_copy_rule_init(netdev, ctx)) {
        return -1;
    }

    if (doca_meta_push_pipe_init(netdev, ctx)) {
        return -1;
    }

    if (doca_meta_push_rule_init(netdev, ctx)) {
        return -1;
    }

    return 0;
}

static void
doca_offload_esw_ctx_uninit(void *ctx_)
{
    struct doca_offload_esw_ctx *ctx = ctx_;

    ovs_rwlock_wrlock(&ctx->port_rwlock);
    /* Eswitch uninit means that multiple pipes will get
     * destroyed. All resizes must have been completed by then. */
    doca_offload_resolve_pipe_resize(ctx);
    /* We can immediately unlock as this port has had all its offload operations
     * disabled. There should not be any further insertion possibly
     * happening if this function is called. As a consequence, there
     * is no chance that another pipe resize could be triggered on this
     * eswitch. */
    ovs_rwlock_unlock(&ctx->port_rwlock);

    atomic_store_explicit(&ovs_doca_eswitch_active_ids[ctx->esw_id], false,
                          memory_order_release);

    /* The fixed rule insertions were counted in the counters of
     * the netdev that issued the eswitch context init.
     *
     * Destroying the fixed rule is done only when the last netdev
     * using this eswitch context is being removed.
     *
     * We cannot keep track of the original init netdev without inducing
     * a circular dependency.
     *
     * So remove the fixed rules without counting the deletions
     * in the uninit netdev. As all netdevs related to this eswitch
     * are meant to be removed after this, the original counts will
     * have been removed once the uninit has finished.
     */
    if (ctx->sample_ctx.root_sample_entry) {
        doca_offload_remove_entry(ctx, AUX_QUEUE, DOCA_FLOW_NO_WAIT,
                                  ctx->sample_ctx.root_sample_entry);
        doca_pipe_group_mask_ctx_unref(ctx->sample_ctx.root_sample_entry_mctx);
    }
    if (ctx->sample_ctx.sample_pipe_entry) {
        doca_offload_remove_entry(ctx, AUX_QUEUE, DOCA_FLOW_NO_WAIT,
                                  ctx->sample_ctx.sample_pipe_entry);
        doca_pipe_group_mask_ctx_unref(ctx->sample_ctx.sample_pipe_entry_mctx);
    }
    if (ctx->sample_ctx.sample_pipe_miss_entry) {
        doca_offload_remove_entry(ctx, AUX_QUEUE, DOCA_FLOW_NO_WAIT,
                                  ctx->sample_ctx.sample_pipe_miss_entry);
        doca_pipe_group_mask_ctx_unref(ctx->sample_ctx.sample_pipe_miss_entry_mctx);
    }
    doca_pipe_group_ctx_unref(ctx->sample_ctx.sample_pipe_ctx);
    if (ctx->sample_ctx.mirror_pipe_entry) {
        doca_offload_remove_entry(ctx, AUX_QUEUE, DOCA_FLOW_NO_WAIT,
                                  ctx->sample_ctx.mirror_pipe_entry);
    }
    doca_flow_pipe_destroy(ctx->sample_ctx.mirror_pipe_ctx.pipe);
    if (ctx->sample_ctx.postmirror_pipe_entry) {
        doca_offload_remove_entry(ctx, AUX_QUEUE, DOCA_FLOW_NO_WAIT,
                                  ctx->sample_ctx.postmirror_pipe_entry);
        doca_pipe_group_mask_ctx_unref(ctx->sample_ctx.postmirror_pipe_entry_mctx);
    }
    if (ctx->sample_ctx.postmirror_lacp_pipe_entry) {
        doca_offload_remove_entry(ctx, AUX_QUEUE, DOCA_FLOW_NO_WAIT,
                                  ctx->sample_ctx.postmirror_lacp_pipe_entry);
        doca_pipe_group_mask_ctx_unref(ctx->sample_ctx.postmirror_lacp_pipe_entry_mctx);
    }
    doca_pipe_group_ctx_unref(ctx->sample_ctx.postmirror_pipe_ctx);
    doca_pipe_group_ctx_unref(ctx->post_meter_pipe_group_ctx);

    if (ctx->ct.pipe) {
        doca_ct_zones_uninit(ctx);
        doca_offload_ct_pipe_uninit(ctx);
    } else if (ctx->shared_ct_counter_id_pool) {
        doca_ct_zones_uninit(ctx);
        doca_ct_pipes_destroy(ctx);
    }

    doca_pipe_group_ctx_unref(ctx->root_pipe_group_ctx);
    ctx->root_pipe_group_ctx = NULL;
    /* DOCA doesn't provide an api to unbind shared counters
     * and they will remain bound until the port is destroyed.
     */
    if (ctx->shared_counter_id_pool) {
        id_fpool_destroy(ctx->shared_counter_id_pool);
        ctx->shared_counter_id_pool = NULL;
    }
    if (ctx->shared_ct_counter_id_pool) {
        id_fpool_destroy(ctx->shared_ct_counter_id_pool);
        ctx->shared_ct_counter_id_pool = NULL;
    }
    if (ctx->shared_mtr_flow_id_pool) {
        /* DOCA doesn't provide an api to unbind shared meters
         * and they will remain bound until the port is destroyed.
         */
        id_fpool_destroy(ctx->shared_mtr_flow_id_pool);
        ctx->shared_mtr_flow_id_pool = NULL;
    }
    if (esw_id_pool) {
        id_fpool_free_id(esw_id_pool, 0, ctx->esw_id);
    }

    doca_hash_pipe_ctx_uninit(ctx, ctx->hash_pipe_ctx);
    doca_pre_miss_rules_uninit(ctx);
    doca_pre_miss_pipe_uninit(ctx);
    doca_post_sw_meter_pipe_uninit(ctx);
    doca_sw_meter_pipe_uninit(ctx);
    doca_meta_transfer_uninit(ctx);

    for (unsigned int qid = 0; qid < MAX_OFFLOAD_QUEUE_NB; qid++) {
        struct destroy_pipe_elem *elem;

        LIST_FOR_EACH_POP (elem, node, &ctx->destroy_pipe_lists[qid]) {
            ovs_rwlock_wrlock(&ctx->port_rwlock);
            doca_flow_pipe_destroy(elem->pipe);
            ovs_rwlock_unlock(&ctx->port_rwlock);
            free(elem);
        }
    }

    if (ctx->gnv_opt_parser.parser) {
        doca_flow_parser_geneve_opt_destroy(ctx->gnv_opt_parser.parser);
        ctx->gnv_opt_parser.parser = NULL;
        if (ovsthread_once_start(&ctx->gnv_opt_parser.once)) {
            ovsthread_once_reset(&ctx->gnv_opt_parser.once);
        }
        ovs_mutex_destroy(&ctx->gnv_opt_parser.once.mutex);
    }

    ovs_rwlock_destroy(&ctx->port_rwlock);
    ovs_mutex_destroy(&ctx->ct_zones_mutex);
    free(ctx->offload_queues);
    ctx->offload_queues = NULL;
    ctx->esw_port = NULL;
}

static int
doca_offload_esw_ctx_init(void *ctx_, void *arg_)
{
    struct netdev *netdev = (struct netdev *) arg_;
    struct doca_offload_esw_ctx *ctx = ctx_;
    struct netdev_offload_dpdk_data *data;
    struct doca_flow_port *doca_port;
    unsigned int n_queues;

    /* Set the esw-ctx reference early.
     * It is required by some inner calls.
     * Beware that is it incomplete and should be handled
     * with care. */
    data = (struct netdev_offload_dpdk_data *)
        ovsrcu_get(void *, &netdev->hw_info.offload_data);
    data->eswitch_ctx = ctx;

    n_queues = ovs_doca_n_offload_queues();
    if (ovs_doca_flow_ct_enabled) {
        n_queues += dpif_netdev_get_n_pmd_threads();
    }
    ctx->offload_queues = xcalloc(n_queues, sizeof(ctx->offload_queues[0]));

    for (unsigned int qid = 0; qid < MAX_OFFLOAD_QUEUE_NB; qid++) {
        ovs_list_init(&ctx->resized_pipe_lists[qid]);
        ovs_list_init(&ctx->destroy_pipe_lists[qid]);
    }

    ovs_refcount_init(&ctx->pipe_resizing);
    /* The shared resources are needed by the
     * sw-meter, which are needed by all other pipes.
     */
    doca_port = netdev_dpdk_doca_port_get(netdev);
    ctx->esw_port = doca_flow_port_switch_get(doca_port);

    shared_counter_id_pool_init(ctx);
    shared_mtr_flow_id_pool_init(ctx);
    ovs_assert(ctx->shared_counter_id_pool);

    /* This mutex is used on error rollback, it must be initialized
     * before any reference to the 'error' label below. */
    ovs_mutex_init(&ctx->ct_zones_mutex);
    ovs_rwlock_init(&ctx->port_rwlock);

    if (doca_bind_shared_meters(ctx)) {
        goto error;
    }

    /* The sw-meter pipes depend on the meta transfer infrastructure. */
    if (doca_meta_transfer_init(netdev, ctx)) {
        goto error;
    }

    /* The sw-meter pipe depends on the post-sw-meter pipe. */
    if (doca_post_sw_meter_pipe_init(netdev, ctx)) {
        goto error;
    }

    /* The sw-meter pipe is referenced by each control and
     * basic pipes as default miss target. Initialize it
     * as soon as possible. */
    if (doca_sw_meter_pipe_init(netdev, ctx)) {
        goto error;
    }

    if (doca_pre_miss_pipe_init(netdev, ctx)) {
        goto error;
    }

    if (doca_pre_miss_rules_init(netdev, ctx)) {
        goto error;
    }
    ctx->root_pipe_group_ctx = doca_pipe_group_ctx_ref(netdev, 0);
    if (ctx->root_pipe_group_ctx == NULL) {
        goto error;
    }

    doca_port = netdev_dpdk_doca_port_get(netdev);
    ctx->esw_port = doca_flow_port_switch_get(doca_port);
    ctx->gnv_opt_parser.once =
        (struct ovsthread_once) OVSTHREAD_ONCE_INITIALIZER;

    if (conntrack_offload_is_enabled()) {
        if (ovs_doca_flow_ct_enabled) {
            if (doca_offload_ct_pipe_init(netdev, ctx)) {
                goto error;
            }
        } else if (ctx->shared_ct_counter_id_pool) {
            if (doca_ct_pipes_init(netdev, ctx)) {
                goto error;
            }
        }
    }

    if (doca_bind_shared_cntrs(ctx)) {
        goto error;
    }

    if (doca_post_meter_pipe_init(netdev, ctx)) {
        goto error;
    }

    ctx->hash_pipe_ctx = doca_hash_pipe_ctx_init(netdev);
    if (!ctx->hash_pipe_ctx) {
        goto error;
    }

    if (doca_sample_post_mirror_pipe_init(netdev, ctx)) {
        goto error;
    }

    if (doca_bind_sample_shared_mirror(ctx)) {
        goto error;
    }

    if (doca_sample_mirror_pipe_init(netdev, ctx)) {
        goto error;
    }

    if (doca_sample_pipe_init(netdev, ctx)) {
        goto error;
    }
    atomic_store_explicit(&ovs_doca_eswitch_active_ids[ctx->esw_id], true,
                          memory_order_release);

    return 0;

error:
    /* Roll-back esw-ctx access on error. */
    data->eswitch_ctx = NULL;
    VLOG_ERR("%s: Failed to init eswitch %d",
             netdev_get_name(netdev),
             netdev_dpdk_get_esw_mgr_port_id(netdev));
    doca_offload_esw_ctx_uninit(ctx);
    return -1;
}

static struct ds *
dump_doca_eswitch(struct ds *s, void *key_, void *ctx_ OVS_UNUSED,
                  void *arg_ OVS_UNUSED)
{
    struct doca_flow_port *esw_port = key_;

    ds_put_format(s, "esw_port=%p, ", esw_port);

    return s;
}

static struct refmap *doca_eswitch_rfm;

static void
doca_eswitch_init(void)
{
    static struct ovsthread_once init_once = OVSTHREAD_ONCE_INITIALIZER;

    if (ovsthread_once_start(&init_once)) {
        doca_eswitch_rfm = refmap_create("doca-eswitch-map",
                                         sizeof(struct doca_flow_port *),
                                         sizeof(struct doca_offload_esw_ctx),
                                         doca_offload_esw_ctx_init,
                                         doca_offload_esw_ctx_uninit,
                                         dump_doca_eswitch, false);

        ovsthread_once_done(&init_once);
    }
}

/* Get the current eswitch context for this netdev,
 * /!\ without taking a reference, and without creating it!
 * The eswitch context must have been initialized once
 * beforehand using 'doca_offload_esw_ctx_ref()' for this netdev.
 */
struct doca_offload_esw_ctx *
doca_offload_esw_ctx_get(const struct netdev *netdev)
{
    struct netdev_offload_dpdk_data *data;

    data = (struct netdev_offload_dpdk_data *)
        ovsrcu_get(void *, &netdev->hw_info.offload_data);
    return data ? data->eswitch_ctx : NULL;
}

static struct doca_offload_esw_ctx *
doca_offload_esw_ctx_ref(struct netdev *netdev)
{
    struct doca_flow_port *doca_port = netdev_dpdk_doca_port_get(netdev);
    struct doca_flow_port *esw_port;

    esw_port = doca_flow_port_switch_get(doca_port);

    doca_eswitch_init();
    return refmap_ref(doca_eswitch_rfm, &esw_port, netdev);
}

static void
doca_offload_esw_ctx_unref(struct doca_offload_esw_ctx *ctx)
{
    refmap_unref(doca_eswitch_rfm, ctx);
}

static bool
add_sw_meter_match(struct netdev *netdev)
{
    struct netdev_offload_dpdk_data *data;
    struct doca_offload_esw_ctx *esw_ctx;
    struct doca_flow_monitor monitor;
    struct doca_flow_match match;
    struct rte_flow_error error;
    struct doca_flow_fwd fwd;
    struct fixed_rule *fr;
    bool ret = true;

    data = (struct netdev_offload_dpdk_data *)
        ovsrcu_get(void *, &netdev->hw_info.offload_data);

    if (!data) {
        return false;
    }

    esw_ctx = doca_offload_esw_ctx_get(netdev);

    memset(&match, 0, sizeof match);
    memset(&monitor, 0, sizeof monitor);

    match.parser_meta.port_meta = netdev_dpdk_get_port_id(netdev);
    match.parser_meta.meter_color = DOCA_FLOW_METER_COLOR_RED;
    monitor.counter_type = DOCA_FLOW_RESOURCE_TYPE_NON_SHARED;

   if (netdev_dpdk_sw_meter_get_dry_run(netdev)) {
        struct doca_basic_pipe_ctx *meta_push_pipe_ctx = NULL;

        meta_push_pipe_ctx = &esw_ctx->meta_transfer_ctx.meta_push_pipe_ctx;
        fwd.type = DOCA_FLOW_FWD_PIPE;
        fwd.next_pipe = meta_push_pipe_ctx->pipe;
    } else {
        fwd.type = DOCA_FLOW_FWD_DROP;
    }

    fr = &data->sw_meter_match;
    if (create_doca_basic_flow_entry(netdev, AUX_QUEUE,
                                     esw_ctx->post_sw_meter_pipe_ctx.pipe,
                                     &match, NULL, &monitor, &fwd, &fr->doh.dfh.flow,
                                     &error, true)) {
        VLOG_ERR("%s: Failed to create sw-meter red rule",
                 netdev_get_name(netdev));
        ret = false;
    }

    return ret;
}

static bool
add_sw_meter_apply(struct netdev *netdev)
{
    struct netdev_offload_dpdk_data *data;
    struct doca_offload_esw_ctx *esw_ctx;
    struct doca_flow_monitor monitor;
    struct doca_flow_match match;
    struct rte_flow_error error;
    struct fixed_rule *fr;
    bool ret = true;
    int port_id;

    data = (struct netdev_offload_dpdk_data *)
        ovsrcu_get(void *, &netdev->hw_info.offload_data);

    if (!data) {
        return false;
    }

    memset(&monitor, 0, sizeof monitor);
    memset(&match, 0, sizeof match);

    esw_ctx = doca_offload_esw_ctx_get(netdev);
    port_id = netdev_dpdk_get_port_id(netdev);

    match.parser_meta.port_meta = port_id;
    monitor.meter_type = DOCA_FLOW_RESOURCE_TYPE_SHARED;
    monitor.shared_meter.shared_meter_id = ovs_doca_sw_meter_id(port_id,
                                                                esw_ctx->esw_id);

    fr = &data->sw_meter_apply;
    if (create_doca_basic_flow_entry(netdev, AUX_QUEUE,
                                     esw_ctx->sw_meter_pipe_ctx.pipe,
                                     &match, NULL, &monitor, NULL, &fr->doh.dfh.flow,
                                     &error, true)) {
        VLOG_ERR("%s: Failed to create sw-meter rule",
                 netdev_get_name(netdev));
        ret = false;
    }

    return ret;
}

static void
dpdk_offload_doca_sw_meter_uninit(struct doca_offload_esw_ctx *esw,
                                  struct netdev *netdev)
{
    struct netdev_offload_dpdk_data *data;

    data = (struct netdev_offload_dpdk_data *)
        ovsrcu_get(void *, &netdev->hw_info.offload_data);

    if (!data) {
        return;
    }

    doca_offload_fixed_rule_uninit(esw, netdev, &data->sw_meter_apply);
    doca_offload_fixed_rule_uninit(esw, netdev, &data->sw_meter_match);
}

/* Configure the shared meter that is used for sw-meter,
 * Insert the 'red-match' rule in post-sw-meter.
 * Insert the 'apply-meter' rule in sw-meter.
 */
static int
dpdk_offload_doca_sw_meter_init(struct doca_offload_esw_ctx *esw,
                                struct netdev *netdev)
{
    struct ofputil_meter_config config;
    struct ofputil_meter_band band;
    ofproto_meter_id dp_meter_id;
    int port_id;

    if (!netdev_dpdk_sw_meter_get(netdev, &config, &band)) {
        return 0;
    }

    port_id = netdev_dpdk_get_port_id(netdev);
    dp_meter_id.uint32 = ovs_doca_sw_meter_dp_ext_id(port_id);
    config.meter_id = dp_meter_id.uint32;

    if (netdev_dpdk_meter_set(dp_meter_id, &config)) {
        VLOG_ERR("%s: Failed to configure shared meter resource for "
                 "sw-meter config.", netdev_get_name(netdev));
        return -1;
    }

    if (!add_sw_meter_match(netdev)) {
        goto error;
    }

    if (!add_sw_meter_apply(netdev)) {
        goto error;
    }

    if (VLOG_IS_DBG_ENABLED()) {
        struct ds s = DS_EMPTY_INITIALIZER;

        if (config.flags & OFPMF13_PKTPS) {
            ds_put_cstr(&s, "pps:");
        } else if (config.flags & OFPMF13_KBPS) {
            ds_put_cstr(&s, "bps:");
            /* The sw-meter configuration is in bps, not Kbps. */
            band.rate *= 1000;
            band.burst_size *= 1000;
        }
        ds_put_format(&s, "%"PRIu32":%"PRIu32,
                      band.rate, band.burst_size);
        VLOG_DBG("%s: Installed sw-meter: %s",
                 netdev_get_name(netdev), ds_cstr(&s));
        ds_destroy(&s);
    }

    return 0;

error:
    dpdk_offload_doca_sw_meter_uninit(esw, netdev);
    return -1;
}

static void
dpdk_offload_doca_aux_tables_uninit(struct netdev *netdev)
{
    struct netdev_offload_dpdk_data *data;

    if (!netdev_dpdk_is_ethdev(netdev)) {
        return;
    }

    data = (struct netdev_offload_dpdk_data *)
        ovsrcu_get(void *, &netdev->hw_info.offload_data);
    if (!data) {
        return;
    }

    dpdk_offload_doca_sw_meter_uninit(data->eswitch_ctx, netdev);

    doca_offload_esw_ctx_unref(data->eswitch_ctx);
    data->eswitch_ctx = NULL;
}

static int
dpdk_offload_doca_aux_tables_init(struct netdev *netdev)
{
    struct netdev_offload_dpdk_data *data;
    struct doca_offload_esw_ctx *ctx;

    if (!netdev_dpdk_is_ethdev(netdev)) {
        return 0;
    }

    ctx = doca_offload_esw_ctx_ref(netdev);
    if (!ctx) {
        VLOG_ERR("%s: Failed to get doca eswitch ctx", netdev_get_name(netdev));
        return -1;
    }

    data = (struct netdev_offload_dpdk_data *)
        ovsrcu_get(void *, &netdev->hw_info.offload_data);
    data->eswitch_ctx = ctx;

    if (dpdk_offload_doca_sw_meter_init(ctx, netdev)) {
        VLOG_ERR("%s: Failed to install sw-meter.", netdev_get_name(netdev));
        return -1;
    }

    return 0;
}

static void
log_conn_rule(struct conn *conn, int dir, int error,
              uint32_t ct_match_zone_id,
              uint32_t ct_action_label_id)
{
    struct ds s;

    if (error) {
        if (VLOG_DROP_ERR(&rl)) {
            return;
        }
    } else {
        if (VLOG_DROP_DBG(&rl)) {
            return;
        }
    }

    ds_init(&s);

    conntrack_offload_dump(&s, conn, dir, ct_match_zone_id,
                           ct_action_label_id, 0);
    if (error) {
        VLOG_ERR("conn create failed: %s", ds_cstr(&s));
    } else {
        VLOG_DBG("conn create: %s", ds_cstr(&s));
    }

    ds_destroy(&s);
}

static struct doca_split_prefix_ctx *
doca_insert_ip6_prefix_conn(struct netdev *netdev,
                            const struct ct_match *ct_match,
                            uint32_t ct_match_zone_id,
                            enum ct_tp_type tp_type,
                            uint32_t *prefix_id)
{
    struct doca_split_prefix_key prefix_key;
    struct doca_split_prefix_ctx *pctx;
    struct doca_split_prefix_arg args;
    struct doca_offload_esw_ctx *ctx;
    struct doca_flow_match *dspec;

    memset(&prefix_key, 0, sizeof(prefix_key));
    dspec = &prefix_key.spec;

    /* IPv6 CT prefix match: port_id + zone + src_ip + next_proto. */
    doca_set_reg_val(&dspec->meta, REG_FIELD_CT_ZONE, ct_match_zone_id);
    dspec->parser_meta.port_meta = netdev_dpdk_get_port_id(netdev);

    dspec->outer.ip6.next_proto = tp_type == CT_TP_TCP ?
                                  IPPROTO_TCP :
                                  IPPROTO_UDP;
    memcpy(dspec->outer.ip6.src_ip, &ct_match->key.src.addr.ipv6,
           sizeof dspec->outer.ip6.src_ip);

    ctx = doca_offload_esw_ctx_get(netdev);
    args.netdev = netdev;
    args.spec = dspec;
    args.mask = &prefix_key.mask;
    args.prefix_is_group = false;
    args.prefix_pipe = ctx->ct_ip6_prefix.pipe;
    args.suffix_pipe = ctx->ct_pipes[CT_NW_IP6][tp_type].pipe;

    prefix_key.prefix_pipe = args.prefix_pipe;

    pctx = split_prefix_ctx_ref(&prefix_key, &args, prefix_id);
    if (!pctx) {
        VLOG_WARN_RL(&rl, "%s: Failed to associate a CT IPv6 prefix with a "
                     "prefix ID", netdev_get_name(netdev));
        return NULL;
    }

    return pctx;
}

static int
doca_offload_insert_conn_dir(struct doca_offload_esw_ctx *esw,
                             unsigned int qid,
                             struct netdev *netdev,
                             struct ct_flow_offload_item ct_offload[1],
                             struct conn_item *ci,
                             struct act_resources *act_res)
{
    struct doca_split_prefix_ctx *pctx = NULL;
    struct doca_flow_header_format *dhdr;
    const struct ct_match *ct_match;
    struct doca_flow_actions dacts;
    struct doca_flow_monitor dmon;
    struct doca_flow_match dspec;
    struct doca_flow_pipe *pipe;
    struct rte_flow_error error;
    enum ct_nw_type nw_type;
    enum ct_tp_type tp_type;
    unsigned int prefix_id;
    uint32_t ct_state_spec;

    ct_match = &ct_offload->ct_match;

    tp_type = ct_match->key.nw_proto == IPPROTO_TCP ? CT_TP_TCP : CT_TP_UDP;
    if (ct_match->key.dl_type == htons(ETH_TYPE_IP)) {
        nw_type = CT_NW_IP4;
    } else if (ct_match->key.dl_type == htons(ETH_TYPE_IPV6)) {
        nw_type = CT_NW_IP6;
        /* No IPv6 conn request should be generated if not enabled. */
        ovs_assert(conntrack_offload_ipv6_is_enabled());
    } else {
        VLOG_DBG_RL(&rl, "Unsupported CT network type.");
        return -1;
    }

    memset(&dspec.meta, 0, sizeof dspec.meta);

    dhdr = &dspec.outer;

    /* IPv4 */
    if (nw_type == CT_NW_IP4) {
        dhdr->ip4.src_ip = (OVS_FORCE doca_be32_t) ct_match->key.src.addr.ipv4;
        dhdr->ip4.dst_ip = (OVS_FORCE doca_be32_t) ct_match->key.dst.addr.ipv4;
        dhdr->ip4.next_proto = ct_match->key.nw_proto;
    /* IPv6 */
    } else {
        /* ip6 src match will be in the prefix rule while dst match in the
         * suffix rule
         */
        pctx = doca_insert_ip6_prefix_conn(netdev, ct_match, act_res->ct_match_zone_id,
                                           tp_type, &prefix_id);
        if (!pctx) {
            return -1;
        }

        memcpy(dhdr->ip6.dst_ip, &ct_match->key.dst.addr.ipv6,
               sizeof dhdr->ip6.dst_ip);
        doca_set_reg_val(&dspec.meta, REG_FIELD_SCRATCH, prefix_id);
    }

    if (tp_type == CT_TP_TCP) {
        dhdr->tcp.l4_port.src_port = (OVS_FORCE doca_be16_t) ct_match->key.src.port;
        dhdr->tcp.l4_port.dst_port = (OVS_FORCE doca_be16_t) ct_match->key.dst.port;
    } else {
        dhdr->udp.l4_port.src_port = (OVS_FORCE doca_be16_t) ct_match->key.src.port;
        dhdr->udp.l4_port.dst_port = (OVS_FORCE doca_be16_t) ct_match->key.dst.port;
    }

    doca_set_reg_val(&dspec.meta, REG_FIELD_CT_ZONE, act_res->ct_match_zone_id);

    dspec.parser_meta.port_meta = netdev_dpdk_get_port_id(netdev);

    dhdr = &dacts.outer;

    /* Common part for all CT, plain and NAT. */

    if (nw_type == CT_NW_IP4) {
        dhdr->l3_type = DOCA_FLOW_L3_TYPE_IP4;
        dhdr->ip4.src_ip = (OVS_FORCE doca_be32_t) ct_match->key.src.addr.ipv4;
        dhdr->ip4.dst_ip = (OVS_FORCE doca_be32_t) ct_match->key.dst.addr.ipv4;
    } else {
        dhdr->l3_type = DOCA_FLOW_L3_TYPE_IP6;
        memcpy(&dhdr->ip6.src_ip, &ct_match->key.src.addr.ipv6,
               sizeof dhdr->ip6.src_ip);
        memcpy(&dhdr->ip6.dst_ip, &ct_match->key.dst.addr.ipv6,
               sizeof dhdr->ip6.dst_ip);
    }

    if (tp_type == CT_TP_TCP) {
        dhdr->l4_type_ext = DOCA_FLOW_L4_TYPE_EXT_TCP;
        dhdr->tcp.l4_port.src_port = (OVS_FORCE doca_be16_t) ct_match->key.src.port;
        dhdr->tcp.l4_port.dst_port = (OVS_FORCE doca_be16_t) ct_match->key.dst.port;
    } else {
        dhdr->l4_type_ext = DOCA_FLOW_L4_TYPE_EXT_UDP;
        dhdr->udp.l4_port.src_port = (OVS_FORCE doca_be16_t) ct_match->key.src.port;
        dhdr->udp.l4_port.dst_port = (OVS_FORCE doca_be16_t) ct_match->key.dst.port;
    }

    /* For NAT translate the relevant fields. */
    if (ct_offload->nat.mod_flags) {
        if (nw_type == CT_NW_IP4) {
            if (ct_offload->nat.mod_flags & NAT_ACTION_SRC) {
                dhdr->ip4.src_ip = (OVS_FORCE doca_be32_t) ct_offload->nat.key.src.addr.ipv4;
            } else if (ct_offload->nat.mod_flags & NAT_ACTION_DST) {
                dhdr->ip4.dst_ip = (OVS_FORCE doca_be32_t) ct_offload->nat.key.dst.addr.ipv4;
            }
        } else {
            if (ct_offload->nat.mod_flags & NAT_ACTION_SRC) {
                memcpy(&dhdr->ip6.src_ip, &ct_offload->nat.key.src.addr.ipv6,
                       sizeof dhdr->ip6.src_ip);
            } else if (ct_offload->nat.mod_flags & NAT_ACTION_DST) {
                memcpy(&dhdr->ip6.dst_ip, &ct_offload->nat.key.dst.addr.ipv6,
                       sizeof dhdr->ip6.dst_ip);
            }
        }
        if (ct_offload->nat.mod_flags & NAT_ACTION_SRC_PORT) {
            if (tp_type == CT_TP_TCP) {
                dhdr->tcp.l4_port.src_port = (OVS_FORCE doca_be16_t) ct_offload->nat.key.src.port;
            } else {
                dhdr->udp.l4_port.src_port = (OVS_FORCE doca_be16_t) ct_offload->nat.key.src.port;
            }
        } else if (ct_offload->nat.mod_flags & NAT_ACTION_DST_PORT) {
            if (tp_type == CT_TP_TCP) {
                dhdr->tcp.l4_port.dst_port = (OVS_FORCE doca_be16_t) ct_offload->nat.key.dst.port;
            } else {
                dhdr->udp.l4_port.dst_port = (OVS_FORCE doca_be16_t) ct_offload->nat.key.dst.port;
            }
        }
    }

    memset(&dmon, 0, sizeof dmon);
    dmon.counter_type = DOCA_FLOW_RESOURCE_TYPE_SHARED;
    dmon.shared_counter.shared_counter_id = act_res->shared_count_ctx->res_id;

    memset(&dacts.meta, 0, sizeof dacts.meta);

    /* CT MARK */
    doca_set_reg_val(&dacts.meta, REG_FIELD_CT_MARK, ct_offload->mark_key);

    /* CT LABEL */
    doca_set_reg_val(&dacts.meta, REG_FIELD_CT_LABEL_ID, act_res->ct_action_label_id);

    /* CT STATE */
    ct_state_spec = ct_offload->ct_state;
    /* If any of the NAT bits is set, set both of them.
     * As the NAT action is executed, the 5-tuple of the packet is modified.
     * A revisit with the post-NAT tuple would miss. */
    if (ct_offload->ct_state & OVS_CS_F_NAT_MASK) {
        ct_state_spec |= OVS_CS_F_NAT_MASK;
    }
    doca_set_reg_val(&dacts.meta, REG_FIELD_CT_STATE, ct_state_spec);

    memset(ci, 0, sizeof *ci);

    dacts.action_idx = 0;
    pipe = esw->ct_pipes[nw_type][tp_type].pipe;
    if (create_doca_basic_flow_entry(netdev, qid, pipe, &dspec,
                                     &dacts, &dmon, NULL, &ci->doca_flow,
                                     &error, false)) {
        VLOG_WARN_RL(&rl, "%s: Failed to create ct entry: Error %d (%s)",
                     netdev_get_name(netdev), error.type, error.message);
        return -1;
    }
    ci->valid = true;
    ci->curr_split_ctx = pctx;

    return 0;
}

static int
dpdk_offload_doca_destroy_conn_dir(struct netdev *netdev,
                                   struct conn_item *ci,
                                   struct rte_flow_error *error)
{
    struct doca_offload_esw_ctx *esw = doca_offload_esw_ctx_get(netdev);
    unsigned int qid = netdev_offload_thread_id();
    doca_error_t err;

    /* Deletion is always synchronous.
     *
     * If async deletion is implemented, aux-table uninit calls deleting
     * entries will use the offload queues in conflict with offload threads
     * polling them during upkeep. It should result in a crash or
     * in a lockup of the queues. */
    err = doca_offload_remove_entry(esw, qid, DOCA_FLOW_NO_WAIT, ci->doca_flow);
    if (err) {
        if (error) {
            error->type = RTE_FLOW_ERROR_TYPE_HANDLE;
            error->message = doca_error_get_descr(err);
        }
        if (err == DOCA_ERROR_IN_USE) {
            return EAGAIN;
        }
        return -1;
    }

    if (ci->curr_split_ctx) {
        split_prefix_ctx_unref(ci->curr_split_ctx);
    }

    /* Netdev can only be NULL during aux tables uninit. */
    if (netdev) {
        dpdk_offload_counter_dec(netdev);
    }

    return 0;
}

static void
doca_flow_ct_translate_match(struct ct_flow_offload_item *item, uint32_t ct_zone_id,
                             struct doca_flow_ct_match *m)
{
    const struct ct_match *ct_match;
    enum ct_nw_type nw_type;

    ct_match = &item->ct_match;
    nw_type = ct_match->key.dl_type == htons(ETH_TYPE_IP) ? CT_NW_IP4 : CT_NW_IP6;

    if (nw_type == CT_NW_IP4) {
        m->ipv4.metadata = reg_field_spec(REG_FIELD_CT_ZONE, ct_zone_id);
        m->ipv4.src_ip = (OVS_FORCE doca_be32_t) ct_match->key.src.addr.ipv4;
        m->ipv4.dst_ip = (OVS_FORCE doca_be32_t) ct_match->key.dst.addr.ipv4;
        m->ipv4.next_proto = ct_match->key.nw_proto;
        m->ipv4.l4_port.src_port = (OVS_FORCE doca_be16_t) ct_match->key.src.port;
        m->ipv4.l4_port.dst_port = (OVS_FORCE doca_be16_t) ct_match->key.dst.port;
    } else {
        m->ipv6.metadata = reg_field_spec(REG_FIELD_CT_ZONE, ct_zone_id);
        memcpy(&m->ipv6.src_ip, &ct_match->key.src.addr.ipv6, sizeof m->ipv6.src_ip);
        memcpy(&m->ipv6.dst_ip, &ct_match->key.dst.addr.ipv6, sizeof m->ipv6.dst_ip);
        m->ipv6.next_proto = ct_match->key.nw_proto;
        m->ipv6.l4_port.src_port = (OVS_FORCE doca_be16_t) ct_match->key.src.port;
        m->ipv6.l4_port.dst_port = (OVS_FORCE doca_be16_t) ct_match->key.dst.port;
    }
}

static void
doca_flow_ct_actions_set(struct doca_flow_ct_actions *a, enum ct_nw_type nw_type, bool nat,
                         uint32_t ct_label, uint32_t ct_mark, uint8_t ct_state)
{
    struct doca_flow_meta ct_mask;

    memset(a, 0, sizeof *a);
    a->resource_type = DOCA_FLOW_RESOURCE_TYPE_NON_SHARED;
    a->data.action_idx = CT_ACTIONS_IDX(nw_type, nat);

    /* If any of the NAT bits is set, set both of them.
     * As the NAT action is executed, the 5-tuple of the packet is modified.
     * A revisit with the post-NAT tuple would miss. */
    if (ct_state & OVS_CS_F_NAT_MASK) {
        ct_state |= OVS_CS_F_NAT_MASK;
    }
    doca_set_reg_val(&a->data.meta, REG_FIELD_CT_STATE, ct_state);
    doca_set_reg_val(&a->data.meta, REG_FIELD_CT_MARK, ct_mark);
    doca_set_reg_val(&a->data.meta, REG_FIELD_CT_LABEL_ID, ct_label);

    /* DOCA-CT has a significant API difference from DOCA-flow regarding
     * meta fields. All register writes must be packed together in adjacent
     * tags, and the offsets are handled by DOCA-CT internally:
     * they must not be handled by the application.
     *
     * In the following example:
     *   struct doca_flow_meta template = {
     *       .u32[2] = 0xffffffff,
     *       .u32[4] = 0xffffffff,
     *   };
     *
     *   struct doca_flow_meta mask = {
     *       .u32[2] = 0xffffffff,
     *       .u32[4] = 0xff00ff00,
     *   };
     *
     *   struct doca_flow_meta value = {
     *       .u32[2] = 0x12345678,
     *       .u32[4] = 0x00120034,
     *   };
     *
     * In DOCA-flow, value.u32[4] should have been 0x12003400;
     * In DOCA-CT, it must instead be set to 0x120034 and DOCA-CT will use
     * the CT pipe mask to set it properly.
     *
     * Once all values have been written to the meta struct and binary OR-ed together,
     * move any field that has an offset where DOCA-CT expects it to be.
     */
    memset(&ct_mask, 0, sizeof ct_mask);
    doca_set_reg_mask(&ct_mask, REG_FIELD_CT_STATE);
    doca_set_reg_mask(&ct_mask, REG_FIELD_CT_MARK);
    doca_set_reg_mask(&ct_mask, REG_FIELD_CT_LABEL_ID);
    for (int i = 0; i < ARRAY_SIZE(a->data.meta.u32); i++) {
        int ctz = ctz32(ct_mask.u32[i]);

        if (ctz && ctz != 32) {
            a->data.meta.u32[i] >>= ctz;
        }
    }
}

static bool
doca_flow_ct_shared_actions_create(struct doca_offload_esw_ctx *esw,
                                   struct shared_ct_actions_key *key,
                                   uint32_t *id, struct rte_flow_error *error)
{
    unsigned int qid = doca_offload_ct_queue_id(netdev_offload_thread_id());
    struct doca_flow_ct_actions actions;
    doca_error_t err;

    /* For plain-CT, the packet header is not modified.
     * The network protocol version does not matter.
     * Here CT_NW_IP4 is arbitrary.
     * Similarly, 'nat' is always false for shared actions.
     */
    doca_flow_ct_actions_set(&actions, CT_NW_IP4, false,
                             key->ct_label, key->ct_mark, key->ct_state);

    err = doca_flow_ct_actions_add_shared(qid, esw->ct.pipe, &actions, 1, id);
    if (DOCA_IS_ERROR(err)) {
        error->type = RTE_FLOW_ERROR_TYPE_UNSPECIFIED;
        error->message = doca_error_get_descr(err);
        return false;
    }
    return true;
}

static void
doca_flow_ct_shared_actions_destroy(struct doca_offload_esw_ctx *esw, uint32_t id)
{
    unsigned int qid = doca_offload_ct_queue_id(netdev_offload_thread_id());
    doca_error_t err;

    err = doca_flow_ct_actions_rm_shared(qid, esw->ct.pipe, &id, 1);
    if (DOCA_IS_ERROR(err)) {
        VLOG_ERR("Failed to delete shared CT actions %" PRIu32": %s", id,
                 doca_error_get_descr(err));
    }
}

static void
doca_flow_ct_translate_actions(struct ct_flow_offload_item *item, uint32_t ct_label_id,
                               struct doca_flow_ct_actions *a)
{
    const struct ct_match *ct_match;
    enum ct_nw_type nw_type;
    bool nat;

    ct_match = &item->ct_match;
    nw_type = ct_match->key.dl_type == htons(ETH_TYPE_IP) ? CT_NW_IP4 : CT_NW_IP6;
    nat = item->nat.mod_flags;

    doca_flow_ct_actions_set(a, nw_type, nat, ct_label_id, item->mark_key, item->ct_state);

    if (!nat) {
        return;
    }

    if (item->nat.mod_flags & NAT_ACTION_SRC) {
        /* SNAT + PAT */
        if (nw_type == CT_NW_IP4) {
            a->data.ip4.src_ip = (OVS_FORCE doca_be32_t) item->nat.key.src.addr.ipv4;
            a->data.ip4.dst_ip = (OVS_FORCE doca_be32_t) ct_match->key.dst.addr.ipv4;
        } else {
            memcpy(&a->data.ip6.src_ip, &item->nat.key.src.addr.ipv6, sizeof a->data.ip6.src_ip);
            memcpy(&a->data.ip6.dst_ip, &ct_match->key.dst.addr.ipv6, sizeof a->data.ip6.dst_ip);
        }
        a->data.l4_port.src_port = (OVS_FORCE doca_be16_t) item->nat.key.src.port;
        a->data.l4_port.dst_port = (OVS_FORCE doca_be16_t) ct_match->key.dst.port;
    } else {
        /* DNAT + PAT */
        if (nw_type == CT_NW_IP4) {
            a->data.ip4.src_ip = (OVS_FORCE doca_be32_t) ct_match->key.src.addr.ipv4;
            a->data.ip4.dst_ip = (OVS_FORCE doca_be32_t) item->nat.key.dst.addr.ipv4;
        } else {
            memcpy(&a->data.ip6.src_ip, &ct_match->key.src.addr.ipv6, sizeof a->data.ip6.src_ip);
            memcpy(&a->data.ip6.dst_ip, &item->nat.key.dst.addr.ipv6, sizeof a->data.ip6.dst_ip);
        }
        a->data.l4_port.src_port = (OVS_FORCE doca_be16_t) ct_match->key.src.port;
        a->data.l4_port.dst_port = (OVS_FORCE doca_be16_t) item->nat.key.dst.port;
    }
}

static int
doca_offload_insert_conn_doca_ct(struct doca_offload_esw_ctx *esw,
                                 unsigned int qid,
                                 struct netdev *netdevs[CT_DIR_NUM],
                                 struct ct_flow_offload_item items[CT_DIR_NUM],
                                 struct conn_item *ci,
                                 struct act_resources *act_res,
                                 bool wait)
{
    struct doca_flow_pipe_entry *entry = NULL;
    doca_error_t err = DOCA_ERROR_UNKNOWN;
    uint32_t ct_entry_flags;
    struct {
        struct doca_flow_ct_match match;
        struct doca_flow_ct_actions actions;
    } dir[CT_DIR_NUM];

    if (items->ct_match.key.dl_type == htons(ETH_TYPE_IPV6)) {
        if (!conntrack_offload_ipv6_is_enabled()) {
            VLOG_ERR_RL(&rl, "%s/%s: CT-IPv6 offload is disabled",
                        netdev_get_name(netdevs[CT_DIR_INIT]),
                        netdev_get_name(netdevs[CT_DIR_REP]));
            return -1;
        }
    }

    memset(dir, 0, sizeof dir);

    for (int i = 0; i < CT_DIR_NUM; i++) {
        doca_flow_ct_translate_match(&items[i], act_res->ct_match_zone_id, &dir[i].match);
        if (act_res->shared_ct_actions_ctx[i]) {
            dir[i].actions.resource_type = DOCA_FLOW_RESOURCE_TYPE_SHARED;
            dir[i].actions.action_handle = act_res->shared_ct_actions_ctx[i]->res_id;
        } else {
            doca_flow_ct_translate_actions(&items[i], act_res->ct_action_label_id, &dir[i].actions);
        }
    }

    ct_entry_flags = wait ? 0 : DOCA_FLOW_CT_ENTRY_FLAGS_NO_WAIT;
    ct_entry_flags |= DOCA_FLOW_CT_ENTRY_FLAGS_COUNTER_ORIGIN;
    ct_entry_flags |= DOCA_FLOW_CT_ENTRY_FLAGS_COUNTER_REPLY;
    ct_entry_flags |= DOCA_FLOW_CT_ENTRY_FLAGS_COUNTER_SHARED;
    ct_entry_flags |= DOCA_FLOW_CT_ENTRY_FLAGS_DIR_ORIGIN;
    ct_entry_flags |= DOCA_FLOW_CT_ENTRY_FLAGS_DIR_REPLY;
    if (items->ct_match.key.dl_type == htons(ETH_TYPE_IPV6)) {
        ct_entry_flags |= DOCA_FLOW_CT_ENTRY_FLAGS_IPV6_ORIGIN;
        ct_entry_flags |= DOCA_FLOW_CT_ENTRY_FLAGS_IPV6_REPLY;
    }

    /* Prepare the entry with NULL matches: it will always miss and always allocate an entry. */
    err = doca_flow_ct_entry_prepare(qid, esw->ct.pipe, DOCA_FLOW_CT_ENTRY_FLAGS_ALLOC_ON_MISS,
                                     NULL, 0, NULL, 0, &entry, NULL);
    if (DOCA_IS_ERROR(err)) {
        VLOG_ERR_RL(&rl, "%s: failed to allocate DOCA-CT entry: %s",
                    netdev_get_name(netdevs[CT_DIR_INIT]), doca_error_get_descr(err));
        return err;
    }

    err = doca_flow_ct_add_entry(qid, esw->ct.pipe, ct_entry_flags,
                                 &dir[CT_DIR_INIT].match, &dir[CT_DIR_REP].match,
                                 &dir[CT_DIR_INIT].actions, &dir[CT_DIR_REP].actions,
                                 0, 0, 0, esw, entry);
    if (DOCA_IS_ERROR(err)) {
        VLOG_ERR_RL(&rl, "%s: failed to insert DOCA-CT entry: %d (%s)",
                    netdev_get_name(netdevs[CT_DIR_INIT]),
                    err, doca_error_get_descr(err));
        goto rollback_entry;
    } else {
        esw->offload_queues[qid].n_waiting_entries++;
    }

    dpdk_offload_counter_inc(netdevs[CT_DIR_INIT]);
    dpdk_offload_counter_inc(netdevs[CT_DIR_REP]);
    ci->doca_flow = entry;

    return 0;

rollback_entry:
    /* Ignore potential error from this call, to keep the current value of 'err'. */
    doca_flow_ct_entry_prepare_rollback(qid, esw->ct.pipe, entry);
    return err;
}

static int
doca_offload_destroy_conn_dirs(struct netdev *netdevs[CT_DIR_NUM], struct conn *conn,
                               struct rte_flow_error *error)
{
    struct doca_offload_esw_ctx *esw;
    struct ct_offload_handle *coh;
    int rets[CT_DIR_NUM];

    memset(rets, 0, sizeof rets);

    esw = doca_offload_esw_ctx_get(netdevs[CT_DIR_INIT] ? netdevs[CT_DIR_INIT]
                                                        : netdevs[CT_DIR_REP]);
    coh = conntrack_offload_get(conn);

    ovs_rwlock_rdlock(&esw->port_rwlock);
    for (int dir = 0; dir < CT_DIR_NUM; dir++) {
        if (!coh->dir[dir].conn_item.valid) {
            continue;
        }
        if (!netdevs[dir]) {
            continue;
        }
        rets[dir] = dpdk_offload_doca_destroy_conn_dir(netdevs[dir],
                                                       &coh->dir[dir].conn_item,
                                                       error);
    }
    ovs_rwlock_unlock(&esw->port_rwlock);
    netdev_offload_dpdk_free_conn_res(&coh->act_resources);
    return rets[CT_DIR_INIT] | rets[CT_DIR_REP];
}

static int
doca_offload_destroy_conn_doca_ct(struct netdev *netdevs[CT_DIR_NUM],
                                  struct conn *conn,
                                  struct rte_flow_error *error)
{
    struct doca_offload_esw_ctx *esw;
    doca_error_t err = DOCA_SUCCESS;
    struct ct_offload_handle *coh;
    struct conn_item *ci;
    unsigned int qid;
    int retries;

    coh = conntrack_offload_get(conn);
    ci = &coh->dir[0].conn_item;
    qid = doca_offload_ct_queue_id(netdev_offload_thread_id());
    esw = doca_offload_esw_ctx_get(netdevs[CT_DIR_INIT] ? netdevs[CT_DIR_INIT]
                                                        : netdevs[CT_DIR_REP]);

    if (ci->doca_flow == NULL) {
        goto free_resources;
    }

    retries = 100;
    while (retries-- > 0) {
        err = doca_flow_ct_rm_entry(qid, esw->ct.pipe,
                                    DOCA_FLOW_CT_ENTRY_FLAGS_NO_WAIT,
                                    ci->doca_flow);
        if (err == DOCA_SUCCESS) {
            esw->offload_queues[qid].n_waiting_entries++;
            err = doca_offload_complete_queue_n(esw, qid,
                                                OVS_DOCA_CT_QUEUE_DEPTH);
            if (netdevs[CT_DIR_INIT]) {
                dpdk_offload_counter_dec(netdevs[CT_DIR_INIT]);
            }
            if (netdevs[CT_DIR_REP]) {
                dpdk_offload_counter_dec(netdevs[CT_DIR_REP]);
            }
        }
        if (err != DOCA_ERROR_IN_PROGRESS) {
            break;
        }
    }

free_resources:
    netdev_offload_dpdk_free_conn_res(&coh->act_resources);
    if (DOCA_IS_ERROR(err)) {
        error->type = RTE_FLOW_ERROR_TYPE_UNSPECIFIED;
        error->message = doca_error_get_descr(err);
        return -1;
    }
    return 0;
}

static int
dpdk_offload_doca_destroy_conn(struct netdev *netdevs[CT_DIR_NUM], struct conn *conn,
                               struct rte_flow_error *error)
{
    if (ovs_doca_flow_ct_enabled) {
        return doca_offload_destroy_conn_doca_ct(netdevs, conn, error);
    } else {
        return doca_offload_destroy_conn_dirs(netdevs, conn, error);
    }
}

static int
dpdk_offload_doca_insert_conns(struct netdev *netdevs[CT_DIR_NUM], struct batch *conns)
{
    struct doca_offload_esw_ctx *esw = doca_offload_esw_ctx_get(netdevs[CT_DIR_INIT]);
    struct rte_flow_error error;
    struct conn *conn;
    unsigned int qid;

    /* If DOCA-CT is enabled:
     *   -> Do not allocate shared counters, they are handled by DOCA-CT.
     *   -> Do try to merge CT actions if possible (plain CT).
     * Else:
     *   -> Allocate one shared counter used in both directions.
     *   -> Impossible to share CT actions with basic pipe CT.
     */
    netdev_offload_dpdk_get_conns_res(netdevs, conns, false,
                                      !ovs_doca_flow_ct_enabled,
                                      ovs_doca_flow_ct_enabled);

    qid = doca_offload_ct_queue_id(netdev_offload_thread_id());

    ovs_rwlock_rdlock(&esw->port_rwlock);
    BATCH_FOREACH_POP (idx, conn, conns) {
        struct ct_flow_offload_item items[CT_DIR_NUM];
        struct ct_offload_handle *coh;
        int ret;

        coh = conntrack_offload_get(conn);
        conntrack_offload_build_add_items(NULL, conn, items, 0);

        if (ovs_doca_flow_ct_enabled) {
            bool last = (idx == batch_size(conns));

            ret = doca_offload_insert_conn_doca_ct(esw, qid, netdevs, items,
                                                   &coh->dir[0].conn_item,
                                                   &coh->act_resources,
                                                   last);
            log_conn_rule(conn, CT_DIR_INIT, ret,
                          coh->act_resources.ct_match_zone_id,
                          coh->act_resources.ct_action_label_id);
            log_conn_rule(conn, CT_DIR_REP, ret,
                          coh->act_resources.ct_match_zone_id,
                          coh->act_resources.ct_action_label_id);
        } else {
            for (int dir = 0; dir < CT_DIR_NUM; dir++) {
                ret = doca_offload_insert_conn_dir(esw, qid, netdevs[dir],
                                                   &items[dir],
                                                   &coh->dir[dir].conn_item,
                                                   &coh->act_resources);
                log_conn_rule(conn, dir, ret,
                              coh->act_resources.ct_match_zone_id,
                              coh->act_resources.ct_action_label_id);
            }
        }

        if (ret) {
            dpdk_offload_doca_destroy_conn(netdevs, conn, &error);
            netdev_offload_dpdk_free_conn_res(&coh->act_resources);
            VLOG_ERR_RL(&rl, "%s/%s: Failed to insert connection %p offload.",
                        netdev_get_name(netdevs[CT_DIR_INIT]),
                        netdev_get_name(netdevs[CT_DIR_REP]), conn);
        } else {
            batch_add(conns, conn);
        }
    }
    ovs_rwlock_unlock(&esw->port_rwlock);
    /* Complete the queue once the lock has been released to avoid
     * deadlock on recursive read-lock. */
    doca_offload_complete_queue_n(esw, qid, ovs_doca_flow_ct_enabled ? OVS_DOCA_CT_QUEUE_DEPTH
                                                                     : OVS_DOCA_QUEUE_DEPTH);

    return -!batch_size(conns);
}

static int
dpdk_offload_doca_packet_hw_entropy(struct netdev *netdev,
                                    struct dp_packet *packet,
                                    uint16_t *entropy)
{
    struct doca_flow_entropy_format header;
    struct doca_flow_port *port;
    uint8_t ip_proto = 0;
    ovs_be16 eth_proto;
    doca_error_t err;
    void *nh;
    void *l4;

    parse_tcp_flags(packet, &eth_proto, NULL, NULL);
    nh = dp_packet_l3(packet);
    l4 = dp_packet_l4(packet);

    if (!netdev_dpdk_is_ethdev(netdev)) {
        return -1;
    }

    memset(&header, 0, sizeof header);

    if (eth_proto == htons(ETH_TYPE_IP) && nh) {
        struct ip_header *ip = nh;

        header.l3_type = DOCA_FLOW_L3_TYPE_IP4;
        header.ip4.src_ip = (OVS_FORCE doca_be32_t) get_16aligned_be32(&ip->ip_src);
        header.ip4.dst_ip = (OVS_FORCE doca_be32_t) get_16aligned_be32(&ip->ip_dst);
        ip_proto = ip->ip_proto;
    } else if (eth_proto == htons(ETH_TYPE_IPV6) && nh) {
        struct ovs_16aligned_ip6_hdr *ip6 = nh;

        header.l3_type = DOCA_FLOW_L3_TYPE_IP6;
        memcpy(header.ip6.src_ip, &ip6->ip6_src, sizeof(ovs_be32[4]));
        memcpy(header.ip6.dst_ip, &ip6->ip6_dst, sizeof(ovs_be32[4]));
        ip_proto = ip6->ip6_nxt;
    } else {
        header.l3_type = DOCA_FLOW_L3_TYPE_NONE;
    }

    if (ip_proto == IPPROTO_TCP && l4) {
        struct tcp_header *tcp = l4;

        header.l4_type_ext = DOCA_FLOW_L4_TYPE_EXT_TCP;
        header.transport.src_port = (OVS_FORCE doca_be16_t) tcp->tcp_src;
        header.transport.dst_port = (OVS_FORCE doca_be16_t) tcp->tcp_dst;
    } else if (ip_proto == IPPROTO_UDP && l4) {
        struct udp_header *udp = l4;

        header.l4_type_ext = DOCA_FLOW_L4_TYPE_EXT_UDP;
        header.transport.src_port = (OVS_FORCE doca_be16_t) udp->udp_src;
        header.transport.dst_port = (OVS_FORCE doca_be16_t) udp->udp_dst;
    } else if (ip_proto == IPPROTO_ICMP) {
        header.l4_type_ext = DOCA_FLOW_L4_TYPE_EXT_ICMP;
    } else if (ip_proto == IPPROTO_ICMPV6) {
        header.l4_type_ext = DOCA_FLOW_L4_TYPE_EXT_ICMP6;
    } else {
        header.l4_type_ext = DOCA_FLOW_L4_TYPE_EXT_NONE;
    }

    port = doca_flow_port_switch_get(netdev_dpdk_doca_port_get(netdev));
    err = doca_flow_port_calc_entropy(port, &header, entropy);
    if (err != DOCA_SUCCESS) {
        return -1;
    }

    return 0;
}

static unsigned int
dpdk_offload_doca_thread_nb(void)
{
    /* Regular offload threads + management thread. */
    return netdev_offload_thread_nb() + 1;
}

struct doca_flow_pipe_entry *
dpdk_offload_sample_entry_get(const struct netdev *netdev)
{
    struct doca_offload_esw_ctx *esw_ctx = doca_offload_esw_ctx_get(netdev);

    return esw_ctx ? esw_ctx->sample_ctx.sample_pipe_entry : NULL;
}

struct ovs_list *
doca_offload_destroy_list_get(struct doca_offload_esw_ctx *esw_ctx)
{
    return &esw_ctx->destroy_pipe_lists[netdev_offload_thread_id()];
}

struct ovs_rwlock *
doca_offload_rwlock_get(const struct netdev *netdev)
{
    struct doca_offload_esw_ctx *esw_ctx = doca_offload_esw_ctx_get(netdev);

    return esw_ctx ? &esw_ctx->port_rwlock : NULL;
}

int
doca_offload_query_entry(struct ovs_rwlock *port_rwlock,
                         struct doca_flow_pipe_entry *entry,
                         struct doca_flow_resource_query *query_stats)
{
    int rv;

    if (port_rwlock) {
        ovs_rwlock_rdlock(port_rwlock);
        rv = doca_flow_resource_query_entry(entry, query_stats);
        ovs_rwlock_unlock(port_rwlock);
    } else {
        rv = doca_flow_resource_query_entry(entry, query_stats);
    }

    return rv;
}

struct dpdk_offload_api dpdk_offload_api_doca = {
    .upkeep = dpdk_offload_doca_upkeep,
    .create = dpdk_offload_doca_create,
    .destroy = dpdk_offload_doca_destroy,
    .query_count = dpdk_offload_doca_query_count,
    .get_packet_recover_info = dpdk_offload_doca_get_pkt_recover_info,
    .insert_conns = dpdk_offload_doca_insert_conns,
    .destroy_conn = dpdk_offload_doca_destroy_conn,
    .query_conn = dpdk_offload_doca_query_conn,
    .reg_fields = dpdk_offload_doca_get_reg_fields,
    .update_stats = dpdk_offload_doca_update_stats,
    .aux_tables_init = dpdk_offload_doca_aux_tables_init,
    .aux_tables_uninit = dpdk_offload_doca_aux_tables_uninit,
    .shared_create = dpdk_offload_doca_shared_create,
    .shared_destroy = dpdk_offload_doca_shared_destroy,
    .shared_query = dpdk_offload_doca_shared_query,
    .packet_hw_hash = dpdk_offload_doca_packet_hw_hash,
    .packet_hw_entropy = dpdk_offload_doca_packet_hw_entropy,
    .thread_nb = dpdk_offload_doca_thread_nb,
};
